//
//  compress.c
//  FF6Tools
//
//  Created by Evan Owen on 2/24/15.
//
//

#include <string.h>
#include "compress.h"

typedef unsigned char byte;
typedef unsigned short word;

unsigned int compress_ff3(unsigned char* destination, const unsigned char* source, size_t length)
{
    byte header = 0;
    byte lineBuffer[17];
    
    int sourcePosition = 0x0800; // start at 0x0800 to ignore the 2K of empty space
    int destinationPosition = 2; // start at 2 so we can fill in the length at the end
    int lineBufferPosition = 1; // start at 1 so we can fill in the header at the end
    int bufferPosition = 0x07DE;
    int currentSequencePosition = 0;
    int longestSequencePosition;
    int currentSequenceLength, longestSequenceLength;
    
    word w;
    byte lineMask = 1;
    
    // create a source buffer preceded by 2K of empty space (this increases compression for some data)
    byte* sourceBuffer = malloc(length + 0x0800);
    memset(sourceBuffer, 0, length + 0x0800);
    memcpy(sourceBuffer + 0x0800, source, length);
    
    while((sourcePosition - 0x0800) < length)
    {
        // find the longest sequence that matches the decompression buffer
        longestSequenceLength = 0;
        longestSequencePosition = 0;
        for(currentSequencePosition = 1; currentSequencePosition <= 0x0800; currentSequencePosition++)
        {
            currentSequenceLength = 0;
            while ((currentSequenceLength < 34) && ((sourcePosition - 0x0800) + currentSequenceLength < length) && (sourceBuffer[sourcePosition + currentSequenceLength - currentSequencePosition] == sourceBuffer[sourcePosition + currentSequenceLength]))
                currentSequenceLength++;
            
            if(currentSequenceLength > longestSequenceLength)
            {
                // this sequence is longer than any others that have been found so far
                longestSequenceLength = currentSequenceLength;
                longestSequencePosition = (bufferPosition - currentSequencePosition) & 0x07FF;
            }
        }

        // check if the longest sequence is compressible
        if (longestSequenceLength >= 3)
        {
            // sequence is compressible - add compressed data to line buffer
            w = (word)(((longestSequenceLength - 3) << 11) | longestSequencePosition);
            lineBuffer[lineBufferPosition++] = (byte)(w);
            w >>= 8;
            lineBuffer[lineBufferPosition++] = (byte)(w);
            sourcePosition += longestSequenceLength;
            bufferPosition += longestSequenceLength;
        }
        else
        {
            // sequence is not compressible - update header byte and add byte to line buffer
            header |= lineMask;
            lineBuffer[lineBufferPosition++] = sourceBuffer[sourcePosition];
            sourcePosition++;
            bufferPosition++;
        }
        bufferPosition &= 0x07FF;
        lineMask <<= 1;
        
        if(lineMask == 0)
        {
            // finished a line, copy it to the destination
            lineBuffer[0] = header;
            memcpy(destination + destinationPosition, lineBuffer, lineBufferPosition);
            destinationPosition += lineBufferPosition;
            header = 0;
            lineBufferPosition = 1;
            lineMask = 1;
        }
    }

    // free the source buffer
    free(sourceBuffer);

    if(lineMask != 1)
    {
        // we're done with all the data but we're still in the middle of a line
        lineBuffer[0] = header;
        memcpy(destination + destinationPosition, lineBuffer, lineBufferPosition);
        destinationPosition += lineBufferPosition;
    }
    
    // now fill in the length and return the total compressed size
    destination[0] = (byte)(destinationPosition);
    destination[1] = (byte)(destinationPosition >> 8);
    
    return destinationPosition;
}

unsigned int decompress_ff3(byte* destination, const byte* source)
{
    int sourcePosition = 2;
    int destinationPosition = 0;
    byte buffer [0x0800];
    memset(buffer, 0, 0x0800);
    const word compressedSize = ((const word*)source)[0];
    int header, pass, repeatCount;
    word w;
    byte b;
    int bufferPosition = 0x07DE;
    
    byte passBuffer [34];
    int passBufferPointer = 0;
    
    while (sourcePosition < compressedSize)
    {
        header = source[sourcePosition++];
        
        for (pass = 0; pass < 8; pass++, header >>= 1)
        {
            passBufferPointer = 0;
            if (header & 1)
            {
                b = source[sourcePosition++];
                passBuffer[passBufferPointer++] = b;
                buffer[bufferPosition++] = b;
                bufferPosition &= 0x07FF;
            }
            else
            {
                w = source[sourcePosition + 1];
                w <<= 8;
                w |= source[sourcePosition];
                sourcePosition += 2;
                repeatCount = (w >> 11) + 3;
                w &= 0x07FF;
                
                for (int i = 0; i < repeatCount; i++)
                {
                    b = buffer[(w + i) & 0x07FF];
                    passBuffer[passBufferPointer++] = b;
                    buffer[bufferPosition++] = b;
                    bufferPosition &= 0x07FF;
                }
            }
            if ((destinationPosition + passBufferPointer) > MAX_COMPRESSION_BUFFER) {
                // maximum buffer length exceeded
                memcpy(destination + destinationPosition, passBuffer, MAX_COMPRESSION_BUFFER - destinationPosition);
                return MAX_COMPRESSION_BUFFER;
            } else {
                // copy this pass to the destination buffer
                memcpy(destination + destinationPosition, passBuffer, passBufferPointer);
                destinationPosition += passBufferPointer;
            }
            
            if (sourcePosition >= compressedSize)
                // reached end of compressed data
                break;
        }
    }
    
    // return the decompressed size
    return destinationPosition;
}

// NOTE: I started writing this function in C to speed up dte optimization, but then I realized
//       that it's actually decoding and encoding all the dialog that makes it so slow.

void optimize_dte(unsigned char* dte, const unsigned char* dialog, size_t length) {
    
    int pos = 0;
    
    // buffer to keep track of the number of occurances of each two-byte sequence (max 65535)
    word sequenceCounts [0x10000];
    
    while (pos < (length - 1)) {
        
        // get the next two bytes
        byte byte1 = dialog[pos];
        byte byte2 = dialog[pos + 1];
        pos++;
        
        // only character 0x20...0x7F can be encoded as DTE
        if (byte1 < 0x20) continue;
        if (byte1 > 0x7F) continue;
        if (byte2 < 0x20) continue;
        if (byte2 > 0x7F) continue;
        
        // OR the bytes together to get the sequence
        int sequence = byte1 | (byte2 << 8);
        word count = sequenceCounts[sequence];
        
        // continue if the max count has been reached
        if (count == 0xFFFF) { continue; }
        
        // increment the sequence count
        sequenceCounts[sequence] = count + 1;
    }
    
    // find the 128 most common sequences
    
    // i haven't written this yet
}

