//
//  BattleScript.swift
//  FF6Tools
//
//  Created by Evan Owen on 5/7/16.
//
//

import Foundation

class BattleScriptEncoding: ROMScriptEncoding {
    
    func numDialogLines(_ stringAssembly: ROMText) -> Int {
        var numLines = 1
        if let regex = try? NSRegularExpression(pattern: "\\\\n", options: []) {
            numLines += regex.numberOfMatches(in: stringAssembly.string, options: [], range: NSMakeRange(0, stringAssembly.string.utf16.count))
        }
        if let regex = try? NSRegularExpression(pattern: "\\\\page", options: []) {
            numLines += regex.numberOfMatches(in: stringAssembly.string, options: [], range: NSMakeRange(0, stringAssembly.string.utf16.count))
        }
        return numLines
    }

    override func willDisassembleScript(_ script: ROMScript) {
        
        if script.identifier == "battleScript" {
            for (index, object) in script.rom.mutableArrayValue(forKey: "battleScriptPointers").enumerated() {
                guard let scriptPointerAssembly = object as? ROMAssembly,
                    let scriptPointer = scriptPointerAssembly.value(forKeyPath: "scriptPointer") as? Int else { continue }
                
                // make a label for each battle script
                let eventName = script.rom.string(atIndex: index, inStringTable: "battleScriptPointers") ?? ("Battle Event " + String(index))
                script.commandLabels[scriptPointer + script.pointerOffset] = eventName
                
                // create a reference for each monster script
                script.addReference(forAssembly: scriptPointerAssembly, encodingName: "battleScript", linkedKeyPath: "scriptPointer")
            }
        } else if script.identifier == "attackScript" {
            for (index, object) in script.rom.mutableArrayValue(forKey: "attackScriptPointers").enumerated() {
                guard let scriptPointerAssembly = object as? ROMAssembly,
                    let scriptPointer = scriptPointerAssembly.value(forKeyPath: "scriptPointer") as? Int else { continue }
                
                // make a label for each attack script
                if script.commandLabels[scriptPointer + script.pointerOffset] == nil {
                    let eventName = script.rom.string(atIndex: index, inStringTable: "attackScriptPointers") ?? ("Attack Script " + String(index))
                    script.commandLabels[scriptPointer + script.pointerOffset] = eventName
                }
                
                // create a reference for each attack script
                script.addReference(forAssembly: scriptPointerAssembly, encodingName: "attackScriptHeader", linkedKeyPath: "scriptPointer")
            }
        }
        
        super.willDisassembleScript(script)
    }
    
    override func didDisassembleCommand(_ command: ROMCommand, fromData data: Data) {
    
        let offset = command.range!.lowerBound
        
        switch command.identifier! {
            
        case "battleScript.animation":
            // pointer to animation
            command.script.addReference(forAssembly: command, encodingName: "attackScriptHeader", linkedKeyPath: "scriptPointer")

        case "battleScript.characterAnimations":
            // pointers to animations for each character
            for character in 0..<24 {
                command.script.addReference(forAssembly: command, encodingName: "attackScriptHeader", linkedKeyPath: "scriptPointer" + String(character))
            }

        case "attackScript.jumpBattleType":
            // pointer for each animation type (3)
            command.script.addReference(forAssembly: command, encodingName: "attackScript", linkedKeyPath: "normalPointer")
            command.script.addReference(forAssembly: command, encodingName: "attackScript", linkedKeyPath: "backPointer")
            command.script.addReference(forAssembly: command, encodingName: "attackScript", linkedKeyPath: "sidePointer")

        case "attackScript.jumpDirection":
            // pointer for each facing direction (2)
            command.script.addReference(forAssembly: command, encodingName: "attackScript", linkedKeyPath: "leftPointer")
            command.script.addReference(forAssembly: command, encodingName: "attackScript", linkedKeyPath: "rightPointer")

        case "attackScript.jumpSubroutine":
            // jump to subroutine
            command.script.addReference(forAssembly: command, encodingName: "attackScript", linkedKeyPath: "scriptPointer")

        case "attackScript.jumpSwdTech":
            // pointer for each swdtech hit (4)
            command.script.addReference(forAssembly: command, encodingName: "attackScript", linkedKeyPath: "scriptPointer0")
            command.script.addReference(forAssembly: command, encodingName: "attackScript", linkedKeyPath: "scriptPointer1")
            command.script.addReference(forAssembly: command, encodingName: "attackScript", linkedKeyPath: "scriptPointer2")
            command.script.addReference(forAssembly: command, encodingName: "attackScript", linkedKeyPath: "scriptPointer3")

        case "attackScript.jumpThread":
            // jump based on thread index
            var count = 0
            var currentOffset = offset + 1
            let firstPointerRange = NSMakeRange(currentOffset, 2)
            var firstPointer = 0
            currentOffset += 2
            count += 1
            (data as NSData).getBytes(&firstPointer, range: firstPointerRange)
            firstPointer += command.script.pointerOffset
            while (true) {
                let pointerRange = NSMakeRange(currentOffset, 2)
                var pointer = 0
                currentOffset += 2
                (data as NSData).getBytes(&pointer, range: pointerRange)
                pointer += command.script.pointerOffset
                if ((pointer < firstPointer) && (pointer >= currentOffset)) {
                    firstPointer = pointer
                }
                if currentOffset > firstPointer { break }
                if currentOffset > pointer { break }
                count += 1
            }
            command.range = offset..<(offset + count * 2 + 1)
            command.disassembleFrom(data)
            
            for thread in 1...count {
                command.script.addReference(forAssembly: command, encodingName: "attackScript", linkedKeyPath: "scriptPointer" + String(thread))
            }
            
        case "attackScript.jumpTarget", "attackScript.jumpAttacker":
            // jump based on target/attacker (5)
            command.script.addReference(forAssembly: command, encodingName: "attackScript", linkedKeyPath: "character1Pointer")
            command.script.addReference(forAssembly: command, encodingName: "attackScript", linkedKeyPath: "character2Pointer")
            command.script.addReference(forAssembly: command, encodingName: "attackScript", linkedKeyPath: "character3Pointer")
            command.script.addReference(forAssembly: command, encodingName: "attackScript", linkedKeyPath: "character4Pointer")
            command.script.addReference(forAssembly: command, encodingName: "attackScript", linkedKeyPath: "monsterPointer")
            
        case "attackScript.jumpMagitek":
            // jump if in magitek mode (2)
            command.script.addReference(forAssembly: command, encodingName: "attackScript", linkedKeyPath: "scriptPointer")
            command.script.addReference(forAssembly: command, encodingName: "attackScript", linkedKeyPath: "magitekPointer")

        case "attackScript.jump":
            // unconditional jump
            command.script.addReference(forAssembly: command, encodingName: "attackScript", linkedKeyPath: "scriptPointer")

        case "attackScript.staticLoop", "attackScript.animatedLoop", "attackScript.frameLoop":
            // repeat
            var currentOffset = offset + 2
            while currentOffset < command.script.range!.upperBound {
                guard let repeatCommand = ROMCommand(command.script.rom, script: command.script, encodingName: command.childEncodingName, data: data, offset: currentOffset) else { break }
                repeatCommand.parent = command
                command.childCommands.add(repeatCommand)
                currentOffset += repeatCommand.assembledLength
                if repeatCommand.identifier == "attackScript.endStaticLoop" { break }
                if repeatCommand.identifier == "attackScript.endAnimatedLoop" { break }
                if repeatCommand.identifier == "attackScript.endOfScript" { break }
                if repeatCommand.identifier == "attackScript.return" { break }
            }
            
        case "attackScript.moveVector":
            // branch backward
            let referenceOffset = command.range!.upperBound - command.intValue(forKey: "offset") - 1
            command.script.addReference(forAssembly: command, encodingName: "attackScript", linkedKeyPath: "offset", atAbsoluteOffset: referenceOffset)

        case "attackScript.branchIfHit":
            // branch forward
            let referenceOffset = command.range!.upperBound + command.intValue(forKey: "offset")
            command.script.addReference(forAssembly: command, encodingName: "attackScript", linkedKeyPath: "offset", atAbsoluteOffset: referenceOffset)

        default: break
        }
    }
    
    override func valueForCommand(_ command: ROMCommand, forKey key: String) -> Any? {
        switch command.identifier! {
            
        case "attackScript.moveThread", "attackScript.moveAttacker", "attackScript.moveTarget", "attackScript.moveThreadIfFlipped":
            if key == "distance" {
                guard let distance = super.valueForCommand(command, forKey: key) as? Int else { break }
                return (distance + 1) as AnyObject?
            }
            
        case "attackScript.jumpThread":
            
            let threadCount = (command.data.count - 1) / 2
            
            if key == "threadCount" {
                return threadCount as AnyObject?
                
            } else if let lastChar = key.characters.last, let index = Int(object: String(lastChar) as AnyObject?) {
                
                if key.hasPrefix("enableThread") {
                    // determine whether this switch should be enabled
                    return (threadCount >= index) as AnyObject?
                } else if index > threadCount {
                    // if the switch is not enabled, return nil so that it won't try to access and invalid range of bytes
                    return nil
                }
            }
            
        default:
            break
        }
        
        return super.valueForCommand(command, forKey: key)
    }

    override func setValue(_ value: Any?, forCommand command: ROMCommand, forKey key: String) {
        switch command.identifier! {
            
        case "attackScript.moveThread", "attackScript.moveAttacker", "attackScript.moveTarget", "attackScript.moveThreadIfFlipped":
            if key == "distance" {
                guard let distance = value as? Int else { break }
                super.setValue((distance - 1) as AnyObject?, forCommand: command, forKey: key)
                return
            }
            
        case "attackScript.moveVector":
            // branch backward
            if key == "offset" {
                guard let eventPointer = value as? Int else { break } // pass non-Int values to superclass
                let offset = command.range!.upperBound + command.script.pointerOffset - eventPointer - 1
                super.setValue(offset as AnyObject?, forCommand: command, forKey: "offset")
                return
            }

        case "attackScript.branchIfHit":
            // branch forward
            if key == "offset" {
                guard let eventPointer = value as? Int else { break } // pass non-Int values to superclass
                let offset = eventPointer + command.script.pointerOffset - command.range!.upperBound
                super.setValue(offset as AnyObject?, forCommand: command, forKey: "offset")
                return
            }

        case "attackScript.jumpThread":
            
            if key == "threadCount" {
                
                guard let newCount = value as? Int else { break }
                let oldCount = (command.data.count - 1) / 2
                
                // break if the value didn't change
                if newCount == oldCount { break }
                
                for index in min(newCount + 1, oldCount + 1)...max(newCount, oldCount) {
                    command.willChangeValue(forKey: "enableThread\(index)")
                    command.willChangeValue(forKey: "scriptPointer\(index)")
                    
                    // save undo data for choices that will be removed
                    if newCount > oldCount { continue }
                    
                    // delete references for choices that will be removed
                    command.setValue(0, forKey: "scriptPointer\(index)")
                    command.references?["scriptPointer" + String(index)] = nil
                }
                
                // change the size of the data
                command.willChangeValue(forKey: "threadCount")
                command.data.count = newCount * 2 + 1
                command.script.updateOffsets()
                command.didChangeValue(forKey: "threadCount")
                
                for index in min(newCount + 1, oldCount + 1)...max(newCount, oldCount) {
                    command.didChangeValue(forKey: "enableThread\(index)")
                    command.didChangeValue(forKey: "scriptPointer\(index)")
                    
                    if newCount < oldCount { continue }
                    
                    // make a copy of the previous choice's reference for each new reference
                    let reference = command.references?["scriptPointer" + String(index - 1)]
                    let copiedReference = reference?.copy() as? ROMReference
                    copiedReference?.linkedKeyPath = "scriptPointer" + String(index)
                    command.references?["scriptPointer" + String(index)] = copiedReference
                    command.setValue(copiedReference, forKey: "scriptPointer\(index)")
                }
                
                return
            }

        default: break
        }
        super.setValue(value, forCommand: command, forKey: key)
    }
    
    override func encodingNameForCommandAfter(_ command: ROMCommand) -> String {
        if command.encodingName == "attackScriptHeader" { return "attackScript" }
        
        return super.encodingNameForCommandAfter(command)
    }
    
    override func numLinesForCommandDescription(_ command: ROMCommand) -> Int {
        
        switch command.identifier! {
        case "battleScript.characterAnimations":
            return 25
            
        case "battleScript.battleDialog":
            let dialogIndex = command.intValue(forKey: "dialog")
            if let dialogText = rom.value(at: dialogIndex, inPropertyWithKey: "battleDialog") as? ROMText {
                return dialogText.numLines + 1// numDialogLines(dialogText) + 1
            }

        case "attackScript.jumpThread":
            let count = (command.data.count - 1) / 2
            return count + 1

        case "attackScript.jumpTarget", "attackScript.jumpAttacker":
            return 6
            
        case "attackScript.characterFrame":
            let leftFrame = command.intValue(forKey: "leftFrame")
            let rightFrame = command.intValue(forKey: "rightFrame")
            if leftFrame == rightFrame
            {
                return 1
            } else {
                return 3
            }

        default: break
        }
        return super.numLinesForCommandDescription(command)
    }
    
    override func descriptionForCommand(_ command: ROMCommand) -> String {
        
        func referenceName(_ key: String = "scriptPointer") -> String {
            guard let reference = command.value(forKey: key) as? ROMReference,
                let referencedCommand = reference.referencedAssembly as? ROMCommand else {
                    return "<Invalid Script Pointer>"
            }
            
            return referencedCommand.label
        }
        
        func characterFrameName(_ key: String = "frame") -> String {
            let frame = command.intValue(forKey: key)
            if frame >= 0x30 {
                // flipped horizontally
                let frameName = rom.string(atIndex: frame - 0x30, inStringTable: "attackCharacterFrames") ?? "Frame " + String(frame - 0x30)
                if frameName.range(of: "Left") != nil {
                    return frameName.replacingOccurrences(of: "Left", with: "Right")
                } else if frameName.range(of: "Right") != nil {
                    return frameName.replacingOccurrences(of: "Right", with: "Left")
                } else if frameName.range(of: "Back") != nil {
                    return frameName.replacingOccurrences(of: "Back", with: "Forward")
                } else if frameName.range(of: "Forward") != nil {
                    return frameName.replacingOccurrences(of: "Forward", with: "Back")
                } else {
                    return frameName + " (Flipped)"
                }
            } else {
                return rom.string(atIndex: frame, inStringTable: "attackCharacterFrames") ?? "Frame " + String(frame)
            }
        }
        
        switch command.identifier! {

            // MARK: battle script descriptions
            
        case "battleScript.characterAnimations":
            var description = "Animations for each character:\n"
            for character in 0..<24 {
                let characterName = command.rom.string(atIndex: character, inStringTable: "battleSpriteNames") ?? "Character " + String(character)
                description += characterName + ": " + referenceName("scriptPointer" + String(character)) + "\n"
            }
            return description
            
        case "battleScript.animation":
            var description = "Animation for "
            let character = command.intValue(forKey: "character")
            let characterName = command.rom.string(atIndex: character, inStringTable: "characterNames") ?? "Character " + String(character)
            let opcode = command.intValue(forKey: "opcode")
            switch opcode {
            case 3...6: description += characterName + ", slot " + String(opcode - 2)
            case 7...12: description += "monster slot " + String(opcode - 6)
            default: return "<Invalid Animation Command>"
            }
            description += ": " + referenceName()
            return description

        case "battleScript.addCharacter":
            let character = command.intValue(forKey: "character")
            let remove = command.boolValue(forKey: "remove")
            let characterName = command.rom.string(atIndex: character, inStringTable: "characterNames") ?? "Character " + String(character)
            return (remove ? "Remove " : "Add ") + characterName + " as a target"

        case "battleScript.showGauge":
            let character = command.intValue(forKey: "character")
            let hide = command.boolValue(forKey: "hide")
            let characterName = command.rom.string(atIndex: character, inStringTable: "characterNames") ?? "Character " + String(character)
            return (hide ? "Hide" : "Show") + " ATB Gauge for " + characterName

        case "battleScript.battleDialog":
            let dialogIndex = command.intValue(forKey: "dialog")
            guard let dialogText = command.rom.value(at: dialogIndex, inPropertyWithKey: "battleDialog") as? ROMText else {
                return "<Invalid battle dialog>"
            }
            return "Display battle dialog:\n" + dialogText.formattedString()

        case "battleScript.eventAnimation":
            let animation = command.intValue(forKey: "animation")
            let animationName = rom.string(atIndex: animation, inStringTable: "battleEventAnimations") ?? "Animation " + String(animation)
            return "Event Animation: " + animationName

        case "battleScript.clearAnimations":
            return "Clear Animations"

        case "battleScript.executeAnimations":
            return "Execute Animations"

        case "battleScript.openDialog":
            return "Open Dialog Window"

        case "battleScript.closeDialog":
            return "Close Dialog Window"

        case "battleScript.endOfScript", "attackScript.endOfScript":
            return "End of Script"

            // MARK: attack script descriptions
            
        case "attackScriptHeader.header":
            let alignment = command.intValue(forKey: "alignment")
            let speed = command.intValue(forKey: "speed")
            switch alignment {
            case 0: return "Speed " + String(speed + 1) + ", Align to Bottom of Target"
            case 1: return "Speed " + String(speed + 1) + ", Align to Center of Target"
            case 2: return "Speed " + String(speed + 1) + ", Align to Top of Target"
            case 3: return "Speed " + String(speed + 1) + ", Align to Layer 3 Thread"
            case 5: return "Speed " + String(speed + 1) + ", Align to Center of Screen"
            case 6: return "Speed " + String(speed + 1) + ", Align to Front of Target"
            case 7: return "Speed " + String(speed + 1) + ", Align to Character"
            default: return "Speed " + String(speed + 1) + ", Alignment " + String(alignment)
            }
            
        case "attackScript.frame":
            let frame = command.intValue(forKey: "frame")
            return "Frame " + String(frame)

        case "attackScript.moveThread":
            let distance = command.intValue(forKey: "distance")
            let direction = command.intValue(forKey: "direction")
            let directionString = rom.string(atIndex: direction, inStringTable: "attackDirections") ?? "<Invalid Direction>"
            return "Move " + directionString + " " + String(distance)

        case "attackScript.moveAttacker":
            let distance = command.intValue(forKey: "distance")
            let direction = command.intValue(forKey: "direction")
            let directionString = rom.string(atIndex: direction, inStringTable: "attackDirections") ?? "<Invalid Direction>"
            return "Move Attacker " + directionString + " " + String(distance)

        case "attackScript.moveTarget":
            let distance = command.intValue(forKey: "distance")
            let direction = command.intValue(forKey: "direction")
            let directionString = rom.string(atIndex: direction, inStringTable: "attackDirections") ?? "<Invalid Direction>"
            return "Move Target " + directionString + " " + String(distance)

        case "attackScript.moveThreadIfFlipped":
            let distance = command.intValue(forKey: "distance")
            let direction = command.intValue(forKey: "direction")
            let directionString = rom.string(atIndex: direction, inStringTable: "attackDirections") ?? "<Invalid Direction>"
            return "Move " + directionString + " " + String(distance) + " if Flipped"

        case "attackScript.speed":
            let speed = command.intValue(forKey: "speed") + 1
            return "Set Animation Speed to " + String(speed)

        case "attackScript.moveToAttacker":
            return "Move to Attacker Position"

        case "attackScript.moveVector":
            let speed = command.intValue(forKey: "speed")
            return "Move Along Vector at Speed " + String(speed) +  " (branch to " + referenceName("offset") + ")"

        case "attackScript.branchIfHit":
            return "Branch to " + referenceName("offset") + " if Attack Didn't Miss"

        case "attackScript.hideShowAttacker":
            let monsters = command.boolValue(forKey: "monsters")
            let show = command.boolValue(forKey: "show")
            return (show ? "Show" : "Hide") + " Attacker Sprite " + (monsters ? "(Monster)" : "(Character)")

        case "attackScript.hideShowTarget":
            let monsters = command.boolValue(forKey: "monsters")
            let show = command.boolValue(forKey: "show")
            return (show ? "Show" : "Hide") + " Target Sprite " + (monsters ? "(Monster)" : "(Character)")

        case "attackScript.orderPriority":
            let invalidate = command.boolValue(forKey: "invalidate")
            if invalidate {
                return "Invalidate character/monster order priority"
            } else {
                return "Validate character/monster order priority"
            }

        case "attackScript.staticLoop":
            let count = command.intValue(forKey: "count")
            return "Static Loop (" + String(count) + " times)"

        case "attackScript.endStaticLoop":
            return "End of Static Loop"

        case "attackScript.animatedLoop":
            let count = command.intValue(forKey: "count")
            return "Animated Loop (" + String(count) + " times)"

        case "attackScript.endAnimatedLoop":
            return "End of Animated Loop"

        case "attackScript.soundEffect":
            let soundEffect = command.intValue(forKey: "soundEffect")
            if soundEffect == 0 {
                return "Play default sound effect"
            } else {
                return "Play sound effect \(soundEffect)"
            }

        case "attackScript.tilePriority":
            let priority = command.intValue(forKey: "priority")
            return "Set tile priority to \(priority)"

        case "attackScript.paletteSwap":
            switch command.intValue(forKey: "opcode") {
            case 0x3B80: return "Swap target's palette with animation palette"
            case 0x3C80: return "Restore target's palette to normal"
            default: return "<Invalid Palette Swap>"
            }

        case "attackScript.jumpThread":
            var description = "Jump based on thread index:"
            let count = (command.data.count - 1) / 2
            for thread in 1...count {
                description += "\n" + String(thread) + ": " + referenceName("scriptPointer" + String(thread))
            }
            return description

        case "attackScript.jumpSubroutine":
            return "Jump to Subroutine: " + referenceName()
            
        case "attackScript.jumpTarget":
            var description = "Jump Based on Target:"
            description += "\nCharacter 1: " + referenceName("character1Pointer")
            description += "\nCharacter 2: " + referenceName("character2Pointer")
            description += "\nCharacter 3: " + referenceName("character3Pointer")
            description += "\nCharacter 4: " + referenceName("character4Pointer")
            description += "\nMonster: " + referenceName("monsterPointer")
            return description

        case "attackScript.jumpAttacker":
            var description = "Jump Based on Attacker:"
            description += "\nCharacter 1: " + referenceName("character1Pointer")
            description += "\nCharacter 2: " + referenceName("character2Pointer")
            description += "\nCharacter 3: " + referenceName("character3Pointer")
            description += "\nCharacter 4: " + referenceName("character4Pointer")
            description += "\nMonster: " + referenceName("monsterPointer")
            return description
            
        case "attackScript.magitekAction":
            let action = command.intValue(forKey: "action")
            let actionName = rom.string(atIndex: action, inStringTable: "attackMagitekActions") ?? "Magitek Action " + String(action)
            let target = command.boolValue(forKey: "target")
            let specificCharacter = command.boolValue(forKey: "specificCharacter")
            let character = command.intValue(forKey: "character")
            let characterName = rom.string(atIndex: character, inStringTable: "battleSpriteNames") ?? "Character " + String(character)
            if specificCharacter {
                return "Set " + characterName + "'s Magitek Action: " + actionName
            } else {
                return "Set " + (target ? "Target" : "Attacker") + "'s Magitek Action: " + actionName
            }

        case "attackScript.characterFrame":
            let leftFrame = command.intValue(forKey: "leftFrame")
            let rightFrame = command.intValue(forKey: "rightFrame")
            if leftFrame == rightFrame {
                return "Set Character Frame: " + characterFrameName("leftFrame")
            } else {
                return "Set Character Frame:\nFacing Left: " + characterFrameName("leftFrame") + "\nFacing Right: " + characterFrameName("rightFrame")
            }

        case "attackScript.return":
            return "Return From Subroutine"
            
        default: return super.descriptionForCommand(command)
        }
    }
}
