//
//  BattleText.swift
//  FF6Tools
//
//  Created by Evan Owen on 2/4/17.
//  Copyright © 2017 Evan Owen. All rights reserved.
//

import Foundation

class BattleTextEncoding: ROMTextEncoding {
    
    override func format(string: String) -> String {
        
        var formattedString = string
        
        // get all of the escape sequences, sorted from longest to shortest
        let escapeSequences = (encodingTable.allKeys as! [String]).filter{ $0.hasPrefix("\\") }.sorted{ $0.characters.count > $1.characters.count }
        
        for sequence in escapeSequences {
            
            switch sequence {
                
            case "\\n":
                // new line
                formattedString = formattedString.replacingOccurrences(of: sequence, with: "\n") // new line
                
            case "\\page":
                // new page
                formattedString = formattedString.replacingOccurrences(of: sequence, with: "\u{12}") // form feed
                
            case "\\char[":
                // character name
                for character in 0..<64 {
                    let characterNameText = rom.value(at: character, inPropertyWithKey: "characterNames") as? ROMText
                    let characterName = characterNameText?.string ?? "Character " + String(character)
                    let nameSequence = "\\char[" + String(character) + "]"
                    formattedString = formattedString.replacingOccurrences(of: nameSequence, with: characterName)
                }
                
            case let nameSequence where nameSequence.hasPrefix("\\char"):
                // character name (GBA)
                let scanner = Scanner(string: nameSequence)
                var characterName = ""
                var characterIndex = 0
                scanner.scanString("\\char", into: nil) // ignore leading "\\char"
                if scanner.scanInt(&characterIndex) == false {
                    rom.log("Invalid character name index")
                    formattedString = formattedString.replacingOccurrences(of: sequence, with: "")
                    break
                }
                
                let nameString = rom.value(at: characterIndex, inPropertyWithKey: "characterNames") as! ROMText
                
                if scanner.scanString("_h", into: nil) {
                    // half-width kana
                    var nameBytes = [UInt8](repeating: 0, count: nameString.assembledLength)
                    nameString.encodedData.getBytes(&nameBytes, length: nameBytes.count)
                    
                    for characterByte in nameBytes {
                        var characterByteInt = Int(characterByte)
                        characterByteInt -= (encodingTable["バ"] as? Int) ?? 0x20
                        characterByteInt += (encodingTable["ﾊﾞ"] as? Int) ?? 0x1F20
                        if let halfWidthCharacter = decodingTable[characterByteInt] as? String {
                            characterName.append(halfWidthCharacter)
                        }
                    }
                } else {
                    // full-width letters
                    characterName = nameString.description
                }
                formattedString = formattedString.replacingOccurrences(of: nameSequence, with: characterName)

            case let sequenceWithParameter where sequenceWithParameter.hasSuffix("["):
                // escape sequence with parameter
                while let sequenceRange = formattedString.range(of: sequenceWithParameter) {
                    var substring = String(formattedString[sequenceRange.lowerBound...])
//                    var substring = formattedString.substring(from: sequenceRange.lowerBound)
                    if let endOfParameter = substring.characters.index(of: "]") {
                        substring = String(substring[...endOfParameter])
//                        substring = substring.substring(to: substring.index(after: endOfParameter))
                        formattedString.removeSubrange(formattedString.range(of: substring)!)
                    } else {
                        formattedString.removeSubrange(sequenceRange)
                    }
                }
                
            default:
                // remove all other escape sequences
                formattedString = formattedString.replacingOccurrences(of: sequence, with: "")
            }
        }
        return formattedString
    }
}
