//
//  EventScript.swift
//  FF6Tools
//
//  Created by Evan Owen on 4/24/16.
//
//

import Foundation

class EventScriptEncoding: ROMScriptEncoding {
    
    var ff6ROM: FF6ROM? { return rom as? FF6ROM }
    
    override func willDisassembleScript(_ script: ROMScript) {
        
        // return if this is not the actual event script
        guard let eventScript = rom.value(forKey: eventScriptKey) as? ROMScript else { return }
        if script !== eventScript { return }
        
        // create references for event triggers
        for (mapIndex, eventTriggerArray) in rom.mutableArrayValue(forKey: "eventTriggers").enumerated() {
            // get the event triggers for each map
            for object in (eventTriggerArray as AnyObject).mutableArrayValue(forKey: "array") {
                
                // get the event pointer for each trigger
                guard let triggerAssembly = object as? ROMAssembly else { continue }
                script.addReference(forAssembly: triggerAssembly, encodingName: mapIndex < 3 ? "worldScript" : "eventScript", linkedKeyPath: "eventPointer")
            }
        }
        
        // create references for npc event pointers
        for npcPropertiesArray in rom.mutableArrayValue(forKey: "npcProperties") {
            // get the npc properties for each map
            for object in (npcPropertiesArray as AnyObject).mutableArrayValue(forKey: "array") {
                
                // get the event pointer for each trigger
                guard let npcAssembly = object as? ROMAssembly,
                    let vehicle = npcAssembly.value(forKey: "vehicle") as? Int,
                    let specialGraphics = npcAssembly.value(forKey: "specialGraphics") as? Int else { continue }
                    
                // skip special NPC graphics
                if vehicle == 0 && specialGraphics == 1 { continue }
                
                script.addReference(forAssembly: npcAssembly, encodingName: "eventScript", linkedKeyPath: "eventPointer")
            }
        }
        
        // create references for map startup events
        for object in rom.mutableArrayValue(forKey: "mapStartupEventPointers") {
            
            // get the event pointer for each trigger
            guard let pointerAssembly = object as? ROMAssembly else { continue }
            script.addReference(forAssembly: pointerAssembly, encodingName: "eventScript", linkedKeyPath: "eventPointer")
        }
        
        // create references for vehicle event pointers
        for object in rom.mutableArrayValue(forKey: "vehicleEventPointers") {
            
            // get the event pointer for each trigger
            guard let pointerAssembly = object as? ROMAssembly else { continue }
            script.addReference(forAssembly: pointerAssembly, encodingName: "vehicleScript", linkedKeyPath: "eventPointer")
        }
        
        // create references for GBA events
        if script.rom.system != .GBA { return }
        
        for object in rom.mutableArrayValue(forKey: "mapStartupEventPointersAdvance") {
            
            // get the event pointer for each trigger
            guard let pointerAssembly = object as? ROMAssembly else { continue }
            script.addReference(forAssembly: pointerAssembly, encodingName: "eventScript", linkedKeyPath: "eventPointer")
        }
        for object in rom.mutableArrayValue(forKey: "npcPropertiesAdvance") {
            
            // get the event pointer for each trigger
            guard let npcAssembly = object as? ROMAssembly,
                let vehicle = npcAssembly.value(forKey: "vehicle") as? Int,
                let specialGraphics = npcAssembly.value(forKey: "specialGraphics") as? Int else { continue }
            
            // skip special NPC graphics
            if vehicle == 0 && specialGraphics == 1 { continue }
            
            script.addReference(forAssembly: npcAssembly, encodingName: "eventScript", linkedKeyPath: "eventPointer")
        }
        for object in rom.mutableArrayValue(forKey: "eventTriggersAdvance") {
            
            // get the event pointer for each trigger
            guard let triggerAssembly = object as? ROMAssembly else { continue }
            
            script.addReference(forAssembly: triggerAssembly, encodingName: "eventScript", linkedKeyPath: "eventPointer")
        }
    }
    
    // i moved this stuff to "didDisassembleCommand:"
//    override func didCreateBlankCommand(command: ROMCommand) {
//        if command.identifier == "eventScript.esper" {
//            command.setValue(0x36, forKey: "esper")
//            
//        } else if command.identifier == "eventScript.objectScript" {
//            guard let endOfScriptCommand = ROMCommand(rom, script: command.script, identifier: "objectScript.endOfScript") else { return }
//            endOfScriptCommand.parent = command
//            command.childCommands.addObject(endOfScriptCommand)
//        }
//    }
    
    override func willAssembleCommand(_ command: ROMCommand) {
        
        switch command.identifier! {
            
        case "eventScript.objectScript":
            let scriptLength = command.assembledLength - 2
            command.setValue(scriptLength, forKey: "length")
            
        default: break
        }
    }
    
    override func didDisassembleCommand(_ command: ROMCommand, fromData data: Data) {

        let offset = command.range!.lowerBound
        
        switch command.identifier! {
            
        case "eventScript.objectScript": // 1.75
            // object script
            var currentOffset = command.range!.lowerBound + 2
            let endOffset = min(currentOffset + command.intValue(forKey: "length"), command.script.range!.upperBound)
            while currentOffset < endOffset {
                guard let objectCommand = ROMCommand(rom, script: command.script, encodingName: command.childEncodingName, data: data, offset: currentOffset) else { break }
                objectCommand.parent = command
                command.childCommands.add(objectCommand)
                currentOffset += objectCommand.assembledLength
            }
            if command.childCommands.count == 0 {
                // if the child array is empty, add an end of script command
                guard let endOfScriptCommand = ROMCommand(rom, script: command.script, identifier: "objectScript.endOfScript") else { break }
                endOfScriptCommand.parent = command
                command.childCommands.add(endOfScriptCommand)
            }
            
        case "eventScript.repeat": // 0.13
            // repeat
            var currentOffset = command.range!.lowerBound + 2
            while currentOffset < command.script.range!.upperBound {
                guard let repeatCommand = ROMCommand(rom, script: command.script, encodingName: command.childEncodingName, data: data, offset: currentOffset) else { break }
                repeatCommand.parent = command
                command.childCommands.add(repeatCommand)
                currentOffset += repeatCommand.assembledLength
                if repeatCommand.identifier == "eventScript.endRepeat" { break }
                if repeatCommand.identifier == "eventScript.endRepeatSwitch" { break }
            }
            if command.childCommands.count == 0 {
                // if the child array is empty, add an end of script command
                guard let endOfScriptCommand = ROMCommand(rom, script: command.script, identifier: "eventScript.endRepeat") else { break }
                endOfScriptCommand.parent = command
                command.childCommands.add(endOfScriptCommand)
            }

        case "eventScript.esper":
            let esper = command.intValue(forKey: "esper")
            if esper < 0 {
                command.setValue(0, forKey: "esper")
            }

        case "eventScript.partyMenu":
            let numParties = command.intValue(forKey: "numParties")
            if numParties < 1 {
                command.setValue(1, forKey: "numParties")
            }

        case "eventScript.changeBackground":
            // change map background
            let height = command.intValue(forKey: "height")
            let width = command.intValue(forKey: "width")
            command.range = offset..<(offset + height * width + 5)
            command.disassembleFrom(data)
            
        case "objectScript.jump": // 0.0
            
            command.script.addReference(forAssembly: command, encodingName: "eventScript", linkedKeyPath: "eventPointer")

        case "eventScript.objectEvent", "eventScript.startTimer", "eventScript.jumpSubroutine", "eventScript.jumpSubroutineRepeat", "eventScript.jumpBattleSwitch", "eventScript.jumpRandom", "worldScript.jumpKeypress", "worldScript.jumpDirection": // 0.19
            
            command.script.addReference(forAssembly: command, encodingName: command.encodingName, linkedKeyPath: "eventPointer")

        case "eventScript.jumpDialog": // 0.21
            // jump based on dialog choice
            var dialogIndex: Int?
            var choiceCount = 2
            
            // find the previous dialog message
            var previousCommand = command.siblings.lastObject as? ROMCommand
            while previousCommand != nil {
                if previousCommand!.identifier == "eventScript.dialog" {
                    // found the previous dialog command
                    dialogIndex = previousCommand!.intValue(forKey: "dialogIndex")
                    break
                }
                previousCommand = previousCommand?.previousCommand
            }
            
            if dialogIndex == nil {
                rom.log("No preceding dialog command found, default to 2 choices")
                command.range = offset..<(offset + choiceCount * 3 + 1)
                command.disassembleFrom(data)
                break
            }
            
            // count the number of choices
            guard let dialogText = command.rom.value(at: dialogIndex!, inPropertyWithKey: "dialog") as? ROMText else { break }
            if let regex = try? NSRegularExpression(pattern: "\\\\choice", options: []) {
                choiceCount = regex.numberOfMatches(in: dialogText.string, options: [], range: NSMakeRange(0, dialogText.string.characters.count))
            }
            
            command.range = offset..<(offset + choiceCount * 3 + 1)
            command.disassembleFrom(data)

            // create a reference for each choice
            if choiceCount == 0 { break } // shouldn't happen unless there is an error during disassembly
            for choice in 1...choiceCount {
                
                command.script.addReference(forAssembly: command, encodingName: "eventScript", linkedKeyPath: "eventPointer" + String(choice))
            }

        case "eventScript.jumpCharacterSwitch": // 0.03
            // jump based on character switch or result of event variable comparison
            let switchCount = command.intValue(forKey: "switchCount")
            command.range = offset..<(offset + switchCount * 3 + 2)
            command.disassembleFrom(data)
            
            // create a reference for each switch
            if switchCount == 0 { break } // shouldn't happen unless there is an error during disassembly
            for index in 1...switchCount {
                command.script.addReference(forAssembly: command, encodingName: "eventScript", linkedKeyPath: "eventPointer" + String(index))
            }

        case "eventScript.jumpSwitch", "worldScript.jumpSwitch", "vehicleScript.jumpSwitch": // 0.44
            // jump based on event switch
            let switchCount = command.intValue(forKey: "switchCount")
            command.range = offset..<(offset + switchCount * 2 + 4)
            command.disassembleFrom(data)
            
            command.script.addReference(forAssembly: command, encodingName: command.encodingName, linkedKeyPath: "eventPointer" + String(switchCount))

        case "objectScript.branch": // 0.05
            // relative branch in an object script
            let opcode = command.intValue(forKey: "opcode")
            var referenceOffset = offset
            if (opcode == 0xFA) || (opcode == 0xFC) {
                // branch backward
                referenceOffset -= command.intValue(forKey: "offset")
            } else {
                // branch forward
                referenceOffset += command.intValue(forKey: "offset")
            }
            command.script.addReference(forAssembly: command, encodingName: "objectScript", linkedKeyPath: "offset", atAbsoluteOffset: referenceOffset)

        case "eventScript.song":
            // play a song (with or without volume)
            if command.intValue(forKey: "opcode") == 0xEF {
                // read the volume byte
                command.range = offset..<(offset + 3)
                command.disassembleFrom(data)
            }

        case "eventScript.soundEffect":
            // play a sound effect (with or without pan)
            if command.intValue(forKey: "opcode") == 0xF5 {
                // read the pan and envelope bytes
                command.range = offset..<(offset + 4)
                command.disassembleFrom(data)
            }

        default: // 0.84
            break
        }
    }
    
    override func encodingNameForCommandAfter(_ command: ROMCommand) -> String {
        
        switch command.identifier! {
        case "eventScript.loadMap", "worldScript.loadMap", "vehicleScript.loadMap":
            let mapIndex = command.intValue(forKey: "mapIndex")
            let vehicle = command.intValue(forKey: "vehicle")
            if [0,1,2,0x01FE,0x01FF].contains(mapIndex) {
                // world map
                if vehicle == 0 {
                    // no vehicle
                    return "worldScript"
                } else {
                    // vehicle
                    return "vehicleScript"
                }
            } else {
                // normal map
                return "eventScript"
            }
            
        case "worldScript.endOfScript", "vehicleScript.endOfScript":
            
            let placeholder = command.script.commandReferences[command.range!.lowerBound + command.assembledLength]
            if let placeholderCommand = placeholder as? ROMCommand {
                return placeholderCommand.encodingName
            } else if let placeholderReference = placeholder as? ROMReference, let placeholderEncoding = placeholderReference.encodingName {
                return placeholderEncoding
            }
            return "eventScript"
            
        default: break
        }
        
        return super.encodingNameForCommandAfter(command)
    }
    
    override func encodingNameForChildrenOfCommand(_ command: ROMCommand) -> String {
        if command.identifier == "eventScript.objectScript" { return "objectScript" }
        return command.encodingName
    }

    override func numLinesForCommandDescription(_ command: ROMCommand) -> Int {

        switch command.identifier! {
            
        case "eventScript.dialog":
            let dialogIndex = command.intValue(forKey: "dialogIndex")
            let dialogText = command.rom.value(at: dialogIndex, inPropertyWithKey: "dialog") as? ROMText
            let numLines = dialogText?.numLines ?? 1
            return numLines + 1
            
        case "eventScript.jumpDialog":
            // one line for each dialog choice
            return 1 + (command.data.count - 1) / 3
            
        case "eventScript.jumpSwitch", "worldScript.jumpSwitch", "vehicleScript.jumpSwitch":
            // one line for each switch tested
            let switchCount = command.intValue(forKey: "switchCount")
            if switchCount == 1 { break }
            return switchCount + 1
            
        case "eventScript.jumpCharacterSwitch":
            // one line for each switch tested
            let switchCount = command.intValue(forKey: "switchCount")
            if switchCount == 0 { break }
            return switchCount + 1
            
        case "eventScript.setParty":
            // one line for each party character
            return 5
            
        default:
            break
        }
        
        return super.numLinesForCommandDescription(command)
    }
    
    override func valueForCommand(_ command: ROMCommand, forKey key: String) -> Any? {
        
        switch command.identifier! {
            
        case "eventScript.modifyHP", "eventScript.modifyMP":
            if key == "setToMax" {
                if let subtract = super.valueForCommand(command, forKey: "subtract") as? Bool,
                    let value = super.valueForCommand(command, forKey: "value") as? Int
                    , subtract && (value == 0x7F) {
                    return true as AnyObject?
                } else { return false as AnyObject? }
            }
            
        case "eventScript.switch":
            if key == "switch" {
                // combine the high and low bytes to get the switch index
                guard let lowByte = super.valueForCommand(command, forKey: key) as? Int else { break }
                let highByte = command.intValue(forKey: "highByte")
                return (lowByte | (highByte << 8)) as AnyObject?
            }
            
        case "objectScript.switch":
            guard let opcode = super.valueForCommand(command, forKey: "opcode") as? Int else { break }
//            let opcode = command.intValue(forKey: "opcode")
            let clear = (opcode >= 0xE4)

            if key == "switch" {
                // combine the high and low bytes to get the switch index
                guard let lowByte = super.valueForCommand(command, forKey: key) as? Int else { break }
                let highByte = opcode - (clear ? 0xE4 : 0xE1)
                return (lowByte | (highByte << 8)) as AnyObject?
            } else if key == "clear" {
                return clear as AnyObject?
            }
            
        case "eventScript.jumpDialog":
            
            let choiceCount = (command.data.count - 1) / 3
            
            if key == "choiceCount" {
                return choiceCount as AnyObject?
                
            } else if let lastChar = key.characters.last, let index = Int(object: String(lastChar) as AnyObject?) {
                
                if key.hasPrefix("enableChoice") {
                    // determine whether this switch should be enabled
                    return (choiceCount >= index) as AnyObject?
                } else if index > choiceCount {
                    // if the switch is not enabled, return nil so that it won't try to access and invalid range of bytes
                    return nil
                }
            }
            
        case "eventScript.jumpSwitch", "worldScript.jumpSwitch", "vehicleScript.jumpSwitch":

            // switch count is 0-indexed, increment it to make it 1-indexed
            guard let intValue = super.valueForCommand(command, forKey: "switchCount") as? Int else { break }
            let switchCount = intValue + 1
            
            if key == "switchCount" {
                return switchCount as AnyObject?
            } else if key.hasPrefix("eventPointer") {
                // eventPointer always accesses the event pointer for the current switch count
                return super.valueForCommand(command, forKey: "eventPointer" + String(switchCount))
                
            } else if let lastChar = key.characters.last, let index = Int(object: String(lastChar) as AnyObject?) {
                
                if key.hasPrefix("enableSwitch") {
                    // determine whether this switch should be enabled
                    return (switchCount >= index) as AnyObject?
                } else if index > switchCount {
                    if key.hasPrefix("switch") {
                        return 0 as AnyObject?
                    } else if key.hasPrefix("negate") {
                        return 1 as AnyObject?
                    }
                    // if the switch is not enabled, return nil so that it won't try to access and invalid range of bytes
                    return nil
                }
            }
            
        case "eventScript.jumpCharacterSwitch":
            
            let switchCount = (command.data.count - 2) / 3

            if let prefixRange = key.range(of: "enableSwitch") ?? key.range(of: "switch") ?? key.range(of: "eventPointer"),
                let index = Int(object: String(key[prefixRange.upperBound...]) as AnyObject?) {
                
                if key.hasPrefix("enableSwitch") {
                    // determine whether this switch should be enabled
                    return (switchCount >= index) as AnyObject?
                } else if index > switchCount {
                    if key.hasPrefix("switch") {
                        return 0 as AnyObject?
                    }
                    // if the switch is not enabled, return nil so that it won't try to access and invalid range of bytes
                    return nil
                }
            }
            
        case "eventScript.song":
            if key == "volume" && command.data.count == 2 {
                return 0xFF as AnyObject?
            }

        case "eventScript.soundEffect":
            if key == "pan" && command.data.count == 2 {
                return 0x80 as AnyObject?
            } else if key == "envelope" && command.data.count == 2 {
                return 0 as AnyObject?
            }
            
        case "objectScript.branch":
            if key == "random" {
                let opcode = command.intValue(forKey: "opcode")
                if (opcode == 0xFA) || (opcode == 0xFB) {
                    return true as AnyObject?
                } else {
                    return false as AnyObject?
                }
            }
            
        case "eventScript.esper":
            if key == "esper" {
                guard let intValue = super.valueForCommand(command, forKey: key) as? Int else { break }
                return (intValue - 0x36) as AnyObject?
            }

        default: break
        }

        return super.valueForCommand(command, forKey: key)

    }
    
    override func setValue(_ value: Any?, forCommand command: ROMCommand, forKey key: String) {
        switch command.identifier! {
            
        case "eventScript.modifyHP", "eventScript.modifyMP":
//            if key == "setToMax" {
//                guard let boolValue = value as? Bool else { return }
//                if boolValue {
//                    super.setValue(true, forCommand: command, forKey: "subtract")
//                    super.setValue(0x7F, forCommand: command, forKey: "value")
//                } else {
//                    super.setValue(false, forCommand: command, forKey: "subtract")
//                    super.setValue(0, forCommand: command, forKey: "value")
//                }
//                return
            if key == "value" {
                guard let intValue = value as? Int else { return }
                command.willChangeValue(forKey: "setToMax")
                if intValue == 0x7F { command.setValue(true, forKey: "subtract") }
                super.setValue(intValue as AnyObject?, forCommand: command, forKey: "value")
                command.didChangeValue(forKey: "setToMax")
                return
            }

        case "objectScript.switch":
            let opcode = command.intValue(forKey: "opcode")
            let clear = (opcode >= 0xE4)

            if key == "switch" {
                
                // combine the high and low byte to set the switch number
                guard let eventSwitch = value as? Int else { break }
                let highByte = eventSwitch >> 8  + (clear ? 0xE4 : 0xE1)
                super.setValue((eventSwitch & 0xFF) as AnyObject?, forCommand: command, forKey: "switch")
                super.setValue(highByte as AnyObject?, forCommand: command, forKey: "opcode")
                return
                
            } else if key == "clear" {
                let highByte = opcode - (clear ? 0xE4 : 0xE1)
                guard let boolValue = value as? Bool else { return }
                command.willChangeValue(forKey: "clear")
                super.setValue((highByte + (boolValue ? 0xE4 : 0xE1)) as AnyObject?, forCommand: command, forKey: "opcode")
                command.didChangeValue(forKey: "clear")
                return
            }

        case "eventScript.switch":
            
            if key == "switch" {
                
                // combine the high and low byte to set the switch number
                guard let eventSwitch = value as? Int else { break }
                let highByte = eventSwitch >> 8
                super.setValue((eventSwitch & 0xFF) as AnyObject?, forCommand: command, forKey: "switch")
                super.setValue(highByte as AnyObject?, forCommand: command, forKey: "highByte")
                return
            }

        case "eventScript.jumpDialog":
            
            if key == "choiceCount" {
                
                guard let newCount = value as? Int else { break }
                let oldCount = (command.data.count - 1) / 3
                
                // break if the value didn't change
                if newCount == oldCount { break }
                
                for index in min(newCount + 1, oldCount + 1)...max(newCount, oldCount) {
                    command.willChangeValue(forKey: "enableChoice\(index)")
                    command.willChangeValue(forKey: "eventPointer\(index)")
                    
                    // save undo data for choices that will be removed
                    if newCount > oldCount { continue }
                    
                    // delete references for choices that will be removed
                    command.setValue(0, forKey: "eventPointer\(index)")
                    command.references?["eventPointer" + String(index)] = nil
                }

                // change the size of the data
                command.willChangeValue(forKey: "choiceCount")
                command.data.count = newCount * 3 + 1
                command.script.updateOffsets()
                command.didChangeValue(forKey: "choiceCount")
                
                for index in min(newCount + 1, oldCount + 1)...max(newCount, oldCount) {
                    command.didChangeValue(forKey: "enableChoice\(index)")
                    command.didChangeValue(forKey: "eventPointer\(index)")

                    if newCount < oldCount { continue }

                    // make a copy of the previous choice's reference for each new reference
                    let reference = command.references?["eventPointer" + String(index - 1)]
                    let copiedReference = reference?.copy() as? ROMReference
                    copiedReference?.linkedKeyPath = "eventPointer" + String(index)
                    command.references?["eventPointer" + String(index)] = copiedReference
                    command.setValue(copiedReference, forKey: "eventPointer\(index)")
                }

                return
            }

        case "eventScript.jumpSwitch", "worldScript.jumpSwitch", "vehicleScript.jumpSwitch":

            let switchCount = (command.data.count - 4) / 2 // intValue + 1

            if key == "switchCount" {
                
                guard let reference = command.value(forKey: "eventPointer") as? ROMReference,
                    let newCount = value as? Int else { return }
                
                // break if the value didn't change
                if newCount == switchCount { break }

                for index in min(newCount + 1, switchCount + 1)...max(newCount, switchCount) {
                    command.willChangeValue(forKey: "enableSwitch\(index)")
                    command.willChangeValue(forKey: "eventPointer\(index)")

                    // save undo data for switches that will be removed
                    if newCount > switchCount { continue }
                    command.setValue(0, forKey: "switch\(index)")
                    command.setValue(1, forKey: "negate\(index)")
                }
                
                // change the length of the data
                command.data.count = newCount * 2 + 4
                command.script.updateOffsets()
                
                // set the switch count and move the event pointer
                super.setValue((newCount - 1) as AnyObject?, forCommand: command, forKey: "switchCount")
                reference.linkedKeyPath = "eventPointer" + String(newCount)
                command.references = ["eventPointer" + String(newCount) : reference]

                for index in min(newCount + 1, switchCount + 1)...max(newCount, switchCount) {
                    command.didChangeValue(forKey: "enableSwitch\(index)")
                    command.didChangeValue(forKey: "eventPointer\(index)")
                    
                    // set new switches to zero
                    if newCount < switchCount { continue }
                    command.setValue(0, forKey: "switch\(index)")
                    command.setValue(1, forKey: "negate\(index)")
                }

                return
                
            } else if let reference = value as? ROMReference , key.hasPrefix("eventPointer") {
                
                // eventPointer always accesses the event pointer for the current switch count
                command.willChangeValue(forKey: "eventPointer")
                command.references = ["eventPointer" + String(switchCount) : reference]
                command.didChangeValue(forKey: "eventPointer")
                return
            }

        case "eventScript.jumpCharacterSwitch":
            
            if key == "switchCount" {
                
                guard let newCount = value as? Int else { break }
                let oldCount = (command.data.count - 2) / 3
                
                if newCount == oldCount { break }
                
                for index in min(newCount + 1, oldCount + 1)...max(newCount, oldCount) {
                    command.willChangeValue(forKey: "enableSwitch\(index)")
                    command.willChangeValue(forKey: "eventPointer\(index)")
                    
                    // save undo data for choices that will be removed
                    if newCount > oldCount { continue }
                    command.setValue(0, forKey: "switch" + String(index))
                    command.setValue(0, forKey: "eventPointer" + String(index))
                    command.references?["eventPointer" + String(index)] = nil
                }
                
                // change the size of the data
                command.data.count = newCount * 3 + 2
                command.script.updateOffsets()
                super.setValue(newCount as AnyObject?, forCommand: command, forKey: "switchCount")
                
                for index in min(newCount + 1, oldCount + 1)...max(newCount, oldCount) {
                    command.didChangeValue(forKey: "enableSwitch\(index)")
                    command.didChangeValue(forKey: "eventPointer\(index)")
                    
                    if newCount < oldCount { continue }
                    
                    // make a copy of the previous choice's reference for each new reference
                    let reference = command.references?["eventPointer" + String(index - 1)]
                    let copiedReference = reference?.copy() as? ROMReference
                    copiedReference?.linkedKeyPath = "eventPointer" + String(index)
                    command.references?["eventPointer" + String(index)] = copiedReference
                    command.setValue(copiedReference as AnyObject?, forKey: "eventPointer\(index)")
                    command.setValue(0 as AnyObject?, forKey: "switch\(index)")
                }
                
                return
            }
            
        case "eventScript.song":
            if key == "volume" {
                guard let intValue = value as? Int else { break }
                command.willChangeValue(forKey: key)
                let length = (intValue == 0xFF) ? 2 : 3
                var dataBytes = [UInt8](repeating: 0, count: length)
                (command.data as NSData).getBytes(&dataBytes, length: length)
                if length == 2 {
                    dataBytes[0] = 0xF0
                } else {
                    dataBytes[0] = 0xEF
                    dataBytes[2] = UInt8(truncatingIfNeeded: intValue)
                }
                command.data = Data(bytes: UnsafePointer<UInt8>(dataBytes), count: length)
                command.script.updateOffsets()
                command.didChangeValue(forKey: key)
                return
            }

        case "eventScript.soundEffect":
            if key == "pan" {
                guard let intValue = value as? Int else { break }
                command.willChangeValue(forKey: key)
                let length = (intValue == 0x80) ? 2 : 4
                var dataBytes = [UInt8](repeating: 0, count: length)
                (command.data as NSData).getBytes(&dataBytes, length: length)
                if length == 2 {
                    dataBytes[0] = 0xF4
                } else {
                    dataBytes[0] = 0xF5
                    dataBytes[2] = UInt8(truncatingIfNeeded: intValue)
                }
                command.data = Data(bytes: UnsafePointer<UInt8>(dataBytes), count: length)
                command.script.updateOffsets()
                command.didChangeValue(forKey: key)
                return
            }
            
        case "objectScript.branch":
            if key == "offset" {
                guard let eventPointer = value as? Int else { break } // pass non-Int values to superclass
                let random = command.boolValue(forKey: "random")
                let offset = eventPointer + command.script.pointerOffset - command.range!.lowerBound
                if offset > 0 {
                    // branch forward
                    super.setValue(offset as AnyObject?, forCommand: command, forKey: "offset")
                    command.setValue(random ? 0xFB : 0xFD, forKey: "opcode")
                } else {
                    // branch backward
                    super.setValue((-offset) as AnyObject?, forCommand: command, forKey: "offset")
                    command.setValue(random ? 0xFA : 0xFC, forKey: "opcode")
//                    command[0] = random ? 0xFA : 0xFC
                }
            } else if key == "random" {
                guard let boolValue = value as? Bool else { return }
                
                // at this point, it doesn't matter if we're going forward or backward
                command.willChangeValue(forKey: "random")
                command.setValue(boolValue ? 0xFA : 0xFC, forKey: "opcode")
//                command[0] = boolValue ? 0xFA : 0xFC
                command.didChangeValue(forKey: "random")
            }
            return

        case "eventScript.esper":
            if key == "esper" && value is Int {
                let intValue = value as! Int
                super.setValue((intValue + 0x36) as AnyObject?, forCommand: command, forKey: key)
                return
            }

        default:
            break
        }
        super.setValue(value, forCommand: command, forKey: key)
    }
    
    override func descriptionForCommand(_ command: ROMCommand) -> String {

        // local helper function definitions
        
        func stringValueForKey(_ key: String, inStringTable stringTable: String) -> String? {
            if command.properties?[key] == nil { return nil }
            let index = command.intValue(forKey: key)
            return stringValueAtIndex(index, inStringTable: stringTable)
        }

        func stringValueAtIndex(_ index: Int, inStringTable stringTable: String) -> String? {
            
            guard let stringTable = rom.value(forKey: "stringTable-" + stringTable) as? ROMStringTable,
                let string = stringTable[index] as? ROMString else { return nil }
            return string.formattedString
        }

        func characterName(_ key: String = "character") -> String {
            return stringValueForKey(key, inStringTable: "characterNames") ?? "<Invalid Character>"
        }
        
        func characterNameForIndex(_ index: Int) -> String {
            return stringValueAtIndex(index, inStringTable: "characterNames") ?? "<Invalid Character>"
        }
        
        func objectName(_ key: String = "object") -> String {
            return stringValueForKey(key, inStringTable: "objectNames") ?? "<Invalid Object>"
        }
        
        func eventName(_ key: String = "eventPointer") -> String {
            guard let reference = command.value(forKey: key) as? ROMReference,
                let referencedCommand = reference.referencedAssembly as? ROMCommand else {
                    return "<Invalid Address>"
            }
            
            return referencedCommand.label
        }
        
        func directionName(_ key: String = "direction") -> String {
            let direction = command.intValue(forKey: key)
            switch direction {
            case 0: return "up"
            case 1: return "right"
            case 2: return "down"
            case 3: return "left"
            default: return "<Invalid direction>"
            }
        }
        
        func vehicleName(_ key: String = "vehicle") -> String {
            let vehicle = command.intValue(forKey: key)
            switch vehicle {
            case 0: return "No vehicle"
            case 1: return "Chocobo"
            case 2: return "Magitek"
            case 3: return "Raft"
            default: return "<Invalid vehicle>"
            }
        }
        
        func switchName(_ key: String = "switch") -> String {
            let eventSwitch = command.intValue(forKey: key)
            return switchNameWithIndex(eventSwitch)
        }
        
        func switchNameWithIndex(_ index: Int) -> String {
            switch index {
            case 0x01A0..<0x01B0:
                if command.previousCommand?.identifier == "eventScript.variable" {
                    // variable compare operation
                    switch index - 0x01A0 {
                    case 0: return "Equal to"
                    case 1: return "Greater than"
                    case 2: return "Less than"
                    default: return "<Invalid comparison operation>"
                    }
                } else if command.previousCommand?.identifier == "eventScript.activePartyNumber" {
                    // active party number
                    return "Party " + String(index - 0x01A0) + " is active"
                } else if command.previousCommand?.identifier == "eventScript.jumpSubroutine",
                    let eventReference = command.previousCommand?.value(forKey: "eventPointer") as? ROMReference,
                    let referencedCommand = eventReference.referencedAssembly as? ROMCommand ,
                    referencedCommand.label == "Check Facing Direction" {
                    // facing direction
                    switch index - 0x01A0 {
                    case 0: return "Facing up"
                    case 1: return "Facing right"
                    case 2: return "Facing down"
                    case 3: return "Facing left"
                    default: return "<Invalid facing direction>"
                    }
                }
                // character name (default)
                return characterNameForIndex(index - 0x01A0)
            case 0x0000..<0x0300: return stringValueAtIndex(index, inStringTable: "eventSwitches") ?? ("Switch " + String(index))
            case 0x0300..<0x0700: return stringValueAtIndex(index - 0x0300, inStringTable: "npcSwitches") ?? ("NPC " + String(index - 0x0300))
            case 0x4000..<0x4300: return stringValueAtIndex(index - 0x4000, inStringTable: "eventSwitchesAdvance") ?? ("Switch " + String(index - 0x4000) + " (GBA)")
            default: return "<Invalid switch>"
            }
        }
        
        func variableName(_ key: String = "variable") -> String {
            let variable = command.intValue(forKey: key)
            return stringValueAtIndex(variable, inStringTable: "eventVariables") ?? ("Variable " + String(describing: index))
        }
        
        func colorName(_ key: String = "color") -> String {
            switch command.intValue(forKey: key) {
            case 0: return "Black"
            case 1: return "Red"
            case 2: return "Green"
            case 3: return "Yellow"
            case 4: return "Blue"
            case 5: return "Magenta"
            case 6: return "Cyan"
            case 7: return "White"
            default: return "<Invalid color>"
            }
        }
        
        func invertedColorName(_ key: String = "color") -> String {
            switch command.intValue(forKey: key) {
            case 0: return "White"
            case 1: return "Cyan"
            case 2: return "Magenta"
            case 3: return "Blue"
            case 4: return "Yellow"
            case 5: return "Green"
            case 6: return "Red"
            case 7: return "Black"
            default: return "<Invalid color>"
            }
        }
        
        var commandDescription: String
        
        switch command.identifier! {
            
            // MARK: event script commands
            
        case "eventScript.objectScript":
            return "Object script for " + objectName()
            
        case "eventScript.waitForObject":
            return "Wait for " + objectName()
            
        case "eventScript.createObject":
            switch command.intValue(forKey: "opcode") {
            case 0x3D: return "Create " + objectName()
            case 0x3E: return "Delete " + objectName()
            default: return "<Invalid opcode>"
            }
            
        case "eventScript.showObject":
            switch command.intValue(forKey: "opcode") {
            case 0x41: return "Show " + objectName()
            case 0x42: return "Hide " + objectName()
            default: return "<Invalid opcode>"
            }
            
        case "eventScript.objectPassability":
            switch command.intValue(forKey: "opcode") {
            case 0x36: return "Disable passability for " + objectName()
            case 0x78: return "Enable passability for " + objectName()
            default: return "<Invalid opcode>"
            }
            
        case "eventScript.objectCollisions":
            switch command.intValue(forKey: "opcode") {
            case 0x7C: return "Enable collisions for " + objectName()
            case 0x7D: return "Disable collisions for " + objectName()
            default: return "<Invalid opcode>"
            }
            
        case "eventScript.objectGraphics":
            let graphics = command.intValue(forKey: "graphics")
            let graphicsName = stringValueForKey("graphics", inStringTable: "mapCharacterGraphics") ?? String(graphics)
//            let graphicsName = rom.stringTable.valueAtIndex(graphics, inPropertyWithKey: "mapCharacterGraphics")?.description ?? String(graphics)
            return "Change " + objectName() + "'s graphics to " + graphicsName
            
        case "eventScript.vehicle":
            return "Set " + objectName() + "'s vehicle to " + vehicleName()
            
        case "eventScript.nameMenu":
            return "Open name change menu for " + characterName()
            
        case "eventScript.partyMenu":
            let numParties = command.intValue(forKey: "numParties")
            if numParties != 1 {
                return "Open party select menu (" + String(numParties) + " parties)"
            }
            return "Open party select menu"
            
        case "eventScript.colosseumMenu":
            return "Open colosseum menu"
            
        case "eventScript.shopMenu":
            let shop = command.intValue(forKey: "shop")
            let shopName = stringValueForKey("shop", inStringTable: "shopProperties") ?? String(shop)
//            let shopName = command.rom.stringTable.valueAtIndex(shop, inPropertyWithKey: "shopProperties")?.description ?? ("Shop " + String(shop))
            return "Open shop: " + shopName
            
        case "eventScript.finalBattleMenu":
            return "Open final battle menu"
            
        case "eventScript.loadMenu":
            return "Open game loading menu"
            
        case "eventScript.restoreSavedGame":
            return "Restore saved game"
            
        case "eventScript.startTimer":
            let timer = command.intValue(forKey: "timer")
            return "Start timer " + String(timer) + ": " + eventName()
            
        case "eventScript.stopTimer":
            let timer = command.intValue(forKey: "timer")
            return "Stop timer " + String(timer)
            
        case "eventScript.resetOverlay":
            return "Reset sprite overlay tiles"
            
        case "eventScript.enablePyramid":
            return "Create a pyramid around " + objectName()
            
        case "eventScript.disablePyramid":
            return "Disable pyramid"
            
        case "eventScript.flashlight":
            let size = command.intValue(forKey: "size")
            return "Set flashlight size to " + String(size)
            
        case "eventScript.mosaic":
            let speed = command.intValue(forKey: "speed")
            return "Mosaic the screen at speed " + String(speed)
            
        case "eventScript.animationFrame":
            let tile = command.intValue(forKey: "tile")
            let frame = command.intValue(forKey: "frame")
            return "Set animation tile " + String(tile) + " to frame " + String(frame)
            
        case "eventScript.animationSpeed":
            let tile = command.intValue(forKey: "tile")
            let speed = command.intValue(forKey: "speed")
            return "Set animation tile " + String(tile) + " to speed " + String(speed)
            
        case "eventScript.objectPalette":
            let palette = command.intValue(forKey: "palette")
            let paletteName = stringValueForKey("palette", inStringTable: "mapSpritePalettes") ?? String(palette)
//            let paletteName = rom.stringTable.valueAtIndex(palette, inPropertyWithKey: "mapSpritePalettes")?.description ?? "Palette " + String(palette)
            return "Change " + objectName() + "'s palette to " + paletteName
            
        case "eventScript.battle", "worldScript.battle", "vehicleScript.battle":
            let battle = command.intValue(forKey: "battle")
            let battleName = stringValueForKey("battle", inStringTable: "eventBattleGroups") ?? "Event Battle " + String(battle)
//            let battleName = rom.stringTable.valueAtIndex(battle, inPropertyWithKey: "eventBattleGroups")?.description ?? "Event Battle " + String(battle)
            return "Battle: " + battleName
            
        case "eventScript.colosseumBattle":
            return "Colosseum battle"
            
        case "eventScript.monster-in-a-box":
            return "Monster-in-a-box battle"
            
        case "eventScript.randomBattle":
            return "Random battle"
            
        case "eventScript.characterProperties":
            return "Set " + characterName() + "'s properties to " + objectName()
            
        case "eventScript.characterName":
            return "Change " + objectName() + "'s name to " + characterName()
            
        case "eventScript.characterPortrait":
            return "Show portrait for " + characterName()
            
        case "eventScript.normalizeLevel":
            return "Normalize " + objectName() + "'s level"
            
        case "eventScript.changeEquipment":
            switch command.intValue(forKey: "opcode") {
            case 0x8D: return "Remove all equipment from " + objectName()
            case 0x9C: return "Equip optimum equipment on " + objectName()
            default: return "<Invalid equipment command>"
            }
            
        case "eventScript.modifyStatus":
            return "Modify " + objectName() + "'s status"
            
        case "eventScript.modifyHP":
            return "Modify " + objectName() + "'s HP"
            
        case "eventScript.modifyMP":
            return "Modify " + objectName() + "'s MP"
            
        case "eventScript.characterParty":
            let party = command.intValue(forKey: "party")
            if party == 0 {
                return "Remove " + objectName() + " from the party"
            }
            return "Add " + objectName() + " to party " + String(party)
            
        case "eventScript.initCharacterObjects":
            return "Initialize character objects"
            
        case "eventScript.setParty":
            commandDescription = "Set party characters: \n"
            for slot in 1...4 {
                let characterName = stringValueForKey("character\(slot)", inStringTable: "mainCharactersEmpty") ?? "<Invalid Character>"
                commandDescription += String(slot) + ". " + characterName
                if slot != 4 { commandDescription += "\n" }
            }
            return commandDescription
            
        case "eventScript.activateParty":
            let party = command.intValue(forKey: "party")
            return "Activate party " + String(party)
            
        case "eventScript.restorePreviousParty":
            return "Restore the previous party"
            
        case "eventScript.resetPreviousParty":
            return "Reset the previous party"
            
        case "eventScript.updateCharacters":
            return "Update character objects"
            
        case "eventScript.createPartyObject":
            return "Create the party object"
            
        case "eventScript.moveParty":
            let x = command.intValue(forKey: "x")
            let y = command.intValue(forKey: "y")
            return "Move the party to (" + String(x) + "," + String(y) + ")"
            
        case "eventScript.item":
            let remove = command.boolValue(forKey: "remove")
            let itemName = stringValueForKey("item", inStringTable: "itemNames") ?? "<Invalid item>"
            return (remove ? "Remove " : "Add ") + itemName + (remove ? " from the inventory" : " to the inventory")
            
        case "eventScript.gp":
            let remove = command.boolValue(forKey: "remove")
            let gp = command.intValue(forKey: "gp")
            return (remove ? "Remove " : "Add ") + String(gp) + (remove ? " GP from the inventory" : " GP to the inventory")
            
        case "eventScript.esper":
            let remove = command.boolValue(forKey: "remove")
            let esperName = stringValueForKey("esper", inStringTable: "esperNames") ?? "<Invalid Esper>"
            return (remove ? "Remove " : "Add ") + esperName + (remove ? " from the inventory" : " to the inventory")
            
        case "eventScript.bumRush":
            let bumRushName = stringValueAtIndex(0x64, inStringTable: "attackProperties") ?? "Bum Rush"
            return "Enable " + bumRushName
            
        case "eventScript.swdTech":
            let swdTechName = stringValueAtIndex(0x07, inStringTable: "battleCommandNames") ?? "SwdTech"
            return "Enable all " + swdTechName + " attacks"
            
        case "eventScript.dialog":
            let dialogIndex = command.intValue(forKey: "dialogIndex")
            let dialogText = command.rom.value(at: dialogIndex, inPropertyWithKey: "dialog") as? ROMText
            var dialogString = dialogText?.formattedString() ?? ""
            
            // replace page breaks with newlines
            dialogString = dialogString.replacingOccurrences(of: "\u{12}", with: "\n")
            
            // replace multiple choice indicators
            var choiceIndex = 1
            while dialogString.contains("\u{5}") {
                dialogString.replaceSubrange(dialogString.range(of: "\u{5}")!, with: String(choiceIndex) + ".")
                choiceIndex += 1
            }
            
            // replace half-width hiragana with normal hiragana
            for hiraganaChar in FF6ROM.hiraganaCharacters.characters {
                let hiraganaSequence = "\\" + String(hiraganaChar)
                dialogString = dialogString.replacingOccurrences(of: hiraganaSequence, with: String(hiraganaChar))
            }
            return "Display dialog message:\n" + dialogString
            
        case "eventScript.waitForDialogWindow":
            return "Wait for dialog window to close"
            
        case "eventScript.waitForKeypress":
            return "Wait for dialog keypress"
            
        case "eventScript.repeat":
            if let lastCommand = command.childCommands.lastObject as? ROMCommand, lastCommand.identifier == "eventScript.endRepeatSwitch" {
//            if (command.childCommands.lastObject as AnyObject).identifier == "eventScript.endRepeatSwitch" {
                return "Repeat while switch is off"
            } else {
                let count = command.intValue(forKey: "count")
                return "Repeat " + String(count) + " times"
            }
            
        case "eventScript.endRepeat":
            return "End repeat"
            
        case "eventScript.endRepeatSwitch":
            return "Repeat if " + switchName() + " == Off"
            
        case "eventScript.jumpRandom":
            return "Jump to " + eventName() + " (50% chance)"
            
        case "eventScript.jumpSubroutine":
            return "Jump to subroutine: " + eventName()
            
        case "eventScript.jumpSubroutineRepeat":
            let repeatCount = command.intValue(forKey: "repeatCount")
            return "Jump to subroutine: " + eventName() + ", " + String(repeatCount) + " times"
            
        case "eventScript.objectEvent":
            return "Set " + objectName() + "'s event to " + eventName()
            
        case "eventScript.jumpBattleSwitch":
            let battleSwitch = command.intValue(forKey: "battleSwitch")
            return "Jump to " + eventName() + " if battle switch " + String(battleSwitch) + " == Off"
            
        case "eventScript.battleSwitch":
            let battleSwitch = command.intValue(forKey: "battleSwitch")
            let clear = command.boolValue(forKey: "clear")
            return "Battle switch " + String(battleSwitch) + " = " + (clear ? "Off" : "On")
            
        case "eventScript.jumpSwitch", "worldScript.jumpSwitch", "vehicleScript.jumpSwitch":
            let switchCount = command.intValue(forKey: "switchCount")
            let eventNameString = eventName("eventPointer\(switchCount)")
            if switchCount == 1 {
                let negate = command.boolValue(forKey: "negate1")
                return "If " + switchName("switch1") + " == " + (negate ? "On" : "Off") + ", jump to " + eventNameString
            }
            
            let and = command.boolValue(forKey: "and")
            commandDescription = "Jump to " + eventNameString + " if " + (and ? "all" : "any") + " of the following conditions are met:\n"
            for bit in 1...switchCount {
                let negate = command.boolValue(forKey: "negate\(bit)")
                commandDescription += String(bit) + ". " + switchName("switch\(bit)") + " == " + (negate ? "On" : "Off")
                if bit != switchCount { commandDescription += "\n" }
            }
            return commandDescription
            
        case "eventScript.jumpDialog":
            commandDescription = "Jump based on dialog choice: \n"
            let choiceCount = (command.data.count - 1) / 3
            if choiceCount == 0 {
                return "<Invalid dialog choice command>"
            }
            for index in 1...choiceCount {
                let eventNameString = eventName("eventPointer\(index)")
                commandDescription += String(index) + ". " + eventNameString
                if index != choiceCount { commandDescription += "\n" }
            }
            return commandDescription
            
        case "eventScript.switch":
            let eventSwitch = command.intValue(forKey: "switch")
            let highByte = 0 //command.intValue(forKey: "highByte")
            let switchName = switchNameWithIndex(eventSwitch | highByte << 8)
            let clear = command.boolValue(forKey: "clear")
            commandDescription = switchName + " = " + (clear ? "Off" : "On")
            return commandDescription
            
        case "eventScript.variable":
            let value = command.intValue(forKey: "value")
            switch command.intValue(forKey: "opcode") {
            case 0xE8: return variableName() + " = " + String(value)
            case 0xE9: return variableName() + " += " + String(value)
            case 0xEA: return variableName() + " -= " + String(value)
            case 0xEB: return "Compare " + variableName() + " to " + String(value)
            default: return "<Invalid variable operation>"
            }
            
        case "eventScript.activePartyCharacters":
            return "Set switches for characters in the active party"
            
        case "eventScript.enabledCharacters":
            return "Set switches for characters with enabled objects"
            
        case "eventScript.initializedCharacters":
            return "Set switches for initialized characters"
            
        case "eventScript.availableCharacters":
            return "Set switches for available characters"
            
        case "eventScript.leadCharacter":
            return "Set switches for the lead character"
            
        case "eventScript.anyPartyCharacters":
            return "Set switches for characters in any party"
            
        case "eventScript.activePartyNumber":
            return "Set switches for the active party number"
            
        case "eventScript.jumpCharacterSwitch":
            
            commandDescription = "Jump to subroutine based on first matching switch:\n"
            let addressCount = (command.data.count - 2) / 3
            for index in 1...addressCount {
                let eventSwitch = command.intValue(forKey: "switch\(index)") + 0x01A0
                let name = switchNameWithIndex(eventSwitch)
                let eventNameString = eventName("eventPointer\(index)")
                commandDescription += name + ": " + eventNameString
                if index != addressCount {
                    commandDescription += "\n"
                }
            }
            return commandDescription
            
        case "eventScript.loadMap", "worldScript.loadMap", "vehicleScript.loadMap":
            let mapName = stringValueForKey("mapIndex", inStringTable: "mapProperties") ?? "<Invalid Map>"
            return "Load map: " + mapName
            
        case "eventScript.parentMap":
            let mapName = stringValueForKey("mapIndex", inStringTable: "mapProperties") ?? "<Invalid Map>"
            return "Set parent map: " + mapName
            
        case "eventScript.partyMap":
            let party = command.intValue(forKey: "party")
            let mapName = stringValueForKey("mapIndex", inStringTable: "mapProperties") ?? "<Invalid Map>"
            return "Move party " + String(party) + " to map: " + mapName
            
        case "eventScript.changeBackground":
            let layer = command.intValue(forKey: "layer")
            let x = command.intValue(forKey: "x")
            let y = command.intValue(forKey: "y")
            let height = command.intValue(forKey: "height")
            let width = command.intValue(forKey: "width")
            return String(format: "Change background layer %d: %d×%d tiles at (%d,%d)", layer + 1, width, height, x, y)
            
        case "eventScript.updateBackground":
            return "Update background"
            
        case "eventScript.endingCinematic":
            let cinematic = command.intValue(forKey: "cinematic")
            return "Show ending cinematic " + String(cinematic)
            
        case "eventScript.theEnd":
            return "Show \"The End\""
            
        case "eventScript.airshipCinematic":
            return "Show airship cinematic"
            
        case "eventScript.trainCinematic":
            return "Show Magitek train ride cinematic"
            
        case "eventScript.worldOfRuinCinematic":
            return "Show World of Ruin cinematic"
            
        case "eventScript.magitekCinematic":
            return "Show Magitek March cinematic"
            
        case "eventScript.titleScreen":
            return "Show title screen"
            
        case "eventScript.floatingIslandCinematic":
            return "Show floating island cinematic"
            
        case "eventScript.modifyPalette", "eventScript.modifyPaletteRange":
            let sprite = command.boolValue(forKey: "sprite")
//            let math = command.intValue(forKey: "math")
//            let color = command.intValue(forKey: "color")
//            let intensity = command.intValue(forKey: "intensity")
            return "Modify " + (sprite ? "sprite" : "background") + " palettes: " + invertedColorName()
            
        case "eventScript.disableColorMath":
            return "Disable color addition/subtraction"
            
        case "eventScript.flashScreen":
            return "Flash the screen: " + colorName()
            
        case "eventScript.colorMath":
            let subtract = command.boolValue(forKey: "subtract")
            return "Set fixed color " + (subtract ? "subtraction: " : "addition: ") + colorName()
            
        case "eventScript.changePalette":
            let vramPalette = command.intValue(forKey: "vramPalette")
            let spritePalette = command.intValue(forKey: "spritePalette")
            return "Change palette " + String(vramPalette) + " to " + String(spritePalette)
            
        case "eventScript.filterPalette":
            return "Filter colors: " + colorName()
            
        case "eventScript.shakeScreen":
            let frequency = command.intValue(forKey: "frequency")
            let amplitude = command.intValue(forKey: "amplitude")
            return "Shake the screen (frequency " + String(frequency) + ", amplitude " + String(amplitude) + ")"
            
        case "eventScript.fadeAuto":
            let fadeOut = command.boolValue(forKey: "fadeOut")
            return fadeOut ? "Fade out" : "Fade in"
            
        case "eventScript.fade":
            let fadeIn = command.boolValue(forKey: "fadeIn")
            let speed = command.intValue(forKey: "speed")
            return (fadeIn ? "Fade in" : "Fade out") + " at speed " + String(speed)
            
        case "eventScript.lockScreen":
            let opcode = command.intValue(forKey: "opcode")
            switch opcode {
            case 0x38: return "Lock the Screen"
            case 0x39: return "Unlock the Screen"
            default: return "<Invalid lock screen command>"
            }
            
        case "eventScript.scroll":
            let opcode = command.intValue(forKey: "opcode")
            var layerName: String
            switch opcode {
            case 0x70, 0x5D: layerName = "layer 1"
            case 0x71, 0x5E: layerName = "layer 2"
            case 0x72, 0x5F: layerName = "layer 3"
            default: layerName = "<Invalid layer>"
            }
            var hSpeed = Int8(truncatingIfNeeded: command.intValue(forKey: "hSpeed"))
            var vSpeed = Int8(truncatingIfNeeded: command.intValue(forKey: "vSpeed"))
            if opcode < 0x70 {
                // adjust negative values for opcodes $5D, $5E, and $5F
                if hSpeed < 0 { hSpeed += 1 }
                if vSpeed < 0 { vSpeed += 1 }
            }
            return String(format: "Scroll %@ at speed (%d,%d)", layerName, hSpeed, vSpeed)
            
        case "eventScript.stopFade":
            return "Stop fading in/out"
            
        case "eventScript.waitForFade":
            return "Wait for fade in/out"
            
        case "eventScript.userControl":
            let opcode = command.intValue(forKey: "opcode")
            switch opcode {
            case 0x3A: return "Enable user control"
            case 0x3B: return "Disable user control"
            default: return "<Invalid user control command>"
            }
            
        case "eventScript.waitAuto":
            switch command.intValue(forKey: "opcode") {
            case 0x91: return "Wait 0.25 seconds"
            case 0x92: return "Wait 0.5 seconds"
            case 0x93: return "Wait 0.75 seconds"
            case 0x94: return "Wait 1 second"
            case 0x95: return "Wait 2 seconds"
            default: return "<Invalid wait command>"
            }
            
        case "eventScript.wait":
            let seconds = command.boolValue(forKey: "seconds")
            let duration = command.intValue(forKey: "duration")
            if seconds {
                return String(format: "Wait %f seconds", Float(duration) / 4)
            } else {
                return String(format: "Wait %d seconds", duration)
            }
            
        case "eventScript.waitSPC":
            return "Wait for SPC"
            
        case "eventScript.song", "eventScript.songWithVolume":
            let songName = stringValueForKey("song", inStringTable: "songs") ?? "<Invalid Song>"
            return "Play song: " + songName
            
        case "eventScript.fadeInSong":
            let songName = stringValueForKey("song", inStringTable: "songs") ?? "<Invalid Song>"
            return "Fade in song: " + songName
            
        case "eventScript.fadeOutSong":
            return "Fade out song"
            
        case "eventScript.fadeInPreviousSong":
            return "Fade in previous song"
            
        case "eventScript.soundEffect", "eventScript.soundEffectWithPan":
            let soundEffectName = stringValueForKey("soundEffect", inStringTable: "gameSoundEffects") ?? "<Invalid Sound Effect>"
            return "Play sound effect: " + soundEffectName
            
        case "eventScript.spcCommand":
            return "SPC command"
            
        case "eventScript.continueSong":
            return "Continue to the next part of the song"
            
        case "eventScript.waitForSPC":
            return "Wait for SPC"
            
        case "eventScript.syncSPC":
            let sync = command.intValue(forKey: "sync")
            return "Sync with SPC position " + String(sync)
            
        case "eventScript.waitForSong":
            return "Wait for the song to end"
            
        case "eventScript.noEffect":
            return "No effect"
            
        case "eventScript.return":
            return "Return from subroutine"
            
        case "eventScript.endOfScript", "objectScript.endOfScript", "worldScript.endOfScript", "vehicleScript.endOfScript":
            return "End of script"
            
            // MARK: object script commands
            
        case "objectScript.action", "worldScript.action":
//            let action = command.intValue(forKey: "action")
            let flipped = command.boolValue(forKey: "flipHorizontal")
            var actionName = stringValueForKey("action", inStringTable: "mapSpriteFrames") ?? "<Invalid Action>"
//            var actionName = rom.stringTable.valueAtIndex(action, inPropertyWithKey: "mapSpriteFrames")?.description ?? ("Action " + String(action))
            if flipped {
                if actionName.range(of: "Left") != nil {
                    actionName = actionName.replacingOccurrences(of: "Left", with: "Right")
                } else {
                    actionName += ", flipped horizontally"
                }
            }
            return "Change action to " + actionName
            
        case "objectScript.move", "worldScript.move":
            let distance = command.intValue(forKey: "distance") + 1
            return String(format: "Move %@ %d tile(s)", directionName(), distance)
            
        case "objectScript.moveDiagonal", "worldScript.moveDiagonal":
            switch command.intValue(forKey: "direction") {
            case 0: return "Move up/right 1×1 tiles"
            case 1: return "Move down/right 1×1 tiles"
            case 2: return "Move down/left 1×1 tiles"
            case 3: return "Move up/left 1×1 tiles"
            case 4: return "Move up/right 2×1 tiles"
            case 5: return "Move up/right 1×2 tiles"
            case 6: return "Move down/right 1×2 tiles"
            case 7: return "Move down/right 2×1 tiles"
            case 8: return "Move down/left 2×1 tiles"
            case 9: return "Move down/left 1×2 tiles"
            case 10: return "Move up/left 1×2 tiles"
            case 11: return "Move up/left 2×1 tiles"
            default: return "<Invalid diagonal direction>"
            }
            
        case "objectScript.speed", "worldScript.speed":
            switch command.intValue(forKey: "speed") {
            case 0: return "Set speed to 25%"
            case 1: return "Set speed to 50%"
            case 2: return "Set speed to 100%"
            case 3: return "Set speed to 200%"
            case 4: return "Set speed to 400%"
            case 5: return "Set speed to 800%"
            default: return "<Invalid speed>"
            }
            
        case "objectScript.walkingAnimation":
            let disable = command.boolValue(forKey: "disable")
            return (disable ? "Disable" : "Enable") + " walking animation"
            
        case "objectScript.direction", "worldScript.direction":
            return "Face " + directionName()
            
        case "objectScript.layerPriority":
            let priority = command.intValue(forKey: "priority")
            switch priority {
            case 0: return "Set object's layer priority to 0 (Default)"
            case 1: return "Set object's layer priority to 1 (Top Sprite Only)"
            case 2: return "Set object's layer priority to 2 (Foreground)"
            case 3: return "Set object's layer priority to 3 (Background)"
            default: return "Invalid layer priority"
            }
            
        case "objectScript.vehicle":
            return "Set vehicle to " + vehicleName()
            
        case "objectScript.showHide":
            let hide = command.boolValue(forKey: "hide")
            return (hide ? "Hide" : "Show") + " object"
            
        case "objectScript.position":
            let x = command.intValue(forKey: "x")
            let y = command.intValue(forKey: "y")
            return String(format: "Move to (%d,%d)", x, y)
            
        case "objectScript.scrollToObject":
            return "Scroll the screen to this object"
            
        case "objectScript.jumpUp":
            let high = command.boolValue(forKey: "high")
            return "Jump " + (high ? "high" : "low")
            
        case "objectScript.wait":
            let duration = command.intValue(forKey: "duration")
            return "Wait " + String(duration * 4) + " frames"
            
        case "objectScript.switch", "worldScript.switch", "vehicleScript.switch":
            let clear = command.boolValue(forKey: "clear")
            return switchName() + " = " + (clear ? "Off" : "On")
            
        case "objectScript.jump":
            return "Jump to " + eventName()
            
        case "objectScript.branch":
            if command.boolValue(forKey: "random") {
                return "Branch to " + eventName("offset") + " (50% chance)"
            } else {
                return "Branch to " + eventName("offset")
            }
            
            // MARK: world script commands
            
        case "worldScript.airshipPosition", "vehicleScript.airshipPosition":
            let x = command.intValue(forKey: "x")
            let y = command.intValue(forKey: "y")
            return String(format: "Move the airship to (%d,%d)", x, y)
            
        case "worldScript.showHide":
            let hide = command.boolValue(forKey: "hide")
            return (hide ? "Hide" : "Show") + " the party"
            
        case "worldScript.jumpKeypress":
            return "If A button is pressed, jump to " + eventName()
            
        case "worldScript.jumpDirection":
            return "If not facing " + directionName() + ", jump to " + eventName()
            
        case "worldScript.screenFade", "vehicleScript.screenFade":
            let fadeOut = command.boolValue(forKey: "fadeOut")
            return (fadeOut ? "Fade out" : "Fade in")
            
        case "worldScript.wait", "vehicleScript.wait":
            let duration = command.intValue(forKey: "duration")
            return "Wait " + String(duration) + " frames"
            
        case "worldScript.miniMap", "vehicleScript.miniMap":
            let show = command.boolValue(forKey: "show")
            return (show ? "Show" : "Hide") + " the mini-map"
            
        case "worldScript.ship":
            return "Change party graphic to ship"
            
        case "worldScript.figaroCastle":
            let submerge = command.boolValue(forKey: "submerge")
            return "Show Figaro Castle " + (submerge ? "submerging" : "emerging")
            
            // MARK: vehicle script commands
            
        case "vehicleScript.unknownC0":
            return "Unknown vehicle script command $C0"
            
        case "vehicleScript.esperTerra":
            return "Change vehicle graphic to esper Terra"
            
        case "vehicleScript.falcon":
            return "Change vehicle graphic to Falcon"
            
        case "vehicleScript.bird":
            return "Change vehicle graphic to bird"
            
        case "vehicleScript.airshipSmoking":
            return "Change vehicle graphic to smoking airship"
            
        case "vehicleScript.airshipCrash":
            return "Show airship crashing cinematic"
            
        case "vehicleScript.lightOfJudgement1":
            return "Show Light of Judgement cinematic 1"
            
        case "vehicleScript.lightOfJudgement2":
            return "Show Light of Judgement cinematic 2"
            
        case "vehicleScript.lightOfJudgement3":
            return "Show Light of Judgement cinematic 3"
            
        case "vehicleScript.explosions":
            return "Show explosions cinematic"
            
        case "vehicleScript.falconWater":
            return "Show Falcon rising from the ocean cinematic"
            
        case "vehicleScript.vectorApproach":
            return "Show Vector approach cinematic"
            
        case "vehicleScript.showVehicle":
            let hide = command.boolValue(forKey: "hide")
            return (hide ? "Hide" : "Show") + " the vehicle"
            
        case "vehicleScript.showArrows":
            return "Show direction arrows"
            
        case "vehicleScript.lockArrows":
            return "Lock direction arrows"
            
        case "vehicleScript.hideArrows":
            return "Hide direction arrows"
            
        case "vehicleScript.moveForward":
            let speed = command.intValue(forKey: "speed")
            return "Move forward with speed " + String(speed)
            
        case "vehicleScript.altitude":
            let altitude = command.intValue(forKey: "altitude")
            return "Set altitude to " + String(altitude)
            
        case "vehicleScript.rotateVehicle":
            let direction = command.intValue(forKey: "direction")
            return "Rotate vehicle in direction " + String(direction)
            
        case "vehicleScript.rotateCamera":
            let direction = command.intValue(forKey: "direction")
            return "Rotate camera in direction " + String(direction)
            
        case "vehicleScript.move":
            commandDescription = "Move"
            if command.boolValue(forKey: "down") { commandDescription += " Down" }
            if command.boolValue(forKey: "up") { commandDescription += " Up" }
            if command.boolValue(forKey: "left") { commandDescription += " Left" }
            if command.boolValue(forKey: "right") { commandDescription += " Right" }
            if command.boolValue(forKey: "forward") { commandDescription += " Forward" }
            if command.boolValue(forKey: "back") { commandDescription += " Back" }
            if command.boolValue(forKey: "sharpTurn") { commandDescription += " Sharp Turn" }
            let duration = command.intValue(forKey: "duration")
            commandDescription += " for " + String(duration) + " frames"
            return commandDescription
            
            // MARK: ff6a event commands

        case "eventScript.ff6a_item":
            let item = command.intValue(forKey: "item")
            let itemName = stringValueForKey("item", inStringTable: "itemNames") ?? "Item " + String(item)
            return "Add " + itemName + " to the inventory (GBA)"

        case "eventScript.ff6a_battle":
            let battle = command.intValue(forKey: "battle")
            let battleName = stringValueForKey("battle", inStringTable: "battleProperties") ?? "Battle " + String(battle)
            return "Battle: " + battleName + " (GBA)"

        case "eventScript.ff6a_switch":
            let switchIndex = command.intValue(forKey: "switch")
            let switchName = switchNameWithIndex(switchIndex + 0x4000)
            switch command.intValue(forKey: "opcode") {
            case 0xE5: return switchName + " = On"
            case 0xE6: return switchName + " = Off"
            default: return "<Invalid Switch Command>"
            }
            
        case "eventScript.ff6a_gameClear":
            return "Game Complete"

        case "eventScript.ff6a_save":
            return "Save Game"

        default:
            return super.descriptionForCommand(command)
        }
    }
}
