//
//  FF6AttackLayer.swift
//  FF6Tools
//
//  Created by Evan Owen on 12/14/16.
//
//

import Foundation

enum FF6AttackLayerType {
    case sprite
    case layer1
    case layer3
    case extra
    case rightHand
    case leftHand
    case hit
    case none
    
    var scriptKey: String {
        switch self {
        case .sprite:       return "spriteScript"
        case .layer1:       return "layer1Script"
        case .layer3:       return "layer3Script"
        case .extra:        return "extraScript"
        case .rightHand:    return "rightHandScript"
        case .leftHand:     return "leftHandScript"
        case .hit:          return "hitScript"
        case .none:         return "invalid"
        }
    }

    var noGraphicsKey: String {
        switch self {
        case .sprite:       return "spriteNoGraphics"
        case .layer1:       return "layer1NoGraphics"
        case .layer3:       return "layer3NoGraphics"
        case .extra:        return "extraNoGraphics"
        default:            return "invalid"
        }
    }

    var paletteKey: String {
        switch self {
        case .sprite:       return "spritePalette"
        case .layer1:       return "layer1Palette"
        case .layer3:       return "layer3Palette"
        case .extra:        return "extraPalette"
        case .rightHand:    return "weaponPalette"
        case .leftHand:     return "weaponPalette"
        case .hit:          return "hitPalette"
        case .none:         return "invalid"
        }
    }
}

class FF6AttackLayer: NSObject, ROMSelectable {
    
    // MARK: ROMSelectable methods
    
    var identifier: String? { return "attackLayer" }
    var properties: NSDictionary? {
        guard let wrappedProperties = graphicsProperties?.properties else { return nil }
        let dummyProperties = NSMutableDictionary(dictionary: wrappedProperties)
        dummyProperties["tilesetIndex"] = 0
        dummyProperties["is2bppInt"] = 0
        dummyProperties["scriptPointer"] = 0
        return dummyProperties
    }
    var nib: String? { return nil }
    
    var graphicsProperties: ROMData?
    var attackScriptPointer: ROMData?
    
    var type: FF6AttackLayerType = .none
    var graphicsIndex = 0xFFFF
    var graphics: Data!
    var frames: Data!
    var tileset = [UInt16](repeating: 0, count: 256)
    var tilesetSize = 0
    var tiles: Data!
    var palette: Data!
    var is2bpp: Bool {
        get { return graphicsProperties?.boolValue(forKey: "is2bpp") ?? false }
        set { graphicsProperties?.setValue(newValue, forKey: "is2bpp") }
    }
    @objc var is2bppInt: Int {
        get { return graphicsProperties?.intValue(forKey: "is2bpp") ?? 0 }
        set { graphicsProperties?.setValue(newValue, forKey: "is2bpp") }
    }
    @objc var frameCount: Int {
        get { return graphicsProperties?.intValue(forKey: "frameCount") ?? 0 }
        set { graphicsProperties?.setValue(newValue, forKey: "frameCount") }
    }
    @objc var frameHeight: Int {
        get {
            guard let height = graphicsProperties?.intValue(forKey: "frameHeight") else { return 1 }
            if height < 1 {
                graphicsProperties?.setValue(1, forKey: "frameHeight")
                return 1
            }
            return height
        }
        set { graphicsProperties?.setValue(newValue, forKey: "frameHeight") }
    }
    @objc var frameWidth: Int {
        get {
            guard let width = graphicsProperties?.intValue(forKey: "frameWidth") else { return 1 }
            if width < 1 {
                graphicsProperties?.setValue(1, forKey: "frameWidth")
                return 1
            }
            return width
        }
        set { graphicsProperties?.setValue(newValue, forKey: "frameWidth") }
    }
    @objc var frameOffset: Int {
        get { return graphicsProperties?.intValue(forKey: "frameOffset") ?? 0 }
        set { graphicsProperties?.setValue(newValue, forKey: "frameOffset") }
    }
    @objc var scriptPointer: ROMReference? {
        get { return attackScriptPointer?.value(forKey: "scriptPointer") as? ROMReference }
        set { attackScriptPointer?.setValue(newValue, forKey: "scriptPointer") }
    }
    @objc var tilesetIndex: Int {
        get {
            if graphicsProperties == nil { return 0 }
            let tileset = graphicsProperties!.intValue(forKey: "tileset")
            let tilesetMSB = graphicsProperties!.boolValue(forKey: "tilesetMSB")
            return tileset + (tilesetMSB ? 0x0100 : 0)
        }
        set {
            graphicsProperties?.setValue(newValue & 0xFF, forKey: "tileset")
            graphicsProperties?.setValue(newValue > 0xFF ? 1 : 0, forKey: "tilesetMSB")
        }
    }
    
    func intValue(forKey key: String) -> Int {
        return graphicsProperties?.intValue(forKey: key) ?? 0
    }
    
    func boolValue(forKey key: String) -> Bool {
        return graphicsProperties?.boolValue(forKey: key) ?? false
    }
    
    override func value(forUndefinedKey key: String) -> Any? {
        return graphicsProperties?.value(forKey: key)
    }
    
    override func setValue(_ value: Any?, forUndefinedKey key: String) {
        graphicsProperties?.setValue(value, forKey: key)
    }
    
    var graphicsKey: String {
        return is2bpp ? "attackGraphics2bpp" : "attackGraphics3bpp"
    }
    
    var tilesetKey: String {
        return is2bpp ? "attackTilesets2bpp" : "attackTilesets3bpp"
    }

    func decodeLayout() {
        
        // get the frame data
        var tileBytes = [UInt16](repeating: 0x0100, count: frameWidth * frameHeight * 4)
        let frameBytes = frames.withUnsafeBytes{ [UInt16](UnsafeBufferPointer(start: $0, count: frames.count / 2)) }
        for tileProperties in frameBytes {
            if tileProperties == 0xFFFF { break }
            let tile = Int(tileProperties & 0x3800) >> 6 | Int(tileProperties & 0x0700) >> 7
            let flip = tileProperties & 0xC000
            let x = Int(tileProperties & 0xF0) >> 4
            let y = Int(tileProperties & 0x0F)
            if x >= frameWidth { break }
            if y >= frameHeight { break }
            var tileOffset = x * 2 + y * frameWidth * 4
            
            switch flip {
            case 0x0000:
                // no flip
                tileBytes[tileOffset + 0] = tileset[tile] ^ flip
                tileBytes[tileOffset + 1] = tileset[tile + 1] ^ flip
                tileOffset += (frameWidth * 2)
                tileBytes[tileOffset + 0] = tileset[tile + 16] ^ flip
                tileBytes[tileOffset + 1] = tileset[tile + 17] ^ flip
            case 0x4000:
                // horizontal flip
                tileBytes[tileOffset + 0] = tileset[tile + 1] ^ flip
                tileBytes[tileOffset + 1] = tileset[tile] ^ flip
                tileOffset += (frameWidth * 2)
                tileBytes[tileOffset + 0] = tileset[tile + 17] ^ flip
                tileBytes[tileOffset + 1] = tileset[tile + 16] ^ flip
            case 0x8000:
                // vertical flip
                tileBytes[tileOffset + 0] = tileset[tile + 16] ^ flip
                tileBytes[tileOffset + 1] = tileset[tile + 17] ^ flip
                tileOffset += (frameWidth * 2)
                tileBytes[tileOffset + 0] = tileset[tile] ^ flip
                tileBytes[tileOffset + 1] = tileset[tile + 1] ^ flip
            case 0xC000:
                // horizontal and vertical flip
                tileBytes[tileOffset + 0] = tileset[tile + 17] ^ flip
                tileBytes[tileOffset + 1] = tileset[tile + 16] ^ flip
                tileOffset += (frameWidth * 2)
                tileBytes[tileOffset + 0] = tileset[tile + 1] ^ flip
                tileBytes[tileOffset + 1] = tileset[tile] ^ flip
            default:
                break
            }
        }

        tiles = Data(buffer: UnsafeBufferPointer(start: tileBytes, count: tileBytes.count * 2))
    }

    var image: CGImage? {
        
        if graphicsIndex == 0x7FFF { return nil }
        if frameWidth == 0 { return nil }
        if frameHeight == 0 { return nil }
        
        // create a PPU
        var ppu = PPU()
        memset(&ppu, 0, MemoryLayout<PPU>.size)
        
        ppu.pal = (palette as NSData).bytes
        ppu.layers.0.format = is2bpp ? SNES_2BPP_TILE : SNES_4BPP_TILE
        ppu.layers.0.height = Int32(frameHeight) * 2
        ppu.layers.0.width = Int32(frameWidth) * 2
        ppu.layers.0.z.0 = Z_LEVEL_BG1_LO
        ppu.layers.0.z.1 = Z_LEVEL_BG1_HI
        ppu.layers.0.gfx = (graphics as NSData).bytes
        ppu.layers.0.tiles = (tiles as NSData).bytes
        ppu.layers.0.main = true
        
        // render the screen
        let dataLength = frameHeight * frameWidth * 256 * MemoryLayout<UInt32>.size
        let data = UnsafeMutablePointer<UInt8>.allocate(capacity: dataLength)
        memset(data, 0, dataLength)
        gfx_render_ppu(data, ppu, Int32(dataLength), Int32(frameWidth) * 16)
        
        // create an image
        let provider = CGDataProvider(dataInfo: nil, data: data, size: dataLength, releaseData: releaseImageData)!
        let bitsPerComponent = 8
        let bitsPerPixel = 32
        let bytesPerRow = frameWidth * 16 * MemoryLayout<UInt32>.size
        let colorSpaceRef = CGColorSpaceCreateDeviceRGB()
        let bitmapInfo = CGBitmapInfo(rawValue: CGBitmapInfo().rawValue | CGImageAlphaInfo.premultipliedLast.rawValue)
        let renderingIntent = CGColorRenderingIntent.defaultIntent
        let image = CGImage(width: frameWidth * 16, height: frameHeight * 16, bitsPerComponent: bitsPerComponent, bitsPerPixel: bitsPerPixel, bytesPerRow: bytesPerRow, space: colorSpaceRef, bitmapInfo: bitmapInfo, provider: provider, decode: nil, shouldInterpolate: true, intent: renderingIntent)
        
        return image
    }

}
