//
//  FF6BattleMonster.swift
//  FF6Tools
//
//  Created by Evan Owen on 3/5/16.
//  Copyright © 2016 Evan Owen. All rights reserved.
//

import Cocoa

class FF6BattleMonster: NSObject, ROMSelectable {
    
    // MARK: ROMSelectable properties
    
    var identifier: String? { return "battleMonster" }
    var properties: NSDictionary? { return ["index" : 0, "present" : 0, "bg1" : 0, "slot" : 0, "verticalOffset" : 0] }
    var nib: String? { return "BattleProperties" }

    // MARK: member variables
    
    let rom: FF6ROM
    let battleProperties: ROMData
    @objc let slot: Int
    
    @objc var index: Int {
        get {
            // get monster index
            var index = battleProperties.intValue(forKey: "monster\(slot)")
            let msb = battleProperties.boolValue(forKey: "monster\(slot)MSB")
            if msb { index += 0x0100 }
            return index
        }
        set {
            willChangeValue(forKey: "isNull")
            willChangeValue(forKey: "isValid")
            stopObservingBattleProperties()
            battleProperties.setValue(newValue & 0xFF, forKey: "monster\(slot)")
            battleProperties.setValue(newValue >= 0x0100 ? 1 : 0, forKey: "monster\(slot)MSB")
            startObservingBattleProperties()
            didChangeValue(forKey: "isNull")
            didChangeValue(forKey: "isValid")
            
            // update the monster's name in the monsterSlots string table
            if let monsterSlotsStringTable = rom.value(forKey: "stringTable-monsterSlots") as? ROMStringTable,
                let monsterSlotString = monsterSlotsStringTable[slot-1] as? ROMString {
                monsterSlotString.rawString = String(slot) + ": " + name
            }
        }
    }
    
    var isNull: Bool { return index == 0x01FF }
    var isValid: Bool { return !isNull }
    var x: Int {
        get { return battleProperties.intValue(forKey: "monster\(slot)X") }
        set { battleProperties.setValue(min(15, max(newValue, 0)), forKey: "monster\(slot)X") }
    }
    var y: Int {
        get { return battleProperties.intValue(forKey: "monster\(slot)Y") }
        set { battleProperties.setValue(min(15, max(newValue, 0)), forKey: "monster\(slot)Y") }
    }
    @objc var present: Bool {
        get { return battleProperties.boolValue(forKey: "monster\(slot)Present") }
        set {
            stopObservingBattleProperties()
            battleProperties.setValue(newValue ? 1 : 0, forKey: "monster\(slot)Present")
            startObservingBattleProperties()
        }
    }
    @objc var bg1: Bool {
        get { return battleProperties.boolValue(forKey: "monster\(slot)BG1") }
        set {
            stopObservingBattleProperties()
            battleProperties.setValue(newValue ? 1 : 0, forKey: "monster\(slot)BG1")
            startObservingBattleProperties()
        }
    }
    var verticalOffset: Int {
        get {
            if isNull { return 0 }
            guard let verticalOffset = rom.value(at: index, inPropertyWithKey: "monsterVerticalOffsets") as? ROMData else { return 0 }
            return verticalOffset.intValue(forKey: "offset")
        }
        set {
            if isNull { return }
            guard let verticalOffset = rom.value(at: index, inPropertyWithKey: "monsterVerticalOffsets") as? ROMData else { return }
            verticalOffset.setValue(newValue, forKey: "offset")
//            (rom.value(at: index, inPropertyWithKey: "monsterVerticalOffsets") as AnyObject).setValue(newValue, forKey: "offset")
        }
    }
    var name: String {
        get {
            if isNull { return "Empty" }
            return rom.string(atIndex: index, inStringTable: "monsterNames") ?? "Monster " + String(index)
        }
    }
    
    var size: (width: Int, height: Int) {
        
        if isNull { return (0, 0) }
        
        return rom.size(ofMonster: index)
    }
    
    init(rom: FF6ROM, battleProperties: ROMData, slot: Int) {
        self.rom = rom
        self.battleProperties = battleProperties
        self.slot = slot

        super.init()
        
        // start observing battleProperties
        startObservingBattleProperties()
    }
    
    deinit {
        // stop observing battleProperties
        stopObservingBattleProperties()
    }
    
    func startObservingBattleProperties() {
        battleProperties.addObserver(self, forKeyPath: "monster\(slot)", options: NSKeyValueObservingOptions.prior, context: &FF6BattleMonster.context)
        battleProperties.addObserver(self, forKeyPath: "monster\(slot)MSB", options: NSKeyValueObservingOptions.prior, context: &FF6BattleMonster.context)
        battleProperties.addObserver(self, forKeyPath: "monster\(slot)Present", options: NSKeyValueObservingOptions.prior, context: &FF6BattleMonster.context)
        battleProperties.addObserver(self, forKeyPath: "monster\(slot)BG1", options: NSKeyValueObservingOptions.prior, context: &FF6BattleMonster.context)
    }

    func stopObservingBattleProperties() {
        battleProperties.removeObserver(self, forKeyPath: "monster\(slot)")
        battleProperties.removeObserver(self, forKeyPath: "monster\(slot)MSB")
        battleProperties.removeObserver(self, forKeyPath: "monster\(slot)Present")
        battleProperties.removeObserver(self, forKeyPath: "monster\(slot)BG1")
    }

    static var context = 0
    
    override func observeValue(forKeyPath keyPath: String?, of object: Any?, change: [NSKeyValueChangeKey : Any]?, context: UnsafeMutableRawPointer?) {
        if context != &FF6BattleMonster.context {
            super.observeValue(forKeyPath: keyPath, of: object, change: change, context: context)
            return
        }
        
        let changeFunc: (String)->() = (change?[NSKeyValueChangeKey.notificationIsPriorKey] != nil) ? willChangeValue : didChangeValue
        if (keyPath == "monster\(slot)") || keyPath == "monster\(slot)MSB" {
            changeFunc("index")
        } else if keyPath == "monster\(slot)Present" {
            changeFunc("present")
        } else if keyPath == "monster\(slot)BG1" {
            changeFunc("bg1")
        }
    }
}
