//
//  FF6BattleView.swift
//  FF6Tools
//
//  Created by Evan Owen on 2/6/16.
//  Copyright © 2016 Evan Owen. All rights reserved.
//

// TODO: show monster names somewhere better
// TODO: add and remove monsters better
// TODO: drag monsters to swap slots
// TODO: don't deselect in battleView and monsterVRAMView when something other than a monster or character is selected
// TODO: copy battles ???
// TODO: copy monsters ???
// TODO: dump to txt/xml file ???
// TODO: test battle with emulator

import Cocoa

class FF6BattleView: FF6View {

    enum BattleType: Int {
        case normal = 0
        case back = 1
        case pincer = 2
        case side = 3
    }

    // MARK: properties

    @objc var battle = 0 {
        willSet {
            stopObservingBattleProperties()
            windowController.deselectObject(battleExperienceGP)
        }
        didSet {
            battleProperties = rom.mutableArrayValue(forKey: "battleProperties").object(at: battle) as? ROMData
            auxBattleProperties = rom.mutableArrayValue(forKey: "auxBattleProperties").object(at: battle) as? ROMData
            battleExperienceGP = FF6BattleExperienceGP(rom: ff6ROM, battleProperties: battleProperties)
            startObservingBattleProperties()
            reloadMonsters()
            reloadCharacters()
            reloadVerticalOffsets()
            battleType = .normal
            validateBattleType()
            needsDisplay = true
        }
    }
    @objc var battleIndexKeyPath: String?
    @objc var backgroundKeyPath: String?
    
    var battleType: BattleType = .normal {
        didSet {
            validateBattleType()
        }
    }
    @objc var background = 0 {
        didSet {
            // validate the background index
            let maxValue = rom.mutableArrayValue(forKey: "battleBackgroundProperties").count - 1
            if background > maxValue { background = 0 }
            if background < 0 { background = maxValue }
            if background != oldValue { needsDisplay = true } // redraw only if the background changed
        }
    }
    var battleExperienceGP: FF6BattleExperienceGP!
    var battleProperties: ROMData!
    var auxBattleProperties: ROMData!
    var zoom: CGFloat { return max(bounds.width / 240, bounds.height / 147) }
    var selectedObject: AnyObject? { didSet { needsDisplay = true } }
    
    var monsters = [FF6BattleMonster]()
    var monsterVerticalOffsets = [ROMData]()
    var monstersSortedBySlot: [FF6BattleMonster] {
        return monsters.sorted {
            $0.slot < $1.slot
        }
    }
    var monstersSortedByPriority: [FF6BattleMonster] {
        return monsters.sorted {
            let y1 = ($0.y + $0.size.height) * 8 + $0.verticalOffset
            let y2 = ($1.y + $1.size.height) * 8 + $1.verticalOffset
            return y1 > y2
        }
    }
    var characters = [FF6BattleCharacter]()
    var charactersSortedByPriority: [FF6BattleCharacter] {
        return characters.sorted {
            rectForCharacter($0).origin.y < rectForCharacter($1).origin.y
        }
    }

    deinit {
        // stop observing battle properties
        stopObservingBattleProperties()
        
        // stop observing all monsters
        for monster in monsters {
            stopObservingMonster(monster)
        }
        
        // clear the monster array
        monsters.removeAll()

        // stop observing all characters
        for character in characters {
            stopObservingCharacter(character)
        }
        
        // clear the monster array
        characters.removeAll()
    }
    
    func validateBattleType() {
        let disableNormal = auxBattleProperties.boolValue(forKey: "disableNormal")
        let disableBack = auxBattleProperties.boolValue(forKey: "disableBack")
        let disablePincer = auxBattleProperties.boolValue(forKey: "disablePincer")
        let disableSide = auxBattleProperties.boolValue(forKey: "disableSide")
        
        if disableNormal && disableBack && disablePincer && disableSide {
            return
        }
        
        if battleType == .normal && disableNormal {
            battleType = .back
        }
        if battleType == .back && disableBack {
            battleType = .pincer
        }
        if battleType == .pincer && disablePincer {
            battleType = .side
        }
        if battleType == .side && disableSide {
            battleType = .normal
        }
    }
    
    func startObservingBattleProperties() {
        auxBattleProperties?.addObserver(self, forKeyPath: "characterAI", options: NSKeyValueObservingOptions.old, context: &FF6BattleView.context)
        auxBattleProperties?.addObserver(self, forKeyPath: "disableNormal", options: NSKeyValueObservingOptions.old, context: &FF6BattleView.context)
        auxBattleProperties?.addObserver(self, forKeyPath: "disableBack", options: NSKeyValueObservingOptions.old, context: &FF6BattleView.context)
        auxBattleProperties?.addObserver(self, forKeyPath: "disablePincer", options: NSKeyValueObservingOptions.old, context: &FF6BattleView.context)
        auxBattleProperties?.addObserver(self, forKeyPath: "disableSide", options: NSKeyValueObservingOptions.old, context: &FF6BattleView.context)
    }
    
    func stopObservingBattleProperties() {
        auxBattleProperties?.removeObserver(self, forKeyPath: "characterAI")
        auxBattleProperties?.removeObserver(self, forKeyPath: "disableNormal", context: &FF6BattleView.context)
        auxBattleProperties?.removeObserver(self, forKeyPath: "disableBack", context: &FF6BattleView.context)
        auxBattleProperties?.removeObserver(self, forKeyPath: "disablePincer", context: &FF6BattleView.context)
        auxBattleProperties?.removeObserver(self, forKeyPath: "disableSide", context: &FF6BattleView.context)
    }
    
    override func awakeFromNib() {
        
        super.awakeFromNib()
        
        if battleIndexKeyPath != nil {
            bind(NSBindingName(rawValue: "battle"), to: windowController, withKeyPath: battleIndexKeyPath!, options: nil)
        }
        if backgroundKeyPath != nil {
            bind(NSBindingName(rawValue: "background"), to: windowController, withKeyPath: backgroundKeyPath!, options: nil)
        }
    }
    
    override func viewDidMoveToWindow() {
        let ff6windowController = windowController as? FF6WindowController
        if let defaultBackground = ff6windowController?.defaultBattleBackground, defaultBackground != 0x3F, backgroundKeyPath == nil {
            self.background = defaultBackground
        } else {
            self.background = 0
        }
    }
    
//    override func resizeWithOldSuperviewSize(oldSize: NSSize) {
//        
//        super.resizeWithOldSuperviewSize(oldSize)
//        
//        let height = frame.width / 240.0 * (rom.isGBA ? 120 : 147)
//        if height > frame.height {
//            let width = frame.height * 240.0 / (rom.isGBA ? 120 : 147)
//            super.setFrameSize(NSMakeSize(width, frame.height))
//        } else {
//            super.setFrameSize(NSMakeSize(frame.width, height))
//        }
//    }
    
//    override func setFrameSize(newSize: NSSize) {
//        let height = newSize.width / 240.0 * (rom.isGBA ? 120 : 147)
//        if height > newSize.height {
//            let width = newSize.height * 240.0 / (rom.isGBA ? 120 : 147)
//            super.setFrameSize(NSMakeSize(width, newSize.height))
//        } else {
//            super.setFrameSize(NSMakeSize(newSize.width, height))
//        }
//    }

    // MARK: NSResponder methods
    
    override func mouseDown(with theEvent: NSEvent) {
        if backgroundKeyPath != nil { return }
        background += 1
    }
    
    // MARK: monster methods
    
    func reloadMonsters() {
        
        // stop observing all monsters
        for monster in monsters {
            stopObservingMonster(monster)
        }
        
        // clear the monster array
        monsters.removeAll()
        
        for slot in 1...6 {
            
            // create a monster object
            let monster = FF6BattleMonster(rom: ff6ROM, battleProperties: self.battleProperties, slot: slot)
            
            // add the monster to the array
            monsters.append(monster)
            startObservingMonster(monster)
        }
    }
    
    func monsterInSlot(_ slot: Int) -> FF6BattleMonster? {
        for monster in monsters {
            if monster.slot == slot {
                return monster
            }
        }
        return nil
    }
    
    func startObservingMonster(_ monster: FF6BattleMonster) {
        monster.battleProperties.addObserver(self, forKeyPath: "monster\(monster.slot)", options: NSKeyValueObservingOptions.old, context: &FF6BattleView.context)
        monster.battleProperties.addObserver(self, forKeyPath: "monster\(monster.slot)MSB", options: NSKeyValueObservingOptions.old, context: &FF6BattleView.context)
        monster.battleProperties.addObserver(self, forKeyPath: "monster\(monster.slot)X", options: NSKeyValueObservingOptions.old, context: &FF6BattleView.context)
        monster.battleProperties.addObserver(self, forKeyPath: "monster\(monster.slot)Y", options: NSKeyValueObservingOptions.old, context: &FF6BattleView.context)
        monster.battleProperties.addObserver(self, forKeyPath: "monster\(monster.slot)Present", options: NSKeyValueObservingOptions.old, context: &FF6BattleView.context)
    }

    func stopObservingMonster(_ monster: FF6BattleMonster) {
        monster.battleProperties.removeObserver(self, forKeyPath: "monster\(monster.slot)")
        monster.battleProperties.removeObserver(self, forKeyPath: "monster\(monster.slot)MSB")
        monster.battleProperties.removeObserver(self, forKeyPath: "monster\(monster.slot)X")
        monster.battleProperties.removeObserver(self, forKeyPath: "monster\(monster.slot)Y")
        monster.battleProperties.removeObserver(self, forKeyPath: "monster\(monster.slot)Present")
    }
    
    func reloadVerticalOffsets() {
        // stop observing vertical offsets
        for verticalOffset in monsterVerticalOffsets {
            verticalOffset.removeObserver(self, forKeyPath: "offset")
        }
        monsterVerticalOffsets.removeAll()

        // observe vertical offsets for all monsters
        for monster in monsters {
            guard let verticalOffset = rom.value(at: monster.index, inPropertyWithKey: "monsterVerticalOffsets") as? ROMData else { continue }
            
            // skip if the offset is already being observed
            if monsterVerticalOffsets.contains(verticalOffset) { continue }
            
            verticalOffset.addObserver(self, forKeyPath: "offset", options: NSKeyValueObservingOptions.old, context: &FF6BattleView.context)
            monsterVerticalOffsets.append(verticalOffset)
        }
    }
    
    func rectForMonster(_ monster: FF6BattleMonster) -> NSRect {
        if monster.isNull { return NSZeroRect }
        
        let monsterSize = monster.size
        var x = 0
        
        // see C1/1481 for x position calculations
        switch battleType {
        case .normal:
            x = (monster.x) * 8
            
        case .back:
            x = 256 - (monster.x + monsterSize.width) * 8
            
        case .pincer:
            if (monster.x + monsterSize.width) * 8 < 0x68 {
                x = (monster.x) * 8 // left side
            } else {
                x = 256 - ((monster.x + monsterSize.width) * 8 - 0x40) // right side
            }
            
        case .side:
            x = monster.x * 8 + 0x30
        }
        
        if background == 44 { x = (monster.x) * 8 }
        
        let y = (monster.y + monsterSize.height) * 8
        let height = monsterSize.height * 8
        let monsterRect = NSMakeRect(CGFloat(x - 8) * zoom, bounds.height - CGFloat(y - 4) * zoom, CGFloat(monsterSize.width * 8) * zoom, CGFloat(height) * zoom)
        return monsterRect
    }
    
    func monsterAtPoint(_ point: NSPoint) -> FF6BattleMonster? {
        for monster in monstersSortedByPriority {
            if rectForMonster(monster).contains(point) {
                return monster
            }
        }
        return nil
    }
    
    func shouldFlipMonster(_ monster: FF6BattleMonster) -> Bool {
        
        if background == 44 { return false }
        
        // see C1/1481 for horizontal flip
        switch battleType {
        case .normal: return false
        case .back:   return true
        case .pincer:
            if (monster.x + monster.size.width) * 8 < 0x68 {
                return false // left side
            } else {
                return true // right side
            }
        case .side:
            if ((monster.x + monster.size.width >> 1) * 8 + 0x30) >= 0x80 {
                return false // facing right
            } else {
                return true // facing left
            }
        }
    }
    
    // MARK: character methods
    
    func reloadCharacters() {
        
        // stop observing all characters
        for character in characters {
            stopObservingCharacter(character)
        }
        
        // clear the character array
        characters.removeAll()
        
        for slot in 1...4 {
            
            // start with a blank character
            let character = FF6BattleCharacter(rom: ff6ROM, auxBattleProperties: auxBattleProperties, slot: slot)
            
            // add the character to the array
            characters.append(character)
            startObservingCharacter(character)
        }
    }
    
    func characterInSlot(_ slot: Int) -> FF6BattleCharacter? {
        for character in characters {
            if character.slot == slot {
                return character
            }
        }
        return nil
    }

    func startObservingCharacter(_ character: FF6BattleCharacter) {
        character.characterSlotData.addObserver(self, forKeyPath: "character", options: NSKeyValueObservingOptions.old, context: &FF6BattleView.context)
        character.characterSlotData.addObserver(self, forKeyPath: "graphics", options: NSKeyValueObservingOptions.old, context: &FF6BattleView.context)
        character.characterSlotData.addObserver(self, forKeyPath: "hidden", options: NSKeyValueObservingOptions.old, context: &FF6BattleView.context)
        character.characterSlotData.addObserver(self, forKeyPath: "enemy", options: NSKeyValueObservingOptions.old, context: &FF6BattleView.context)
        character.characterSlotData.addObserver(self, forKeyPath: "x", options: NSKeyValueObservingOptions.old, context: &FF6BattleView.context)
        character.characterSlotData.addObserver(self, forKeyPath: "y", options: NSKeyValueObservingOptions.old, context: &FF6BattleView.context)
    }
    
    func stopObservingCharacter(_ character: FF6BattleCharacter) {
        character.characterSlotData.removeObserver(self, forKeyPath: "character")
        character.characterSlotData.removeObserver(self, forKeyPath: "graphics")
        character.characterSlotData.removeObserver(self, forKeyPath: "hidden")
        character.characterSlotData.removeObserver(self, forKeyPath: "enemy")
        character.characterSlotData.removeObserver(self, forKeyPath: "x")
        character.characterSlotData.removeObserver(self, forKeyPath: "y")
    }

    func rectForCharacter(_ character: FF6BattleCharacter) -> NSRect {
        
        var rect = NSRect(x: 0, y: 0, width: 16 * zoom, height: 32 * zoom)
        
        if character.hasAI {
            // get position from AI data
            rect.origin.x = min(CGFloat(character.x * 2 - 10), 232) * zoom
            rect.origin.y = bounds.height - min(CGFloat(20 + character.y * 2), 160) * zoom
            return rect
        }
        
        // character with no AI (see C1/3A26 and C1/3ABE)
        
        var xOffsets: [CGFloat] = [187, 195, 203, 210]
        var yOffsets: [CGFloat] = [88, 106, 124, 142]
        
        if battleType == .back {
            xOffsets = [37, 29, 21, 14]
        } else if battleType == .pincer {
            xOffsets = [112, 112, 112, 112]
        } else if battleType == .side {
            xOffsets = [191, 210, 33, 14]
            yOffsets = [98, 142, 98, 142]
        }
        if background == 44 {
            // magitek train ride adjustment
            yOffsets = [100, 114, 128, 142]
            xOffsets = [187, 195, 203, 210]
        }
        
        rect.origin.x = xOffsets[character.slot - 1] * zoom
        rect.origin.y = bounds.height - yOffsets[character.slot - 1] * zoom
        
        return rect
    }
    
    func characterAtPoint(_ point: NSPoint) -> FF6BattleCharacter? {
        for character in characters {
            if rectForCharacter(character).contains(point) {
                return character
            }
        }
        return nil
    }
    
    func shouldFlipCharacter(_ character: FF6BattleCharacter) -> Bool {

        if character.hasAI { return character.enemy }
        if background == 44 { return false }

        switch battleType {
        case .normal: return false
        case .back:   return true
        case .pincer: return (character.slot == 1 || character.slot == 3)
        case .side:   return (character.slot == 3 || character.slot == 4)
        }
    }
    
    // MARK: drawing methods
    
    override func draw(_ dirtyRect: NSRect) {
        super.draw(dirtyRect)
        
        // fill the rect with black (in case the background doesn't draw)
        NSColor.black.set()
        NSBezierPath.fill(dirtyRect)
        
        // draw the battle background
        drawBackgroundinRect(dirtyRect)

        // draw monsters in reverse order of priority (this will draw from the back-most to the front-most monster)
        for monster in monstersSortedByPriority.reversed() {
            drawMonster(monster)
        }

        // draw characters
        for character in charactersSortedByPriority.reversed() {
            drawCharacterShadow(character)
            drawCharacter(character)
        }
    }
    
    func drawBackgroundinRect(_ rect: NSRect) {
        // draw the battle background
        guard let iref = ff6ROM.battleBackgroundImage(background) else {
            rom.log("Unable to draw battle background")
            return
        }
        let height: CGFloat = ff6ROM.isGBA ? 120 : 147
        let width: CGFloat = 240
        let imageSize = NSMakeSize(32 * 8 * zoom, 32 * 8 * zoom)
        let image = NSImage(cgImage: iref, size: imageSize) // this is an NSImage of the entire 32x32 background image scaled to the zoom factor
        let srcRect = NSMakeRect(8 * zoom, 104 * zoom, width * zoom, height * zoom) // this is a rect with only the portion of the image that will be drawn on the screen
        
        // draw the background image
//        let imagePoint = NSMakePoint(0.0, bounds.height - height * zoom)
        NSGraphicsContext.current?.imageInterpolation = imageInterpolation
//        NSGraphicsContext.current()?.imageInterpolation = NSImageInterpolation.none
        image.draw(at: NSZeroPoint, from: srcRect, operation: NSCompositingOperation.sourceOver, fraction: 1.0)
    }
    
    func drawMonster(_ monster: FF6BattleMonster) {
        
        if monster.isNull { return }
        let rect = rectForMonster(monster)
        
        // create a monster image
        guard let iref = ff6ROM.monsterImage(monster.index) else {
            rom.log("Unable to create monster image")
            return
        }
        
        // draw the image
        let imageHeight = CGFloat(iref.height)
        let imageWidth = CGFloat(iref.width)
        var image = NSImage(cgImage: iref, size: NSMakeSize(imageWidth, imageHeight)) // this is an NSImage scaled to the zoom factor
        var imageRect = NSMakeRect(0, imageHeight - rect.height / zoom, rect.width / zoom, rect.height / zoom)
        var imageFraction: CGFloat = monster.present ? 1.0 : 0.5

        // tint the monster if it is selected
        if monster === selectedObject {
            image.tintWithColor(NSColor.keyboardFocusIndicatorColor)
            // don't make selected monsters as transparent
            if imageFraction == 0.5 {
                imageFraction = 0.75
            }
        }
        
        // flip the monster horizontally if needed
        if shouldFlipMonster(monster) {
            image = image.flippedHorizontally
            imageRect.origin.x = imageWidth - rect.width / zoom
        }

        NSGraphicsContext.current?.imageInterpolation = imageInterpolation
        image.draw(in: rect, from: imageRect, operation: NSCompositingOperation.sourceOver, fraction: imageFraction)
    }
    
    func drawCharacter(_ character: FF6BattleCharacter) {
        
        let rect = rectForCharacter(character)
        var graphics = character.slot - 1
        if character.hasAI { graphics = character.graphics }
        if graphics == 0xFF { graphics = character.character }
        
        // create a character image
        guard let iref = ff6ROM.battleCharacterImage(graphics, frame: 0x11) else {
            rom.log("Unable to create character image")
            return
        }
        
        // draw the image
        let imageHeight = CGFloat(iref.height)
        let imageWidth = CGFloat(iref.width)
        var image = NSImage(cgImage: iref, size: NSMakeSize(imageWidth, imageHeight)) // this is an NSImage scaled to the zoom factor
        let imageRect = NSMakeRect(0, imageHeight - rect.height / zoom, rect.width / zoom, rect.height / zoom)
        var imageFraction: CGFloat = 1.0

        if character.hasAI && character.hidden {
            // make character transparent if hidden
            imageFraction = 0.5
        }

        // tint the character if it is selected
        if character === selectedObject {
            image.tintWithColor(NSColor.keyboardFocusIndicatorColor)
            // don't make selected characters as transparent
            if imageFraction == 0.5 {
                imageFraction = 0.75
            }
        }

        // flip the character horizontally if needed
        if shouldFlipCharacter(character) {
            image = image.flippedHorizontally
        }
        
        NSGraphicsContext.current?.imageInterpolation = imageInterpolation
        image.draw(in: rect, from: imageRect, operation: NSCompositingOperation.sourceOver, fraction: imageFraction)
    }
    
    func drawCharacterShadow(_ character: FF6BattleCharacter) {
        
        let rect = rectForCharacter(character).offsetBy(dx: 0, dy: -24 * zoom)

        // create a shadow image
        guard let iref = ff6ROM.characterShadowImage() else {
            rom.log("Unable to create character shadow")
            return
        }
        
        // draw the image
        let imageHeight = CGFloat(iref.height) * zoom
        let imageWidth = CGFloat(iref.width) * zoom
        let image = NSImage(cgImage: iref, size: NSMakeSize(imageWidth, imageHeight)) // this is an NSImage scaled to the zoom factor
        let imageRect = NSMakeRect(0, imageHeight - rect.height, rect.width, rect.height)
        
        NSGraphicsContext.current?.imageInterpolation = imageInterpolation
//        NSGraphicsContext.current()?.imageInterpolation = NSImageInterpolation.none
        image.draw(in: rect, from: imageRect, operation: NSCompositingOperation.sourceOver, fraction: 1.0)

    }
    
    override func observeValue(forKeyPath keyPath: String?, of object: Any?, change: [NSKeyValueChangeKey : Any]?, context: UnsafeMutableRawPointer?) {
        if context != &FF6BattleView.context {
            super.observeValue(forKeyPath: keyPath, of: object, change: change, context: context)
        }

        if keyPath?.hasPrefix("monster") == true {
            reloadVerticalOffsets()
        } else if keyPath == "characterAI" {
            reloadCharacters()
        } else if keyPath?.hasPrefix("disable") == true {
            validateBattleType()
        }

        // always redraw if an observed object changed
        needsDisplay = true
    }
    
//    static var context = 0
}

// wrapper class for showing the total GP and experience for a battle

class FF6BattleExperienceGP: NSObject, ROMSelectable {
    
    // MARK: ROMSelectable properties
    
    var identifier: String? { return "battleExperienceGP" }
    var properties: NSDictionary? { return ["totalGP" : 0, "totalExperience" : 0] }
    var nib: String? { return nil }
    
    // MARK: member variables
    
    let rom: FF6ROM
    let battleProperties: ROMData
    var observedMonsterProperties = [ROMData]()
    
    @objc dynamic var totalExperience: Int {
        get {
            var experience = 0
            for monsterProperties in observedMonsterProperties {
                experience += monsterProperties.intValue(forKey: "experience")
            }
            return experience
        }
        set { }
    }
    
    @objc dynamic var totalGP: Int {
        get {
            var gp = 0
            for monsterProperties in observedMonsterProperties {
                gp += monsterProperties.intValue(forKey: "gold")
            }
            return gp
        }
        set { }
    }
    
    init(rom: FF6ROM, battleProperties: ROMData) {
        self.rom = rom
        self.battleProperties = battleProperties
        
        super.init()
        for slot in 1...6 {
            battleProperties.addObserver(self, forKeyPath: "monster\(slot)", options: NSKeyValueObservingOptions.old, context: &FF6BattleExperienceGP.context)
            battleProperties.addObserver(self, forKeyPath: "monster\(slot)MSB", options: NSKeyValueObservingOptions.old, context: &FF6BattleExperienceGP.context)
        }
        startObservingMonsters()
    }
    
    deinit {
        for slot in 1...6 {
            battleProperties.removeObserver(self, forKeyPath: "monster\(slot)", context: &FF6BattleExperienceGP.context)
            battleProperties.removeObserver(self, forKeyPath: "monster\(slot)MSB", context: &FF6BattleExperienceGP.context)
        }
        stopObservingMonsters()
    }
    
    func startObservingMonsters() {
        
        for slot in 1...6 {
            let monster = FF6BattleMonster(rom: rom, battleProperties: battleProperties, slot: slot)
            guard let monsterProperties = rom.value(at: monster.index, inPropertyWithKey: "monsterProperties") as? ROMData else { continue }
            monsterProperties.addObserver(self, forKeyPath: "experience", options: NSKeyValueObservingOptions.old, context: &FF6BattleExperienceGP.context)
            monsterProperties.addObserver(self, forKeyPath: "gold", options: NSKeyValueObservingOptions.old, context: &FF6BattleExperienceGP.context)
            observedMonsterProperties.append(monsterProperties)
        }
    }
    
    func stopObservingMonsters() {
        for monsterProperties in observedMonsterProperties {
            monsterProperties.removeObserver(self, forKeyPath: "experience", context: &FF6BattleExperienceGP.context)
            monsterProperties.removeObserver(self, forKeyPath: "gold", context: &FF6BattleExperienceGP.context)
        }
        observedMonsterProperties.removeAll()
    }
    
    override func observeValue(forKeyPath keyPath: String?, of object: Any?, change: [NSKeyValueChangeKey : Any]?, context: UnsafeMutableRawPointer?) {
        if context != &FF6BattleExperienceGP.context {
            super.observeValue(forKeyPath: keyPath, of: object, change: change, context: context)
        }

        // reload monsters
        stopObservingMonsters()
        startObservingMonsters()

        // notifiy observers that the experience and GP changed
        totalExperience = 0
        totalGP = 0
    }
    
    static var context = 0
}

extension NSImage {
    
    var flippedHorizontally: NSImage {
        
        let flippedImage = NSImage(size: size)
        
        flippedImage.lockFocus()
        let userInterpolationInt = UserDefaults.standard.integer(forKey: imageInterpolationKey)
        let imageInterpolation = NSImageInterpolation(rawValue: UInt(userInterpolationInt)) ?? NSImageInterpolation.none
        NSGraphicsContext.current?.imageInterpolation = imageInterpolation // this has no effect for some reason
        
        // clear the image
        NSColor.clear.set()
        NSMakeRect(0, 0, size.width, size.height).fill()
        
        // make a horizontally flipped transform
        var hFlipTransform = AffineTransform.identity
        hFlipTransform.translate(x: size.width, y: 0)
        hFlipTransform.scale(x: -1.0, y: 1.0)
        (hFlipTransform as NSAffineTransform).concat()
        
        // draw the flipped image
        draw(in: NSMakeRect(0, 0, size.width, size.height))
        flippedImage.unlockFocus()
        
        return flippedImage
    }
    
    func tintWithColor(_ color: NSColor, alpha: CGFloat = 0.5) {
        
        lockFocus()
        let userInterpolationInt = UserDefaults.standard.integer(forKey: imageInterpolationKey)
        let imageInterpolation = NSImageInterpolation(rawValue: UInt(userInterpolationInt)) ?? NSImageInterpolation.none
        NSGraphicsContext.current?.imageInterpolation = imageInterpolation // this has no effect for some reason
        color.withAlphaComponent(alpha).set()
        NSMakeRect(0, 0, size.width, size.height).fill(using: NSCompositingOperation.sourceAtop)
        unlockFocus()
    }
}
