//
//  FF6DialogView.swift
//  FF6Tools
//
//  Created by Evan Owen on 9/27/15.
//
//

import Cocoa

class FF6DialogView: FF6View {
    
    // MARK: instance variables
    
    var dialogEncoding: ROMTextEncoding? {
        return rom.textEncodings["dialog"] as? ROMTextEncoding
    }
    var textEncoding: ROMTextEncoding? {
        if textEncodingName != nil, let encoding = rom.textEncodings[textEncodingName!] as? ROMTextEncoding {
            return encoding
        }
        return dialogEncoding
    }
    @objc var assemblyIdentifier: String?
    @objc var textEncodingName: String?
    var windowLayer: CGLayer?
    var wallpaperIndex = 1 {
        didSet {
            // invalidate the window layer and redraw
            windowLayer = nil
            needsDisplay = true
        }
    }
    var textLayers: [CGLayer]?
    @objc var textAssembly: ROMText? {
        didSet {
            // return if the object is already selected
            if textAssembly === oldValue { return }
            
            // redraw the window
            windowLayer = nil
            
            // stop observing the old string
            oldValue?.removeObserver(self, forKeyPath: "string", context: &FF6DialogView.context)
            
            // select this string in the dialog view
            dialogString = textAssembly
            
            // observe the new string
            textAssembly?.addObserver(self, forKeyPath: "string", options: NSKeyValueObservingOptions.old, context: &FF6DialogView.context)

        }
    }
    var dialogString: AnyObject? {
        didSet {
            // invalidate the text layers and redraw
            textLayers = nil
            needsDisplay = true
            if dialogString is String { return }
            if oldValue !== dialogString && superview is NSClipView {
                // scroll to the top if the dialog string changed
                scroll(NSMakePoint(0.0, 0.0))
            }
        }
    }
    var windowHeight: Int {
        if assemblyIdentifier == "dialog" { return isGBA ? 6 : 9 }
        if assemblyIdentifier == "mapTitles" { return 4 }
        if assemblyIdentifier?.hasSuffix("Descriptions") == true { return 5 }
        if assemblyIdentifier == "battleMessages" { return 4 }
        if assemblyIdentifier == "monsterDialog" { return 4 }
        if assemblyIdentifier == "battleDialog" {
            if let numLines = textAssembly?.numLines {
                return numLines * 2 + 2
            }
            return 9
        }
        return 9
    }
    var pageSize: NSSize { return NSSize(width: 30 * 8, height: Int(windowHeight) * 8) }
    var zoom: CGFloat {
        if let clipView = superview as? NSClipView {
            return clipView.documentVisibleRect.width / pageSize.width
        }
        return 1.0
    }
    
    deinit {
        textAssembly?.removeObserver(self, forKeyPath: "string", context: &FF6DialogView.context)
    }
    
    override func awakeFromNib() {
        if windowController == nil || assemblyIdentifier == nil { return }
        bind(NSBindingName(rawValue: "textAssembly"), to: windowController!, withKeyPath: assemblyIdentifier!, options: nil)
    }

    override var isFlipped: Bool { return true }

    override func mouseDown(with event: NSEvent) {
        // increment the wallpaper index when clicked
        wallpaperIndex += 1
        if wallpaperIndex > 7 { wallpaperIndex = 0 }
    }
    
    // MARK: drawing methods
    
    override func setFrameSize(_ newSize: NSSize) {
        
        // set the frame size based on the superview width and the number of pages
        if let clipView = superview as? NSClipView  {
            let numPages = textLayers?.count ?? 1
            var frameSize = NSSize(width: clipView.documentVisibleRect.width, height: CGFloat(numPages) * pageSize.height * zoom)
            if frameSize.height == 0 { frameSize.height = pageSize.height * zoom }
            super.setFrameSize(frameSize)
            return
        }
        super.setFrameSize(newSize)
    }
    
    override func draw(_ dirtyRect: NSRect) {
        super.draw(dirtyRect)

//        if rom == nil { return }
        if windowLayer == nil { drawWindow() }
        if textLayers == nil { textLayers = [CGLayer](); drawText() }

        // get the current graphics context and set the interpolation mode
        guard let currentContext = NSGraphicsContext.current else { return }
        currentContext.imageInterpolation = imageInterpolation

        // update the frame size based on the number of pages
        setFrameSize(frame.size)

        // draw an empty window if there are no pages
        if textLayers!.count == 0 {
            let pageRect = NSRect(x: 0, y: 0, width: pageSize.width * zoom, height: pageSize.height * zoom)
            currentContext.cgContext.draw(windowLayer!, in: pageRect)
            return
        }
        
        // draw each page in the view
        for (page, textLayer) in textLayers!.enumerated() {
            let pageRect = NSRect(x: 0, y: CGFloat(page) * pageSize.height * zoom, width: pageSize.width * zoom, height: pageSize.height * zoom)
            currentContext.cgContext.draw(windowLayer!, in: pageRect)
            currentContext.cgContext.draw(textLayer, in: pageRect)
        }
    }

    func drawWindow() {
        
        // get the current graphics context
        guard let currentContext = NSGraphicsContext.current else { return }
        
        // load menu window graphics
        guard let graphicsData = rom.value(at: wallpaperIndex, inPropertyWithKey: "menuWindowGraphics") as? ROMData else { return }
        let gfxLength = 0x0700
        let gfx = UnsafeMutablePointer<UInt8>.allocate(capacity: gfxLength)
        memset(gfx, 0, gfxLength)
        let graphicsBytes: [UInt8] = graphicsData.data.withUnsafeBytes{ [UInt8](UnsafeBufferPointer(start: $0, count: graphicsData.data.count)) }
        gfx_decode(gfx, graphicsBytes, Int32(graphicsBytes.count), isGBA ? LINEAR_4BPP : SNES_4BPP)
        let windowGraphics = Data(bytes: UnsafePointer<UInt8>(gfx), count: gfxLength)
        gfx.deinitialize(); gfx.deallocate(capacity: gfxLength)
        
        // load menu window palette
        guard let paletteData = rom.value(at: wallpaperIndex, inPropertyWithKey: "menuWindowPalettes") as? ROMData else { return }
        let palLength = 16 * MemoryLayout<UInt32>.size // 16 colors
        let pal = UnsafeMutablePointer<UInt8>.allocate(capacity: palLength)
        memset(pal, 0, palLength)
        let paletteBytes: [UInt8] = paletteData.data.withUnsafeBytes{ [UInt8](UnsafeBufferPointer(start: $0, count: paletteData.data.count)) }
        gfx_decode(pal, paletteBytes, Int32(paletteBytes.count), BGR555)
        let windowPalette = Data(bytes: UnsafePointer<UInt8>(pal), count: palLength)
        pal.deinitialize(); pal.deallocate(capacity: palLength)
        
        // create menu window tile formation
        var tileFormation = [UInt16]()
        for windowRow in 0..<windowHeight {
            switch windowRow {
            case 0:
                // top row
                tileFormation.append(contentsOf: [16, 17, 18, 17, 18, 17, 18, 17, 18, 17, 18, 17, 18, 17, 18, 17, 18, 17, 18, 17, 18, 17, 18, 17, 18, 17, 18, 17, 18, 19])
                
            case windowHeight - 1:
                // bottom row
                tileFormation.append(contentsOf: [24, 25, 26, 25, 26, 25, 26, 25, 26, 25, 26, 25, 26, 25, 26, 25, 26, 25, 26, 25, 26, 25, 26, 25, 26, 25, 26, 25, 26, 27])
                
            case let row1 where row1 % 4 == 1:
                tileFormation.append(contentsOf: [20,  0,  1,  2,  3,  0,  1,  2,  3,  0,  1,  2,  3,  0,  1,  2,  3,  0,  1,  2,  3,  0,  1,  2,  3,  0,  1,  2,  3, 21])

            case let row1 where row1 % 4 == 2:
                tileFormation.append(contentsOf: [22,  4,  5,  6,  7,  4,  5,  6,  7,  4,  5,  6,  7,  4,  5,  6,  7,  4,  5,  6,  7,  4,  5,  6,  7,  4,  5,  6,  7, 23])

            case let row1 where row1 % 4 == 3:
                tileFormation.append(contentsOf: [20,  8,  9, 10, 11,  8,  9, 10, 11,  8,  9, 10, 11,  8,  9, 10, 11,  8,  9, 10, 11,  8,  9, 10, 11,  8,  9, 10, 11, 21])

            case let row1 where row1 % 4 == 0:
                tileFormation.append(contentsOf: [22, 12, 13, 14, 15, 12, 13, 14, 15, 12, 13, 14, 15, 12, 13, 14, 15, 12, 13, 14, 15, 12, 13, 14, 15, 12, 13, 14, 15, 23])

            default:
                tileFormation.append(contentsOf: [20,  0,  1,  2,  3,  0,  1,  2,  3,  0,  1,  2,  3,  0,  1,  2,  3,  0,  1,  2,  3,  0,  1,  2,  3,  0,  1,  2,  3, 21])
            }
        }
        let tileData = Data(buffer: UnsafeBufferPointer(start: &tileFormation, count: tileFormation.count * 2))
        
        // create the ppu
        let width = Int(pageSize.width)
        let height = Int(pageSize.height)
        var ppu = PPU()
        memset(&ppu, 0, MemoryLayout<PPU>.size)
        ppu.pal = (windowPalette as NSData).bytes
        ppu.flipped = true
        ppu.layers.0.format = SNES_4BPP_TILE
        ppu.layers.0.height = Int32(height / 8)
        ppu.layers.0.width = Int32(width / 8)
        ppu.layers.0.z.0 = Z_LEVEL_BG1_LO
        ppu.layers.0.z.1 = Z_LEVEL_BG1_HI
        ppu.layers.0.gfx = (windowGraphics as NSData).bytes
        ppu.layers.0.tiles = (tileData as NSData).bytes
        ppu.layers.0.main = true
        
        // render the screen
        let dataLength = width * height * MemoryLayout<UInt32>.size
        let data = UnsafeMutablePointer<UInt8>.allocate(capacity: dataLength)
        memset(data, 0, dataLength)
        gfx_render_ppu(data, ppu, Int32(dataLength), Int32(width))
        
        // draw the dialog window
        let provider = CGDataProvider(dataInfo: nil, data: data, size: dataLength, releaseData: releaseImageData)!
        let bitsPerComponent = 8
        let bitsPerPixel = 32
        let bytesPerRow = width * MemoryLayout<UInt32>.size
        let colorSpaceRef = CGColorSpaceCreateDeviceRGB()
        let bitmapInfo = CGBitmapInfo(rawValue: CGBitmapInfo().rawValue | CGImageAlphaInfo.premultipliedLast.rawValue)
        let renderingIntent = CGColorRenderingIntent.defaultIntent
        let windowImage = CGImage(width: width, height: height, bitsPerComponent: bitsPerComponent, bitsPerPixel: bitsPerPixel, bytesPerRow: bytesPerRow, space: colorSpaceRef, bitmapInfo: bitmapInfo, provider: provider, decode: nil, shouldInterpolate: false, intent: renderingIntent)!
        let windowRect = NSRect(x: 0, y: 0, width: width, height: height)
        windowLayer = CGLayer(currentContext.cgContext, size: windowRect.size, auxiliaryInfo: nil)
        windowLayer?.context?.draw(windowImage, in: windowRect)
        
        // free memory
        data.deinitialize(); data.deallocate(capacity: dataLength)
    }
    
    func drawText() {

        if let text = dialogString as? ROMText {
            drawText(text)
            
        } else if let string = dialogString as? String {
            
            guard var formattedString = textEncoding?.format(string: string) else { return }

            // draw each page of text
            if textLayers == nil { textLayers = [CGLayer]() }
            textLayers!.removeAll()
            while let pageLayer = drawPage(&formattedString) {
                textLayers!.append(pageLayer)
            }
            needsDisplay = true
        }
    }
    
    func drawText(_ text: ROMText) {
        // format the text
        var formattedString = text.formattedString()
        
        // draw each page of text
        if textLayers == nil { textLayers = [CGLayer]() }
        textLayers!.removeAll()
        while let pageLayer = drawPage(&formattedString) {
            textLayers!.append(pageLayer)
        }
        needsDisplay = true

    }
    
    func drawPage(_ string: inout String) -> CGLayer? {
        
        if isGBA {
            return drawPageGBA(&string)
        }
        
        var xIndent: CGFloat = 8.0
        if assemblyIdentifier?.hasSuffix("Descriptions") == true {
            xIndent = 5.0
        } else if (assemblyIdentifier == "battleMessages") || (assemblyIdentifier == "monsterDialog") || (assemblyIdentifier == "battleDialog") {
            xIndent = 4.0
        }

        var yIndent: CGFloat = 9.0
        if (assemblyIdentifier == "battleMessages") || (assemblyIdentifier == "monsterDialog") || (assemblyIdentifier == "battleDialog") {
            yIndent = 11.0
        }
        
        var linesPerPage = 4
        if assemblyIdentifier == "monsterDialog" {
            linesPerPage = 1
        } else if assemblyIdentifier == "battleDialog" {
            linesPerPage = (windowHeight - 2) / 2
        }

        if string.isEmpty { return nil }
        guard let currentContext = NSGraphicsContext.current else { return nil }
        guard let pageLayer = CGLayer(currentContext.cgContext, size: pageSize, auxiliaryInfo: nil) else { return nil }
        
        var currentPoint = NSPoint(x: xIndent, y: yIndent)

        func drawWord(_ word: String, atPoint point: NSPoint) {
            var currentPoint = point
            
            for character in word.characters {
                drawCharacter(character, atPoint: currentPoint)
                currentPoint.x += widthOfCharacter(character)
            }
        }
        
        func drawCharacter(_ character: Character, atPoint point: NSPoint) {
            guard let encodedValue = dialogEncoding?.encodingTable[String(character)] as? Int else {
                rom.log("Invalid character: \(character)")
                return
            }
            
            drawCharacterWithEncodedValue(encodedValue, atPoint: point)
        }
        
        func drawCharacterWithEncodedValue(_ encodedValue: Int, atPoint point: NSPoint) {
            if encodedValue >= 0x1C00 {
                drawCharacterWithEncodedValue(encodedValue - 0x1B00, atPoint: point)
                return
            }
            
            guard let characterGraphicsData = rom.value(at: encodedValue - 0x20, inPropertyWithKey: "bigTextGraphics") as? ROMData else {
                rom.log("Invalid dialog character value \(encodedValue)")
                return
            }
            drawCharacterWithGraphics(characterGraphicsData.data, atPoint: point)
        }
        
        func drawCharacterWithGraphics(_ characterGraphicsData: Data, atPoint point: NSPoint) {
            if characterGraphicsData.count != 11 * MemoryLayout<UInt16>.size {
                rom.log("Invalid dialog character graphics")
                return
            }
            var characterGraphics = [UInt16](repeating: 0, count: 11)
            (characterGraphicsData as NSData).getBytes(&characterGraphics, length: 11 * MemoryLayout<UInt16>.size)
            
            for (y, var line) in characterGraphics.enumerated() {
                var x = 0
                while line != 0 {
                    if (line & 0x8000) != 0 {
                        pageLayer.context?.setFillColor(gray: 1.0, alpha: 1.0)
                        pageLayer.context?.fill(NSRect(x: point.x + CGFloat(x), y: point.y + CGFloat(y), width: 1, height: 1))
                        pageLayer.context?.setFillColor(gray: 0.0, alpha: 1.0)
                        pageLayer.context?.fill(NSRect(x: point.x + CGFloat(x + 1), y: point.y + CGFloat(y + 1), width: 1, height: 1))
                    }
                    x += 1
                    line <<= 1
                }
            }
        }
        
        func widthOfWord(_ word: String) -> CGFloat {
            var width: CGFloat = 0
            for character in word.characters {
                width += widthOfCharacter(character)
            }
            return width
        }
        
        func widthOfCharacter(_ character: Character) -> CGFloat {
            guard let encodedValue = dialogEncoding?.encodingTable[String(character)] as? Int else {
                return 0.0
            }
            return widthOfCharacterWithEncodedValue(encodedValue)
        }
        
        func widthOfCharacterWithEncodedValue(_ encodedValue: Int) -> CGFloat {
            if encodedValue >= 0x1C00 {
                return widthOfCharacterWithEncodedValue(encodedValue - 0x1B00)
            }
            
            guard let bigTextWidths = rom.value(at: encodedValue, inPropertyWithKey: "bigTextWidths") as? ROMData else { return 0.0 }
            return CGFloat(bigTextWidths.intValue(forKey: "width"))
        }

        var currentLine = 0
        
        // get the space character
        var spaceChar: Character = " "
        if dialogEncoding?.encodingTable["\u{3000}"] != nil {
            // use the full-width space if it exists in the encoding table
            spaceChar = "\u{3000}"
        }

        while !string.isEmpty {
            switch string.characters.first! {
                
            case spaceChar:
                // space
                currentPoint.x += widthOfCharacter(spaceChar)
                string.remove(at: string.startIndex)
                
            case "\u{5}":
                // multiple choice indicator
                currentPoint.x += widthOfCharacterWithEncodedValue(0xFF)
                string.remove(at: string.startIndex)
                
            case "\n":
                // new line
                currentPoint.x = xIndent
                currentPoint.y += 15
                string.remove(at: string.startIndex)
                currentLine += 1
                if currentLine >= linesPerPage { return pageLayer }
                
            case "\u{12}":
                // new page
                string.remove(at: string.startIndex)
                return pageLayer
                
            case "\\":
                // half-width hiragana
                string.remove(at: string.startIndex)
                if string.isEmpty { break }
                let hiraganaChar = string.characters.first!
                if !FF6ROM.hiraganaCharacters.contains(String(hiraganaChar)) { continue }
                guard let encodingTable = dialogEncoding?.encodingTable as? [String : Int],
                    var encodedValue = encodingTable[String(hiraganaChar)] else {
                    rom.log("Invalid character: \"\(hiraganaChar)\"")
                    continue
                }
                encodedValue -= encodingTable["バ"] ?? 0x20
                encodedValue += encodingTable["ﾊﾞ"] ?? 0x1F20
                drawCharacterWithEncodedValue(encodedValue, atPoint: currentPoint)
                currentPoint.x += widthOfCharacterWithEncodedValue(encodedValue)
                string.remove(at: string.startIndex)
                
            default:
                // normal character
                var word: String!
                if spaceChar != " " {
                    // jap-style characters (separates lines anywhere)
                    word = String(string.characters.first!)
                } else if let nextWhiteSpace = string.rangeOfCharacter(from: CharacterSet(charactersIn: " \u{3000}\n\u{5}\u{12}\0")) {
                    // found whitespace
                    word = String(string[..<nextWhiteSpace.lowerBound])
//                    word = string.substring(to: nextWhiteSpace.lowerBound)
                } else {
                    // there is no whitespace before the end of the string
                    word = string
                }
                
                let width = widthOfWord(word)
                if (currentPoint.x + width) >= 226 { // C0/7FA0 says 224, but this matches the actual game for some reason
                    currentPoint.x = xIndent
                    currentPoint.y += 15
                    currentLine += 1
                    if currentLine >= linesPerPage { return pageLayer }
                }
                drawWord(word, atPoint: currentPoint)
                currentPoint.x += width
                string.removeSubrange(string.range(of: word)!)
            }
        }
        return pageLayer
    }
    
    func drawPageGBA(_ string: inout String) -> CGLayer? {
        if string.isEmpty { return nil }
        guard let currentContext = NSGraphicsContext.current else {
            return nil
        }
        guard let pageLayer = CGLayer(currentContext.cgContext, size: pageSize, auxiliaryInfo: nil) else {
            return nil
        }
        
        var currentPoint = NSPoint(x: 8, y: 5)
        
        func drawWord(_ word: String, atPoint point: NSPoint) {
            var currentPoint = point
            
            for character in word.characters {
                drawCharacter(character, atPoint: currentPoint)
                currentPoint.x += widthOfCharacter(character)
            }
        }
        
        func drawCharacter(_ character: Character, atPoint point: NSPoint) {
            guard let encodedValue = dialogEncoding?.encodingTable[String(character)] as? Int else { // rom.textEncodingTables["dialog"]?[String(character)] else {
                rom.log("Invalid character: \(character)")
                return
            }
            
            drawCharacterWithEncodedValue(encodedValue, atPoint: point)
        }
        
        func drawCharacterWithEncodedValue(_ encodedValue: Int, atPoint point: NSPoint) {
            if encodedValue >= 0x0100 {
//                drawCharacterWithEncodedValue(encodedValue - 0x1B00, atPoint: point)
                return
            }
            
            guard let characterGraphicsData = rom.value(at: encodedValue, inPropertyWithKey: "bigTextAdvance") as? ROMData else {
                rom.log("Invalid dialog character value \(encodedValue)")
                return
            }
            drawCharacterWithGraphics(characterGraphicsData.data, atPoint: point)
        }
        
        func drawCharacterWithGraphics(_ characterGraphicsData: Data, atPoint point: NSPoint) {
            
            var width = 0
            var bytesPerLine = 0
            (characterGraphicsData as NSData).getBytes(&width, length: 1)
            (characterGraphicsData as NSData).getBytes(&bytesPerLine, range: NSMakeRange(1, 1))
            if characterGraphicsData.count != 2 + 12 * bytesPerLine {
                rom.log("Invalid dialog character graphics")
                return
            }
            var characterGraphics = [UInt8](repeating: 0, count: 12 * bytesPerLine)
            (characterGraphicsData as NSData).getBytes(&characterGraphics, range: NSMakeRange(2, characterGraphics.count))
            
            for y in 0..<12 {
                for x in 0..<bytesPerLine {
                    var line = characterGraphics[y * bytesPerLine + x]
                    for pixel in 0..<4 {
                        let color = line & 0x03
                        if color == 1 {
                            pageLayer.context?.setFillColor(gray: 1.0, alpha: 1.0)
                            pageLayer.context?.fill(NSRect(x: point.x + CGFloat(x * 4 + pixel), y: point.y + CGFloat(y), width: 1, height: 1))
                        } else if color == 2 {
                            pageLayer.context?.setFillColor(gray: 0.0, alpha: 1.0)
                            pageLayer.context?.fill(NSRect(x: point.x + CGFloat(x * 4 + pixel), y: point.y + CGFloat(y), width: 1, height: 1))
                        }
                        line >>= 2
                    }
                }
            }
        }
        
        func widthOfWord(_ word: String) -> CGFloat {
            var width: CGFloat = 0
            for character in word.characters {
                width += widthOfCharacter(character)
            }
            return width
        }
        
        func widthOfCharacter(_ character: Character) -> CGFloat {
            guard let encodedValue = dialogEncoding?.encodingTable[String(character)] as? Int else { // rom.textEncodingTables["dialog"]?[String(character)] else {
                return 0.0
            }
            return widthOfCharacterWithEncodedValue(encodedValue)
        }
        
        func widthOfCharacterWithEncodedValue(_ encodedValue: Int) -> CGFloat {
            if encodedValue >= 0x0100 {
                return 0//widthOfCharacterWithEncodedValue(encodedValue - 0x1B00)
            }
            
            guard let characterGraphicsData = rom.value(at: encodedValue, inPropertyWithKey: "bigTextAdvance") as? ROMData else {
                rom.log("Invalid dialog character value \(encodedValue)")
                return 0.0
            }

            var width = 0
            let _ = characterGraphicsData.data.copyBytes(to: UnsafeMutableBufferPointer<Int>(start: &width, count: 1), from: 0..<1)
            return CGFloat(width)
        }
        
        var currentLine = 0
        
        // get the space character
        var spaceChar: Character = " "
        if dialogEncoding?.encodingTable["\u{3000}"] != nil {
            // use the full-width space if it exists in the encoding table
            spaceChar = "\u{3000}"
        }
        
        while !string.isEmpty {
            switch string.characters.first! {
                
            case spaceChar:
                // space
                currentPoint.x += widthOfCharacter(spaceChar)
                string.remove(at: string.startIndex)
                
            case "\u{5}":
                // multiple choice indicator
                currentPoint.x += widthOfCharacterWithEncodedValue(0xFF)
                string.remove(at: string.startIndex)
                
            case "\n":
                // new line
                currentPoint.x = 8
                currentPoint.y += 12
                string.remove(at: string.startIndex)
                currentLine += 1
                if currentLine > 2 { return pageLayer }
                
            case "\u{12}":
                // new page
                string.remove(at: string.startIndex)
                return pageLayer
                
            case "\\":
                // half-width hiragana
                string.remove(at: string.startIndex)
                if string.isEmpty { break }
                let hiraganaChar = string.characters.first!
                if !FF6ROM.hiraganaCharacters.contains(String(hiraganaChar)) { continue }
                guard let encodingTable = dialogEncoding?.encodingTable as? [String : Int],
                    var encodedValue = encodingTable[String(hiraganaChar)] else {
                    rom.log("Invalid character: \"\(hiraganaChar)\"")
                    continue
                }
                encodedValue -= encodingTable["バ"] ?? 0x20
                encodedValue += encodingTable["ﾊﾞ"] ?? 0x1F20
                drawCharacterWithEncodedValue(encodedValue, atPoint: currentPoint)
                currentPoint.x += widthOfCharacterWithEncodedValue(encodedValue)
                string.remove(at: string.startIndex)
                
            default:
                // normal character
                var word: String!
                if spaceChar != " " {
                    // jap-style characters (separates lines anywhere)
                    word = String(string.characters.first!)
                } else if let nextWhiteSpace = string.rangeOfCharacter(from: CharacterSet(charactersIn: " \u{3000}\n\u{5}\u{12}\0")) {
                    // found whitespace
                    word = String(string[..<nextWhiteSpace.lowerBound])
//                    word = string.substring(to: nextWhiteSpace.lowerBound)
                } else {
                    // there is no whitespace before the end of the string
                    word = string
                }
                
                let width = widthOfWord(word)
                if (currentPoint.x + width) >= 226 { // C0/7FA0 says 224, but this matches the actual game for some reason
                    currentPoint.x = 8
                    currentPoint.y += 12
                    currentLine += 1
                    if currentLine > 2 { return pageLayer }
                }
                drawWord(word, atPoint: currentPoint)
                currentPoint.x += width
                string.removeSubrange(string.range(of: word)!)
            }
        }
        return pageLayer
    }
    
    override func controlTextDidEndEditing(_ obj: Notification) {
        if let textField = obj.object as? NSTextField, let text = dialogString as? ROMText {
            // update the text field if any invalid characters were removed
            textField.stringValue = text.string
        }
    }
    
    override func controlTextDidChange(_ obj: Notification) {
        if let textField = obj.object as? NSTextField {
            // update the dialog view
            dialogString = textField.stringValue as AnyObject?
        }
    }
    
    override func observeValue(forKeyPath keyPath: String?, of object: Any?, change: [NSKeyValueChangeKey : Any]?, context: UnsafeMutableRawPointer?) {
        if context != &FF6DialogView.context {
            super.observeValue(forKeyPath: keyPath, of: object, change: change, context: context)
        }
        
        guard let text = object as? ROMText else { return }
        
        text.removeObserver(self, forKeyPath: "string", context: &FF6DialogView.context)
        dialogString = text
        windowLayer = nil
        text.addObserver(self, forKeyPath: "string", options: NSKeyValueObservingOptions.old, context: &FF6DialogView.context)
    }
    
//    static var context = 0
}
