//
//  FF6MapEditor.swift
//  FF6Tools
//
//  Created by Evan Owen on 7/26/16.
//
//

import Cocoa

class FF6MapEditor: FF6MapView {

    @IBOutlet var triggerMenu: NSMenu?
    
    // selection
    var selectedTrigger: FF6MapTrigger?
    var selectedTiles: [UInt8] = [0x73, 0, 0, 1, 1, 0]
    var selectedLayerIndex = 3
    var selectedLayer: FF6MapLayer {
        if mapIndex < 3 { return bg1 }
        switch selectedLayerIndex {
        case 0: return bg1
        case 1: return bg2
        case 2: return bg3
        case 3: return overlay // probably shouldn't have this here
        default:
            rom.log("Invalid layer")
            return bg1
        }
    }
    
    @objc dynamic var showScreenBoundary: Bool = false { didSet { needsDisplay = true } }
    var screenLayer: CGLayer?
    var screenLayerPosition = (x: 0, y: 0)
    
    var tilePropertiesData: ROMData?
    var maskLayer: CGLayer?
    var tileMask: FF6MapTileMask = .noMask
    
    enum FF6MapTileMask: Int {
        case noMask = 0
        case spriteOverlay = 1
        case passability = 2
        case npcPassability = 3
        case layerPriority = 4
    }

    @IBAction func showLayer(_ sender: AnyObject) {
        guard let button = sender as? NSButton, let identifier = button.identifier else { return }
        self.setValue(button.state, forKey: identifier.rawValue)
    }
    
    @IBAction func selectLayer(_ sender: AnyObject) {
        // sender is a segmented control
        guard let layerButton = sender as? NSSegmentedControl else { return }
        
        // return if the layer didn't change
        if selectedLayerIndex == layerButton.selectedSegment { return }

        selectedLayerIndex = layerButton.selectedSegment
        selectedTiles = [0x73, 0, 0, 1, 1, 0]
        tilesetView?.selectedTiles = selectedTiles
        tilesetView?.tilesetLayer = nil
        tilesetView?.needsDisplay = true
    }

    @IBAction func selectMask(_ sender: AnyObject) {
        guard let menuItem = sender as? NSMenuItem else { return }
        
        tileMask = FF6MapTileMask.init(rawValue: menuItem.tag) ?? .noMask
        maskLayer = nil
        needsDisplay = true
        
        tilesetView?.maskLayer = nil
        tilesetView?.needsDisplay = true
    }
    
    override func loadMap() {
        super.loadMap()

        maskLayer = nil
        tilesetView?.tilesetLayer = nil
        tilesetView?.maskLayer = nil
        tilesetView?.needsDisplay = true
        
        // stop observing the previous map's tile properties
        stopObservingTileProperties()
        tilePropertiesData = nil
        
        // select the map's tile properties
        if mapIndex < 3 { return }
        guard let tilePropertiesIndex = mapProperties?.intValue(forKey: "tileProperties") else { return }
        tilePropertiesData = rom.value(at: tilePropertiesIndex, inPropertyWithKey: "mapTileProperties") as? ROMData

        // start observing tile properties
        startObservingTileProperties()
    }
    
    func stopObservingTileProperties() {
        if tilePropertiesData == nil { return }
        
        guard let byte1 = tilePropertiesData?.value(forKey: "byte1") as? ROMArray,
            let byte2 = tilePropertiesData?.value(forKey: "byte2") as? ROMArray else { return }
        
        for object in byte1.array {
            guard let tile = object as? ROMData else { continue }
            stopObserving(object: tile)
        }

        for object in byte2.array {
            guard let tile = object as? ROMData else { continue }
            stopObserving(object: tile)
        }
    }

    func startObservingTileProperties() {
        if tilePropertiesData == nil { return }

        guard let byte1 = tilePropertiesData?.value(forKey: "byte1") as? ROMArray,
            let byte2 = tilePropertiesData?.value(forKey: "byte2") as? ROMArray else { return }
        
        for object in byte1.array {
            guard let tile = object as? ROMData else { continue }
            startObserving(object: tile)
        }
        
        for object in byte2.array {
            guard let tile = object as? ROMData else { continue }
            startObserving(object: tile)
        }
    }

    override func observeValue(forKeyPath keyPath: String?, of object: Any?, change: [NSKeyValueChangeKey : Any]?, context: UnsafeMutableRawPointer?) {
        if context != &FF6MapView.context {
            super.observeValue(forKeyPath: keyPath, of: object, change: change, context: context)
            return
        }
        
        if let romData = object as? ROMData, romData.value(forKey: "stairsUL") != nil || romData.value(forKey: "passableNPC") != nil {
            // tile properties have changed
            maskLayer = nil
            needsDisplay = true
            tilesetView?.maskLayer = nil
            tilesetView?.needsDisplay = true
        } else {
            super.observeValue(forKeyPath: keyPath, of: object, change: change, context: context)
        }
    }

    // MARK: drawing methods
    
    override func draw(_ dirtyRect: NSRect) {
        
        super.draw(dirtyRect)
        
        // draw the tile mask
        if tileMask != .noMask && mapIndex >= 3 {
            if maskLayer == nil { updateMaskLayer() }
            if maskLayer != nil {
                
                for x in stride(from: 0, to: 128, by: bg1.width) {
                    for y in stride(from: 0, to: 64, by: bg1.height) {
                        let maskRect = NSMakeRect(CGFloat(x) * 16 * zoom, CGFloat(y) * 16 * zoom, CGFloat(bg1.width) * 16 * zoom, CGFloat(bg1.height) * 16 * zoom)
                        NSGraphicsContext.current?.cgContext.draw(maskLayer!, in: maskRect)
                    }
                }
            }
        }

        // draw the screen boundary
        if showScreenBoundary && mapIndex >= 3 {
            if screenLayer == nil {
                updateScreenLayer()
            }
            NSGraphicsContext.current?.cgContext.draw(screenLayer!, in: bounds)
        }

        // draw the cursor
        if selectedLayerIndex == 3 {
            // trigger cursor
            
            // return if the selected trigger is not on this map
            if selectedTrigger == nil || !triggers.contains(selectedTrigger!) {
                selectedTrigger = nil
                return
            }
            
            if let triggerRect = rectForTrigger(selectedTrigger!), let triggerColor = selectedTrigger?.type.color {
                let x = Int(triggerRect.origin.x / zoom)
                let y = Int((bounds.height - triggerRect.origin.y - triggerRect.size.height) / zoom)
                let width = Int(triggerRect.size.width / zoom)
                let height = Int(triggerRect.size.height / zoom)
                drawCursor(x: x, y: y, width: width, height: height, color: triggerColor)
            }
            
        } else {
            // tile cursor
            
            // get cursor geometry
            if selectedTiles.count < 5 { return }
            let x = Int(selectedTiles[1])
            let y = Int(selectedTiles[2])
            let width = Int(selectedTiles[3])
            let height = Int(selectedTiles[4])
            
            // get cursor color (based on selected layer)
            let cursorColor: NSColor
            switch selectedLayerIndex {
            case 0: cursorColor = NSColor.green // bg1
            case 1: cursorColor = NSColor.blue // bg2
            case 2: cursorColor = NSColor.red // bg3
            default: cursorColor = NSColor.green
            }
            
            drawCursor(x: x * 16, y: y * 16, width: width * 16, height: height * 16, color: cursorColor)
        }
    }
    
    func updateMaskLayer(forTiles tileData: Data? = nil) {

        if tileMask == .noMask { return }
        if mapIndex < 3 { return }

        // create the mask layer if it is not valid
        if maskLayer == nil {
            if let currentContext = NSGraphicsContext.current?.cgContext {
                maskLayer = CGLayer(currentContext, size: CGSize(width: bg1.width * 16, height: bg1.height * 16), auxiliaryInfo: nil)
            } else {
                // unable to get current context, this was probably not called from drawRect
                rom.log("Unable to create mask layer")
                return
            }
        }
        let context = maskLayer?.context

        var x = 0
        var y = 0
        var height = bg1.height
        var width = bg1.width
        
        if tileData != nil && maskLayer != nil {
            let tileBytes: [UInt8] = tileData!.withUnsafeBytes{ [UInt8](UnsafeBufferPointer(start: $0, count: tileData!.count)) }
            
            // read the tile data header
            if tileBytes.count <= 5 {
                rom.log("Selected tile data is missing a header")
            } else {
                
                // redraw only a portion of the mask layer
                x = Int(tileBytes[1]) & 0x7F
                y = Int(tileBytes[2]) & 0x3F
                width = Int(tileBytes[3])
                height = Int(tileBytes[4])
            }
        }

        // get tile properties
        guard let tilePropertiesIndex = mapProperties?.intValue(forKey: "tileProperties"),
            let tilePropertiesData = rom.value(at: tilePropertiesIndex, inPropertyWithKey: "mapTileProperties") as? ROMData else { return }
        
        // assemble tile properties data to capture any changes that have been made
        var dummyData = Data()
        tilePropertiesData.assembleTo(&dummyData)
        let tilePropertiesBytes: [UInt8] = tilePropertiesData.data.withUnsafeBytes{ [UInt8](UnsafeBufferPointer(start: $0, count: tilePropertiesData.data.count)) }
        
        // create the mask graphics
        guard let maskGraphicsPath = Bundle.main.path(forResource: "MaskGraphics", ofType: "bin"),
            let maskGraphics = NSData(contentsOfFile: maskGraphicsPath) else { return }
        
        // create the ppu
        var ppu = PPU()
        memset(&ppu, 0, MemoryLayout<PPU>.size)
        
        // create the mask palette
        let palette: [UInt32] = [
            0, 0x80000000, 0x40000000, 0x20000000, // 0 = black
            0, 0x800000FF, 0x400000FF, 0x200000FF, // 1 = red
            0, 0x8000FF00, 0x4000FF00, 0x2000FF00, // 2 = green
            0, 0x8000FFFF, 0x4000FFFF, 0x2000FFFF, // 3 = yellow
            0, 0x80FF0000, 0x40FF0000, 0x20FF0000, // 4 = blue
            0, 0x80FF00FF, 0x40FF00FF, 0x20FF00FF, // 5 = magenta
            0, 0x80FFFF00, 0x40FFFF00, 0x20FFFF00, // 6 = cyan
            0, 0x80FFFFFF, 0x40FFFFFF, 0x20FFFFFF  // 7 = white
        ]
        ppu.pal = UnsafeRawPointer(palette)
        
        // set up the ppu layer
        ppu.layers.0.format = SNES_2BPP_TILE
        ppu.layers.0.height = Int32(height * 2)
        ppu.layers.0.width = Int32(width * 2)
        ppu.layers.0.v_scroll = 0
        ppu.layers.0.h_scroll = 0
        ppu.layers.0.z.0 = Z_LEVEL_MASK
        ppu.layers.0.z.1 = Z_LEVEL_MASK
        ppu.layers.0.main = true
        
        // set the ppu graphics and tiles based on the selected mask
        var maskTiles = [UInt16](repeating: 0, count: height * width * 4)
        switch tileMask {
        case .noMask: return
        case .spriteOverlay:
            // set up overlay
            ppu.layers.0.gfx = (overlay.graphics as NSData).bytes
            ppu.layers.0.tiles = (overlay.tiles as NSData).bytes
            ppu.layers.0.height = Int32(bg1.height * 2)
            ppu.layers.0.width = Int32(bg1.width * 2)
            ppu.layers.0.v_scroll = Int32(y * 2)
            ppu.layers.0.h_scroll = Int32(x * 2)
            
        case .passability:
            
            let bg1Layout: [UInt8] = bg1.layout.withUnsafeBytes{ [UInt8](UnsafeBufferPointer(start: $0, count: bg1.layout.count)) }
            
            for row in 0..<height {
                for col in 0..<width {
                    
                    if (y + row) * bg1.width + (col + x) >= bg1Layout.count { break }
                    if row * width + col >= maskTiles.count / 4 { break }
                    
                    let layoutTile = bg1Layout[(y + row) * bg1.width + (col + x)]
                    
                    let byte1 = tilePropertiesBytes[Int(layoutTile)]
                    let byte2 = tilePropertiesBytes[Int(layoutTile) + 256]
                    let tileProperties = Int(byte1) | Int(byte2) << 8
                    
                    // skip impassable tiles
                    if tileProperties == 0xFFF7 { continue }
                    
                    // set the tile based on passability
                    var tiles = [UInt16]()
                    switch byte2 & 0x0F {
                    case 0x01: tiles = [4, 5, 4, 5]
                    case 0x02: tiles = [5, 4, 5, 4]
                    case 0x03: tiles = [5, 5, 5, 5]
                    case 0x04: tiles = [4, 4, 6, 6]
                    case 0x05: tiles = [4, 5, 6, 7]
                    case 0x06: tiles = [5, 4, 7, 6]
                    case 0x07: tiles = [5, 5, 7, 7]
                    case 0x08: tiles = [6, 6, 4, 4]
                    case 0x09: tiles = [6, 7, 4, 5]
                    case 0x0A: tiles = [7, 6, 5, 4]
                    case 0x0B: tiles = [7, 7, 5, 5]
                    case 0x0C: tiles = [6, 6, 6, 6]
                    case 0x0D: tiles = [6, 7, 6, 7]
                    case 0x0E: tiles = [7, 6, 7, 6]
                    case 0x0F: tiles = [7, 7, 7, 7]
                    default:   tiles = [0, 0, 0, 0]
                    }
                    
                    // set v/h flip for each tile
                    tiles[1] |= 0x4000
                    tiles[2] |= 0x8000
                    tiles[3] |= 0xC000
                    
                    // stairs tile
                    switch byte1 & 0xC0 {
                    case 0x40: tiles = [0x0008, 0x0009, 0x000A, 0x000B] // up/right
                    case 0x80: tiles = [0x4009, 0x4008, 0x400B, 0x400A] // up/left
                    default: break
                    }
                    
                    // ladder tile
                    if (byte2 & 0x40) != 0 {
                        tiles = [0x000C, 0x400C, 0x800C, 0xC00C]
                    }
                    
                    // door tile
                    if (byte1 & 0x20) != 0 {
                        tiles = [0x000D, 0x400D, 0x000E, 0x000F]
                    }
                    
                    // counter tile
                    if tileProperties == 0x0007 {
                        tiles = [0x0410, 0x4410, 0x8410, 0xC410]
                    }
                    
                    // set the palette based on z-level
                    var paletteByte: UInt16 = 0
                    switch byte1 & 0x07 {
                    case 1: paletteByte = 0x1000 // upper (blue)
                    case 2: paletteByte = 0x0400 // lower (red)
                    case 3: paletteByte = 0x1400 // stairs (magenta)
                    case 4: paletteByte = 0x0800 // bridge (green)
                    default: break // other
                    }
                    tiles[0] |= paletteByte
                    tiles[1] |= paletteByte
                    tiles[2] |= paletteByte
                    tiles[3] |= paletteByte
                    
                    var tileOffset = col * 2 + row * width * 4
                    
                    maskTiles[tileOffset] = tiles[0]
                    maskTiles[tileOffset + 1] = tiles[1]
                    tileOffset += width * 2
                    maskTiles[tileOffset] = tiles[2]
                    maskTiles[tileOffset + 1] = tiles[3]
                }
            }
            
            ppu.layers.0.gfx = maskGraphics.bytes
            ppu.layers.0.tiles = UnsafeRawPointer(maskTiles)
            
        case .npcPassability:
            
            let bg1Layout: [UInt8] = bg1.layout.withUnsafeBytes{ [UInt8](UnsafeBufferPointer(start: $0, count: bg1.layout.count)) }

            for row in 0..<height {
                for col in 0..<width {
                    
                    if (y + row) * bg1.width + (col + x) >= bg1Layout.count { break }
                    if row * width + col >= maskTiles.count / 4 { break }
                    
                    let layoutTile = bg1Layout[(y + row) * bg1.width + (col + x)]
                    
                    let byte1 = tilePropertiesBytes[Int(layoutTile)]
                    let byte2 = tilePropertiesBytes[Int(layoutTile) + 256]
                    let tileProperties = Int(byte1) | Int(byte2) << 8
                    
                    // skip impassable and counter tiles
                    if (tileProperties == 0xFFF7) || (tileProperties == 0x0007) { continue }

                    // skip tiles that are not passable by npcs
                    if byte2 & 0x80 == 0 { continue }
                    
                    // set the palette based on z-level
                    var paletteByte: UInt16 = 0
                    switch byte1 & 0x07 {
                    case 1: paletteByte = 0x1000 // upper (blue)
                    case 2: paletteByte = 0x0400 // lower (red)
                    case 3: paletteByte = 0x1400 // stairs (magenta)
                    case 4: paletteByte = 0x0800 // bridge (green)
                    default: break // other
                    }

                    var tileOffset = col * 2 + row * width * 4
                    maskTiles[tileOffset] = 0x0007 | paletteByte
                    maskTiles[tileOffset + 1] = 0x4007 | paletteByte
                    tileOffset += width * 2
                    maskTiles[tileOffset] = 0x8007 | paletteByte
                    maskTiles[tileOffset + 1] = 0xC007 | paletteByte
                }
            }

            ppu.layers.0.gfx = maskGraphics.bytes
            ppu.layers.0.tiles = UnsafeRawPointer(maskTiles)
            
        case .layerPriority:
            
            let bg1Layout: [UInt8] = bg1.layout.withUnsafeBytes{ [UInt8](UnsafeBufferPointer(start: $0, count: bg1.layout.count)) }
            
            for row in 0..<height {
                for col in 0..<width {
                    
                    if (y + row) * bg1.width + (col + x) >= bg1Layout.count { break }
                    if row * width + col >= maskTiles.count / 4 { break }
                    
                    let layoutTile = bg1Layout[(y + row) * bg1.width + (col + x)]
                    
                    let byte1 = tilePropertiesBytes[Int(layoutTile)]
                    let byte2 = tilePropertiesBytes[Int(layoutTile) + 256]
                    let tileProperties = Int(byte1) | Int(byte2) << 8
                    
                    // skip impassable and counter tiles
                    if (tileProperties == 0xFFF7) || (tileProperties == 0x0007) { continue }
                    
                    var tile: UInt16 = 0x0000
                    switch byte1 & 0x18 {
                    case 0x08: tile = 0x0401
                    case 0x10: tile = 0x0801
                    case 0x18: tile = 0x1401
                    default: break
                    }
                    
                    var tileOffset = col * 2 + row * width * 4
                    maskTiles[tileOffset] = tile
                    maskTiles[tileOffset + 1] = tile
                    tileOffset += width * 2
                    maskTiles[tileOffset] = tile
                    maskTiles[tileOffset + 1] = tile
                }
            }
            
            ppu.layers.0.gfx = maskGraphics.bytes
            ppu.layers.0.tiles = UnsafeRawPointer(maskTiles)

        }
        
        // render the ppu
        let imageHeight = height * 16
        let imageWidth = width * 16
        let dataLength = imageHeight * imageWidth * 4
        let data = UnsafeMutablePointer<UInt8>.allocate(capacity: dataLength)
        memset(data, 0, dataLength)
        gfx_render_ppu(data, ppu, Int32(dataLength), Int32(imageWidth))
        
        // create an image
        let provider = CGDataProvider(dataInfo: nil, data: data, size: dataLength, releaseData: releaseImageData)!
        let bitsPerComponent = 8
        let bitsPerPixel = 32
        let bytesPerRow = imageWidth * MemoryLayout<UInt32>.size
        let colorSpaceRef = CGColorSpaceCreateDeviceRGB()
        let bitmapInfo = CGBitmapInfo(rawValue: CGBitmapInfo().rawValue | CGImageAlphaInfo.premultipliedLast.rawValue)
        let renderingIntent = CGColorRenderingIntent.defaultIntent
        let iref = CGImage(width: imageWidth, height: imageHeight, bitsPerComponent: bitsPerComponent, bitsPerPixel: bitsPerPixel, bytesPerRow: bytesPerRow, space: colorSpaceRef, bitmapInfo: bitmapInfo, provider: provider, decode: nil, shouldInterpolate: true, intent: renderingIntent)
        
        // draw the image to the mask layer
        let maskRect = NSMakeRect(CGFloat(x) * 16, CGFloat(bg1.height - y - height) * 16, CGFloat(imageWidth), CGFloat(imageHeight))
        context?.clear(maskRect)
        context?.draw(iref!, in: maskRect)
        
        // free memory
        data.deinitialize(); data.deallocate(capacity: dataLength)
    }
    
    func updateScreenLayer() {
        
        if showScreenBoundary == false { return }
        if mapIndex < 3 { return }
        
        let width = 128 * 16
        let height = 64 * 16
        
        // clear the screen boundary layer
        if screenLayer == nil {
            guard let currentContext = NSGraphicsContext.current?.cgContext else { return }
            screenLayer = CGLayer(currentContext, size: CGSize(width: CGFloat(width) * zoom, height: CGFloat(height) * zoom), auxiliaryInfo: nil)
        }
        
        let screenLayerContext = screenLayer?.context
        
        let x = CGFloat(screenLayerPosition.x - 7)
        let y = CGFloat(screenLayerPosition.y + 6) + 0.5
        
        let screenRect = NSMakeRect(x * 16 * zoom, bounds.height - y * 16 * zoom, 15 * 16 * zoom, 13 * 16 * zoom)
        
        screenLayerContext?.setFillColor(NSColor(deviceWhite: 0.0, alpha: 0.5).cgColor)
        screenLayerContext?.setBlendMode(CGBlendMode.destinationAtop)
        screenLayerContext?.fill(bounds)
        screenLayerContext?.setFillColor(NSColor(deviceWhite: 0.0, alpha: 0.0).cgColor)
        screenLayerContext?.setBlendMode(CGBlendMode.destinationAtop)
        screenLayerContext?.fill(screenRect)
    }

    // MARK: NSResponder methods
    
    override var acceptsFirstResponder: Bool { return true }
    
    override func mouseDown(with theEvent: NSEvent) {
        
        window?.makeFirstResponder(self)

        var event: NSEvent! = theEvent
        
        // get the point that was clicked
        let clickedPoint = convert(theEvent.locationInWindow, from: nil)
        if !bounds.contains(clickedPoint) { return }
        
        showCursor = true

        if selectedLayerIndex == 3 {
            // try to select a trigger
            let triggerArray = triggersAtPoint(clickedPoint)
            
            if selectedTrigger != nil && triggerArray.contains(selectedTrigger!) && selectedTrigger !== triggerArray.last {
                // cycle through triggers stacked at this point
                selectedTrigger = triggerArray[triggerArray.index(of: selectedTrigger!)! + 1]
            } else {
                // select the first trigger in the stack (or nil)
                selectedTrigger = triggerArray.first
            }
            
            if selectedTrigger == nil {
                // no triggers at that point
                selectedTiles = [0x73, 0, 0, 0, 0]
                
                if windowController?.objectIsSelected(mapProperties) == false {
                    // select map properties
                    windowController.deselectAllObjects()
                    windowController.selectObject(mapProperties)
//                    windowController?.selectObject(atIndex: mapIndex, inArray: rom.value(forKey: "mapProperties") as AnyObject?)
                }
                return
                
            } else if windowController?.objectIsSelected(selectedTrigger) == false {
                // select the trigger in the window controller
                windowController.deselectAllObjects()
                windowController.selectObject(selectedTrigger, identifierOverride: nil)

                // update the cursor
                needsDisplay = true
            }
        }
        
        let oldX = selectedTrigger?.x
        let oldY = selectedTrigger?.y
        
        var triggerHasNotMoved = true
        while event != nil {
            
            switch event.type {
            case .leftMouseUp:
                // update the layout data in the rom when the mouse is released
                if (selectedLayerIndex != 3) && (selectedLayer.layoutIndex != -1) {
                    setMapLayoutData(selectedLayer.layout, forIndex: selectedLayer.layoutIndex)
                }
                return
                
            case .leftMouseDragged, .leftMouseDown:
                // get the dragged/clicked point
                let draggedPoint = convert(event.locationInWindow, from: nil)
                if !bounds.contains(draggedPoint) {
                    break
                }
                
                if selectedLayerIndex == 3 && selectedTrigger != nil {

                    // get the delta from the original clicked point
                    var deltaX = draggedPoint.x - clickedPoint.x + clickedPoint.x.truncatingRemainder(dividingBy: (16 * zoom))
                    var deltaY = draggedPoint.y - clickedPoint.y + clickedPoint.y.truncatingRemainder(dividingBy: (16 * zoom))
                    if deltaX < 0 { deltaX -= (16 * zoom) }
                    if deltaY < 0 { deltaY -= (16 * zoom) }
                    
                    // limit the position to valid values (this shouldn't be needed because i already checked to make sure the point was in the bounds)
                    let newX = min(255, max(oldX! + Int(deltaX / zoom / 16.0), 0))
                    let newY = min(255, max(oldY! - Int(deltaY / zoom / 16.0), 0))
                    
                    // break if the monster didn't move
                    if ((newX == selectedTrigger!.x) && (newY == selectedTrigger!.y)) { break }
                    
                    // move the selected trigger (only allow undo the first time)
                    setTriggerPosition(x: newX, y: newY, triggerAssembly: selectedTrigger!.triggerAssembly, canUndo: triggerHasNotMoved)
                    triggerHasNotMoved = false

                } else if selectedTiles.count >= 5 { // selectedLayerIndex < 3 {

                    let draggedX = Int(draggedPoint.x / zoom / 16.0)
                    let draggedY = Int((bounds.height - draggedPoint.y) / zoom / 16.0)

                    // continue if the cursor hasn't moved
                    let oldX = Int(selectedTiles[1])
                    let oldY = Int(selectedTiles[2])
                    if (oldX == draggedX) && (oldY == draggedY) && (event.type != .leftMouseDown) { break }

                    selectedTiles[1] = UInt8(draggedX)
                    selectedTiles[2] = UInt8(draggedY)

                    // set the tiles
                    var tiles = selectedTiles
                    
                    // update the header
                    let x = Int(draggedPoint.x / zoom / 16.0) % selectedLayer.width
                    let y = Int((bounds.height - draggedPoint.y) / zoom / 16.0) % selectedLayer.height
                    tiles[1] = UInt8(x)
                    tiles[2] = UInt8(y | selectedLayerIndex << 6)
                    
                    // pass a copy of the updated data to setTiles
                    let tileData = Data(buffer: UnsafeBufferPointer<UInt8>(start: &tiles, count: tiles.count))
                    if mapIndex < 3 {
                        setWorldTiles(tileData, forMapIndex: mapIndex)
                    } else {
                        setTiles(tileData)
                    }
                }
                
            default:
                break
            }
            
            // get the next event
            event = window?.nextEvent(matching: [.leftMouseUp, .leftMouseDragged])
        }
    }
    
    override func rightMouseDown(with theEvent: NSEvent) {
        
        window?.makeFirstResponder(self)

        var event: NSEvent! = theEvent

        let clickedPoint = convert(theEvent.locationInWindow, from: nil)
        if !bounds.contains(clickedPoint) { return }
        
        showCursor = true

        let clickedX = Int(clickedPoint.x / zoom / 16.0)
        let clickedY = Int((bounds.height - clickedPoint.y) / zoom / 16.0)
        
        // update the selected tiles (also saves the cursor position for creating a new trigger)
        let clickedTile = tileAt(point: clickedPoint)
        selectedTiles = [0x73, UInt8(clickedX), UInt8(clickedY), 1, 1, clickedTile]
        tilesetView?.selectedTiles = [0x73, clickedTile % 16, (clickedTile / 16), 1, 1, clickedTile]
        
        // select tile properties if a bg1 tile was selected
        selectTileProperties(tile: Int(clickedTile))
        
        needsDisplay = true

        if selectedLayerIndex == 3 {
            // try to select a trigger
            selectedTrigger = triggerAtPoint(clickedPoint)
            
            if selectedTrigger == nil {
                windowController?.deselectAllObjects()
            } else if windowController?.objectIsSelected(selectedTrigger) == false {
                // select the trigger in the window controller
                windowController?.deselectAllObjects()
                windowController?.selectObject(selectedTrigger, identifierOverride: nil)
            }
            
            // update and show the trigger menu
            if triggerMenu != nil {
                triggerMenu!.update()
                NSMenu.popUpContextMenu(triggerMenu!, with: theEvent, for: self)
            }
            
            return
        }

        // keep track of where the mouse gets dragged
        var draggedX = clickedX
        var draggedY = clickedY

        while event != nil {
            
            switch event.type {
            case .rightMouseUp:
                // return if the right mouse button is up
                return
                
            case .rightMouseDragged:
                let draggedPoint = convert(event.locationInWindow, from: nil)
                
                // don't update if the dragged coordinates didn't change
                let newX = Int(draggedPoint.x / zoom / 16.0)
                let newY = Int((bounds.height - draggedPoint.y) / zoom / 16.0)
                if (draggedX == newX) && (draggedY == newY) { break }
                draggedX = newX
                draggedY = newY
                
                // get the selected rectangle
                let minX = max(min(draggedX, clickedX), 0)
                let minY = max(min(draggedY, clickedY), 0)
                let maxX = min(max(draggedX, clickedX), 255)
                let maxY = min(max(draggedY, clickedY), 255)
                let width = maxX - minX + 1
                let height = maxY - minY + 1
                
                // create the new tile selection
                selectedTiles = [0x73, UInt8(minX), UInt8(minY), UInt8(width), UInt8(height)]
                for y in 0..<height {
                    for x in 0..<width {
                        selectedTiles.append(tileAtCoordinates(Int(minX + x), Int(minY + y)))
                    }
                }
                tilesetView?.selectedTiles = [0x73, 0, 0, 0, 0]
                
                needsDisplay = true
                
            default: break
            }
            
            // get the next event
            event = window?.nextEvent(matching: [.rightMouseUp, .rightMouseDragged])
        }
    }
    
    override func mouseMoved(with theEvent: NSEvent) {

        let cursorPoint = convert(theEvent.locationInWindow, from: nil)
        let x = Int(cursorPoint.x / 16 / zoom)
        let y = Int((bounds.height - cursorPoint.y) / 16 / zoom)

        if selectedLayerIndex == 3 {
            // don't update the cursor if the trigger layer is selected
            
        } else if !bounds.contains(cursorPoint) || selectedTiles.count < 5 {
            // hide the cursor if the mouse left the view
            showCursor = false
            
        } else {
            // update the cursor
            selectedTiles[1] = UInt8(truncatingIfNeeded: x)
            selectedTiles[2] = UInt8(truncatingIfNeeded: y)
            showCursor = true
            needsDisplay = true
        }

        // update the screen boundary layer
        if screenLayerPosition != (x, y) {
            screenLayerPosition = (x, y)
            screenLayer = nil
            if showScreenBoundary {
                needsDisplay = true
            }
        }
    }
    
    override func mouseEntered(with theEvent: NSEvent) {
        window?.makeFirstResponder(self)
    }
    
    override func mouseExited(with theEvent: NSEvent) {
        if selectedLayerIndex != 3 {
            // hide the tile cursor when the mouse exits
            showCursor = false
            needsDisplay = true
        }
    }
    
    override func updateTrackingAreas() {
        for area in trackingAreas {
            removeTrackingArea(area)
        }
        window?.acceptsMouseMovedEvents = true
        
        let opts: NSTrackingArea.Options = [NSTrackingArea.Options.mouseEnteredAndExited, NSTrackingArea.Options.activeAlways]
        let trackingArea = NSTrackingArea(rect: self.bounds, options: opts, owner: self, userInfo: nil)
        addTrackingArea(trackingArea)
        super.updateTrackingAreas()
    }

    // MARK: edit tiles
    
    @objc func setMapLayoutData(_ layoutData: Data, forIndex layoutIndex: Int) {
        
        // never change layout index 0
        // never use this method to change a world map (index -1)
        if layoutIndex <= 0 { return }
        
        // save undo data
        guard let mapLayoutArray = rom.value(forKeyPath: "mapLayouts") as? ROMArray else { return }
        let oldLayoutData = mapLayoutArray[layoutIndex].value(forKey: "data") as! Data
        if layoutData == oldLayoutData { return }
        
        if let undo = windowController?.undoManager {
            let target = undo.prepare(withInvocationTarget: self) as AnyObject
            target.setMapLayoutData(oldLayoutData,  forIndex: layoutIndex)
            if (!undo.isUndoing) {
                undo.setActionName("Change Map Tiles")
            }
        }
        
        // set the new data
        stopObserving(object: bg1.layoutData)
        stopObserving(object: bg2.layoutData)
        stopObserving(object: bg3.layoutData)
        mapLayoutArray[layoutIndex].setValue(layoutData, forKey: "data")
        startObserving(object: bg1.layoutData)
        startObserving(object: bg2.layoutData)
        startObserving(object: bg3.layoutData)

        // update tile data and redraw the map if any layer on the current map has this layout index
        if bg1.layoutIndex == layoutIndex {
            bg1.layout = layoutData
            bg1.decodeLayout()
            overlay.layout = layoutData
            overlay.decodeLayout()
            mapLayer = nil
            maskLayer = nil
            needsDisplay = true
        }
        if bg2.layoutIndex == layoutIndex {
            bg2.layout = layoutData
            bg2.decodeLayout()
            mapLayer = nil
            needsDisplay = true
        }
        if bg3.layoutIndex == layoutIndex {
            bg3.layout = layoutData
            bg3.decodeLayout()
            mapLayer = nil
            needsDisplay = true
        }
    }
    
    func setTiles(_ tileData: Data) {
        
        var tiles = [UInt8](repeating: 0, count: tileData.count)
        tileData.copyBytes(to: &tiles, count: tileData.count)

        // read the tile data header
        if tiles.count <= 5 {
            rom.log("Selected tile data is missing a header")
            return
        }
        var x = Int(tiles[1])
        var y = Int(tiles[2])
        let width = Int(tiles[3])
        let height = Int(tiles[4])
        var layerIndex = 0
        if mapIndex >= 3 {
            layerIndex = y >> 6
            x &= 0x7F
            y &= 0x3F
        }
        if tiles.count != 5 + width * height {
            rom.log("Invalid selected tile data size")
            return
        }
        tiles.removeSubrange(0..<5)
        
        // edit the tile layout for the affected layer
        var layer: FF6MapLayer!
        switch layerIndex {
        case 0: layer = bg1
        case 1: layer = bg2
        case 2: layer = bg3
        default:
            rom.log("Invalid layer")
            return
        }
        var layoutData = Data(layer.layout!)
        
        x %= layer.width
        y %= layer.height
        let chunkWidth = (x + width) > layer.width ? layer.width - x : width
        let chunkHeight = (y + height) > layer.height ? layer.height - y : height
        for row in 0..<chunkHeight {
            // change one row at a time
            let firstTile = x + (y + row) * layer.width
            let lastTile = min(firstTile + chunkWidth, layoutData.count)
            if lastTile <= firstTile { continue }
            let tileCount = lastTile - firstTile
            layoutData.replaceSubrange(firstTile..<lastTile, with: tiles[0..<tileCount])
            tiles.removeSubrange(0..<width)
        }
        
        if bg1.layoutIndex == layer.layoutIndex {
            bg1.layout = layoutData
            bg1.decodeLayout()
            overlay.layout = layoutData
            overlay.decodeLayout()
            updateMaskLayer(forTiles: tileData)
        }
        if bg2.layoutIndex == layer.layoutIndex {
            bg2.layout = layoutData
            bg2.decodeLayout()
        }
        if bg3.layoutIndex == layer.layoutIndex {
            bg3.layout = layoutData
            bg3.decodeLayout()
        }
        
        // redraw a portion of the map
        updateMapLayerInRect(NSMakeRect(CGFloat(x) * 16 * zoom, bounds.height - CGFloat(y) * 16 * zoom, CGFloat(width) * zoom * 16, CGFloat(height) * zoom * 16))
        needsDisplay = true
    }
    
    @objc func setWorldTiles(_ tileData: Data, forMapIndex mapIndex: Int) {
        
        var tiles = [UInt8](repeating: 0, count: tileData.count)
        tileData.copyBytes(to: &tiles, count: tileData.count)
        
        // read the tile data header
        if tiles.count <= 5 {
            rom.log("Selected tile data is missing a header")
            return
        }
        
        // get the tile layout for the selected world map
        var worldLayoutKey: String?
        var mapSize = 256
        switch mapIndex {
        case 0: worldLayoutKey = "worldOfBalanceMapLayout"
        case 1: worldLayoutKey = "worldOfRuinMapLayout"
        case 2: worldLayoutKey = "serpentTrenchMapLayout"; mapSize = 128
        default: return
        }
        if worldLayoutKey == nil { return }
        guard let worldLayout = rom.value(forKey: worldLayoutKey!) as? ROMData else { return }
        
        // create a copy of the layout data (if you don't copy it, there's a Swift bug that can fuck it up later)
        var layoutData = Data(worldLayout.data)
        
        // read the location and size of the new tiles
        let x = Int(tiles[1]) % mapSize
        let y = Int(tiles[2]) % mapSize
        let width = Int(tiles[3])
        let height = Int(tiles[4])
        if tiles.count != 5 + width * height {
            rom.log("Invalid selected tile data size")
            return
        }
        tiles.removeSubrange(0..<5)
        
        // copy tiles to map layout (ignore any tiles that fall beyond the edge of the map)
        let chunkWidth = (x + width) > mapSize ? mapSize - x : width
        let chunkHeight = (y + height) > mapSize ? mapSize - y : height
        var oldTiles: [UInt8] = [0x73, UInt8(x), UInt8(y), UInt8(chunkWidth), UInt8(chunkHeight)]
        
        var somethingChanged = false
        for row in 0..<chunkHeight {
            let firstTile = x + (y + row) * mapSize
            let lastTile = min(firstTile + chunkWidth, layoutData.count)
            if lastTile <= firstTile { continue }
            let tileRange: Range<Int> = firstTile..<lastTile

            // copy each row of the old map layout
            oldTiles.append(contentsOf: layoutData.subdata(in: tileRange))
            
            // don't change anything if the tiles aren't changing
            let newTiles = Data(bytes: UnsafePointer<UInt8>(tiles), count: chunkWidth)
            if layoutData.subdata(in: tileRange) == newTiles {
                tiles.removeSubrange(0..<width)
                continue
            }
            somethingChanged = true
            
            // set each row of the new layout
            layoutData.replaceSubrange(tileRange, with: newTiles)
            if mapIndex == self.mapIndex {
                // update bg1 if this is for the current map
                bg1.decodePartialWorldLayout(newTiles, tileRange: tileRange)
            }
            tiles.removeSubrange(0..<width)
        }

        // redraw even if nothing changed (to update the cursor)
        needsDisplay = true

        // return if nothing changed
        if somethingChanged == false { return }
        
        // update the world layout (in the rom and for the layer)
        worldLayout.data = layoutData
        bg1.layout = layoutData
        
        // save undo data
        if let undo = windowController?.undoManager {
            let target = undo.prepare(withInvocationTarget: self) as AnyObject
            target.setWorldTiles(Data(buffer: UnsafeBufferPointer<UInt8>(start: &oldTiles, count: oldTiles.count)), forMapIndex: mapIndex)
            if (!undo.isUndoing) {
                undo.setActionName("Change World Map Tiles")
            }
        }
        
        // redraw a portion of the map
        updateMapLayerInRect(NSMakeRect(CGFloat(x) * 16 * zoom, bounds.height - CGFloat(y) * 16 * zoom, CGFloat(width) * zoom * 16, CGFloat(height) * zoom * 16))
        needsDisplay = true
    }
    
    func tileAt(point: NSPoint) -> UInt8 {
        let x = Int(point.x / zoom / 16.0)
        let y = Int((bounds.height - point.y) / zoom / 16.0)
        return tileAtCoordinates(x, y)
    }
    
    func tileAtCoordinates(_ x: Int, _ y: Int) -> UInt8 {
        if selectedLayerIndex == 3 { return 0 }
        
        let layer = selectedLayer
        let x = x % layer.width
        let y = y % layer.height
        let layoutData = layer.layout as NSData
        let tileRange = NSMakeRange(y * layer.width + x, 1)
        if layoutData.length < NSMaxRange(tileRange) { return 0 }
        var tile: UInt8 = 0
        layoutData.getBytes(&tile, range: tileRange)
        
        return tile
    }
    
    func selectTileProperties(tile: Int) {
        
        if mapIndex < 2 {
            // select world map tile properties
            guard let worldTileProperties = rom.value(at: mapIndex, inPropertyWithKey: "worldMapTileProperties") as? ROMArray,
                let tileData = worldTileProperties[tile] as? ROMData else { return }
            windowController?.deselectAllObjects()
            switch mapIndex {
            case 0: tileData.identifier = "worldOfBalanceTileProperties"
            case 1: tileData.identifier = "worldOfRuinTileProperties"
            default: break
            }
            windowController.selectObject(tileData)
            return
        } else if mapIndex == 2 {
            // serpent trench has no tile properties
            return
        } else if selectedLayer.type != .bg1 {
            // tile properties are only on bg1
            return
        }
        
        // select tile properties
        guard let tilePropertiesIndex = mapProperties?.intValue(forKey: "tileProperties"),
            let tileProperties = rom.value(at: tilePropertiesIndex, inPropertyWithKey: "mapTileProperties") as? ROMData,
            let byte1Array = tileProperties.value(forKey: "byte1") as? ROMArray,
            let byte2Array = tileProperties.value(forKey: "byte2") as? ROMArray,
            let byte1 = byte1Array[tile] as? ROMData,
            let byte2 = byte2Array[tile] as? ROMData else { return }
        
        windowController.deselectAllObjects()
        windowController.selectObject(FF6MapTileProperties(rom: ff6ROM, byte1: byte1, byte2: byte2), identifierOverride: nil)
        windowController.selectObject(byte1)
        windowController.selectObject(byte2)
    }
    
    // MARK: edit triggers

    @IBAction func cut(_ sender: AnyObject) {
        
        if selectedTrigger == nil { return }
        copy(sender)
        
        // remove the trigger from the rom array
        guard let triggerArrayName = selectedTrigger!.type.arrayName,
            let triggerArray = rom.value(at: mapIndex, inPropertyWithKey: triggerArrayName) as? ROMArray else { return }
        removeTrigger(selectedTrigger!.triggerAssembly, fromArray: triggerArray)
        
        selectedTrigger = nil
        windowController?.deselectAllObjects()
    }

    @IBAction func copy(_ sender: AnyObject) {
        guard let triggerAssembly = selectedTrigger?.triggerAssembly, let triggerArrayName = selectedTrigger?.type.arrayName else { return }
        
        var triggerData = Data()
        triggerAssembly.assembleTo(&triggerData)
        
        // clear the pasteboard
        let pasteboard = NSPasteboard.general
        pasteboard.clearContents()
        
        // add the trigger data to the pasteboard
        let pasteboardItem = NSPasteboardItem()
        pasteboardItem.setString(triggerArrayName, forType: NSPasteboard.PasteboardType(rawValue: "com.ekowen86.FF6Tools.FF6MapTrigger.type"))
        pasteboardItem.setData(triggerData, forType: NSPasteboard.PasteboardType(rawValue: "com.ekowen86.FF6Tools.FF6MapTrigger"))
        pasteboard.writeObjects([pasteboardItem])
    }
    
    @IBAction func paste(_ sender: AnyObject) {
        // get the pasteboard
        guard let pastedItems = NSPasteboard.general.readObjects(forClasses: [NSPasteboardItem.self], options: nil) as? [NSPasteboardItem] else { return }
        if pastedItems.count == 0 { return }
        
        // make sure the encoding matches
        guard let triggerArrayName = pastedItems[0].string(forType: NSPasteboard.PasteboardType(rawValue: "com.ekowen86.FF6Tools.FF6MapTrigger.type")),
            let triggerData = pastedItems[0].data(forType: NSPasteboard.PasteboardType(rawValue: "com.ekowen86.FF6Tools.FF6MapTrigger")),
            let triggerArray = rom.value(at: mapIndex, inPropertyWithKey: triggerArrayName) as? ROMArray,
            let triggerAssembly = triggerArray.createNewAssembly() else { return }
        
        triggerAssembly.disassembleFrom(triggerData)
        
        if selectedTiles.count == 6 {
            let x = Int(selectedTiles[1])
            let y = Int(selectedTiles[2])
            triggerAssembly.setValue(x, forKeyPath: "x")
            triggerAssembly.setValue(y, forKeyPath: "y")
        }

        insertTrigger(triggerAssembly, inArray: triggerArray, atIndex: NSNotFound)
    }
    
    override func keyDown(with theEvent: NSEvent) {
        
        // there's got to be a better way to do this
        if theEvent.keyCode == 0x33 {
            delete(self)
        } else {
            super.keyDown(with: theEvent)
        }
    }
    
    @IBAction func delete(_ sender: AnyObject) {
        if selectedTrigger == nil { return }
        
        // remove the trigger from the rom array
        guard let triggerArrayName = selectedTrigger!.type.arrayName,
            let triggerArray = rom.value(at: mapIndex, inPropertyWithKey: triggerArrayName) as? ROMArray else { return }
        removeTrigger(selectedTrigger!.triggerAssembly, fromArray: triggerArray)
        
        selectedTrigger = nil
        windowController?.deselectAllObjects()
    }
    
    func rectForTrigger(_ trigger: FF6MapTrigger) -> NSRect? {
        
        let x = trigger.triggerAssembly.intValue(forKey: "x")
        let y = trigger.triggerAssembly.intValue(forKey: "y")
        
        if trigger.isLong {
            // long entrance trigger
            let length = trigger.triggerAssembly.intValue(forKey: "length")
            let vertical = trigger.triggerAssembly.boolValue(forKey: "vertical")
            if vertical {
                return NSMakeRect(CGFloat(x) * 16 * zoom, bounds.height - CGFloat(y + length + 1) * 16 * zoom, 16 * zoom, 16 * CGFloat(length + 1) * zoom)
            } else {
                return NSMakeRect(CGFloat(x) * 16 * zoom, bounds.height - CGFloat(y + 1) * 16 * zoom, 16 * CGFloat(length + 1) * zoom, 16 * zoom)
            }
        } else {
            // all other triggers
            return NSMakeRect(CGFloat(x) * 16 * zoom, bounds.height - CGFloat(y + 1) * 16 * zoom, 16 * zoom, 16 * zoom)
        }
    }
    
    func triggersAtPoint(_ point: NSPoint) -> [FF6MapTrigger] {
        var triggerArray = [FF6MapTrigger]()
        
        for trigger in triggers {
            if let triggerRect = rectForTrigger(trigger), triggerRect.contains(point) {
                triggerArray.append(trigger)
            }
        }
        return triggerArray
    }
    
    func triggerAtPoint(_ point: NSPoint) -> FF6MapTrigger? {
        return triggersAtPoint(point).first
    }
    
    @objc func setTriggerPosition(x: Int, y: Int, triggerAssembly: ROMAssembly, canUndo: Bool = true) {
        
        // save undo data
        if canUndo {
            if let undo = windowController?.undoManager {
                let target = undo.prepare(withInvocationTarget: self) as AnyObject
                let oldX = triggerAssembly.value(forKey: "x") as! Int
                let oldY = triggerAssembly.value(forKey: "y") as! Int
                target.setTriggerPosition(x: oldX, y: oldY, triggerAssembly: triggerAssembly, canUndo: true)
                if (!undo.isUndoing) {
                    undo.setActionName("Move Trigger")
                }
            }
        }
        
        // set the trigger's position
        if let trigger = triggerForAssembly(triggerAssembly) {
            // trigger is on this map
            windowController?.stopObservingAllObjects()
            trigger.triggerAssembly.setValue(x, forKeyPath: "x")
            trigger.triggerAssembly.setValue(y, forKeyPath: "y")
            windowController?.startObservingAllObjects()
            triggerLayer = nil
            needsDisplay = true
        } else {
            // trigger is not on this map
            triggerAssembly.setValue(x, forKeyPath: "x")
            triggerAssembly.setValue(y, forKeyPath: "y")
        }
    }
    
    @objc func insertTrigger(_ triggerAssembly: ROMAssembly, inArray array: ROMArray, atIndex index: Int) {
        if index == NSNotFound {
            array.array.add(triggerAssembly)
        } else {
            array.array.insert(triggerAssembly, at: index)
        }
        
        // reload all triggers
        loadTriggers()
        
        // check if the trigger is on the current map
        if let mapTrigger = triggerForAssembly(triggerAssembly) {
            // select the trigger, redraw the cursor and triggers
            selectedTrigger = mapTrigger
            windowController?.deselectAllObjects()
            windowController?.selectObject(mapTrigger, identifierOverride: nil)
            triggerLayer = nil
            needsDisplay = true
        }
        
        // save undo data
        if let undo = windowController?.undoManager {
            let target = undo.prepare(withInvocationTarget: self) as AnyObject
            target.removeTrigger(triggerAssembly, fromArray: array)
            if (!undo.isUndoing) {
                undo.setActionName("Insert Trigger")
            }
        }
    }
    
    @objc func removeTrigger(_ triggerAssembly: ROMAssembly, fromArray array: ROMArray) {
        
        // save undo info
        if let undo = windowController?.undoManager {
            let target = undo.prepare(withInvocationTarget: self) as AnyObject
            target.insertTrigger(triggerAssembly, inArray: array, atIndex: array.array.index(of: triggerAssembly))
            if (!undo.isUndoing) {
                undo.setActionName("Remove Trigger")
            }
        }
        
        // check if the trigger is on the current map
        for trigger in triggers {
            if trigger.triggerAssembly !== triggerAssembly { continue }
            // redraw the cursor and triggers
            mouseMoved(with: NSEvent())
            triggerLayer = nil
            needsDisplay = true
        }
        
        // remove the trigger from the map's trigger array
        array.array.remove(triggerAssembly)
        
        // reload all triggers
        loadTriggers()
    }
    
    @IBAction func addTrigger(_ sender: AnyObject) {
        
        guard let menuItem = sender as? NSMenuItem,
            let type = TriggerType(rawValue: menuItem.tag),
            let triggerArrayName = type.arrayName,
            let triggerArray = rom.value(at: mapIndex, inPropertyWithKey: triggerArrayName) as? ROMArray else { return }
        
        // create a new trigger at the cursor position
        guard let triggerAssembly = triggerArray.createNewAssembly() else { return }
        if selectedTiles.count < 5 { return }
        let x = Int(selectedTiles[1])
        let y = Int(selectedTiles[2])
        triggerAssembly.setValue(x, forKeyPath: "x")
        triggerAssembly.setValue(y, forKeyPath: "y")
        
        insertTrigger(triggerAssembly, inArray: triggerArray, atIndex: NSNotFound)
    }
    
    @IBAction func deleteTrigger(_ sender: AnyObject) {
        
        // delete the selected trigger
        let triggerToRemove: FF6MapTrigger! = sender as? FF6MapTrigger ?? selectedTrigger
        if triggerToRemove == nil { return }
        
        // remove the trigger from the rom array
        guard let triggerArrayName = triggerToRemove.type.arrayName,
            let triggerArray = rom.value(at: mapIndex, inPropertyWithKey: triggerArrayName) as? ROMArray else { return }
        removeTrigger(triggerToRemove.triggerAssembly, fromArray: triggerArray)
    }

}

// wrapper class for map tile properties

class FF6MapTileProperties: NSObject, ROMSelectable {
    
    // MARK: ROMSelectable properties
    
    var identifier: String? { return "mapTilePropertiesWrapper" }
    var properties: NSDictionary? { return nil }
    var nib: String? { return nil }
    
    // MARK: member variables
    
    let rom: FF6ROM
    let byte1: ROMData
    let byte2: ROMData
    
    @objc dynamic var isImpassable: Bool {
        get {
            var rawValue: UInt16 = 0
            byte1.data.withUnsafeBytes{(bytes: UnsafePointer<UInt8>)->Void in
                rawValue = UInt16(bytes[0])
            }
            byte2.data.withUnsafeBytes{(bytes: UnsafePointer<UInt8>)->Void in
                rawValue |= UInt16(bytes[0]) << 8
            }
            return rawValue == 0xFFF7
        }
        set {
            stopObservingTile()
            willChangeValue(forKey: "isCounter")

            guard let byte1Properties = byte1.properties else { return }
            guard let byte2Properties = byte2.properties else { return }

            for (key, _) in byte1Properties {
                guard let identifier = key as? String else { continue }
                byte1.willChangeValue(forKey: identifier)
            }
            
            for (key, _) in byte2Properties {
                guard let identifier = key as? String else { continue }
                byte2.willChangeValue(forKey: identifier)
            }

            var rawValue1 = newValue ? 0xF7 : 0
            var rawValue2 = newValue ? 0xFF : 0
            
            byte1.data = Data(bytes: &rawValue1, count: 1)
            byte2.data = Data(bytes: &rawValue2, count: 1)
            
            for (key, _) in byte1Properties {
                guard let identifier = key as? String else { continue }
                byte1.didChangeValue(forKey: identifier)
            }
            
            for (key, _) in byte2Properties {
                guard let identifier = key as? String else { continue }
                byte2.didChangeValue(forKey: identifier)
            }

            didChangeValue(forKey: "isCounter")
            startObservingTile()
        }
    }
    
    @objc dynamic var isCounter: Bool {
        get {
            var rawProperties: UInt16 = 0
            byte1.data.withUnsafeBytes{(bytes: UnsafePointer<UInt8>)->Void in
                rawProperties = UInt16(bytes[0])
            }
            byte2.data.withUnsafeBytes{(bytes: UnsafePointer<UInt8>)->Void in
                rawProperties |= UInt16(bytes[0]) << 8
            }
            return rawProperties == 0x0007
        }
        set {
            stopObservingTile()
            willChangeValue(forKey: "isImpassable")
            
            guard let byte1Properties = byte1.properties else { return }
            guard let byte2Properties = byte2.properties else { return }
            
            for (key, _) in byte1Properties {
                guard let identifier = key as? String else { continue }
                byte1.willChangeValue(forKey: identifier)
            }
            
            for (key, _) in byte2Properties {
                guard let identifier = key as? String else { continue }
                byte2.willChangeValue(forKey: identifier)
            }
            
            var rawValue1 = newValue ? 0x07 : 0
            var rawValue2 = 0
            
            byte1.data = Data(bytes: &rawValue1, count: 1)
            byte2.data = Data(bytes: &rawValue2, count: 1)
            
            for (key, _) in byte1Properties {
                guard let identifier = key as? String else { continue }
                byte1.didChangeValue(forKey: identifier)
            }
            
            for (key, _) in byte2Properties {
                guard let identifier = key as? String else { continue }
                byte2.didChangeValue(forKey: identifier)
            }
            
            didChangeValue(forKey: "isImpassable")
            startObservingTile()
        }
    }
    
    init(rom: FF6ROM, byte1: ROMData, byte2: ROMData) {
        self.rom = rom
        self.byte1 = byte1
        self.byte2 = byte2
        
        super.init()
        
        startObservingTile()
    }
    
    deinit {
        stopObservingTile()
    }
    
    func startObservingTile() {
        
        guard let byte1Properties = byte1.properties else { return }
        guard let byte2Properties = byte2.properties else { return }
        
        for (key, _) in byte1Properties {
            guard let identifier = key as? String else { continue }
            byte1.addObserver(self, forKeyPath: identifier, options: NSKeyValueObservingOptions.prior, context: &FF6MapTileProperties.context)
        }

        for (key, _) in byte2Properties {
            guard let identifier = key as? String else { continue }
            byte2.addObserver(self, forKeyPath: identifier, options: NSKeyValueObservingOptions.prior, context: &FF6MapTileProperties.context)
        }
    }

    func stopObservingTile() {
        
        guard let byte1Properties = byte1.properties else { return }
        guard let byte2Properties = byte2.properties else { return }
        
        for (key, _) in byte1Properties {
            guard let identifier = key as? String else { continue }
            byte1.removeObserver(self, forKeyPath: identifier, context: &FF6MapTileProperties.context)
        }
        
        for (key, _) in byte2Properties {
            guard let identifier = key as? String else { continue }
            byte2.removeObserver(self, forKeyPath: identifier, context: &FF6MapTileProperties.context)
        }
    }

    override func observeValue(forKeyPath keyPath: String?, of object: Any?, change: [NSKeyValueChangeKey : Any]?, context: UnsafeMutableRawPointer?) {
        if context != &FF6MapTileProperties.context {
            super.observeValue(forKeyPath: keyPath, of: object, change: change, context: context)
        }
        
        // notify observers that impassable or counter may have changed
        if change?[NSKeyValueChangeKey.notificationIsPriorKey] != nil {
            willChangeValue(forKey: "isImpassable")
            willChangeValue(forKey: "isCounter")
        } else {
            didChangeValue(forKey: "isImpassable")
            didChangeValue(forKey: "isCounter")
        }
    }
    
    static var context = 0
}
