//
//  FF6MapLayer.swift
//  FF6Tools
//
//  Created by Evan Owen on 7/30/15.
//
//

import Foundation

enum FF6MapLayerType {
    case bg1
    case bg2
    case bg3
    case overlay
    case world
}

struct FF6MapLayer {
    var type = FF6MapLayerType.bg1
    var layoutIndex = 0
    var layoutData: ROMData?
    var layout: Data!
    var graphics: Data!
    var tileset: Data!
    var tilesetData: ROMData?
    var paletteAssignmentData: ROMData?
    var tiles: Data!
    var width = 16
    var height = 16
    var horizontalShift: Int8 = 0
    var verticalShift: Int8 = 0
    var format: Int32 = 0
    var bg3Priority: Bool = false

    mutating func decodeLayout() {
        switch type {
        case .bg1, .bg2:    decodeNormalLayout()
        case .bg3:          decodeBG3Layout()
        case .overlay:      decodeOverlayLayout()
        case .world:        decodeWorldLayout()
        }
    }

    fileprivate mutating func decodeNormalLayout() {
        
        // get the tileset
        let tilesetLength = 0x0800
        var tilesetBytes = [UInt8](repeating: 0, count: tilesetLength)
        tileset.copyBytes(to: &tilesetBytes, count: tilesetLength)
        
        // get the layout
        let layoutLength = height * width
        var layoutBytes = [UInt8](repeating: 0, count: layoutLength)
        layout.copyBytes(to: &layoutBytes, count: layoutLength)
        
        // read data for each tile
        let tileDataLength = height * width * 4 * MemoryLayout<UInt16>.size
        var tileData = [UInt8](repeating: 0, count: tileDataLength)
        let mask = width - 1
        for (index, tile) in layoutBytes.enumerated() {
            if index > layout.count { break }
            let intTile = Int(tile)
            var tileOffset = (index & mask) << 2
            tileOffset += (index & ~mask) << 3
            tileData[tileOffset + 0] = tilesetBytes[intTile + 0x0000]
            tileData[tileOffset + 1] = tilesetBytes[intTile + 0x0400]
            tileData[tileOffset + 2] = tilesetBytes[intTile + 0x0100]
            tileData[tileOffset + 3] = tilesetBytes[intTile + 0x0500]
            tileOffset += (width << 2)
            tileData[tileOffset + 0] = tilesetBytes[intTile + 0x0200]
            tileData[tileOffset + 1] = tilesetBytes[intTile + 0x0600]
            tileData[tileOffset + 2] = tilesetBytes[intTile + 0x0300]
            tileData[tileOffset + 3] = tilesetBytes[intTile + 0x0700]
        }
        tiles = Data(bytes: UnsafePointer<UInt8>(tileData), count: tileDataLength)
    }
    
    fileprivate mutating func decodeBG3Layout() {
        
        // get the tileset
        let tilesetLength = 0x40
        var tilesetBytes = [UInt8](repeating: 0, count: tilesetLength)
        tileset.copyBytes(to: &tilesetBytes, count: tilesetLength)
        
        // get the layout
        let layoutLength = height * width
        var layoutBytes = [UInt8](repeating: 0, count: layoutLength)
        layout.copyBytes(to: &layoutBytes, count: layoutLength)
        
        // read data for each tile
        let tileDataLength = layoutLength * 4 * MemoryLayout<UInt16>.size
        var tileData = [UInt8](repeating: 0, count: tileDataLength)
        let mask = width - 1
        for (index, tile) in layoutBytes.enumerated() {
            if index > layout.count { break }
            var tileOffset = (index & mask) << 2
            tileOffset += ((index & ~mask) << 3)
            var hiByte: UInt8 = 0

            if format == SNES_2BPP_TILE {
                hiByte = tilesetBytes[Int(tile) & 0x3F] & 0x1C
                if bg3Priority { hiByte |= 0x20 }
                hiByte |= (tile & 0xC0)
            } else if format == GBA_4BPP_TILE {
                hiByte = tilesetBytes[Int(tile) & 0x3F] & 0x70
                if bg3Priority { hiByte |= 0x80 }
                hiByte |= (tile & 0xC0) >> 4
            }

            switch (tile & 0xC0) {
            case 0x00:
                // no flip
                tileData[tileOffset + 0] = (tile & 0x3F) << 2
                tileData[tileOffset + 1] = hiByte
                tileData[tileOffset + 2] = (tile & 0x3F) << 2 + 1
                tileData[tileOffset + 3] = hiByte
                tileOffset += (width << 2)
                tileData[tileOffset + 0] = (tile & 0x3F) << 2 + 2
                tileData[tileOffset + 1] = hiByte
                tileData[tileOffset + 2] = (tile & 0x3F) << 2 + 3
                tileData[tileOffset + 3] = hiByte
            case 0x40:
                // horizontal flip
                tileData[tileOffset + 0] = (tile & 0x3F) << 2 + 1
                tileData[tileOffset + 1] = hiByte
                tileData[tileOffset + 2] = (tile & 0x3F) << 2
                tileData[tileOffset + 3] = hiByte
                tileOffset += (width << 2)
                tileData[tileOffset + 0] = (tile & 0x3F) << 2 + 3
                tileData[tileOffset + 1] = hiByte
                tileData[tileOffset + 2] = (tile & 0x3F) << 2 + 2
                tileData[tileOffset + 3] = hiByte
            case 0x80:
                // vertical flip
                tileData[tileOffset + 0] = (tile & 0x3F) << 2 + 2
                tileData[tileOffset + 1] = hiByte
                tileData[tileOffset + 2] = (tile & 0x3F) << 2 + 3
                tileData[tileOffset + 3] = hiByte
                tileOffset += (width << 2)
                tileData[tileOffset + 0] = (tile & 0x3F) << 2
                tileData[tileOffset + 1] = hiByte
                tileData[tileOffset + 2] = (tile & 0x3F) << 2 + 1
                tileData[tileOffset + 3] = hiByte
            case 0xC0:
                // horizontal and vertical flip
                tileData[tileOffset + 0] = (tile & 0x3F) << 2 + 3
                tileData[tileOffset + 1] = hiByte
                tileData[tileOffset + 2] = (tile & 0x3F) << 2 + 2
                tileData[tileOffset + 3] = hiByte
                tileOffset += (width << 2)
                tileData[tileOffset + 0] = (tile & 0x3F) << 2 + 1
                tileData[tileOffset + 1] = hiByte
                tileData[tileOffset + 2] = (tile & 0x3F) << 2 + 0
                tileData[tileOffset + 3] = hiByte
            default:
                break
            }
        }
        tiles = Data(bytes: UnsafePointer<UInt8>(tileData), count: tileDataLength)
    }
    
    fileprivate mutating func decodeOverlayLayout() {
        
        // get the tileset
        let tilesetLength = 0x100
        var tilesetBytes = [UInt8](repeating: 0, count: tilesetLength)
        tileset.copyBytes(to: &tilesetBytes, count: tilesetLength)
        
        // get the layout
        let layoutLength = height * width
        var layoutBytes = [UInt8](repeating: 0, count: layoutLength)
        layout.copyBytes(to: &layoutBytes, count: layoutLength)
        
        // read data for each tile
        let tileDataLength = layoutLength * 4 * MemoryLayout<UInt16>.size
        var tileData = [UInt8](repeating: 0, count: tileDataLength)
        let mask = width - 1
        for (index, bg1Tile) in layoutBytes.enumerated() {
            if index > layout.count { break }
            var overlayTile = tilesetBytes[Int(bg1Tile)]
            if overlayTile == 0xFF { continue }
            var tileOffset = (index & mask) << 2
            tileOffset += ((index & ~mask) << 3)
            var hiByte: UInt8 = 0x08 // mask palette 2 (green)
            hiByte |= (overlayTile & 0xC0)
            let temp = (overlayTile & 0x0E) << 1
            overlayTile &= 0x20
            overlayTile |= temp
            overlayTile += 4
            switch (hiByte & 0xC0) {
            case 0x00:
                // no flip
                tileData[tileOffset + 0] = overlayTile
                tileData[tileOffset + 1] = hiByte
                tileData[tileOffset + 2] = overlayTile + 1
                tileData[tileOffset + 3] = hiByte
                tileOffset += (width << 2)
                tileData[tileOffset + 0] = overlayTile + 2
                tileData[tileOffset + 1] = hiByte
                tileData[tileOffset + 2] = overlayTile + 3
                tileData[tileOffset + 3] = hiByte
            case 0x40:
                // horizontal flip
                tileData[tileOffset + 0] = overlayTile + 1
                tileData[tileOffset + 1] = hiByte
                tileData[tileOffset + 2] = overlayTile
                tileData[tileOffset + 3] = hiByte
                tileOffset += (width << 2)
                tileData[tileOffset + 0] = overlayTile + 3
                tileData[tileOffset + 1] = hiByte
                tileData[tileOffset + 2] = overlayTile + 2
                tileData[tileOffset + 3] = hiByte
            case 0x80:
                // vertical flip
                tileData[tileOffset + 0] = overlayTile + 2
                tileData[tileOffset + 1] = hiByte
                tileData[tileOffset + 2] = overlayTile + 3
                tileData[tileOffset + 3] = hiByte
                tileOffset += (width << 2)
                tileData[tileOffset + 0] = overlayTile
                tileData[tileOffset + 1] = hiByte
                tileData[tileOffset + 2] = overlayTile + 1
                tileData[tileOffset + 3] = hiByte
            case 0xC0:
                // horizontal and vertical flip
                tileData[tileOffset + 0] = overlayTile + 3
                tileData[tileOffset + 1] = hiByte
                tileData[tileOffset + 2] = overlayTile + 2
                tileData[tileOffset + 3] = hiByte
                tileOffset += (width << 2)
                tileData[tileOffset + 0] = overlayTile + 1
                tileData[tileOffset + 1] = hiByte
                tileData[tileOffset + 2] = overlayTile
                tileData[tileOffset + 3] = hiByte
            default:
                break
            }
        }
        tiles = Data(bytes: UnsafePointer<UInt8>(tileData), count: tileDataLength)
    }
    
    fileprivate mutating func decodeWorldLayout() {
        // get the tileset
        let tilesetLength = 0x0400
        var tilesetBytes = [UInt8](repeating: 0, count: tilesetLength)
        tilesetData?.data.copyBytes(to: &tilesetBytes, count: tilesetLength)
//        tileset.copyBytes(to: &tilesetBytes, count: tilesetLength)

        // get the palette assignments
        var tilePalettes = [UInt8](repeating: 0, count: 0x0100)
        if format == SNES_4BPP_TILE {
            let paletteDataLength = 0x80
            var paletteBytes = [UInt8](repeating: 0, count: paletteDataLength)
//            tileset.copyBytes(to: &paletteBytes, from: 0x2400..<0x2480)
            paletteAssignmentData?.data.copyBytes(to: &paletteBytes, count: paletteDataLength)
            for (index, palette) in paletteBytes.enumerated() {
                tilePalettes [index << 1] = (palette & 0x0F) << 2
                tilePalettes [index << 1 + 1] = (palette & 0xF0) >> 2
            }
        }
        
        // get the layout
        let layoutLength = height * width
        var layoutBytes = [UInt8](repeating: 0, count: layoutLength)
        layout.copyBytes(to: &layoutBytes, count: layoutLength)
        
        // read data for each tile
        let tileDataLength = height * width * 4 * MemoryLayout<UInt16>.size
        var tileData = [UInt8](repeating: 0, count: tileDataLength)
        let mask = width - 1
        for (index, tile) in layoutBytes.enumerated() {
            if index > layout.count { break }
            let intTile = Int(tile) << 2
            var tileIndex: UInt8!
            var tileOffset = (index & mask) << 2
            tileOffset += (index & ~mask) << 3
            
            tileIndex = tilesetBytes[intTile + 0]
            tileData[tileOffset + 0] = tileIndex
            tileData[tileOffset + 1] = tilePalettes[Int(tileIndex)]

            tileIndex = tilesetBytes[intTile + 1]
            tileData[tileOffset + 2] = tileIndex
            tileData[tileOffset + 3] = tilePalettes[Int(tileIndex)]

            tileOffset += (width << 2)
            
            tileIndex = tilesetBytes[intTile + 2]
            tileData[tileOffset + 0] = tileIndex
            tileData[tileOffset + 1] = tilePalettes[Int(tileIndex)]
            
            tileIndex = tilesetBytes[intTile + 3]
            tileData[tileOffset + 2] = tileIndex
            tileData[tileOffset + 3] = tilePalettes[Int(tileIndex)]
        }
        tiles = Data(bytes: UnsafePointer<UInt8>(tileData), count: tileDataLength)
    }

    mutating func decodePartialWorldLayout(_ layout: Data, tileRange: Range<Int>) {
        // get the tileset
        let tilesetLength = 0x0400
        var tilesetBytes = [UInt8](repeating: 0, count: tilesetLength)
        tilesetData?.data.copyBytes(to: &tilesetBytes, count: tilesetLength)
//        tileset.copyBytes(to: &tilesetBytes, count: tilesetLength)
        
        // get the palette assignments
        var tilePalettes = [UInt8](repeating: 0, count: 0x0100)
        if format == SNES_4BPP_TILE {
            let paletteDataLength = 0x80
            var paletteBytes = [UInt8](repeating: 0, count: paletteDataLength)
//            tileset.copyBytes(to: &paletteBytes, from: 0x2400..<(0x2400 + paletteDataLength))
            paletteAssignmentData?.data.copyBytes(to: &paletteBytes, count: paletteDataLength)
            for (index, palette) in paletteBytes.enumerated() {
                tilePalettes [index << 1] = (palette & 0x0F) << 2
                tilePalettes [index << 1 + 1] = (palette & 0xF0) >> 2
            }
        }
        
        // get the layout
        let layoutLength = tileRange.count
        var layoutBytes = [UInt8](repeating: 0, count: layoutLength)
        layout.copyBytes(to: &layoutBytes, count: layoutLength)
//        (layout as NSData).getBytes(&layoutBytes, length: layoutLength)
        
        // read data for each tile
        let tileDataLength = height * width * 4 * MemoryLayout<UInt16>.size
        var tileData = [UInt8](repeating: 0, count: tileDataLength)

        // get the existing tile bytes
        tiles.copyBytes(to: &tileData, count: tileDataLength)

        let mask = width - 1
        for (var index, tile) in layoutBytes.enumerated() {
            if index > layout.count { break }
            index += tileRange.lowerBound
            let intTile = Int(tile) << 2
            var tileIndex: UInt8!
            var tileOffset = (index & mask) << 2
            tileOffset += (index & ~mask) << 3
            
            tileIndex = tilesetBytes[intTile + 0]
            tileData[tileOffset + 0] = tileIndex
            tileData[tileOffset + 1] = tilePalettes[Int(tileIndex)]
            
            tileIndex = tilesetBytes[intTile + 1]
            tileData[tileOffset + 2] = tileIndex
            tileData[tileOffset + 3] = tilePalettes[Int(tileIndex)]
            
            tileOffset += (width << 2)
            
            tileIndex = tilesetBytes[intTile + 2]
            tileData[tileOffset + 0] = tileIndex
            tileData[tileOffset + 1] = tilePalettes[Int(tileIndex)]
            
            tileIndex = tilesetBytes[intTile + 3]
            tileData[tileOffset + 2] = tileIndex
            tileData[tileOffset + 3] = tilePalettes[Int(tileIndex)]
        }
        tiles = Data(bytes: UnsafePointer<UInt8>(tileData), count: tileDataLength)
    }
}
