//
//  FF6MapPreview.swift
//  FF6Tools
//
//  Created by Evan Owen on 7/31/16.
//
//

import Cocoa

class FF6MapPreview: FF6MapView {

    @objc var assemblyIdentifier: String?
    @objc var assembly: AnyObject?
    @objc var mapIndexKeyPath: String?
    @objc var xKeyPath: String?
    @objc var yKeyPath: String?
    override var zoom: CGFloat { get { return max(bounds.width / 240, bounds.height / 208) } set { } }
    
    override func awakeFromNib() {
        if windowController == nil || assemblyIdentifier == nil || mapIndexKeyPath == nil { return }
        bind(NSBindingName(rawValue: "mapIndex"), to: windowController!, withKeyPath: mapIndexKeyPath!, options: nil)
        bind(NSBindingName(rawValue: "assembly"), to: windowController!, withKeyPath: assemblyIdentifier!, options: nil)
    }
    
    func getPosition() -> (x: Int, y: Int)? {
        if windowController == nil || assemblyIdentifier == nil || xKeyPath == nil || yKeyPath == nil { return nil }
            guard let x = assembly?.value(forKeyPath: xKeyPath!) as? Int,
            let y = assembly?.value(forKeyPath: yKeyPath!) as? Int else { return nil }

        return (x,y)
    }
    
    @objc func setPosition(_ x: Int, _ y: Int, forObject object: AnyObject?, canUndo: Bool) {
        if windowController == nil || assemblyIdentifier == nil || xKeyPath == nil || yKeyPath == nil { return }
        
        // if assembly is nil, use the assembly bound to the window controller
        let assembly = object ?? self.assembly
        
        guard let oldX = assembly?.value(forKeyPath: xKeyPath!) as? Int,
            let oldY = assembly?.value(forKeyPath: yKeyPath!) as? Int else { return }
        
        // save undo data
        if canUndo {
            if let undo = windowController?.undoManager {
                let target = undo.prepare(withInvocationTarget: self) as AnyObject
                target.setPosition(oldX, oldY, forObject: assembly, canUndo: true)
                if (!undo.isUndoing) {
                    undo.setActionName("Move Trigger Destination")
                }
            }
        }

        // set the new position
        windowController?.stopObserving(object: assembly)
        assembly?.setValue(x, forKeyPath: xKeyPath!)
        assembly?.setValue(y, forKeyPath: yKeyPath!)
        windowController?.startObserving(object: assembly)
        needsDisplay = true
    }
    
    override func draw(_ dirtyRect: NSRect) {
        
        // get the cursor position
        let position = getPosition() ?? (7,6)
        
        // fill the back area with black
        NSColor.black.set()
        NSBezierPath.fill(dirtyRect)
        NSGraphicsContext.current?.imageInterpolation = imageInterpolation
        
        // draw the map
        if mapLayer == nil { updateMapLayerInRect(NSZeroRect) }
        let mapSize = mapLayer?.size
        let mapRect = NSMakeRect(zoom * 16 * CGFloat(7 - position.x), 16 * zoom * (CGFloat(position.y) + 6.5) - (mapSize?.height)! * zoom, (mapSize?.width)! * zoom, (mapSize?.height)! * zoom)
        NSGraphicsContext.current?.cgContext.draw(mapLayer!, in: mapRect)

        // draw the triggers
        if triggerLayer == nil { updateTriggerLayer() }
        let triggerLayerSize = triggerLayer?.size
        let triggerLayerRect = NSMakeRect(zoom * 16 * CGFloat(7 - position.x), 16 * zoom * (CGFloat(position.y) + 6.5) - (triggerLayerSize?.height)!, (triggerLayerSize?.width)!, (triggerLayerSize?.height)!)
        NSGraphicsContext.current?.cgContext.draw(triggerLayer!, in: triggerLayerRect)

        // draw the cursor
        showCursor = true
        drawCursor(x: 7 * 16, y: 7 * 16 - 8, width: 16, height: 16, color: NSColor.orange)
    }
    
    override func setFrameSize(_ newSize: NSSize) {
        // redraw the trigger any time the frame size changes
        triggerLayer = nil
        super.setFrameSize(newSize)
    }
    
    override func mouseDown(with theEvent: NSEvent) {
        var event = theEvent
        let position = getPosition() ?? (7,6)
        
        // get the point that was clicked
        let clickedPoint = convert(event.locationInWindow, from: nil)
        if !bounds.contains(clickedPoint) { return }
        
        // save the original position
        let oldX = position.x
        let oldY = position.y
        
        var positionHasNotChanged = true
        while true {
            
            switch event.type {
            case .leftMouseUp:
                return
                
            case .leftMouseDragged, .leftMouseDown:
                
                // get the dragged/clicked point
                let draggedPoint = convert(event.locationInWindow, from: nil)
                
                // get the delta from the original clicked point
                let deltaX = (draggedPoint.x - clickedPoint.x) / zoom / 8.0
                let deltaY = (draggedPoint.y - clickedPoint.y) / zoom / 8.0
                
                // limit the position to valid values
                let newX = min(255, max(oldX + Int(deltaX), 0))
                let newY = min(255, max(oldY - Int(deltaY), 0))
                
                // break if the position didn't change
                if (newX == position.x) && (newY == position.y) && positionHasNotChanged { break }
                
                // set the new position
                setPosition(newX, newY, forObject: nil, canUndo: positionHasNotChanged)
                positionHasNotChanged = false
                
            default:
                break
            }
            event = window!.nextEvent(matching: [.leftMouseUp, .leftMouseDragged])!
        }
    }

}
