//
//  FF6MapTilesetView.swift
//  FF6Tools
//
//  Created by Evan Owen on 7/12/15.
//
//

import Cocoa

class FF6MapTilesetView: FF6View {
    
    @IBOutlet weak var mapEditor: FF6MapEditor!

    var zoom: CGFloat { return frame.width / 256.0 }
    var tilesetLayer: CGLayer?
    var maskLayer: CGLayer?
    var selectedTiles: [UInt8] = [0x73, 0, 0, 1, 1, 0] {
        didSet {
            needsDisplay = true
        }
    }
    var layerIndex: Int { return mapEditor.selectedLayerIndex }
    
    func updateTilesetLayer() {
        
        var layer = mapEditor.selectedLayer
        layer.width = 16
        layer.height = 16
        let layout = (0..<256).map{ UInt8($0) }
        layer.layout = Data(bytes: layout, count: 256)
        layer.decodeLayout()

        // create the ppu
        var ppu = PPU()
        memset(&ppu, 0, MemoryLayout<PPU>.size)
        ppu.pal = (mapEditor.mapPalette! as NSData).bytes
        
        ppu.layers.0.format = layer.format
        ppu.layers.0.height = 32
        ppu.layers.0.width = 32
        ppu.layers.0.z.0 = Z_LEVEL_BG1_LO
        ppu.layers.0.z.1 = Z_LEVEL_BG1_HI
        ppu.layers.0.gfx = (layer.graphics as NSData).bytes
        ppu.layers.0.tiles = (layer.tiles as NSData).bytes
        ppu.layers.0.main = true
        
        // render the screen
        let dataLength = 256 * 256 * MemoryLayout<UInt32>.size
        let data = UnsafeMutablePointer<UInt8>.allocate(capacity: dataLength)
        memset(data, 0, dataLength)
        gfx_render_ppu(data, ppu, Int32(dataLength), 256)
        
        // create an image
        let provider = CGDataProvider(dataInfo: nil, data: data, size: dataLength, releaseData: releaseImageData)!
        let bitsPerComponent = 8
        let bitsPerPixel = 32
        let bytesPerRow = 256 * MemoryLayout<UInt32>.size
        let colorSpaceRef = CGColorSpaceCreateDeviceRGB()
        let bitmapInfo = CGBitmapInfo(rawValue: CGBitmapInfo().rawValue | CGImageAlphaInfo.premultipliedLast.rawValue)
        let iref = CGImage(width: 256, height: 256, bitsPerComponent: bitsPerComponent, bitsPerPixel: bitsPerPixel, bytesPerRow: bytesPerRow, space: colorSpaceRef, bitmapInfo: bitmapInfo, provider: provider, decode: nil, shouldInterpolate: true, intent: .defaultIntent)
        
        // draw the image
        let imageRect = CGRect(x: 0, y: 0, width: 256, height: 256)
        if let currentContext = NSGraphicsContext.current?.cgContext {
            tilesetLayer = CGLayer(currentContext, size: imageRect.size, auxiliaryInfo: nil)
            tilesetLayer?.context?.draw(iref!, in: imageRect)
        }
        
        // free memory
        data.deinitialize(); data.deallocate(capacity: dataLength)
    }
    
    func updateMaskLayer() {
        
        if mapEditor.tileMask == .noMask { return }
        if mapEditor.mapIndex < 3 { return }
        
        let height = 16
        let width = 16

        // create the mask layer if it is not valid
        if maskLayer == nil {
            if let currentContext = NSGraphicsContext.current?.cgContext {
                maskLayer = CGLayer(currentContext, size: CGSize(width: width * 16, height: height * 16), auxiliaryInfo: nil)
            } else {
                // unable to get current context, this was probably not called from drawRect
                rom.log("Unable to create mask layer")
                return
            }
        }
        let context = maskLayer?.context
        
        // get tile properties
        guard let tilePropertiesIndex = mapEditor.mapProperties?.intValue(forKey: "tileProperties"),
            let tilePropertiesData = rom.value(at: tilePropertiesIndex, inPropertyWithKey: "mapTileProperties") as? ROMData else { return }
        
        // assemble tile properties data to capture any changes that have been made
        var dummyData = Data()
        tilePropertiesData.assembleTo(&dummyData)
        let tilePropertiesBytes: [UInt8] = tilePropertiesData.data.withUnsafeBytes{ [UInt8](UnsafeBufferPointer(start: $0, count: tilePropertiesData.data.count)) }
        
        // create the mask graphics
        guard let maskGraphicsPath = Bundle.main.path(forResource: "MaskGraphics", ofType: "bin"),
            let maskGraphics = NSData(contentsOfFile: maskGraphicsPath) else { return }
        
        // create the ppu
        var ppu = PPU()
        memset(&ppu, 0, MemoryLayout<PPU>.size)
        
        // create the mask palette
        let palette: [UInt32] = [
            0, 0x80000000, 0x40000000, 0x20000000, // 0 = black
            0, 0x800000FF, 0x400000FF, 0x200000FF, // 1 = red
            0, 0x8000FF00, 0x4000FF00, 0x2000FF00, // 2 = green
            0, 0x8000FFFF, 0x4000FFFF, 0x2000FFFF, // 3 = yellow
            0, 0x80FF0000, 0x40FF0000, 0x20FF0000, // 4 = blue
            0, 0x80FF00FF, 0x40FF00FF, 0x20FF00FF, // 5 = magenta
            0, 0x80FFFF00, 0x40FFFF00, 0x20FFFF00, // 6 = cyan
            0, 0x80FFFFFF, 0x40FFFFFF, 0x20FFFFFF  // 7 = white
        ]
        ppu.pal = UnsafeRawPointer(palette)
        
        // set up the ppu layer
        ppu.layers.0.format = SNES_2BPP_TILE
        ppu.layers.0.height = Int32(height * 2)
        ppu.layers.0.width = Int32(width * 2)
        ppu.layers.0.v_scroll = 0
        ppu.layers.0.h_scroll = 0
        ppu.layers.0.z.0 = Z_LEVEL_MASK
        ppu.layers.0.z.1 = Z_LEVEL_MASK
        ppu.layers.0.main = true
        
        // set the ppu graphics and tiles based on the selected mask
        var maskTiles = [UInt16](repeating: 0, count: height * width * 4)
        let layout = (0..<256).map{ UInt8($0) }
        switch mapEditor.tileMask {
        case .noMask: return
        case .spriteOverlay:
            // set up overlay
            var overlay = mapEditor.overlay
            overlay.width = width
            overlay.height = height
            overlay.layout = Data(bytes: layout, count: 256)
            overlay.decodeLayout()

            ppu.layers.0.gfx = (overlay.graphics as NSData).bytes
            ppu.layers.0.tiles = (overlay.tiles as NSData).bytes
            ppu.layers.0.height = Int32(height * 2)
            ppu.layers.0.width = Int32(width * 2)
            ppu.layers.0.v_scroll = 0
            ppu.layers.0.h_scroll = 0
            
        case .passability:
            
            for row in 0..<height {
                for col in 0..<width {
                    
                    if row * 16 + col >= layout.count { break }
                    if row * width + col >= maskTiles.count / 4 { break }
                    
                    let layoutTile = layout[row * width + col]
                    
                    let byte1 = tilePropertiesBytes[Int(layoutTile)]
                    let byte2 = tilePropertiesBytes[Int(layoutTile) + 256]
                    let tileProperties = Int(byte1) | Int(byte2) << 8
                    
                    // skip impassable tiles
                    if tileProperties == 0xFFF7 { continue }
                    
                    // set the tile based on passability
                    var tiles = [UInt16]()
                    switch byte2 & 0x0F {
                    case 0x01: tiles = [4, 5, 4, 5]
                    case 0x02: tiles = [5, 4, 5, 4]
                    case 0x03: tiles = [5, 5, 5, 5]
                    case 0x04: tiles = [4, 4, 6, 6]
                    case 0x05: tiles = [4, 5, 6, 7]
                    case 0x06: tiles = [5, 4, 7, 6]
                    case 0x07: tiles = [5, 5, 7, 7]
                    case 0x08: tiles = [6, 6, 4, 4]
                    case 0x09: tiles = [6, 7, 4, 5]
                    case 0x0A: tiles = [7, 6, 5, 4]
                    case 0x0B: tiles = [7, 7, 5, 5]
                    case 0x0C: tiles = [6, 6, 6, 6]
                    case 0x0D: tiles = [6, 7, 6, 7]
                    case 0x0E: tiles = [7, 6, 7, 6]
                    case 0x0F: tiles = [7, 7, 7, 7]
                    default:   tiles = [0, 0, 0, 0]
                    }
                    
                    // set v/h flip for each tile
                    tiles[1] |= 0x4000
                    tiles[2] |= 0x8000
                    tiles[3] |= 0xC000
                    
                    // stairs tile
                    switch byte1 & 0xC0 {
                    case 0x40: tiles = [0x0008, 0x0009, 0x000A, 0x000B] // up/right
                    case 0x80: tiles = [0x4009, 0x4008, 0x400B, 0x400A] // up/left
                    default: break
                    }
                    
                    // ladder tile
                    if (byte2 & 0x40) != 0 {
                        tiles = [0x000C, 0x400C, 0x800C, 0xC00C]
                    }
                    
                    // door tile
                    if (byte1 & 0x20) != 0 {
                        tiles = [0x000D, 0x400D, 0x000E, 0x000F]
                    }
                    
                    // counter tile
                    if tileProperties == 0x0007 {
                        tiles = [0x0410, 0x4410, 0x8410, 0xC410]
                    }
                    
                    // set the palette based on z-level
                    var paletteByte: UInt16 = 0
                    switch byte1 & 0x07 {
                    case 1: paletteByte = 0x1000 // upper (blue)
                    case 2: paletteByte = 0x0400 // lower (red)
                    case 3: paletteByte = 0x1400 // stairs (magenta)
                    case 4: paletteByte = 0x0800 // bridge (green)
                    default: break // other
                    }
                    tiles[0] |= paletteByte
                    tiles[1] |= paletteByte
                    tiles[2] |= paletteByte
                    tiles[3] |= paletteByte
                    
                    var tileOffset = col * 2 + row * width * 4
                    
                    maskTiles[tileOffset] = tiles[0]
                    maskTiles[tileOffset + 1] = tiles[1]
                    tileOffset += width * 2
                    maskTiles[tileOffset] = tiles[2]
                    maskTiles[tileOffset + 1] = tiles[3]
                }
            }
            
            ppu.layers.0.gfx = maskGraphics.bytes
            ppu.layers.0.tiles = UnsafeRawPointer(maskTiles)
            
        case .npcPassability:
            
            for row in 0..<height {
                for col in 0..<width {
                    
                    if row * 16 + col >= layout.count { break }
                    if row * width + col >= maskTiles.count / 4 { break }
                    
                    let layoutTile = layout[row * width + col]
                    
                    let byte1 = tilePropertiesBytes[Int(layoutTile)]
                    let byte2 = tilePropertiesBytes[Int(layoutTile) + 256]
                    let tileProperties = Int(byte1) | Int(byte2) << 8
                    
                    // skip impassable and counter tiles
                    if (tileProperties == 0xFFF7) || (tileProperties == 0x0007) { continue }
                    
                    if byte2 & 0x80 == 0 { continue }
                    
                    // set the palette based on z-level
                    var paletteByte: UInt16 = 0
                    switch byte1 & 0x07 {
                    case 1: paletteByte = 0x1000 // upper (blue)
                    case 2: paletteByte = 0x0400 // lower (red)
                    case 3: paletteByte = 0x1400 // stairs (magenta)
                    case 4: paletteByte = 0x0800 // bridge (green)
                    default: break // other
                    }
                    
                    var tileOffset = col * 2 + row * width * 4
                    maskTiles[tileOffset] = 0x0007 | paletteByte
                    maskTiles[tileOffset + 1] = 0x4007 | paletteByte
                    tileOffset += width * 2
                    maskTiles[tileOffset] = 0x8007 | paletteByte
                    maskTiles[tileOffset + 1] = 0xC007 | paletteByte
                }
            }
            
            ppu.layers.0.gfx = maskGraphics.bytes
            ppu.layers.0.tiles = UnsafeRawPointer(maskTiles)
            
        case .layerPriority:
            
            for row in 0..<height {
                for col in 0..<width {
                    
                    if row * 16 + col >= layout.count { break }
                    if row * width + col >= maskTiles.count / 4 { break }
                    
                    let layoutTile = layout[row * width + col]
                    
                    let byte1 = tilePropertiesBytes[Int(layoutTile)]
                    let byte2 = tilePropertiesBytes[Int(layoutTile) + 256]
                    let tileProperties = Int(byte1) | Int(byte2) << 8
                    
                    // skip impassable and counter tiles
                    if (tileProperties == 0xFFF7) || (tileProperties == 0x0007) { continue }
                    
                    var tile: UInt16 = 0x0000
                    switch byte1 & 0x18 {
                    case 0x08: tile = 0x0401 // upper sprite priority (red)
                    case 0x10: tile = 0x0801 // lower sprite priority (green)
                    case 0x18: tile = 0x1401 // upper & lower sprite priority (magenta)
                    default: break
                    }
                    
                    var tileOffset = col * 2 + row * width * 4
                    maskTiles[tileOffset] = tile
                    maskTiles[tileOffset + 1] = tile
                    tileOffset += width * 2
                    maskTiles[tileOffset] = tile
                    maskTiles[tileOffset + 1] = tile
                }
            }
            
            ppu.layers.0.gfx = maskGraphics.bytes
            ppu.layers.0.tiles = UnsafeRawPointer(maskTiles)

        }
        
        // render the ppu
        let imageHeight = height * 16
        let imageWidth = width * 16
        let dataLength = imageHeight * imageWidth * 4
        let data = UnsafeMutablePointer<UInt8>.allocate(capacity: dataLength)
        memset(data, 0, dataLength)
        gfx_render_ppu(data, ppu, Int32(dataLength), Int32(imageWidth))
        
        // create an image
        let provider = CGDataProvider(dataInfo: nil, data: data, size: dataLength, releaseData: releaseImageData)!
        let bitsPerComponent = 8
        let bitsPerPixel = 32
        let bytesPerRow = imageWidth * MemoryLayout<UInt32>.size
        let colorSpaceRef = CGColorSpaceCreateDeviceRGB()
        let bitmapInfo = CGBitmapInfo(rawValue: CGBitmapInfo().rawValue | CGImageAlphaInfo.premultipliedLast.rawValue)
        let renderingIntent = CGColorRenderingIntent.defaultIntent
        let iref = CGImage(width: imageWidth, height: imageHeight, bitsPerComponent: bitsPerComponent, bitsPerPixel: bitsPerPixel, bytesPerRow: bytesPerRow, space: colorSpaceRef, bitmapInfo: bitmapInfo, provider: provider, decode: nil, shouldInterpolate: true, intent: renderingIntent)
        
        // draw the image to the mask layer
        let maskRect = NSMakeRect(0, 0, CGFloat(imageWidth), CGFloat(imageHeight))
        context?.clear(maskRect)
        context?.draw(iref!, in: maskRect)
        
        // free memory
        data.deinitialize(); data.deallocate(capacity: dataLength)
    }

    override func draw(_ dirtyRect: NSRect) {
        super.draw(dirtyRect)

        // fill the rect with the background color
        let tilesetRect = NSMakeRect(0, frame.height - 256 * zoom, 256 * zoom, 256 * zoom)
        backgroundColor.set()
        NSBezierPath.fill(tilesetRect)

        // update the tileset layer
        if layerIndex > 2 { return }
        if tilesetLayer == nil { updateTilesetLayer() }

        // draw the tileset layer
        NSGraphicsContext.current?.imageInterpolation = imageInterpolation
        NSGraphicsContext.current?.cgContext.draw(tilesetLayer!, in: tilesetRect)

        // draw the tile mask
        if mapEditor.tileMask != .noMask && mapEditor.mapIndex >= 3 && mapEditor.selectedLayer.type == .bg1 {
            if maskLayer == nil { updateMaskLayer() }
            if maskLayer != nil {
                NSGraphicsContext.current?.cgContext.draw(maskLayer!, in: tilesetRect)
            }
        }

        // draw the selected tile cursor
        let highlightColor: NSColor
        switch mapEditor.selectedLayer.type {
        case .bg1: highlightColor = NSColor.green
        case .bg2: highlightColor = NSColor.blue
        case .bg3: highlightColor = NSColor.red
        case .world: highlightColor = NSColor.green
        case .overlay: highlightColor = NSColor.white
        }
        drawCursor(highlightColor)
    }
    
    func drawCursor(_ highlightColor: NSColor) {
        // get the selected tiles
        if selectedTiles.count <= 5 { return }
        let x = Int(selectedTiles[1] & 0x7F)
        let y = Int(selectedTiles[2] & 0x3F)
        let width = Int(selectedTiles[3])
        let height = Int(selectedTiles[4])
        if selectedTiles.count != 5 + width * height {
            rom.log("Invalid map layout chunk data size")
            return
        }
        
        // convert the selection to screen coordinates
        let left = round(CGFloat(x) * 16 * zoom) + 0.5
        let right = round(CGFloat(x + width) * 16 * zoom) - 0.5
        let top = round(CGFloat(y) * 16 * zoom) + 0.5
        let bottom = round(CGFloat(y + height) * 16 * zoom) - 0.5
        var cursorRect = NSMakeRect(left, frame.height - bottom, right - left, bottom - top)
        
        // draw the outside border (black)
        NSColor.black.set()
        NSBezierPath.stroke(cursorRect)

        // draw the outer highlight
        cursorRect = cursorRect.insetBy(dx: 1.0, dy: 1.0)
        highlightColor.set()
        NSBezierPath.stroke(cursorRect)

        // draw the inner highlight (white)
        cursorRect = cursorRect.insetBy(dx: 1.0, dy: 1.0)
        NSColor.white.set()
        NSBezierPath.stroke(cursorRect)

        // draw the inner border (black)
        cursorRect = cursorRect.insetBy(dx: 1.0, dy: 1.0)
        NSColor.black.set()
        NSBezierPath.stroke(cursorRect)
    }
    
    override func mouseDown(with theEvent: NSEvent) {
        let clickedPoint = convert(theEvent.locationInWindow, from: nil)
        if !bounds.contains(clickedPoint) { return }
        let clickedX = UInt8(clickedPoint.x / zoom / 16)
        let clickedY = UInt8((bounds.height - clickedPoint.y) / zoom / 16)
        let clickedTile = clickedX + clickedY * 16
        var draggedTile = clickedTile
        selectedTiles = [0x73, clickedX, clickedY, 1, 1, clickedTile]
        mapEditor.selectedTiles = selectedTiles
        
        // select tile properties if a bg1 tile was selected
        mapEditor.selectTileProperties(tile: Int(clickedTile))

        while true {
            let nextEvent = window?.nextEvent(matching: [.leftMouseUp, .rightMouseUp, .leftMouseDragged, .rightMouseDragged])
            if nextEvent == nil { break }
            
            switch (nextEvent!.type) {
            case .leftMouseUp, .rightMouseUp:
                return
                
            case .leftMouseDragged, .rightMouseDragged:
                let draggedPoint = convert(nextEvent!.locationInWindow, from: nil)
                let draggedX = UInt8(min(max(draggedPoint.x / zoom / 16, 0), 15))
                let draggedY = UInt8(min(max((bounds.height - draggedPoint.y) / zoom / 16, 0), 15))
                if draggedTile == draggedX + draggedY * 16 { continue }
                draggedTile = draggedX + draggedY * 16
                
                // get the selected rectangle
                let minX = min(draggedX, clickedX)
                let minY = min(draggedY, clickedY)
                let maxX = max(draggedX, clickedX)
                let maxY = max(draggedY, clickedY)
                let width = maxX - minX + 1
                let height = maxY - minY + 1
                
                // create the new tile selection
                var newSelection = [0x73, minX, minY, width, height]
                for y in 0..<height {
                    for x in 0..<width {
                        newSelection.append(minX + x + (minY + y) * 16)
                    }
                }
                selectedTiles = newSelection
                mapEditor.selectedTiles = selectedTiles
                
            default:
                break
            }
        }
    }
    
    override func rightMouseDown(with theEvent: NSEvent) {
        // left and right click have the same effect
        mouseDown(with: theEvent)
    }
    
}
