//
//  FF6MapView.swift
//  FF6Tools
//
//  Created by Evan Owen on 4/24/15.
//
//

// TODO: only create a new PPU when a map changes

import Cocoa

class FF6MapView: FF6View {
    
    @IBOutlet weak var tilesetView: FF6MapTilesetView?

    // map layers and triggers
    var bg1 = FF6MapLayer()
    var bg2 = FF6MapLayer()
    var bg3 = FF6MapLayer()
    var overlay = FF6MapLayer()
    var triggers = [FF6MapTrigger]()
    
    // map data
    var mapProperties: ROMData?
    var colorMath: ROMData?
    var mapPalette: Data?
    var mapWidth = 0
    var mapHeight = 0
    var parallax = 0
    
    // drawing stuff
    var zoom: CGFloat = 2.0
    var mapLayer: CGLayer?
    var triggerLayer: CGLayer?

    @objc dynamic var showLayer1: Bool = true { didSet { needsDisplay = true; mapLayer = nil } }
    @objc dynamic var showLayer2: Bool = true { didSet { needsDisplay = true; mapLayer = nil } }
    @objc dynamic var showLayer3: Bool = true { didSet { needsDisplay = true; mapLayer = nil } }
    @objc dynamic var showTriggers: Bool = true { didSet { needsDisplay = true; triggerLayer = nil } }
    @objc dynamic var showCursor: Bool = false { didSet { needsDisplay = true } }

    @objc dynamic var mapIndex = 3 {
        didSet {
            loadMap()
            loadTriggers()
            
            // reset the frame size
            if let clipView = superview as? NSClipView {
                setFrameSize(clipView.documentVisibleRect.size)
                if mapIndex != oldValue {
                    // scroll to the origin if the map index changed
                    scroll(NSMakePoint(0.0, frame.height))
                }
            }
        }
    }
    
    // MARK: initialization
    
    func loadTriggers() {
        // stop observing any old triggers
        for trigger in triggers {
            stopObserving(object: trigger.triggerAssembly)
        }
        
        // remove any existing triggers
        triggers.removeAll()
        
        func loadTriggersWithIdentifier(_ identifier: String) {
            guard let triggerArray = rom.value(at: mapIndex, inPropertyWithKey: identifier) as? ROMArray else { return }
            for triggerAssembly in triggerArray.array {
                triggers.append(FF6MapTrigger(triggerArray: triggerArray, triggerAssembly: triggerAssembly as! ROMData))
                startObserving(object: triggerAssembly as AnyObject?)
            }
        }

        func loadTriggersWithIdentifierAdvance(_ identifier: String) {
            if !isGBA { return }
            guard let triggerArray = rom.value(forKey: identifier) as? ROMArray else { return }
            for triggerAssembly in triggerArray.array {
                if (triggerAssembly as AnyObject).value(forKeyPath: "mapIndex") as? Int != mapIndex { continue }
                triggers.append(FF6MapTrigger(triggerArray: triggerArray, triggerAssembly: triggerAssembly as! ROMData))
                startObserving(object: triggerAssembly as AnyObject?)
            }
        }
        
        loadTriggersWithIdentifier("eventTriggers")
        loadTriggersWithIdentifier("shortEntranceTriggers")
        loadTriggersWithIdentifier("longEntranceTriggers")
        loadTriggersWithIdentifier("treasureProperties")
        loadTriggersWithIdentifier("npcProperties")
        loadTriggersWithIdentifierAdvance("eventTriggersAdvance")
        loadTriggersWithIdentifierAdvance("npcPropertiesAdvance")
        
        for trigger in triggers {
            switch trigger.type {

            case .specialNPC:
                // set the special graphics flag
                trigger.triggerAssembly.setValue(1, forKey: "specialGraphics")
                
            case .npc, .event, .worldEvent:
                if mapIndex < 3 && trigger.type == .event {
                    // identify event triggers on world maps
                    trigger.type = .worldEvent
                }
                // create a reference for event triggers and npcs
                if trigger.triggerAssembly.references == nil {
                    let eventScript = rom.value(forKey: "eventScript") as? ROMScript
                    _ = eventScript?.addReference(forAssembly: trigger.triggerAssembly, encodingName: mapIndex < 3 ? "worldScript" : "eventScript", linkedKeyPath: "eventPointer")
                }

            default: break
            }
        }
        
        // redraw triggers
        triggerLayer = nil
        needsDisplay = true
    }
    
    func triggerForAssembly(_ triggerAssembly: ROMAssembly?) -> FF6MapTrigger? {
        
        if triggerAssembly == nil { return nil }
        
        // check if the trigger is on the current map
        for trigger in triggers {
            if trigger.triggerAssembly === triggerAssembly {
                return trigger
            }
        }
        return nil
    }
    
    func loadMap() {
        
        stopObserving(object: mapProperties)
        mapProperties = nil
        stopObserving(object: colorMath)
        colorMath = nil
        stopObserving(object: bg1.layoutData)
        bg1.layoutData = nil
        stopObserving(object: bg2.layoutData)
        bg2.layoutData = nil
        stopObserving(object: bg3.layoutData)
        bg3.layoutData = nil
        
        stopObserving(object: bg1.tilesetData)
        bg1.tilesetData = nil
        stopObserving(object: bg2.tilesetData)
        bg2.tilesetData = nil

        stopObserving(object: bg1.paletteAssignmentData)
        bg1.paletteAssignmentData = nil

        // load the new map
        if mapIndex < 3 {
            loadWorldMap(mapIndex)
        } else {
            loadNormalMap(mapIndex)
        }
        
        // find the largest layer dimensions
        let maxWidth = max(bg1.width, bg2.width, bg3.width)
        let maxHeight = max(bg1.height, bg2.height, bg3.height)
        
        // if the map dimensions are 0 (loop), set the dimensions to match the biggest layer
        if mapWidth == 0 { mapWidth = maxWidth } else { mapWidth += 1 }
        if mapHeight == 0 { mapHeight = maxHeight } else { mapHeight += 1 }
        
        // redraw the map (the whole thing)
        mapLayer = nil
        needsDisplay = true
    }
    
    func loadNormalMap(_ mapIndex: Int) {
        
        // don't try to load parent map or previous map
        if mapIndex >= 510 { return }
        
        // get map properties and map data
        mapProperties = rom.value(at: mapIndex, inPropertyWithKey: "mapProperties") as? ROMData
        startObserving(object: mapProperties)
        guard let graphicsIndex1 = Int(object: mapProperties, keyPath: "graphics1"),
            let graphicsIndex2 = Int(object: mapProperties, keyPath: "graphics2"),
            let graphicsIndex3 = Int(object: mapProperties, keyPath: "graphics3"),
            let graphicsIndex4 = Int(object: mapProperties, keyPath: "graphics4"),
            let bg3GraphicsIndex = Int(object: mapProperties, keyPath: "bg3Graphics"),
            let bg1TilesetIndex = Int(object: mapProperties, keyPath: "bg1Tileset"),
            let bg2TilesetIndex = Int(object: mapProperties, keyPath: "bg2Tileset"),
            let bg1LayoutIndex = Int(object: mapProperties, keyPath: "bg1Layout"),
            let bg2LayoutIndex = Int(object: mapProperties, keyPath: "bg2Layout"),
            let bg3LayoutIndex = Int(object: mapProperties, keyPath: "bg3Layout"),
            let bg1HClip = Int(object: mapProperties, keyPath: "bg1HorizontalClip"),
            let bg1VClip = Int(object: mapProperties, keyPath: "bg1VerticalClip"),
            let bg2HClip = Int(object: mapProperties, keyPath: "bg2HorizontalClip"),
            let bg2VClip = Int(object: mapProperties, keyPath: "bg2VerticalClip"),
            let bg3HClip = Int(object: mapProperties, keyPath: "bg3HorizontalClip"),
            let bg3VClip = Int(object: mapProperties, keyPath: "bg3VerticalClip"),
            let bg3Priority = Int(object: mapProperties, keyPath: "bg3Priority"),
            let overlayIndex = Int(object: mapProperties, keyPath: "overlay"),
            let paletteIndex = Int(object: mapProperties, keyPath: "palette"),
            let animationDataIndex = Int(object: mapProperties, keyPath: "animation"),
            let bg3AnimationIndex = Int(object: mapProperties, keyPath: "bg3Animation"),
            let colorMathIndex = Int(object: mapProperties, keyPath: "colorMath"),
            let bg2HorizontalShift = Int(object: mapProperties, keyPath: "bg2HorizontalShift"),
            let bg2VerticalShift = Int(object: mapProperties, keyPath: "bg2VerticalShift"),
            let bg3HorizontalShift = Int(object: mapProperties, keyPath: "bg3HorizontalShift"),
            let bg3VerticalShift = Int(object: mapProperties, keyPath: "bg3VerticalShift"),
            let width = Int(object: mapProperties, keyPath: "width"),
            let height = Int(object: mapProperties, keyPath: "height"),
            let _parallax = Int(object: mapProperties, keyPath: "parallax"),
            let animationGraphics = rom.value(forKeyPath: "mapAnimationGraphics") as? ROMData,
            let graphics1 = rom.value(at: graphicsIndex1, inPropertyWithKey: "mapGraphics") as? ROMData,
            let graphics2 = rom.value(at: graphicsIndex2, inPropertyWithKey: "mapGraphics") as? ROMData,
            let graphics3 = rom.value(at: graphicsIndex3, inPropertyWithKey: "mapGraphics") as? ROMData,
            let graphics4 = rom.value(at: graphicsIndex4, inPropertyWithKey: "mapGraphics") as? ROMData,
            let bg3GraphicsData = rom.value(at: bg3GraphicsIndex, inPropertyWithKey: "mapBG3Graphics") as? ROMData,
            let bg1Tileset = rom.value(at: bg1TilesetIndex, inPropertyWithKey: "mapTilesets") as? ROMData,
            let bg2Tileset = rom.value(at: bg2TilesetIndex, inPropertyWithKey: "mapTilesets") as? ROMData,
            let bg1Layout = rom.value(at: bg1LayoutIndex, inPropertyWithKey: "mapLayouts") as? ROMData,
            let bg2Layout = rom.value(at: bg2LayoutIndex, inPropertyWithKey: "mapLayouts") as? ROMData,
            let bg3Layout = rom.value(at: bg3LayoutIndex, inPropertyWithKey: "mapLayouts") as? ROMData,
            let palette = rom.value(at: paletteIndex, inPropertyWithKey: "mapPalettes") as? ROMData,
            let animationProperties = rom.value(at: animationDataIndex, inPropertyWithKey: "mapAnimationProperties") as? ROMArray,
            let overlayData = rom.value(at: overlayIndex, inPropertyWithKey: "mapOverlayProperties") as? ROMData,
            let overlayGraphics = rom.value(forKey: "mapOverlayGraphics") as? ROMData,
            let overlayTileFormation = rom.value(forKey: "mapOverlayLayouts") as? ROMArray else {
                rom.log("Unable to read data for map \(mapIndex)")
                return
        }
        
        // TODO: do something better with these
        mapWidth = width
        mapHeight = height
        
        parallax = _parallax // TODO: read the actual parallax properties
        colorMath = rom.value(at: colorMathIndex, inPropertyWithKey: "mapColorMath") as? ROMData
        startObserving(object: colorMath)
        
        // decode background graphics
        let gfxLength = 0x10000
        let gfx = UnsafeMutablePointer<UInt8>.allocate(capacity: gfxLength)
        memset(gfx, 0, gfxLength)
        graphics1.data.withUnsafeBytes{(bytes: UnsafePointer<UInt8>)->Void in
            gfx_decode(gfx.advanced(by: 0x0000), bytes, Int32(graphics1.data.count), isGBA ? LINEAR_4BPP : SNES_4BPP)
        }
        graphics2.data.withUnsafeBytes{(bytes: UnsafePointer<UInt8>)->Void in
            gfx_decode(gfx.advanced(by: 0x4000), bytes, Int32(graphics1.data.count), isGBA ? LINEAR_4BPP : SNES_4BPP)
        }
        graphics3.data.withUnsafeBytes{(bytes: UnsafePointer<UInt8>)->Void in
            gfx_decode(gfx.advanced(by: 0x6000), bytes, Int32(graphics1.data.count), isGBA ? LINEAR_4BPP : SNES_4BPP)
        }

        if graphicsIndex3 != graphicsIndex4 {
            graphics4.data.withUnsafeBytes{(bytes: UnsafePointer<UInt8>)->Void in
                gfx_decode(gfx.advanced(by: 0x8000), bytes, Int32(graphics1.data.count), isGBA ? LINEAR_4BPP : SNES_4BPP)
            }
        }
        for (index, animatedTile) in animationProperties.array.enumerated() {
            let frame1 = (animatedTile as AnyObject).value(forKeyPath: "frame1") as! Int
            let _animationGraphics = animationGraphics.data
            _animationGraphics.withUnsafeBytes{(bytes: UnsafePointer<UInt8>)->Void in
                let animatedTilePointer = bytes.advanced(by: frame1)
                gfx_decode(gfx.advanced(by: 0xA000 + index * 0x100), animatedTilePointer, 0x80, isGBA ? LINEAR_4BPP : SNES_4BPP)
            }
        }
        let bgGraphics = Data(bytes: gfx, count: gfxLength)
        gfx.deinitialize(); gfx.deallocate(capacity: gfxLength)
        
        // decode bg3 graphics
        let gfx3Length = 0x4000
        let gfx3 = UnsafeMutablePointer<UInt8>.allocate(capacity: gfx3Length)
        memset(gfx3, 0, gfx3Length)
        let _bg3GraphicsData = bg3GraphicsData.data
        _bg3GraphicsData.withUnsafeBytes{(bytes: UnsafePointer<UInt8>)->Void in
            gfx_decode(gfx3, bytes.advanced(by: 0x40), Int32(bg3GraphicsData.data.count - 0x40), isGBA ? LINEAR_4BPP : SNES_2BPP)
        }
        if bg3AnimationIndex != 0 {
            // decode bg3 animation graphics
            let bg3AnimationGraphics = rom.value(at: bg3AnimationIndex - 1, inPropertyWithKey: "mapBG3AnimationGraphics") as! ROMData
            let bg3AnimationData = rom.value(at: bg3AnimationIndex - 1, inPropertyWithKey: "mapBG3AnimationProperties") as! ROMData
            let size = bg3AnimationData.intValue(forKey: "size")
            let frame1 = bg3AnimationData.intValue(forKey: "frame1")
            let _bg3AnimationGraphics = bg3AnimationGraphics.data
            _bg3AnimationGraphics.withUnsafeBytes{(bytes: UnsafePointer<UInt8>)->Void in
                gfx_decode(gfx3, bytes.advanced(by: frame1), Int32(size), isGBA ? LINEAR_4BPP : SNES_2BPP)
            }
        }
        let bg3Graphics = Data(bytes: gfx3, count: gfx3Length)
        gfx3.deinitialize(); gfx3.deallocate(capacity: gfx3Length)
        
        // decode overlay graphics
        let gfxOverlayLength = 0x1100
        let gfxOverlay = UnsafeMutablePointer<UInt8>.allocate(capacity: gfxOverlayLength)
        memset(gfxOverlay, 0, gfxOverlayLength)
        var overlayTiles = [UInt8](repeating: 0, count: 16)
        overlayData.data.copyBytes(to: &overlayTiles, count: 16)
        for (index, tile) in overlayTiles.enumerated() {
            let intTile = Int(tile)
            if intTile > overlayTileFormation.array.count { continue }
            var tilePointers = [Int](repeating: 0, count: 4)
            tilePointers[0] = (overlayTileFormation.array[intTile] as! ROMData).intValue(forKey: "tilePointer1")
            tilePointers[1] = (overlayTileFormation.array[intTile] as! ROMData).intValue(forKey: "tilePointer2")
            tilePointers[2] = (overlayTileFormation.array[intTile] as! ROMData).intValue(forKey: "tilePointer3")
            tilePointers[3] = (overlayTileFormation.array[intTile] as! ROMData).intValue(forKey: "tilePointer4")
            let _overlayGraphics = overlayGraphics.data
            _overlayGraphics.withUnsafeBytes{(bytes: UnsafePointer<UInt8>)->Void in
                gfx_decode(gfxOverlay.advanced(by: (index + 1) * 256 + 0x00), bytes.advanced(by: tilePointers[0]), 8, NES_1BPP)
                gfx_decode(gfxOverlay.advanced(by: (index + 1) * 256 + 0x40), bytes.advanced(by: tilePointers[1]), 8, NES_1BPP)
                gfx_decode(gfxOverlay.advanced(by: (index + 1) * 256 + 0x80), bytes.advanced(by: tilePointers[2]), 8, NES_1BPP)
                gfx_decode(gfxOverlay.advanced(by: (index + 1) * 256 + 0xC0), bytes.advanced(by: tilePointers[3]), 8, NES_1BPP)
            }
        }
        overlay.graphics = Data(bytes: gfxOverlay, count: gfxOverlayLength)
        gfxOverlay.deinitialize(); gfxOverlay.deallocate(capacity: gfxOverlayLength)
        
        // decode palette
        let palLength = 0x100 * MemoryLayout<UInt32>.size
        let pal = UnsafeMutablePointer<UInt8>.allocate(capacity: palLength)
        memset(pal, 0, palLength)
        palette.data.withUnsafeBytes{(bytes: UnsafePointer<UInt8>)->Void in
            gfx_decode(pal, bytes, Int32(palette.data.count), BGR555)
        }
        var backAreaColor = 0xFF000000 // set the back area color to black (C0/267D)
        memcpy(pal, &backAreaColor, 4)
        if isGBA {
            // i can't remember why this is here but i must have needed it
            var transparentColor = 0
            memcpy(pal.advanced(by: 16), &transparentColor, 4)
            memcpy(pal.advanced(by: 32), &transparentColor, 4)
            memcpy(pal.advanced(by: 48), &transparentColor, 4)
        }
        mapPalette = Data(bytes: pal, count: palLength)
        pal.deinitialize(); pal.deallocate(capacity: palLength)
        
        // initialize each layer
        bg1.type = .bg1
        bg1.tilesetData = bg1Tileset
        bg1.tileset = bg1Tileset.data
        bg1.graphics = bgGraphics
        bg1.layoutIndex = bg1LayoutIndex
        bg1.layoutData = bg1Layout
        bg1.layout = bg1Layout.data
        bg1.height = 1 << (4 + bg1VClip) // (C0/17E7)
        bg1.width = 1 << (4 + bg1HClip)
        bg1.format = isGBA ? GBA_4BPP_TILE : SNES_4BPP_TILE
        if bg1.layoutIndex == 0 {
            // create a blank NSData for layout 0
            bg1.layout = Data(count: bg1.height * bg1.width * 4 * MemoryLayout<UInt16>.size)
            bg1.layoutData = nil
        }
        bg1.decodeLayout()
        startObserving(object: bg1.layoutData)
        startObserving(object: bg1.tilesetData)
        
        bg2.type = .bg2
        bg2.tilesetData = bg2Tileset
        bg2.tileset = bg2Tileset.data
        bg2.graphics = bgGraphics
        bg2.horizontalShift = Int8(truncatingIfNeeded: bg2HorizontalShift)
        bg2.verticalShift = Int8(truncatingIfNeeded: bg2VerticalShift)
        bg2.layoutIndex = bg2LayoutIndex
        bg2.layoutData = bg2Layout
        bg2.layout = bg2Layout.data
        bg2.height = 1 << (4 + bg2VClip)
        bg2.width = 1 << (4 + bg2HClip)
        bg2.format = isGBA ? GBA_4BPP_TILE : SNES_4BPP_TILE
        if bg2.layoutIndex == 0 {
            // create a blank NSData for layout 0
            bg2.layout = Data(count: bg2.height * bg2.width * 4 * MemoryLayout<UInt16>.size)
            bg2.layoutData = nil
        }
        bg2.decodeLayout()
        startObserving(object: bg2.layoutData)
        startObserving(object: bg2.tilesetData)
        
        bg3.type = .bg3
        bg3.graphics = bg3Graphics
        bg3.horizontalShift = Int8(truncatingIfNeeded: bg3HorizontalShift)
        bg3.verticalShift = Int8(truncatingIfNeeded: bg3VerticalShift)
        bg3.tileset = bg3GraphicsData.data // the bg3 tileset is just the first 64 bytes of the graphics data
        bg3.layoutIndex = bg3LayoutIndex
        bg3.layoutData = bg3Layout
        bg3.layout = bg3Layout.data
        bg3.height = 1 << (4 + bg3VClip)
        bg3.width = 1 << (4 + bg3HClip)
        bg3.format = isGBA ? GBA_4BPP_TILE : SNES_2BPP_TILE
        bg3.bg3Priority = bg3Priority == 0 ? false : true
        if bg3.layoutIndex == 0 {
            // create a blank NSData for layout 0
            bg3.layout = Data(count: bg3.height * bg3.width * 4 * MemoryLayout<UInt16>.size)
            bg3.layoutData = nil
        }
        bg3.decodeLayout()
        startObserving(object: bg3.layoutData)
        
        // decode overlay tile layout
        overlay.type = .overlay
        overlay.height = bg1.height
        overlay.width = bg1.width
        overlay.format = SNES_4BPP_TILE
        overlay.layoutIndex = bg1.layoutIndex
        overlay.tileset = overlayData.data.subdata(in: 16..<272)
        overlay.layout = bg1.layout
        overlay.decodeLayout()
        
        // set up the tilesets as tile layouts
        let graphicsDefinition = NSMutableDictionary()
        graphicsDefinition["0x0000"] = "mapGraphics[\(graphicsIndex1)]"
        graphicsDefinition["0x4000"] = "mapGraphics[\(graphicsIndex2)]"
        graphicsDefinition["0x6000"] = "mapGraphics[\(graphicsIndex3)]"
        if graphicsIndex3 != graphicsIndex4 {
            graphicsDefinition["0x8000"] = "mapGraphics[\(graphicsIndex4)]"
        }
        let paletteDefinition: NSDictionary = ["0x00" : "mapPalettes[\(paletteIndex)]"]

        let bg1TilesetLayout = bg1Tileset as? ROMTileLayout
        bg1TilesetLayout?.graphics = graphicsDefinition
        bg1TilesetLayout?.palette = paletteDefinition
        
        let bg2TilesetLayout = bg2Tileset as? ROMTileLayout
        bg2TilesetLayout?.graphics = graphicsDefinition
        bg2TilesetLayout?.palette = paletteDefinition
    }
    
    func loadWorldMap(_ mapIndex: Int) {
        
        var mapName: String!
        switch mapIndex {
        case 0: mapName = "worldOfBalance"
        case 1: mapName = "worldOfRuin"
        case 2: mapName = "serpentTrench"
        default:
            rom.log("Invalid world map")
            return
        }
        
        guard let graphicsData = rom.value(forKey: mapName + "Graphics") as? ROMData,
            let graphics = graphicsData.value(forKey: "graphics") as? ROMData,
            let palette = rom.value(forKey: mapName + "Palette") as? ROMData,
            let layout = rom.value(forKey: mapName + "MapLayout") as? ROMData else { return }
        
        // decode palette
        let palLength = 0x100 * MemoryLayout<UInt32>.size
        let pal = UnsafeMutablePointer<UInt8>.allocate(capacity: palLength)
        memset(pal, 0, palLength)
        palette.data.withUnsafeBytes{(bytes: UnsafePointer<UInt8>)->Void in
            gfx_decode(pal, bytes, Int32(palette.data.count), BGR555)
        }
        var backAreaColor = 0xFF000000 // set the back area color to black
        memcpy(pal, &backAreaColor, 4)
        mapPalette = Data(bytes: UnsafePointer<UInt8>(pal), count: palLength)
        pal.deinitialize(); pal.deallocate(capacity: palLength)
        
        // decode world map graphics
        let gfxLength = 0x4000
        let gfx = UnsafeMutablePointer<UInt8>.allocate(capacity: gfxLength)
        memset(gfx, 0, gfxLength)
        bg1.tileset = graphics.data
        let _graphics = graphics.data
        _graphics.withUnsafeBytes{(bytes: UnsafePointer<UInt8>)->Void in
            gfx_decode(gfx, bytes, isGBA ? 0x4000 : 0x2000, isGBA ? LINEAR_8BPP : LINEAR_4BPP)
        }
        bg1.graphics = Data(bytes: gfx, count: gfxLength)
        gfx.deinitialize(); gfx.deallocate(capacity: gfxLength)
        
        bg1.type = .world
        bg1.layoutData = layout
        bg1.tilesetData = graphicsData.value(forKey: "tileset") as? ROMData
        bg1.paletteAssignmentData = graphicsData.value(forKey: "paletteAssignment") as? ROMData
        bg1.layoutIndex = -1
        bg1.height = mapIndex == 2 ? 128 : 256
        bg1.width = bg1.height
        bg1.format = isGBA ? GBA_4BPP_TILE : SNES_4BPP_TILE
        bg1.layout = layout.data
        bg1.decodeLayout()
        startObserving(object: bg1.layoutData)
        startObserving(object: bg1.tilesetData)
        startObserving(object: bg1.paletteAssignmentData)

        mapWidth = bg1.width
        mapHeight = bg1.height
    }
    
    // MARK: object observing methods
    
    func startObserving(object: AnyObject?) {
        
        guard let properties = object?.value(forKey: "properties") as? [String : AnyObject] else { return }
        for (propertyName, _) in properties {
            object?.addObserver(self, forKeyPath: propertyName, options: NSKeyValueObservingOptions.new, context: &FF6MapView.context)
        }
    }

    func stopObserving(object: AnyObject?) {
        
        guard let properties = object?.value(forKey: "properties") as? [String : AnyObject] else { return }
        for (propertyName, _) in properties {
            object?.removeObserver(self, forKeyPath: propertyName, context: &FF6MapView.context)
        }
    }

    override func observeValue(forKeyPath keyPath: String?, of object: Any?, change: [NSKeyValueChangeKey : Any]?, context: UnsafeMutableRawPointer?) {
        if context != &FF6MapView.context {
            super.observeValue(forKeyPath: keyPath, of: object, change: change, context: context)
            return
        }
        
        if object == nil {
            return
        } else if object as AnyObject? === mapProperties || object as AnyObject? === colorMath {
            // object is this map's properties or color math
            tilesetView?.tilesetLayer = nil
            tilesetView?.needsDisplay = true
            loadMap()
        } else if let assembly = object as? ROMAssembly, triggerForAssembly(assembly) != nil {
            // object is a trigger that is on this map
            triggerLayer = nil
            needsDisplay = true
        } else if let assembly = object as? ROMData, assembly.identifier == "mapLayouts" {
            // object is a map layout (size changed)
            loadMap()
        } else if let assembly = object as? ROMData, assembly.identifier == "mapTilesets" || assembly.identifier == "tileset" || assembly.identifier == "graphics" || assembly.identifier == "paletteAssignment" {
            // reload map and tileset view
            loadMap()
            tilesetView?.needsDisplay = true
        } else {
            // otherwise, just redraw
            needsDisplay = true
        }
    }

    // MARK: NSView methods

    override var isOpaque: Bool { return true }
    override var preservesContentDuringLiveResize: Bool { return true }
    
    override func setFrameSize(_ newSize: NSSize) {
        
        // TODO: make this less ugly - it really fucks it up for subclasses
        
        if !(superview is NSClipView) {
            super.setFrameSize(newSize)
            return
        }
        
        if mapIndex < 2 {
            super.setFrameSize(NSMakeSize(256 * 16 * zoom, 256 * 16 * zoom))
        } else if mapIndex == 2 {
            super.setFrameSize(NSMakeSize(128 * 16 * zoom, 128 * 16 * zoom))
        } else {
            super.setFrameSize(NSMakeSize(128 * 16 * zoom, 64 * 16 * zoom))
        }
    }
    
    override func resize(withOldSuperviewSize oldSize: NSSize) {
        super.resize(withOldSuperviewSize: oldSize)
        if let clipView = superview as? NSClipView {
            setFrameSize(clipView.documentVisibleRect.size)
        }
    }
    
    // MARK: drawing methods
    
    override func draw(_ dirtyRect: NSRect) {
        super.draw(dirtyRect)
        
        // fill the view with black
        if let clipView = superview as? NSClipView, let scrollView = clipView.superview as? NSScrollView {
            scrollView.backgroundColor = NSColor.black
        }
        NSColor.black.set()
        NSBezierPath.fill(dirtyRect)
        NSGraphicsContext.current?.imageInterpolation = imageInterpolation
        
        // draw map
        if mapLayer == nil { updateMapLayerInRect(NSZeroRect) }
        let maxWidth = CGFloat(max(bg1.width, bg2.width, bg3.width))
        let maxHeight = CGFloat(max(bg1.height, bg2.height, bg3.height))

        for x in stride(from: CGFloat(0), to: 128, by: maxWidth) {
            for y in stride(from: CGFloat(0), to: 64, by: maxHeight) {
                let mapRect = NSMakeRect(x * 16 * zoom, y * 16 * zoom, maxWidth * 16 * zoom, maxHeight * 16 * zoom)
                NSGraphicsContext.current?.cgContext.draw(mapLayer!, in: mapRect)
            }
        }

        // draw triggers
        if showTriggers {
            if triggerLayer == nil { updateTriggerLayer() }
            NSGraphicsContext.current?.cgContext.draw(triggerLayer!, in: bounds)
        }
    }
    
    func drawCursor(x: Int, y: Int, width: Int, height: Int, color: NSColor) {
        
        if showCursor == false { return }
        
        // convert the selection to screen coordinates
        let left = round(CGFloat(x) * zoom) + 0.5
        let right = round(CGFloat(x + width) * zoom) - 0.5
        let top = round(CGFloat(y) * zoom) + 0.5
        let bottom = round(CGFloat(y + height) * zoom) - 0.5
        let cursorRect = NSMakeRect(left, frame.height - bottom, right - left, bottom - top)

        // stroke 4 rectangles to make the cursor
        NSColor.black.set(); NSBezierPath.stroke(cursorRect)
        color.set();         NSBezierPath.stroke(cursorRect.insetBy(dx: 1, dy: 1))
        NSColor.white.set(); NSBezierPath.stroke(cursorRect.insetBy(dx: 2, dy: 2))
        NSColor.black.set(); NSBezierPath.stroke(cursorRect.insetBy(dx: 3, dy: 3))
    }
    
    func updateMapLayerInRect(_ rect: NSRect) {
        
        // create the map layer if it is not valid
        if mapLayer == nil {
            if let currentContext = NSGraphicsContext.current?.cgContext {
                let maxWidth = max(bg1.width, bg2.width, bg3.width)
                let maxHeight = max(bg1.height, bg2.height, bg3.height)
                mapLayer = CGLayer(currentContext, size: CGSize(width: maxWidth * 16, height: maxHeight * 16), auxiliaryInfo: nil)
            } else {
                // unable to get current context, this was probably not called from drawRect
                rom.log("Unable to create map layer")
                return
            }
        }
        
        // TODO: this comes up a lot and I really need to fix it
        if mapPalette == nil { return }

        if mapIndex < 3 {
            updateWorldMapInRect(rect)
            return
        }
        
        // rect parameter is in unflipped bounds coordinates
        var x = 0
        var y = 0
        var width = mapWidth * 16
        var height = mapHeight * 16
        
        if !rect.isEmpty {
            // image size (rounded to the nearest 16 pixels)
            x = Int(rect.minX / 16 / zoom) * 16
            x %= width
            y = Int((bounds.height - rect.minY) / 16 / zoom) * 16
            y %= height
            width = Int(rect.width / 16 / zoom) * 16
            height = Int(rect.height / 16 / zoom) * 16
        }
        
        let imageRect = NSMakeRect(CGFloat(x), CGFloat(max(bg1.height, bg2.height, bg3.height) * 16 - y - height), CGFloat(width), CGFloat(height))

        // create the ppu
        var ppu = PPU()
        memset(&ppu, 0, MemoryLayout<PPU>.size)
        ppu.pal = (mapPalette! as NSData).bytes
        ppu.back_area = true
        ppu.subtract = colorMath?.boolValue(forKey: "subtraction") ?? false
        ppu.half = colorMath?.boolValue(forKey: "halfMath") ?? false
        
        // set up bg1
        ppu.layers.0.format = bg1.format
        ppu.layers.0.height = Int32(bg1.height * 2)
        ppu.layers.0.width = Int32(bg1.width * 2)
        ppu.layers.0.v_scroll = Int32(y / 8)
        ppu.layers.0.h_scroll = Int32(x / 8)
        ppu.layers.0.z.0 = Z_LEVEL_BG1_LO
        ppu.layers.0.z.1 = Z_LEVEL_BG1_HI
        ppu.layers.0.gfx = (bg1.graphics as NSData).bytes
        ppu.layers.0.tiles = (bg1.tiles as NSData).bytes
        ppu.layers.0.main = showLayer1 // bg1 is always enabled in main screen
        ppu.layers.0.sub = showLayer1 && colorMath?.boolValue(forKey: "subScreenLayer1") ?? false
        ppu.layers.0.math = showLayer1 && colorMath?.boolValue(forKey: "affectLayer1") ?? false
        
        // set up bg2
        ppu.layers.1.format = bg2.format
        ppu.layers.1.height = Int32(bg2.height * 2)
        ppu.layers.1.width = Int32(bg2.width * 2)
        if parallax == 0 {
            ppu.layers.1.v_scroll = Int32(bg2.verticalShift) * 2 + Int32(y / 8)
            ppu.layers.1.h_scroll = Int32(bg2.horizontalShift) * 2 + Int32(x / 8)
        } else {
            ppu.layers.1.v_scroll = Int32(y / 8)
            ppu.layers.1.h_scroll = Int32(x / 8)
        }
        ppu.layers.1.z.0 = Z_LEVEL_BG2_LO
        ppu.layers.1.z.1 = Z_LEVEL_BG2_HI
        ppu.layers.1.gfx = (bg2.graphics as NSData).bytes
        ppu.layers.1.tiles = (bg2.tiles as NSData).bytes
        ppu.layers.1.main = showLayer2 && colorMath?.boolValue(forKey: "mainScreenLayer2") ?? true
        ppu.layers.1.sub = showLayer2 && colorMath?.boolValue(forKey: "subScreenLayer2") ?? false
        ppu.layers.1.math = showLayer2 && colorMath?.boolValue(forKey: "affectLayer2") ?? false
        
        // set up bg3
        ppu.layers.2.format = bg3.format
        ppu.layers.2.height = Int32(bg3.height * 2)
        ppu.layers.2.width = Int32(bg3.width * 2)
        if parallax == 0 {
            ppu.layers.2.v_scroll = Int32(bg3.verticalShift * 2) + Int32(y / 8)
            ppu.layers.2.h_scroll = Int32(bg3.horizontalShift * 2) + Int32(x / 8)
        } else {
            ppu.layers.2.v_scroll = Int32(y / 8)
            ppu.layers.2.h_scroll = Int32(x / 8)
        }
        ppu.layers.2.z.0 = Z_LEVEL_BG3_LO
        ppu.layers.2.z.1 = Z_LEVEL_BG3_HI_P // always high priority bg3
        ppu.layers.2.gfx = (bg3.graphics as NSData).bytes
        ppu.layers.2.tiles = (bg3.tiles as NSData).bytes
        ppu.layers.2.main = showLayer3 && colorMath?.boolValue(forKey: "mainScreenLayer3") ?? true
        ppu.layers.2.sub = showLayer3 && colorMath?.boolValue(forKey: "subScreenLayer3") ?? false
        ppu.layers.2.math = showLayer3 && colorMath?.boolValue(forKey: "affectLayer3") ?? false
        
        // render the map
        let dataLength = height * width * MemoryLayout<UInt32>.size
        let data = UnsafeMutablePointer<UInt8>.allocate(capacity: dataLength)
        memset(data, 0, dataLength)
        gfx_render_ppu(data, ppu, Int32(dataLength), Int32(width))
        
        // create an image
        let provider = CGDataProvider(dataInfo: nil, data: data, size: dataLength, releaseData: releaseImageData)!
        let bitsPerComponent = 8
        let bitsPerPixel = 32
        let bytesPerRow = width * MemoryLayout<UInt32>.size
        let colorSpaceRef = CGColorSpaceCreateDeviceRGB()
        let bitmapInfo = CGBitmapInfo(rawValue: CGBitmapInfo().rawValue | CGImageAlphaInfo.premultipliedLast.rawValue)
        let renderingIntent = CGColorRenderingIntent.defaultIntent
        let iref = CGImage(width: width, height: height, bitsPerComponent: bitsPerComponent, bitsPerPixel: bitsPerPixel, bytesPerRow: bytesPerRow, space: colorSpaceRef, bitmapInfo: bitmapInfo, provider: provider, decode: nil, shouldInterpolate: true, intent: renderingIntent)
        
        // draw the image to the map layer
        let layerContext = mapLayer?.context
        layerContext?.draw(iref!, in: imageRect)
        
        // free memory
        data.deinitialize(); data.deallocate(capacity: dataLength)
    }
    
    func updateWorldMapInRect(_ rect: NSRect) {
        
        // rect parameter is in unflipped bounds coordinates
        var x = 0
        var y = 0
        var width = mapWidth * 16
        var height = mapHeight * 16
        
        if !rect.isEmpty {
            // image size (rounded to the nearest 16 pixels)
            x = Int(rect.minX / 16 / zoom) * 16
            x %= width
            y = Int((bounds.height - rect.minY) / 16 / zoom) * 16
            y %= height
            width = Int(rect.width / 16 / zoom) * 16
            height = Int(rect.height / 16 / zoom) * 16
        }
        
        let imageRect = NSMakeRect(CGFloat(x), CGFloat(bg1.height * 16 - y - height), CGFloat(width), CGFloat(height))
        
        // create the ppu
        var ppu = PPU()
        memset(&ppu, 0, MemoryLayout<PPU>.size)
        ppu.pal = (mapPalette! as NSData).bytes
        ppu.back_area = true
        
        // set up bg1
        ppu.layers.0.format = bg1.format
        ppu.layers.0.height = Int32(bg1.height * 2)
        ppu.layers.0.width = Int32(bg1.width * 2)
        ppu.layers.0.v_scroll = Int32(y / 8)
        ppu.layers.0.h_scroll = Int32(x / 8)
        ppu.layers.0.z.0 = Z_LEVEL_BG1_LO
        ppu.layers.0.z.1 = Z_LEVEL_BG1_HI
        ppu.layers.0.gfx = (bg1.graphics as NSData).bytes
        ppu.layers.0.tiles = (bg1.tiles as NSData).bytes
        ppu.layers.0.main = true // bg1 is always enabled in main screen
        
        // render the map
        let dataLength = height * width * MemoryLayout<UInt32>.size
        let data = UnsafeMutablePointer<UInt8>.allocate(capacity: dataLength)
        memset(data, 0, dataLength)
        gfx_render_ppu(data, ppu, Int32(dataLength), Int32(width))
        
        // create an image
        let provider = CGDataProvider(dataInfo: nil, data: data, size: dataLength, releaseData: releaseImageData)!
        let bitsPerComponent = 8
        let bitsPerPixel = 32
        let bytesPerRow = width * MemoryLayout<UInt32>.size
        let colorSpaceRef = CGColorSpaceCreateDeviceRGB()
        let bitmapInfo = CGBitmapInfo(rawValue: CGBitmapInfo().rawValue | CGImageAlphaInfo.premultipliedLast.rawValue)
        let renderingIntent = CGColorRenderingIntent.defaultIntent
        let iref = CGImage(width: width, height: height, bitsPerComponent: bitsPerComponent, bitsPerPixel: bitsPerPixel, bytesPerRow: bytesPerRow, space: colorSpaceRef, bitmapInfo: bitmapInfo, provider: provider, decode: nil, shouldInterpolate: true, intent: renderingIntent)
        
        // draw the image to the map layer
        let layerContext = mapLayer?.context
        layerContext?.draw(iref!, in: imageRect)

        // free memory
        data.deinitialize(); data.deallocate(capacity: dataLength)
    }
    
    func updateTriggerLayer() {
        
        if !showTriggers { return }
        
        // clear the trigger layer
        if let currentContext = NSGraphicsContext.current?.cgContext {
            if mapIndex < 2 {
                triggerLayer = CGLayer(currentContext, size: CGSize(width: 256 * 16 * zoom, height: 256 * 16 * zoom), auxiliaryInfo: nil)
            } else if mapIndex == 2 {
                triggerLayer = CGLayer(currentContext, size: CGSize(width: 128 * 16 * zoom, height: 128 * 16 * zoom), auxiliaryInfo: nil)
            } else {
                triggerLayer = CGLayer(currentContext, size: CGSize(width: 128 * 16 * zoom, height: 64 * 16 * zoom), auxiliaryInfo: nil)
            }
        } else {
            return
        }

        // draw all of the triggers
        for trigger in triggers {
            drawTrigger(trigger)
        }
        
        // draw all of the npcs in order by y value
        for trigger in triggers.sorted(by: {$0.y < $1.y}) {
            if trigger.type == .npc || trigger.type == .specialNPC {
                drawNPC(trigger)
            }
        }
    }
    
    func drawTrigger(_ trigger: FF6MapTrigger) {
        if trigger.type == .none { return }

        // calculate the position of the trigger rectangle
        let mapHeight = CGFloat(mapIndex < 3 ? 256 : 64) // TODO: this is causing triggers not to draw on the serpent trench, which i don't really care about right now
        let triggerLayerContext = triggerLayer?.context

        func drawTriggerRect(x: Int, y: Int) {
            
            let left = round(CGFloat(x) * 16 * zoom) + 0.5
            let right = round(CGFloat(x + 1) * 16 * zoom) - 0.5
            let top = round(CGFloat(y) * 16 * zoom) + 0.5
            let bottom = round(CGFloat(y + 1) * 16 * zoom) - 0.5
            let triggerRect = NSMakeRect(left, mapHeight * 16 * zoom - bottom, right - left, bottom - top)
            
            let triggerPath = CGPath(roundedRect: NSInsetRect(triggerRect, 2, 2), cornerWidth: zoom * 2, cornerHeight: zoom * 2, transform: nil)
            
            // draw the transparent colored rectangle
            triggerLayerContext?.setFillColor(trigger.type.color.cgColor.copy(alpha: 0.5)!)
            triggerLayerContext?.addPath(triggerPath)
            triggerLayerContext?.fillPath()
            
            // draw the white border
            let borderColor = CGColor(red: 1.0, green: 1.0, blue: 1.0, alpha: 1.0)
            triggerLayerContext?.setStrokeColor(borderColor)
            triggerLayerContext?.addPath(triggerPath)
            triggerLayerContext?.strokePath()
        }

        if trigger.isLong {
            // long entrance tiles
            let length = trigger.triggerAssembly.intValue(forKey: "length")
            let vertical = trigger.triggerAssembly.boolValue(forKey: "vertical")

            // draw each tile
            for tile in 0...length {
                drawTriggerRect(x: trigger.x + (vertical ? 0 : tile), y: trigger.y + (vertical ? tile : 0))
            }
        } else {
            // all other trigger types
            drawTriggerRect(x: trigger.x, y: trigger.y)
        }
    }
    
    func drawNPC(_ npc: FF6MapTrigger) {
        // read npc properties
        var x = CGFloat(npc.x) * 16
        var y = CGFloat(npc.y) * 16 + 8
        let spriteIndex = npc.triggerAssembly.intValue(forKey: "graphics")
        let paletteIndex = npc.triggerAssembly.intValue(forKey: "palette")
        let specialAnimation = npc.triggerAssembly.intValue(forKey: "specialAnimation")
        let vehicle = npc.triggerAssembly.intValue(forKey: "vehicle")
        let showRider = npc.triggerAssembly.boolValue(forKey: "showRider")
        let facingDirection = npc.triggerAssembly.intValue(forKey: "facingDirection")
        var frameIndex = 0
        var is32x32 = false
        var hFlip = false
        if npc.type == .specialNPC {
            // special NPC graphics
            is32x32 = npc.triggerAssembly.boolValue(forKey: "is32x32")
            hFlip = npc.triggerAssembly.boolValue(forKey: "horizontalFlip")
            let shiftDown = npc.triggerAssembly.boolValue(forKey: "offsetDirection")
            let isSlave = npc.triggerAssembly.boolValue(forKey: "isSlave")
            let positionOffset = npc.triggerAssembly.intValue(forKey: "positionOffset")
            
            if isSlave {
                // slave npc
                let masterIndex = npc.triggerAssembly.intValue(forKey: "masterNPC")
                if masterIndex >= npc.triggerArray.count { return }
                let masterNPC = FF6MapTrigger(triggerArray: npc.triggerArray, triggerAssembly: npc.triggerArray[masterIndex] as! ROMData)
                x = CGFloat(masterNPC.x) * 16
                y = CGFloat(masterNPC.y) * 16 + 8

                if shiftDown    { y += CGFloat(positionOffset * 16) }
                else            { x += CGFloat(positionOffset * 16) }

            } else {
                // not a slave npc
                if shiftDown    { y += CGFloat(positionOffset * 2) }
                else            { x += CGFloat(positionOffset * 2) }
            }

            if is32x32 {
                y += 16
            }

        } else if specialAnimation != 0 {
            // special NPC animation
            if specialAnimation == 2 { frameIndex = 0x32 }
            if specialAnimation == 3 { frameIndex = 0x28 }

        } else {
            // normal NPC
            var facingFrames = [0x04, 0x47, 0x01, 0x07]
            if (vehicle == 1) || (vehicle == 2) {
                facingFrames[1] = 0x6E
                facingFrames[3] = 0x2E
            }
            frameIndex = facingFrames[facingDirection]
            if (frameIndex & 0x40) != 0 { hFlip = true }
        }
        
        // get the sprite graphics and tile formation
        let mapSpriteGraphics = rom.value(forKey: "mapSpriteGraphics") as! ROMData
        let mapSpriteTileFormation = rom.value(forKey: "mapSpriteTileFormation") as! ROMArray
        let tileFormation = mapSpriteTileFormation[frameIndex & 0x3F] as! ROMArray
        
        // get a pointer to the sprite graphics
        let mapSpriteGraphicsPointersLo = rom.value(forKey: "mapSpriteGraphicsPointersLo") as! ROMArray
        let mapSpriteGraphicsPointersHi = rom.value(forKey: "mapSpriteGraphicsPointersHi") as! ROMArray
        let gfxPointerLo = mapSpriteGraphicsPointersLo[spriteIndex].value(forKey: "pointer") as! Int
        let gfxPointerHi = mapSpriteGraphicsPointersHi[spriteIndex].value(forKey: "pointer") as! Int
        var gfxPointer = gfxPointerLo | (gfxPointerHi << 16)
        gfxPointer &= 0x00FFFFFF
        gfxPointer -= isGBA ? 0xF60000 : rom.mapMode.unmapAddress(mapSpriteGraphics.range!.lowerBound)
        
        // decode graphics
        let tileCount = is32x32 ? 16 : (npc.type == .specialNPC) ? 4 : 6
        let gfxLength = 0x40 * tileCount
        let gfx = UnsafeMutablePointer<UInt8>.allocate(capacity: gfxLength)
        memset(gfx, 0, gfxLength)
        var tileData = [UInt16](repeating: 0, count: tileCount)
        for tile in 0..<tileCount {
            let tileOffset = (npc.type == .specialNPC) ? (tile * 0x20) : tileFormation[tile].value(forKey: "offset") as! Int
            mapSpriteGraphics.data.withUnsafeBytes{(bytes: UnsafePointer<UInt8>)->Void in
                gfx_decode(gfx.advanced(by: tile * 0x40), bytes.advanced(by: gfxPointer + tileOffset), 0x20, isGBA ? LINEAR_4BPP : SNES_4BPP)
            }
            if hFlip {
                tileData[tile ^ (is32x32 ? 3 : 1)] = UInt16(tile | 0x4000)
            } else {
                tileData[tile] = UInt16(tile)
            }
        }
        let spriteGraphics = Data(bytes: UnsafePointer<UInt8>(gfx), count: gfxLength)
        gfx.deinitialize(); gfx.deallocate(capacity: gfxLength)
        
        // decode palette
        let mapSpritePalettes = rom.value(forKey: "mapSpritePalettes") as! ROMArray
        let paletteData = mapSpritePalettes[paletteIndex].value(forKey: "data") as! Data
        let vehiclePaletteData = mapSpritePalettes[vehicle == 3 ? 11 : 7].value(forKey: "data") as! Data
        let palLength = 0x20 * MemoryLayout<UInt32>.size
        let pal = UnsafeMutablePointer<UInt8>.allocate(capacity: palLength)
        memset(pal, 0, palLength)
        gfx_decode(pal, (paletteData as NSData).bytes, Int32(paletteData.count), BGR555)
        gfx_decode(pal.advanced(by: 0x10 * MemoryLayout<UInt32>.size), (vehiclePaletteData as NSData).bytes, Int32(vehiclePaletteData.count), BGR555)
        let spritePalette = Data(bytes: UnsafePointer<UInt8>(pal), count: palLength)
        pal.deinitialize(); pal.deallocate(capacity: palLength)
        
        // create the ppu
        var ppu = PPU()
        memset(&ppu, 0, MemoryLayout<PPU>.size)
        ppu.pal = (spritePalette as NSData).bytes
        
        // image size (in pixels)
        var width = 16
        var height = 24
        if npc.type == .specialNPC {
            width = is32x32 ? 32 : 16
            height = width
        }
        
        func drawVehicle() {
            if vehicle == 0 { return }
            if npc.type != .npc { return }
            if specialAnimation != 0 { return }
            
            var vehicleWidth = 32
            let vehicleHeight = 32
            var vehicleX = x
            let vehicleY = y
            
            // decode vehicle graphics
            let vehicleGraphicsData = rom.value(forKey: "vehicleGraphics") as! ROMData
            let gfxVehicleLength = 0x8000
            let gfxVehicle = UnsafeMutablePointer<UInt8>.allocate(capacity: gfxVehicleLength)
            memset(gfxVehicle, 0, gfxVehicleLength)
            let _vehicleGraphicsData = vehicleGraphicsData.data
            _vehicleGraphicsData.withUnsafeBytes{(bytes: UnsafePointer<UInt8>)->Void in
                gfx_decode(gfxVehicle, bytes, Int32(vehicleGraphicsData.data.count), isGBA ? LINEAR_4BPP : SNES_4BPP)
            }
            let vehicleGraphics = Data(bytes: gfxVehicle, count: gfxVehicleLength)
            gfxVehicle.deinitialize(); gfxVehicle.deallocate(capacity: gfxVehicleLength)

            // get the vehicle tile layout
            var vehicleTiles = [UInt16](repeating: 0, count: 16)
            if vehicle == 1 {
                // chocobo
                y -= 12
                switch facingDirection {
                    
                case 0: // up
                    vehicleTiles = [
                        0x042C, 0x042D,
                        0x043C, 0x043D,
                        0x042E, 0x042F,
                        0x043E, 0x043F]
                    vehicleWidth = 16
                    height = 16
                    y -= 8
                    
                case 1: // right
                    vehicleTiles = [
                        0x4449, 0x4448, 0x4445, 0x4444,
                        0x4459, 0x4458, 0x4455, 0x4454,
                        0x444B, 0x444A, 0x4447, 0x4446,
                        0x445B, 0x445A, 0x4457, 0x4456]
                    vehicleX -= 8
                    x -= 3
                    
                case 2: // down
                    vehicleTiles = [
                        0x0422, 0x0423,
                        0x0432, 0x0433,
                        0x0424, 0x0425,
                        0x0434, 0x0435]
                    vehicleWidth = 16
                    height = 16
                    y -= 9
                    
                case 3: // left
                    vehicleTiles = [
                        0x0444, 0x0445, 0x0448, 0x0449,
                        0x0454, 0x0455, 0x0458, 0x0459,
                        0x0446, 0x0447, 0x044A, 0x044B,
                        0x0456, 0x0457, 0x045A, 0x045B]
                    vehicleX -= 8
                    x += 3
                    
                default: break
                }
                
            } else if vehicle == 2 {
                // magitek
                
                vehicleX -= 8
                height = 16
                y -= 22
                
                switch facingDirection {
                case 0: vehicleTiles = [
                    0x048C, 0x048D, 0x448D, 0x448C,
                    0x049C, 0x049D, 0x449D, 0x449C,
                    0x048E, 0x048F, 0x448F, 0x448E,
                    0x049E, 0x049F, 0x449F, 0x449E]
                    
                case 1: vehicleTiles = [
                    0x44AB, 0x44AA, 0x44A9, 0x44A8,
                    0x44BB, 0x44BA, 0x44B9, 0x44B8,
                    0x44AF, 0x44AE, 0x44AD, 0x44AC,
                    0x44BF, 0x44BE, 0x44BD, 0x44BC]
                    
                case 2: vehicleTiles = [
                    0x0480, 0x0481, 0x4481, 0x4480,
                    0x0490, 0x0491, 0x4491, 0x4490,
                    0x0482, 0x0483, 0x4483, 0x4482,
                    0x0492, 0x0493, 0x4493, 0x4492]
                    
                case 3: vehicleTiles = [
                    0x04A8, 0x04A9, 0x04AA, 0x04AB,
                    0x04B8, 0x04B9, 0x04BA, 0x04BB,
                    0x04AC, 0x04AD, 0x04AE, 0x04AF,
                    0x04BC, 0x04BD, 0x04BE, 0x04BF]
                    
                default: break
                }
                
            } else if vehicle == 3 {
                // raft
                vehicleX -= 8
                y -= 16
                switch facingDirection {
                case 0: vehicleTiles = [
                    0x0400, 0x0401, 0x0404, 0x0405,
                    0x0410, 0x0411, 0x0414, 0x0415,
                    0x0402, 0x0403, 0x0406, 0x0407,
                    0x0412, 0x0413, 0x0416, 0x0417]
                    
                case 1: vehicleTiles = [
                    0x0408, 0x0409, 0x040C, 0x040D,
                    0x0418, 0x0419, 0x041C, 0x041D,
                    0x040A, 0x040B, 0x040E, 0x040F,
                    0x041A, 0x041B, 0x041E, 0x041F]
                    
                case 2: vehicleTiles = [
                    0x0400, 0x0401, 0x0404, 0x0405,
                    0x0410, 0x0411, 0x0414, 0x0415,
                    0x0402, 0x0403, 0x0406, 0x0407,
                    0x0412, 0x0413, 0x0416, 0x0417]
                    
                case 3: vehicleTiles = [
                    0x0408, 0x0409, 0x040C, 0x040D,
                    0x0418, 0x0419, 0x041C, 0x041D,
                    0x040A, 0x040B, 0x040E, 0x040F,
                    0x041A, 0x041B, 0x041E, 0x041F]
                    
                default: break
                }
                
            }
            
            // set up the ppu
            ppu.layers.0.main = true
            ppu.layers.0.format = SNES_4BPP_TILE // this works for the gba version too
            ppu.layers.0.gfx = (vehicleGraphics as NSData).bytes
            ppu.layers.0.height = Int32(vehicleHeight / 8)
            ppu.layers.0.width = Int32(vehicleWidth / 8)
            ppu.layers.0.z.0 = Z_LEVEL_SPRITE_0
            ppu.layers.0.z.1 = Z_LEVEL_SPRITE_1
            ppu.layers.0.z.2 = Z_LEVEL_SPRITE_2
            ppu.layers.0.z.3 = Z_LEVEL_SPRITE_3
            ppu.layers.0.tiles = UnsafeRawPointer(vehicleTiles)

            // render the sprite
            let dataLength = vehicleHeight * vehicleWidth * MemoryLayout<UInt32>.size
            let data = UnsafeMutablePointer<UInt8>.allocate(capacity: dataLength)
            memset(data, 0, dataLength)
            gfx_render_ppu(data, ppu, Int32(dataLength), Int32(vehicleWidth))
            
            // create an image
            let provider = CGDataProvider(dataInfo: nil, data: data, size: vehicleHeight * vehicleWidth * 4, releaseData: releaseImageData)!
            let bitsPerComponent = 8
            let bitsPerPixel = 32
            let bytesPerRow = 4 * vehicleWidth
            let colorSpaceRef = CGColorSpaceCreateDeviceRGB()
            let bitmapInfo = CGBitmapInfo(rawValue: CGBitmapInfo().rawValue | CGImageAlphaInfo.premultipliedLast.rawValue)
            let renderingIntent = CGColorRenderingIntent.defaultIntent
            let iref = CGImage(width: vehicleWidth, height: vehicleHeight, bitsPerComponent: bitsPerComponent, bitsPerPixel: bitsPerPixel, bytesPerRow: bytesPerRow, space: colorSpaceRef, bitmapInfo: bitmapInfo, provider: provider, decode: nil, shouldInterpolate: true, intent: renderingIntent)
            
            // draw the image to the trigger layer
            let layerContext = triggerLayer?.context
            layerContext?.draw(iref!, in: CGRect(x: vehicleX * zoom, y: (64 * 16 - vehicleY) * zoom, width: CGFloat(vehicleWidth) * zoom, height: CGFloat(vehicleHeight) * zoom))
            
            // free memory
            data.deinitialize(); data.deallocate(capacity: dataLength)

        }
        
        func drawSprite() {
            
            if vehicle != 0 && !showRider && specialAnimation == 0 { return }
            
            // draw sprite
            ppu.layers.0.main = true
            ppu.layers.0.format = SNES_4BPP_TILE // this works for the gba version too
            ppu.layers.0.gfx = (spriteGraphics as NSData).bytes
            ppu.layers.0.height = Int32(height / 8)
            ppu.layers.0.width = Int32(width / 8)
            ppu.layers.0.z.0 = Z_LEVEL_SPRITE_0
            ppu.layers.0.z.1 = Z_LEVEL_SPRITE_1
            ppu.layers.0.z.2 = Z_LEVEL_SPRITE_2
            ppu.layers.0.z.3 = Z_LEVEL_SPRITE_3
            ppu.layers.0.tiles = UnsafeRawPointer(tileData)
            
            // render the sprite
            let dataLength = height * width * MemoryLayout<UInt32>.size
            let data = UnsafeMutablePointer<UInt8>.allocate(capacity: dataLength)
            memset(data, 0, dataLength)
            gfx_render_ppu(data, ppu, Int32(dataLength), Int32(width))
            
            // create an image
            let provider = CGDataProvider(dataInfo: nil, data: data, size: height * width * 4, releaseData: releaseImageData)!
            let bitsPerComponent = 8
            let bitsPerPixel = 32
            let bytesPerRow = 4 * width
            let colorSpaceRef = CGColorSpaceCreateDeviceRGB()
            let bitmapInfo = CGBitmapInfo(rawValue: CGBitmapInfo().rawValue | CGImageAlphaInfo.premultipliedLast.rawValue)
            let renderingIntent = CGColorRenderingIntent.defaultIntent
            let iref = CGImage(width: width, height: height, bitsPerComponent: bitsPerComponent, bitsPerPixel: bitsPerPixel, bytesPerRow: bytesPerRow, space: colorSpaceRef, bitmapInfo: bitmapInfo, provider: provider, decode: nil, shouldInterpolate: true, intent: renderingIntent)
            
            // draw the image to the trigger layer
            let layerContext = triggerLayer?.context
            layerContext?.draw(iref!, in: CGRect(x: x * zoom, y: (64 * 16 - y) * zoom, width: CGFloat(width) * zoom, height: CGFloat(height) * zoom))
            
            // free memory
            data.deinitialize(); data.deallocate(capacity: dataLength)
        }
        
        func drawTail() {
            
            // draw chocobo tail or head

            if vehicle != 1 { return }
            if specialAnimation != 0 { return }

            let tailWidth = 16
            let tailHeight = 16
            let tailX = x
            var tailY = y + 12

            // get the vehicle tile layout
            var tailTiles = [UInt16](repeating: 0, count: 16)
            
            if facingDirection == 0 {
                // facing up, draw tail
                tailTiles = [
                    0x042A, 0x042B,
                    0x043A, 0x043B]
                tailY += 1
            } else if facingDirection == 2 {
                // facing down, draw head
                tailTiles = [
                    0x0420, 0x0421,
                    0x0430, 0x0431]
            } else {
                // facing left or right, don't draw anything
                return
            }

            // decode vehicle graphics
            let vehicleGraphicsData = rom.value(forKey: "vehicleGraphics") as! ROMData
            let gfxVehicleLength = 0x8000
            let gfxVehicle = UnsafeMutablePointer<UInt8>.allocate(capacity: gfxVehicleLength)
            memset(gfxVehicle, 0, gfxVehicleLength)
            let _vehicleGraphicsData = vehicleGraphicsData.data
            _vehicleGraphicsData.withUnsafeBytes{(bytes: UnsafePointer<UInt8>)->Void in
                gfx_decode(gfxVehicle, bytes, Int32(vehicleGraphicsData.data.count), isGBA ? LINEAR_4BPP : SNES_4BPP)
            }
            let vehicleGraphics = Data(bytes: gfxVehicle, count: gfxVehicleLength)
            gfxVehicle.deinitialize(); gfxVehicle.deallocate(capacity: gfxVehicleLength)
            
            // set up the ppu
            ppu.layers.0.main = true
            ppu.layers.0.format = SNES_4BPP_TILE // this works for the gba version too
            ppu.layers.0.gfx = (vehicleGraphics as NSData).bytes
            ppu.layers.0.height = Int32(tailHeight / 8)
            ppu.layers.0.width = Int32(tailWidth / 8)
            ppu.layers.0.z.0 = Z_LEVEL_SPRITE_0
            ppu.layers.0.z.1 = Z_LEVEL_SPRITE_1
            ppu.layers.0.z.2 = Z_LEVEL_SPRITE_2
            ppu.layers.0.z.3 = Z_LEVEL_SPRITE_3
            ppu.layers.0.tiles = UnsafeRawPointer(tailTiles)
            
            // render the sprite
            let dataLength = tailHeight * tailWidth * MemoryLayout<UInt32>.size
            let data = UnsafeMutablePointer<UInt8>.allocate(capacity: dataLength)
            memset(data, 0, dataLength)
            gfx_render_ppu(data, ppu, Int32(dataLength), Int32(tailWidth))
            
            // create an image
            let provider = CGDataProvider(dataInfo: nil, data: data, size: tailHeight * tailWidth * 4, releaseData: releaseImageData)!
            let bitsPerComponent = 8
            let bitsPerPixel = 32
            let bytesPerRow = 4 * tailWidth
            let colorSpaceRef = CGColorSpaceCreateDeviceRGB()
            let bitmapInfo = CGBitmapInfo(rawValue: CGBitmapInfo().rawValue | CGImageAlphaInfo.premultipliedLast.rawValue)
            let renderingIntent = CGColorRenderingIntent.defaultIntent
            let iref = CGImage(width: tailWidth, height: tailHeight, bitsPerComponent: bitsPerComponent, bitsPerPixel: bitsPerPixel, bytesPerRow: bytesPerRow, space: colorSpaceRef, bitmapInfo: bitmapInfo, provider: provider, decode: nil, shouldInterpolate: true, intent: renderingIntent)
            
            // draw the image to the trigger layer
            let layerContext = triggerLayer?.context
            layerContext?.draw(iref!, in: CGRect(x: tailX * zoom, y: (64 * 16 - tailY) * zoom, width: CGFloat(tailWidth) * zoom, height: CGFloat(tailHeight) * zoom))
            
            // free memory
            data.deinitialize(); data.deallocate(capacity: dataLength)
        }

        drawVehicle()
        drawSprite()
        drawTail()

        // something is wrong with swift here. for some reason it can't crop a sprite image to 16x16 but it can crop to 8x16
//        let cropRect1 = CGRect(x: 0, y: 0, width: 8, height: 16)
//        let spriteRect1 = NSMakeRect((x + 16) * zoom, (64 * 16 - y) * zoom, 8 * zoom, 16 * zoom)
//        if let spriteImage1 = rom?.npcImage(spriteIndex, frame: frameIndex, palette: paletteIndex)?.cropping(to: cropRect1) {
//            triggerLayer?.context?.draw(spriteImage1, in: spriteRect1)
//        }
//
//        let cropRect2 = CGRect(x: 8, y: 0, width: 8, height: 16)
//        let spriteRect2 = NSMakeRect((x + 24) * zoom, (64 * 16 - y) * zoom, 8 * zoom, 16 * zoom)
//        if let spriteImage2 = rom?.npcImage(spriteIndex, frame: frameIndex, palette: paletteIndex)?.cropping(to: cropRect2) {
//            triggerLayer?.context?.draw(spriteImage2, in: spriteRect2)
//        }
    }
}
