//
//  FF6ROM.swift
//  FF6Tools
//
//  Created by Evan Owen on 2/21/15.
//
//

// japanese translations: http://kwhazit.ucoz.net/trans/ff6/index.html
// more translations: http://finalfantasy.wikia.com/wiki/User:Gilgameshkun/Final_Fantasy_VI_Script

import Foundation

let itemNamesKey = "itemNames"
let eventScriptKey = "eventScript"

class FF6ROM: ROM {

    // MARK: instance variables
    
    @objc var isGBA: Bool { return system == .GBA ? true : false }

    required init?(contentsOfURL url: URL) {
        super.init(contentsOfURL: url)
        
        // create script encodings
        for (name, encodingType) in [
            "eventScript" : EventScriptEncoding.self,
            "objectScript" : EventScriptEncoding.self,
            "worldScript" : EventScriptEncoding.self,
            "vehicleScript" : EventScriptEncoding.self,
            "battleScript" : BattleScriptEncoding.self,
            "attackScript" : BattleScriptEncoding.self,
            "attackScriptHeader" : BattleScriptEncoding.self,
            "monsterScript" : MonsterScriptEncoding.self,
            "trainScript" : TrainScriptEncoding.self,
            "songScript" : SongScriptEncoding.self,
            "songScriptHeader" : SongScriptEncoding.self] {
            guard let encoding = scriptEncodings[name] as? ROMScriptEncoding else { continue }
            scriptEncodings[name] = encodingType.self.init(self, definition: encoding.createDefinition())
        }

        // create text encodings
        for (name, encodingType) in [
            "dialog" : DialogTextEncoding.self,
            "nullTerminated" : BattleTextEncoding.self,
            "battle" : BattleTextEncoding.self,
            "bigText" : BattleTextEncoding.self] {
                guard let encoding = textEncodings[name] as? ROMTextEncoding else { continue }
                textEncodings[name] = encodingType.self.init(self, definition: encoding.createDefinition())
        }
    }

    override func write(to url: URL) throws {
        
        // for testing relocation
//        expand(toLength: 0x400000)
//        let relocatedAssembly = value(forKeyPath: "instrumentBRR") as! ROMAssembly
//        unusedMemory.insert(integersIn: relocatedAssembly.range!)
//        relocate(relocatedAssembly, to: mapMode.mapAddress(0xF01234))
//        relocatedAssembly.updateReferences()
//        
//        let relocatedPointerTable = value(forKeyPath: "pointerTable-instrumentBRR") as! ROMAssembly
//        unusedMemory.insert(integersIn: relocatedPointerTable.range!)
//        relocate(relocatedPointerTable, to: mapMode.mapAddress(0xF51453))
//        relocatedPointerTable.updateReferences()
        
//        for key in memoryMap.allKeys {
//            guard let identifier = key as? String else { continue }
//            guard let assembly = value(forKey: identifier) as? ROMAssembly else { continue }
//        }
        
        // split dialog into two banks
        let combinedDialogArray = splitDialogBanks()
        
        // TODO: fix this so I can edit battle background graphics
        subAssemblies.removeObject(forKey: "battleBackgroundGraphics")
        subAssemblies.removeObject(forKey: "mapGraphics")

        // auto-relocate and assemble everything
        assemble()
        
        // combine the dialog banks back together
        if combinedDialogArray != nil,
            let dialog = subAssemblies["dialog"] as? ROMArray,
            let dialogBank2 = value(forKey: "dialogBank2") as? ROMArray {
            
            dialog.array = combinedDialogArray!
            dialogBank2.array.removeAllObjects()
        }

        // print check values
        log(String(format: "CRC32: 0x%08X", ROM.calculateCRC32(data)))
        if system == .SFC {
            updateSNESChecksum()
            log(String(format: "SNES Checksum (calculated): 0x%04X", ROM.calculateChecksum(data)))
        }
        
        try super.write(to: url)
    }
    
    // MARK: NSKeyValueCoding methods
    
    override func value(forKey key: String) -> Any? {
        
        if key == "smallMonsterGraphicsMaps" {
            let monsterGraphicsMaps = value(forKey: "monsterGraphicsMaps") as? ROMData
            return monsterGraphicsMaps?.value(forKey: "smallMaps")
        } else if key == "largeMonsterGraphicsMaps" {
            let monsterGraphicsMaps = value(forKey: "monsterGraphicsMaps") as? ROMData
            return monsterGraphicsMaps?.value(forKey: "largeMaps")
        }
        
        // key with indexed array
        if let bracketIndex = key.characters.index(of: "[") {
            let arrayKey = String(key[..<bracketIndex])
            var bracketString = String(key[bracketIndex...])
            bracketString.characters.removeFirst()
            bracketString.characters.removeLast()
            let arrayIndex = Int(bracketString) ?? 0
            
            return value(at: arrayIndex, inPropertyWithKey: arrayKey)
        }
        
        let keyComponents = key.components(separatedBy: "-")

        if let sub: AnyObject = subAssemblies.object(forKey: key) as AnyObject? {
            // assembly already exists in subData
            return sub
            
        } else if let definition = memoryMap[key] as? NSMutableDictionary {
            
            // TODO: call the super class to create and disassemble the new assembly
            
            // key is in the memory map
            definition["identifier"] = key
            guard let newAssembly = ROMObject.createObject(self, definition: definition) as? ROMAssembly else { return nil }
            subAssemblies[key] = newAssembly
            
            log(String(format: "Disassembling: %@", newAssembly.name ?? key))
            newAssembly.disassembleFrom(data)
            
            if (key == "dialog") && (memoryMap["dialogBank2"] != nil) {
                // read the other dialog bank and combine the two
                if let dialog = newAssembly as? ROMArray, let dialogBank2 = value(forKey: "dialogBank2") as? ROMArray {
                    dialog.array.addObjects(from: dialogBank2.array as [AnyObject])
                    dialogBank2.array.removeAllObjects()
                }
                
            } else if (system == .GBA) && (key == "serpentTrenchMapLayout") {
                // fix gba serpent trench map layout length
                if let serpentTrenchMapLayout = newAssembly as? ROMData, serpentTrenchMapLayout.data.count > 0x4000 {
                    var serpentTrenchMapLayoutData = serpentTrenchMapLayout.data.subdata(in: 0..<0x4000)
                    serpentTrenchMapLayout.data = serpentTrenchMapLayoutData
                    
                    // compress the layout to get its compressed length
                    ROMCompression.FF6.compress(&serpentTrenchMapLayoutData)

                    // mark free memory as unused
                    let freeStart = serpentTrenchMapLayout.range!.lowerBound + serpentTrenchMapLayoutData.count
                    let freeEnd = serpentTrenchMapLayout.range!.upperBound
                    unusedMemory.insert(integersIn: freeStart..<freeEnd)

                    // update the assembly's range
                    serpentTrenchMapLayout.range = serpentTrenchMapLayout.range!.lowerBound..<freeStart
                }
            } else if key == "battleBackgroundGraphics" {

                // fix battle background graphics that use map graphics
                if let battleBackgroundGraphics = newAssembly as? ROMArray,
                    let battleBackgroundGraphicsPointerTable = battleBackgroundGraphics.pointerTable,
                    let mapGraphics = value(forKey: "mapGraphics") as? ROMArray,
                    let mapGraphicsPointerTable = mapGraphics.pointerTable {
                    
                    for (index, pointer) in battleBackgroundGraphicsPointerTable.pointers.enumerated() {
                        
                        // break if index is out of bounds
                        if index >= battleBackgroundGraphics.array.count { break }
                        
                        // skip pointers that are within the battle background graphics range
                        if battleBackgroundGraphics.range!.contains(pointer) { continue }
                        
                        let graphicsToFix = battleBackgroundGraphics.array[index] as! ROMData
                        if system == .GBA {
                            let offsetFix = battleBackgroundGraphicsPointerTable.pointerOffset! - mapGraphicsPointerTable.pointerOffset!
                            let start = pointer & 0x7FFFFF - offsetFix
                            let end = start + 0x2000
                            graphicsToFix.range = start..<end
                        } else {
                            graphicsToFix.range = pointer..<(pointer + 0x2000)
                        }
                        graphicsToFix.compression = .None
                        graphicsToFix.disassembleFrom(data)
                    }
                }
            } else if key == "attackGraphicsFrames" {
                
                let attackGraphicsFrames = newAssembly as! ROMArray
                
                // append extra frames for FF6 Advance
                if let attackGraphicsFramesAdvance = value(forKey: "attackGraphicsFramesAdvance") as? ROMArray {
                    attackGraphicsFrames.array.addObjects(from: attackGraphicsFramesAdvance.array as [AnyObject])
                    attackGraphicsFramesAdvance.array.removeAllObjects()
                }
                
                // add an 0xFFFF terminator to each frame
                for object in attackGraphicsFrames.array {
                    guard let frameData = object as? ROMData else { continue }
                    var frameBytes = [UInt8](repeating: 0, count: frameData.data.count)
                    frameData.data.copyBytes(to: &frameBytes, count: frameData.data.count)
                    if (frameBytes.count < 2) || (frameBytes[frameData.data.count - 1] != 0xFF) || (frameBytes[frameData.data.count - 2] != 0xFF) {
                        frameBytes.append(0xFF)
                        frameBytes.append(0xFF)
                    }
                    frameData.data = Data(frameBytes)
                }
            }
            
            // disassemble next assembly
            if let nextAssemblyIdentifier = newAssembly.nextAssembly,
                let nextAssembly = value(forKey: nextAssemblyIdentifier) as? ROMAssembly {
                // make sure the assemblies are correctly linked to one another
                if nextAssembly.previousAssembly != newAssembly.identifier {
                    nextAssembly.previousAssembly = nil
                    newAssembly.nextAssembly = nil
                }
            }
            // disassemble previous assembly
            if let previousAssemblyIdentifier = newAssembly.previousAssembly,
                let previousAssembly = value(forKey: previousAssemblyIdentifier) as? ROMAssembly {
                // make sure the assemblies are correctly linked to one another
                if previousAssembly.nextAssembly != newAssembly.identifier {
                    previousAssembly.nextAssembly = nil
                    newAssembly.previousAssembly = nil
                }
            }

            return newAssembly
         
        } else if keyComponents.first == "stringTable" && keyComponents.count == 2 {
            let identifier = keyComponents[1]
            
            if let stringTable = subAssemblies[key] as? ROMStringTable {
                // string table has already been created
                return stringTable.stringArray
                
            } else if let definition = stringTables[identifier] as? NSMutableDictionary {
                // string table is defined but hasn't been created yet
                definition["identifier"] = identifier
                let newStringTable = ROMStringTable(self, definition: definition)
                subAssemblies[key] = newStringTable
                
                log("Creating String Table: " + (newStringTable.name ?? identifier))
                
                return newStringTable
            }
        } else if keyComponents.first == "pointerTable" && keyComponents.count == 2 {
            let identifier = keyComponents[1]
            
            if let assembly = subAssemblies[identifier] as? ROMAssembly {
                // string table has already been created
                return assembly.pointerTable
            }
        }
        
        return super.value(forKey: key)
    }
    
    override func expand(toLength length: Int) {
        super.expand(toLength: length)
        if system != .SFC { return }
        
        // fix the ROM size in the SNES header
        if let snesHeader = value(forKey: "snesHeader") as? ROMData {
            
            let mbit = 1024 * 1024 / 8 // number of bytes in one megabit
            
            switch (data.count) {
            case  3*mbit...4*mbit:  snesHeader.setValue(0x09, forKeyPath: "romSize")
            case  4*mbit...8*mbit:  snesHeader.setValue(0x0A, forKeyPath: "romSize")
            case  8*mbit...16*mbit: snesHeader.setValue(0x0B, forKeyPath: "romSize")
            case 16*mbit...32*mbit: snesHeader.setValue(0x0C, forKeyPath: "romSize")
            case 32*mbit...64*mbit: snesHeader.setValue(0x0D, forKeyPath: "romSize")
            default:
                log("Invalid SNES ROM Size")
                break
            }
        }
    }
    
    // MARK: dialog methods

    func splitDialogBanks() -> NSMutableArray? {
        
        // fail if dialog has not been accessed
        if subAssemblies["dialog"] == nil { return nil }
        
        // split dialog into two banks
        guard let bank1 = value(forKey: "dialog") as? ROMArray,
            let bank2 = value(forKey: "dialogBank2") as? ROMArray,
            let bankIncrement = value(forKey: "dialogBankIncrement") as? ROMData else { return nil }
        
        // save the combined dialog array
        let combinedArray = bank1.array
        
        // assemble the dialog to a dummy data to update the pointer table
        var dummyData = Data()
        bank1.assembleTo(&dummyData)
        
        // find the dialog index where the bank changes
        let pointers = bank1.pointerTable!.pointers
        var bankIncrementIndex = 0
        let firstBank = pointers[0] >> 16
        for pointer in pointers {
            if firstBank != (pointer >> 16) { break }
            bankIncrementIndex += 1
        }
        bankIncrement.data = NSMutableData(bytes: &bankIncrementIndex, length: 2) as Data
        
        // update bank 2 range
        bank2.array = NSMutableArray(array: combinedArray.subarray(with: NSMakeRange(bankIncrementIndex, bank1.array.count - bankIncrementIndex)))
        let bank2Start = pointers[bankIncrementIndex]
        let bank2End = bank2.range!.upperBound
        bank2.range = bank2Start..<bank2End
        
        // update bank 2 pointer table range
        let bank2PointerStart = bank1.pointerTable!.range!.lowerBound + bankIncrementIndex * bank1.pointerTable!.pointerSize
        let bank2PointerEnd = bank2PointerStart + bank2.array.count * bank1.pointerTable!.pointerSize
        bank2.pointerTable!.range = bank2PointerStart..<bank2PointerEnd
        bank2.pointerTable?.pointerOffset = bank1.pointerTable!.pointerOffset! + 0x010000
        
        // update bank 1 range
        bank1.array = NSMutableArray(array: combinedArray.subarray(with: NSMakeRange(0, bankIncrementIndex)))
        let bank1Start = bank1.range!.lowerBound
        let bank1End = pointers[bankIncrementIndex]
        bank1.range = bank1Start..<bank1End
        
        // update bank1 pointer table range
        let bank1PointerStart = bank1.pointerTable!.range!.lowerBound
        let bank1PointerEnd = bank1PointerStart + bank1.array.count * bank1.pointerTable!.pointerSize
        bank1.pointerTable!.range = bank1PointerStart..<bank1PointerEnd
        
        return combinedArray
    }
    
    func optimizeDTE() -> Bool {
        
        // this works but it's really slow and kinda ugly
        
        // get the encoding/decoding tables for dte (only the first 128 characters)
        guard let dteTable = value(forKey: "dteTable") as? ROMData,
            let dialog = value(forKey: "dialog") as? ROMArray,
            let dteEncoding = textEncodings["dteTable"] as? ROMTextEncoding,
            let dialogEncoding = textEncodings["dialog"] as? ROMTextEncoding else {
            log("DTE table not found")
            return false
        }
        
        // encode every dialog string without dte
        for object in dialog.array {
            guard let dialogString = object as? ROMText else { continue }
            let dummy = dialogString.string
            dialogString.encoding = dteEncoding
            dialogString.string = ""
            dialogString.string = dummy
        }
        
        // assemble the uncompressed dialog
        var uncompressedDialogData = Data()
        dialog.assembleTo(&uncompressedDialogData)
        var uncompressedDialogBytes: [UInt8] = uncompressedDialogData.withUnsafeBytes{ [UInt8](UnsafeBufferPointer(start: $0, count: uncompressedDialogData.count)) }

        // count the character pair combinations
        var dteCombinations = [Int : Int]()
        for i in 0..<(uncompressedDialogBytes.count - 1) {
            let firstByte = uncompressedDialogBytes[i]
            let secondByte = uncompressedDialogBytes[i + 1]
            if firstByte < 0x20 { continue }
            if firstByte > 0x7F { continue }
            if secondByte < 0x20 { continue }
            if secondByte > 0x7F { continue }
            let combinedBytes = Int(firstByte) | (Int(secondByte) << 8)
            let count = dteCombinations[combinedBytes] ?? 0
            dteCombinations[combinedBytes] = count + 1
        }

        // sort the combinations to get the 128 most common ones
        var sortedCombinations = dteCombinations.sorted{ $0.1 > $1.1 }
        sortedCombinations.removeLast(sortedCombinations.count - 128)
        
        // update the dialog encoding table and the dte table data
        let newData = NSMutableData()
        
        // add the new values
        let encodingTable = NSMutableDictionary(dictionary: dteEncoding.encodingTable)
        let decodingTable = NSMutableDictionary(dictionary: dteEncoding.decodingTable)
        let characterTable = NSMutableDictionary()
        for (var combination, _) in sortedCombinations {
            let index = newData.length / 2 + 0x80
            let combinationData = NSData(bytes: &combination, length: 2)
            let decodedString = dteEncoding.decode(data: combinationData as Data)
            characterTable[String(format: "0x%02X", index)] = decodedString
            decodingTable[index] = decodedString
            encodingTable[decodedString] = index
            newData.append(combinationData as Data)
        }

        // update the DTE table data
        dteTable.data = newData as Data
        characterTables["dte"] = characterTable
        
        // set the new dialog encoding and decoding tables
        dialogEncoding.encodingTable = encodingTable
        dialogEncoding.decodingTable = decodingTable
        
        // re-encode every dialog string with the new dte table
        for object in dialog.array {
            guard let dialogString = object as? ROMText else { continue }
            let dummy = dialogString.string
            dialogString.encoding = dialogEncoding
            dialogString.string = ""
            dialogString.string = dummy
        }
        
        return true
    }
    
    static let hiraganaCharacters = "ぁあぃいぅうぇえぉおかがきぎくぐけげこごさざしじすずせぜそぞただちぢっつづてでとどなにぬねのはばぱひびぴふぶぷへべぺほぼぽまみむめもゃやゅゆょよらりるれろゎわゐゑをんゔゕゖじどこいえも"

    // MARK: convenience image methods
    
    func monsterImage(_ index: Int) -> CGImage? {
        
        var monsterIndex = index
        
        if index == 262 {
            // phantom train
            return phantomTrainImage()
        }
        
        if isGBA && monsterIndex >= 384 {
            // skip esper graphics data
            monsterIndex += 36
        }
        
        // load the graphics data
        guard let graphicsData = value(at: monsterIndex, inPropertyWithKey: "monsterGraphicsProperties") as? ROMData else { return nil }
        let graphicsPointer = graphicsData.intValue(forKey: "graphicsPointer") << (isGBA ? 4 : 3)
        var paletteIndex = graphicsData.intValue(forKey: "palette")
        var is3bpp = false
        if !isGBA {
            // swap palette bytes (SNES only)
            paletteIndex = Int(UInt16(paletteIndex).byteSwapped)
            is3bpp = graphicsData.boolValue(forKey: "is3bpp")
        }
        let isLargeMap = graphicsData.boolValue(forKey: "largeMap")
        let mapIndex = graphicsData.intValue(forKey: "mapIndex")

        // image and tile size
        let width = isLargeMap ? 128 : 64
        let height = isLargeMap ? 128 : 64
        let tileSize = is3bpp ? 0x18 : 0x20
        
        // make a buffer for the raw graphics
        let rawLength = Int(height * width / 64) * tileSize
        let rawGraphics = UnsafeMutablePointer<UInt8>.allocate(capacity: rawLength)
        memset(UnsafeMutableRawPointer(rawGraphics), 0, rawLength)
        
        // get the graphics and graphics map
        guard let graphics = value(forKeyPath: "monsterGraphics.data") as? Data else { return nil }
        
        var graphicsMap: Data
        if isGBA {
            // gba graphics maps are stored in sequential order, regardless of size
            guard let monsterGraphicsMaps = value(forKey: "monsterGraphicsMaps") as? ROMData else { return nil }
            let mapSize = isLargeMap ? 32 : 8
            let start = mapIndex << 3
            let end = start + mapSize
            graphicsMap = monsterGraphicsMaps.data.subdata(in: start..<end)
        } else {
            // small and large graphics maps are stored separately on snes
            let graphicsMapsKey = isLargeMap ? "largeMonsterGraphicsMaps" : "smallMonsterGraphicsMaps"
            guard let mapData = value(at: mapIndex, inPropertyWithKey: graphicsMapsKey) as? ROMData else { return nil }
            graphicsMap = mapData.data
        }
        
        // load the raw graphics to the buffer
        var mapRow: UInt16 = 0
        var moldOffset = 0
        var graphicsOffset = graphicsPointer
        let bytesPerTileRow = isLargeMap ? 2 : 1
        var tile = 0
        for _ in 0..<(height / 8) {
            let moldRange = NSMakeRange(moldOffset, bytesPerTileRow)
            (graphicsMap as NSData).getBytes(&mapRow, range: moldRange)
            if !isLargeMap || !isGBA {
                mapRow = mapRow.byteSwapped
            }
            for _ in 0..<(width / 8) {
                if (mapRow & 0x8000) != 0 {
                    let tileRange = NSMakeRange(graphicsOffset, tileSize)
                    (graphics as NSData).getBytes(rawGraphics + tile * tileSize, range: tileRange)
                    graphicsOffset += tileSize
                }
                mapRow <<= 1
                tile += 1
            }
            moldOffset += bytesPerTileRow
        }
        
        // make the color palette
        guard let monsterPalettes = value(forKey: "monsterPalettes") as? ROMArray else { return nil }
        var paletteData: Data
        if is3bpp {
            paletteData = (monsterPalettes[Int(paletteIndex)] as! ROMData).data
        } else {
            let combinedPaletteData = NSMutableData()
            let firstPalette = monsterPalettes[Int(paletteIndex)] as! ROMData
            let secondPalette = monsterPalettes[Int(paletteIndex) + 1] as! ROMData
            combinedPaletteData.append(firstPalette.data)
            combinedPaletteData.append(secondPalette.data)
            paletteData = combinedPaletteData as Data
        }
        
        let colorPalette = UnsafeMutablePointer<UInt16>.allocate(capacity: 16)
        memset(UnsafeMutableRawPointer(colorPalette), 0, 32)
        (paletteData as NSData).getBytes(colorPalette, length: paletteData.count)
        
        // convert palette to 32-bit ARGB
        let decodedPaletteLength = 16 * 4
        let decodedPalette = UnsafeMutablePointer<UInt8>.allocate(capacity: decodedPaletteLength)
        gfx_decode(decodedPalette, (paletteData as NSData).bytes, 32, BGR555)
        colorPalette.deinitialize(); colorPalette.deallocate(capacity: 32)
        let monsterPaletteData = Data(bytes: UnsafePointer<UInt8>(decodedPalette), count: decodedPaletteLength)
        decodedPalette.deinitialize(); decodedPalette.deallocate(capacity: decodedPaletteLength)
        
        // convert graphics to RGB bitmap
        let decodedGraphicsSize = height * width
        let decodedGraphics = UnsafeMutablePointer<UInt8>.allocate(capacity: decodedGraphicsSize)
        let tileFormat = isGBA ? LINEAR_4BPP : (is3bpp ? SNES_3BPP : SNES_4BPP)
        gfx_decode(decodedGraphics, rawGraphics, Int32(rawLength), tileFormat)
        rawGraphics.deinitialize(); rawGraphics.deallocate(capacity: rawLength)
        let monsterGraphicsData = Data(bytes: UnsafePointer<UInt8>(decodedGraphics), count: decodedGraphicsSize)
        decodedGraphics.deinitialize(); decodedGraphics.deallocate(capacity: decodedGraphicsSize)
        
        // render the image
        let dataLength = height * width * MemoryLayout<UInt32>.size
        let data = UnsafeMutablePointer<UInt32>.allocate(capacity: dataLength)
        memset(data, 0, dataLength * MemoryLayout<UInt32>.size)
        gfx_render(data, (monsterGraphicsData as NSData).bytes, (monsterPaletteData as NSData).bytes, Int32(monsterGraphicsData.count), Int32(width))
        let imageData = Data(bytes: UnsafeRawPointer(data), count: dataLength)
        data.deinitialize(); data.deallocate(capacity: dataLength)

        // create an image
        let provider = CGDataProvider(dataInfo: nil, data: (imageData as NSData).bytes, size: dataLength, releaseData: releaseImageData)!
        let bitsPerComponent = 8
        let bitsPerPixel = 32
        let bytesPerRow = width * MemoryLayout<UInt32>.size
        let colorSpaceRef = CGColorSpaceCreateDeviceRGB()
        let bitmapInfo = CGBitmapInfo(rawValue: CGBitmapInfo().rawValue | CGImageAlphaInfo.premultipliedLast.rawValue)
        let renderingIntent = CGColorRenderingIntent.defaultIntent
        return CGImage(width: width, height: height, bitsPerComponent: bitsPerComponent, bitsPerPixel: bitsPerPixel, bytesPerRow: bytesPerRow, space: colorSpaceRef, bitmapInfo: bitmapInfo, provider: provider, decode: nil, shouldInterpolate: true, intent: renderingIntent)

    }
    
    func size(ofMonster index: Int) -> (width: Int, height: Int) {
        
        // load the graphics data
        var graphicsIndex = index
        
        if index == 262 {
            // phantom train
            return (16, 16)
        }

        if index >= 384 && isGBA { graphicsIndex += 36 }
        guard let graphicsData = value(at: graphicsIndex, inPropertyWithKey: "monsterGraphicsProperties") as? ROMData else { return (0,0) }
        let isLargeMap = graphicsData.boolValue(forKey: "largeMap")
        let mapIndex = graphicsData.intValue(forKey: "mapIndex")
        
        // get the graphics map
        var graphicsMap: Data
        if isGBA {
            // gba graphics maps are stored in sequential order, regardless of size
            guard let monsterGraphicsMaps = rom.value(forKey: "monsterGraphicsMaps") as? ROMData else { return (0,0) }
            let mapSize = isLargeMap ? 32 : 8
            let start = mapIndex << 3
            let end = start + mapSize
            graphicsMap = monsterGraphicsMaps.data.subdata(in: start..<end)
        } else {
            // small and large graphics maps are stored separately on snes
            let graphicsMapsKey = isLargeMap ? "largeMonsterGraphicsMaps" : "smallMonsterGraphicsMaps"
            guard let mapData = value(at: mapIndex, inPropertyWithKey: graphicsMapsKey) as? ROMData else { return (0,0) }
            graphicsMap = mapData.data
        }
        
        // map size
        let mapSize = isLargeMap ? 128 : 64
        
        // monster width and height
        var width = 0
        var height = 0
        
        var currentRowWidth = 0
        var mapRow: UInt16 = 0
        var moldOffset = 0
        let bytesPerTileRow = isLargeMap ? 2 : 1
        for y in 0..<(mapSize / 8) {
            let moldRange = NSMakeRange(moldOffset, bytesPerTileRow)
            (graphicsMap as NSData).getBytes(&mapRow, range: moldRange)
            if mapRow != 0 { height = y + 1 }
            if !isLargeMap || !isGBA {
                mapRow = mapRow.byteSwapped
            }
            for x in 0..<(mapSize / 8) {
                if (mapRow & 0x8000) != 0 {
                    currentRowWidth = x + 1
                }
                mapRow <<= 1
            }
            width = max(width, currentRowWidth)
            moldOffset += bytesPerTileRow
        }
        
        return (width, height)
    }
    
    func phantomTrainImage() -> CGImage? {
        guard let battleBackgroundProperties = value(at: 50, inPropertyWithKey: "battleBackgroundProperties") as? ROMData else { return nil }
        
        // decode graphics
        let graphics1 = battleBackgroundProperties.intValue(forKey: "graphics1")
        let graphics3 = battleBackgroundProperties.intValue(forKey: "graphics3")
        let gfxLength = 0x8000
        let gfx = UnsafeMutablePointer<UInt8>.allocate(capacity: gfxLength)
        memset(gfx, 0, gfxLength)
        if graphics1 != 0xFF {
            guard let graphicsData1 = value(at: graphics1 & 0x7F, inPropertyWithKey: "battleBackgroundGraphics") as? ROMData else { return nil }
            graphicsData1.data.withUnsafeBytes{ (bytes: UnsafePointer<UInt8>) in
                gfx_decode(gfx.advanced(by: isGBA ? 0x4000 : 0x0000), bytes, 0x2000, isGBA ? LINEAR_4BPP : SNES_4BPP)
            }
        }
        if graphics3 != 0xFF {
            guard let graphicsData3 = value(at: graphics3 & 0x7F, inPropertyWithKey: "battleBackgroundGraphics") as? ROMData else { return nil }
            graphicsData3.data.withUnsafeBytes{ (bytes: UnsafePointer<UInt8>) in
                gfx_decode(gfx.advanced(by: isGBA ? 0x2000 : 0x6000), bytes, 0x1000, isGBA ? LINEAR_4BPP : SNES_4BPP)
            }
        }
        let backgroundGraphics = Data(bytes: UnsafePointer<UInt8>(gfx), count: gfxLength)
        gfx.deinitialize(); gfx.deallocate(capacity: gfxLength)
        
        // decode layout (only the first one is used)
        let layout = battleBackgroundProperties.intValue(forKey: "layout1")
        guard let layoutData = value(at: layout, inPropertyWithKey: "battleBackgroundLayouts") as? ROMData else { return nil }
        var layoutBytes: [UInt16] = layoutData.data.withUnsafeBytes{ [UInt16](UnsafeBufferPointer(start: $0, count: layoutData.data.count / 2)) }
        for index in 0..<layoutBytes.count {
            layoutBytes[index] &= 0x01FF
        }
        let backgroundTiles = Data(bytes: &layoutBytes, count: layoutBytes.count * 2)
        
        // load the graphics data
        guard let graphicsData = value(at: 262, inPropertyWithKey: "monsterGraphicsProperties") as? ROMData else { return nil }
        var paletteIndex = graphicsData.intValue(forKey: "palette")
        var is3bpp = false
        if !isGBA {
            // swap palette bytes (SNES only)
            paletteIndex = Int(UInt16(paletteIndex).byteSwapped)
            is3bpp = graphicsData.boolValue(forKey: "is3bpp")
        }

        // make the color palette
        guard let monsterPalettes = value(forKey: "monsterPalettes") as? ROMArray else { return nil }
        var paletteData: Data
        if is3bpp {
            paletteData = (monsterPalettes[Int(paletteIndex)] as! ROMData).data
        } else {
            let combinedPaletteData = NSMutableData()
            let firstPalette = monsterPalettes[Int(paletteIndex)] as! ROMData
            let secondPalette = monsterPalettes[Int(paletteIndex) + 1] as! ROMData
            combinedPaletteData.append(firstPalette.data)
            combinedPaletteData.append(secondPalette.data)
            paletteData = combinedPaletteData as Data
        }
        
        let colorPalette = UnsafeMutablePointer<UInt16>.allocate(capacity: 16)
        memset(UnsafeMutableRawPointer(colorPalette), 0, 32)
        (paletteData as NSData).getBytes(colorPalette, length: paletteData.count)
        
        // decode palette
        let palLength = 0x100 * MemoryLayout<UInt32>.size
        let pal = UnsafeMutablePointer<UInt8>.allocate(capacity: palLength)
        memset(pal, 0, palLength)
        paletteData.withUnsafeBytes{ (bytes: UnsafePointer<UInt8>) in
            gfx_decode(pal, bytes, Int32(paletteData.count), BGR555)
        }
        let backgroundPalette = Data(bytes: UnsafePointer<UInt8>(pal), count: palLength)
        pal.deinitialize(); pal.deallocate(capacity: palLength)
        
        // create the ppu
        var ppu = PPU()
        memset(&ppu, 0, MemoryLayout<PPU>.size)
        ppu.pal = (backgroundPalette as NSData).bytes
        
        // set up bg2
        ppu.layers.1.format = isGBA ? GBA_4BPP_TILE : SNES_4BPP_TILE
        ppu.layers.1.height = 16
        ppu.layers.1.width = 16
        ppu.layers.1.z.0 = Z_LEVEL_BG2_LO
        ppu.layers.1.z.1 = Z_LEVEL_BG2_HI
        ppu.layers.1.h_scroll = 0
        ppu.layers.1.v_scroll = 0
        ppu.layers.1.gfx = (backgroundGraphics as NSData).bytes
        ppu.layers.1.tiles = (backgroundTiles as NSData).bytes
        ppu.layers.1.main = true
        
        // render the background
        let height = 16 * 8
        let width = 16 * 8
        let dataLength = height * width * MemoryLayout<UInt32>.size
        let data = UnsafeMutablePointer<UInt8>.allocate(capacity: dataLength)
        memset(data, 0, dataLength)
        gfx_render_ppu(data, ppu, Int32(dataLength), Int32(width))
        let imageData = Data(bytes: UnsafePointer<UInt8>(data), count: dataLength)
        data.deinitialize(); data.deallocate(capacity: dataLength)
        
        // create an image
        let provider = CGDataProvider(dataInfo: nil, data: (imageData as NSData).bytes, size: dataLength, releaseData: releaseImageData)!
        let bitsPerComponent = 8
        let bitsPerPixel = 32
        let bytesPerRow = width * MemoryLayout<UInt32>.size
        let colorSpaceRef = CGColorSpaceCreateDeviceRGB()
        let bitmapInfo = CGBitmapInfo(rawValue: CGBitmapInfo().rawValue | CGImageAlphaInfo.premultipliedLast.rawValue)
        let renderingIntent = CGColorRenderingIntent.defaultIntent
        return CGImage(width: width, height: height, bitsPerComponent: bitsPerComponent, bitsPerPixel: bitsPerPixel, bytesPerRow: bytesPerRow, space: colorSpaceRef, bitmapInfo: bitmapInfo, provider: provider, decode: nil, shouldInterpolate: true, intent: renderingIntent)
    }

    func battleBackgroundImage(_ index: Int) -> CGImage? {
        guard let battleBackgroundProperties = value(at: index, inPropertyWithKey: "battleBackgroundProperties") as? ROMData else { return nil }
        
        // decode graphics
        let graphics1 = battleBackgroundProperties.intValue(forKey: "graphics1")
        let graphics2 = battleBackgroundProperties.intValue(forKey: "graphics2")
        let graphics3 = battleBackgroundProperties.intValue(forKey: "graphics3")
        let gfxLength = 0x12000
        let gfx = UnsafeMutablePointer<UInt8>.allocate(capacity: gfxLength)
        memset(gfx, 0, gfxLength)
        if graphics1 != 0xFF {
            guard let graphicsData1 = value(at: graphics1 & 0x7F, inPropertyWithKey: "battleBackgroundGraphics") as? ROMData else { return nil }
            graphicsData1.data.withUnsafeBytes{ (bytes: UnsafePointer<UInt8>) in
                gfx_decode(gfx.advanced(by: isGBA ? 0x4000 : 0x0000), bytes, Int32(graphicsData1.data.count), isGBA ? LINEAR_4BPP : SNES_4BPP)
            }
        }
        if graphics2 != 0xFF {
            guard let graphicsData2 = value(at: graphics2 & 0x7F, inPropertyWithKey: "battleBackgroundGraphics") as? ROMData else { return nil }
            graphicsData2.data.withUnsafeBytes{ (bytes: UnsafePointer<UInt8>) in
                gfx_decode(gfx.advanced(by: isGBA ? 0x6000 : 0x2000), bytes, Int32(graphicsData2.data.count), isGBA ? LINEAR_4BPP : SNES_4BPP)
            }
        }
        if graphics3 != 0xFF {
            guard let graphicsData3 = value(at: graphics3 & 0x7F, inPropertyWithKey: "battleBackgroundGraphics") as? ROMData else { return nil }
            graphicsData3.data.withUnsafeBytes{ (bytes: UnsafePointer<UInt8>) in
                gfx_decode(gfx.advanced(by: isGBA ? 0x2000 : 0xE000), bytes, Int32(graphicsData3.data.count), isGBA ? LINEAR_4BPP : SNES_4BPP)
            }
        }
        let backgroundGraphics = Data(bytes: UnsafePointer<UInt8>(gfx), count: gfxLength)
        gfx.deinitialize(); gfx.deallocate(capacity: gfxLength)
        
        // decode layout (only the first one is used)
        let layout = battleBackgroundProperties.intValue(forKey: "layout1")
        guard let layoutData = value(at: layout, inPropertyWithKey: "battleBackgroundLayouts") as? ROMData else { return nil }
        let backgroundTiles = layoutData.data
        
        // decode palette
        let palette = battleBackgroundProperties.intValue(forKey: "palette")
        guard let paletteData = value(at: palette, inPropertyWithKey: "battleBackgroundPalettes") as? ROMData else { return nil }
        let palLength = 0x100 * MemoryLayout<UInt32>.size
        let pal = UnsafeMutablePointer<UInt8>.allocate(capacity: palLength)
        memset(pal, 0, palLength)
        paletteData.data.withUnsafeBytes{ (bytes: UnsafePointer<UInt8>) in
            gfx_decode(pal.advanced(by: 0x0140), bytes, Int32(paletteData.data.count), BGR555)
        }
        var backAreaColor = 0xFF000000
        memcpy(pal, &backAreaColor, 4)
        let backgroundPalette = Data(bytes: UnsafePointer<UInt8>(pal), count: palLength)
        pal.deinitialize(); pal.deallocate(capacity: palLength)
        
        // create the ppu
        var ppu = PPU()
        memset(&ppu, 0, MemoryLayout<PPU>.size)
        ppu.pal = (backgroundPalette as NSData).bytes
        ppu.back_area = true
        
        // set up bg2
        ppu.layers.1.format = isGBA ? GBA_4BPP_TILE : SNES_4BPP_TILE
        ppu.layers.1.height = 32
        ppu.layers.1.width = 32
        ppu.layers.1.z.0 = Z_LEVEL_BG2_LO
        ppu.layers.1.z.1 = Z_LEVEL_BG2_HI
        ppu.layers.1.h_scroll = 0
        ppu.layers.1.v_scroll = 0
        ppu.layers.1.gfx = (backgroundGraphics as NSData).bytes
        ppu.layers.1.tiles = (backgroundTiles as NSData).bytes
        ppu.layers.1.main = true
        
        // render the background
        let height = 32 * 8
        let width = 32 * 8
        let dataLength = height * width * MemoryLayout<UInt32>.size
        let data = UnsafeMutablePointer<UInt8>.allocate(capacity: dataLength)
        memset(data, 0, dataLength)
        gfx_render_ppu(data, ppu, Int32(dataLength), Int32(width))
        let imageData = Data(bytes: UnsafePointer<UInt8>(data), count: dataLength)
        data.deinitialize(); data.deallocate(capacity: dataLength)
        
        // create an image
        let provider = CGDataProvider(dataInfo: nil, data: (imageData as NSData).bytes, size: dataLength, releaseData: releaseImageData)!
        let bitsPerComponent = 8
        let bitsPerPixel = 32
        let bytesPerRow = width * MemoryLayout<UInt32>.size
        let colorSpaceRef = CGColorSpaceCreateDeviceRGB()
        let bitmapInfo = CGBitmapInfo(rawValue: CGBitmapInfo().rawValue | CGImageAlphaInfo.premultipliedLast.rawValue)
        let renderingIntent = CGColorRenderingIntent.defaultIntent
        return CGImage(width: width, height: height, bitsPerComponent: bitsPerComponent, bitsPerPixel: bitsPerPixel, bytesPerRow: bytesPerRow, space: colorSpaceRef, bitmapInfo: bitmapInfo, provider: provider, decode: nil, shouldInterpolate: true, intent: renderingIntent)
    }
    
    func npcImage(_ index: Int, frame: Int, palette: Int) -> CGImage? {
        
        let hFlip = (frame & 0x40) != 0
        
        // get the sprite graphics and tile formation
        let mapSpriteGraphics = value(forKey: "mapSpriteGraphics") as! ROMData
        let mapSpriteTileFormation = value(forKey: "mapSpriteTileFormation") as! ROMArray
        let tileFormation = mapSpriteTileFormation[frame & 0x3F] as! ROMArray
        
        // get a pointer to the sprite graphics
        let mapSpriteGraphicsPointersLo = rom?.value(forKey: "mapSpriteGraphicsPointersLo") as! ROMArray
        let mapSpriteGraphicsPointersHi = rom?.value(forKey: "mapSpriteGraphicsPointersHi") as! ROMArray
        let gfxPointerLo = mapSpriteGraphicsPointersLo[index].value(forKey: "pointer") as! Int
        let gfxPointerHi = mapSpriteGraphicsPointersHi[index].value(forKey: "pointer") as! Int
        var gfxPointer = gfxPointerLo | (gfxPointerHi << 16)
        gfxPointer &= 0x00FFFFFF
        gfxPointer -= isGBA ? 0xF60000 : rom!.mapMode.unmapAddress(mapSpriteGraphics.range!.lowerBound)
        
        // decode graphics
        let tileCount = 6
        let gfxLength = 0x40 * tileCount
        let gfx = UnsafeMutablePointer<UInt8>.allocate(capacity: gfxLength)
        memset(gfx, 0, gfxLength)
        var tileData = [UInt16](repeating: 0, count: tileCount)
        for tile in 0..<tileCount {
            let tileOffset = tileFormation[tile].value(forKey: "offset") as! Int
            mapSpriteGraphics.data.withUnsafeBytes{(bytes: UnsafePointer<UInt8>)->Void in
                gfx_decode(gfx.advanced(by: tile * 0x40), bytes.advanced(by: gfxPointer + tileOffset), 0x20, isGBA ? LINEAR_4BPP : SNES_4BPP)
            }
            if hFlip {
                tileData[tile ^ 1] = UInt16(tile | 0x4000)
            } else {
                tileData[tile] = UInt16(tile)
            }
        }
        let spriteGraphics = Data(bytes: UnsafePointer<UInt8>(gfx), count: gfxLength)
        gfx.deinitialize(); gfx.deallocate(capacity: gfxLength)
        
        // decode palette
        let mapSpritePalettes = rom?.value(forKey: "mapSpritePalettes") as! ROMArray
        let paletteData = mapSpritePalettes[palette].value(forKey: "data") as! Data
        let palLength = 0x10 * MemoryLayout<UInt32>.size
        let pal = UnsafeMutablePointer<UInt8>.allocate(capacity: palLength)
        memset(pal, 0, palLength)
        gfx_decode(pal, (paletteData as NSData).bytes, Int32(paletteData.count), BGR555)
        let spritePalette = Data(bytes: UnsafePointer<UInt8>(pal), count: palLength)
        pal.deinitialize(); pal.deallocate(capacity: palLength)
        
        // create the ppu
        var ppu = PPU()
        memset(&ppu, 0, MemoryLayout<PPU>.size)
        ppu.pal = (spritePalette as NSData).bytes
        
        // image size (in pixels)
        let width = 16
        let height = 24

        // draw sprite
        ppu.layers.0.main = true
        ppu.layers.0.format = SNES_4BPP_TILE // this works for the gba version too
        ppu.layers.0.gfx = (spriteGraphics as NSData).bytes
        ppu.layers.0.height = Int32(height / 8)
        ppu.layers.0.width = Int32(width / 8)
        ppu.layers.0.z.0 = Z_LEVEL_SPRITE_0
        ppu.layers.0.z.1 = Z_LEVEL_SPRITE_1
        ppu.layers.0.z.2 = Z_LEVEL_SPRITE_2
        ppu.layers.0.z.3 = Z_LEVEL_SPRITE_3
        ppu.layers.0.tiles = UnsafeRawPointer(tileData)
        
        // render the sprite
        let dataLength = height * width * MemoryLayout<UInt32>.size
        let data = UnsafeMutablePointer<UInt8>.allocate(capacity: dataLength)
        memset(data, 0, dataLength)
        gfx_render_ppu(data, ppu, Int32(dataLength), Int32(width))
        
        // create an image
        let provider = CGDataProvider(dataInfo: nil, data: data, size: height * width * 4, releaseData: releaseImageData)!
        let bitsPerComponent = 8
        let bitsPerPixel = 32
        let bytesPerRow = 4 * width
        let colorSpaceRef = CGColorSpaceCreateDeviceRGB()
        let bitmapInfo = CGBitmapInfo(rawValue: CGBitmapInfo().rawValue | CGImageAlphaInfo.premultipliedLast.rawValue)
        let renderingIntent = CGColorRenderingIntent.defaultIntent
        let iref = CGImage(width: width, height: height, bitsPerComponent: bitsPerComponent, bitsPerPixel: bitsPerPixel, bytesPerRow: bytesPerRow, space: colorSpaceRef, bitmapInfo: bitmapInfo, provider: provider, decode: nil, shouldInterpolate: true, intent: renderingIntent)
        
        // free memory
        data.deinitialize(); data.deallocate(capacity: dataLength)

        return iref
    }
    
    func battleCharacterImage(_ index: Int, frame: Int) -> CGImage? {
        // default frame (facing forward)
        var hFlip = false
        var frame = frame
        if frame > 0x30 {
            frame -= 0x30
            hFlip = true
        }
        
        // image size (in pixels)
        let width = 16
        let height = 32
        
        // get the sprite graphics, tile formation, and palette
        let mapSpriteGraphics = value(forKey: "mapSpriteGraphics") as! ROMData
        let tileFormationArray = mutableArrayValue(forKey: "battleSpriteTileFormation")
        if frame >= tileFormationArray.count { return nil }
        let tileFormation = tileFormationArray.object(at: frame) as! ROMArray
        
        let paletteArray = mutableArrayValue(forKey: "battleSpritePaletteIndexes")
        if index >= paletteArray.count { return nil }
        let paletteProperty = paletteArray.object(at: index) as! ROMProperty
        let paletteIndex = paletteProperty.property
        
        // load the sprite graphics pointer table
        let battleSpriteGraphicsPointerTable = value(forKey: "battleSpriteGraphicsPointers") as! ROMPointerTable
        
        // create a dummy graphics array to hold the pointer table
        let dummyGraphicsArray = ROMArray(self)
        dummyGraphicsArray.range = mapSpriteGraphics.range
        dummyGraphicsArray.pointerTable = battleSpriteGraphicsPointerTable
        dummyGraphicsArray.initialCount = battleSpriteGraphicsPointerTable.pointers.count
        dummyGraphicsArray.disassembleFrom(self.data)
        let dummyGraphics = dummyGraphicsArray[index] as! ROMData

        // decode graphics
        let tileCount = 8
        let gfxLength = 0x40 * tileCount
        let gfx = UnsafeMutablePointer<UInt8>.allocate(capacity: gfxLength)
        memset(gfx, 0, gfxLength)
        var tileData = [UInt16](repeating: 0, count: tileCount)
        for tile in 0..<tileCount {
            let tileOffset = tileFormation[tile].value(forKey: "offset") as! Int
            if tileOffset == 0xFFFF { continue } // skip blank tiles
            dummyGraphics.data.withUnsafeBytes{ (bytes: UnsafePointer<UInt8>) in
                gfx_decode(gfx.advanced(by: tile * 0x40), bytes.advanced(by: tileOffset), 0x20, isGBA ? LINEAR_4BPP : SNES_4BPP)
            }
            if hFlip {
                tileData[tile ^ 1] = UInt16(tile | 0x4000)
            } else {
                tileData[tile] = UInt16(tile)
            }
        }
        
        let spriteGraphics = Data(bytes: UnsafePointer<UInt8>(gfx), count: gfxLength)
        gfx.deinitialize(); gfx.deallocate(capacity: gfxLength)

        // decode palette
        let mapSpritePalettes = value(forKey: "battleSpritePalettes") as! ROMArray
        let paletteData = mapSpritePalettes[paletteIndex].value(forKey: "data") as! Data
        let palLength = 0x10 * MemoryLayout<UInt32>.size
        let pal = UnsafeMutablePointer<UInt8>.allocate(capacity: palLength)
        memset(pal, 0, palLength)
        gfx_decode(pal, (paletteData as NSData).bytes, Int32(paletteData.count), BGR555)
        let spritePalette = Data(bytes: UnsafePointer<UInt8>(pal), count: palLength)
        pal.deinitialize(); pal.deallocate(capacity: palLength)
        
        // create the ppu
        var ppu = PPU()
        memset(&ppu, 0, MemoryLayout<PPU>.size)
        ppu.pal = (spritePalette as NSData).bytes
        
        ppu.layers.0.main = true
        ppu.layers.0.format = SNES_4BPP_TILE // this works for the gba version too
        ppu.layers.0.gfx = (spriteGraphics as NSData).bytes
        ppu.layers.0.height = Int32(height / 8)
        ppu.layers.0.width = Int32(width / 8)
        ppu.layers.0.z.0 = Z_LEVEL_SPRITE_0
        ppu.layers.0.z.1 = Z_LEVEL_SPRITE_1
        ppu.layers.0.z.2 = Z_LEVEL_SPRITE_2
        ppu.layers.0.z.3 = Z_LEVEL_SPRITE_3
        ppu.layers.0.tiles = UnsafeRawPointer(tileData)
        
        // render the sprite
        let dataLength = height * width * MemoryLayout<UInt32>.size
        let data = UnsafeMutablePointer<UInt8>.allocate(capacity: dataLength)
        memset(data, 0, dataLength)
        gfx_render_ppu(data, ppu, Int32(dataLength), Int32(width))
        let imageData = Data(bytes: UnsafePointer<UInt8>(data), count: dataLength)
        data.deinitialize(); data.deallocate(capacity: dataLength)

        // create an image
        let provider = CGDataProvider(dataInfo: nil, data: (imageData as NSData).bytes, size: height * width * 4, releaseData: releaseImageData)!
        let bitsPerComponent = 8
        let bitsPerPixel = 32
        let bytesPerRow = 4 * width
        let colorSpaceRef = CGColorSpaceCreateDeviceRGB()
        let bitmapInfo = CGBitmapInfo(rawValue: CGBitmapInfo().rawValue | CGImageAlphaInfo.premultipliedLast.rawValue)
        let renderingIntent = CGColorRenderingIntent.defaultIntent
        let iref = CGImage(width: width, height: height, bitsPerComponent: bitsPerComponent, bitsPerPixel: bitsPerPixel, bytesPerRow: bytesPerRow, space: colorSpaceRef, bitmapInfo: bitmapInfo, provider: provider, decode: nil, shouldInterpolate: true, intent: renderingIntent)
        return iref
    }
    
    func characterShadowImage() -> CGImage? {
        
        // image size (in pixels)
        let width = 16
        let height = 16

        // decode the shadow palette
        guard let shadowPaletteData = value(forKey: "battleStatusPalettes") as? ROMData else { return nil }
        let palLength = 0x40 * MemoryLayout<UInt32>.size
        let pal = UnsafeMutablePointer<UInt8>.allocate(capacity: palLength)
        memset(pal, 0, palLength)
        for index in 0..<4 {
            let _shadowPaletteData = shadowPaletteData.data
            _shadowPaletteData.withUnsafeBytes{ (bytes: UnsafePointer<UInt8>) in
                gfx_decode(pal.advanced(by: (index * 16 + 12) * 4), bytes.advanced(by: 8 * index), 8, BGR555)
            }
        }
        let shadowPalette = Data(bytes: UnsafePointer<UInt8>(pal), count: palLength)
        pal.deinitialize(); pal.deallocate(capacity: palLength)
        
        // decode shadow tile graphics
        let shadowGraphicsBufferLength = 0x4000
        let shadowGraphicsBuffer = UnsafeMutablePointer<UInt8>.allocate(capacity: shadowGraphicsBufferLength)
        guard let statusGraphics = value(forKey: "battleStatusGraphics") as? ROMData else { return nil }
        statusGraphics.data.withUnsafeBytes{ (bytes: UnsafePointer<UInt8>) in
            gfx_decode(shadowGraphicsBuffer, bytes, Int32(statusGraphics.data.count), SNES_4BPP)
        }
        let shadowGraphics = Data(bytes: UnsafePointer<UInt8>(shadowGraphicsBuffer), count: shadowGraphicsBufferLength)
        shadowGraphicsBuffer.deinitialize(); shadowGraphicsBuffer.deallocate(capacity: shadowGraphicsBufferLength)
        let shadowGraphicsTiles: [UInt16] = [0x000C, 0x000D, 0x001C, 0x001D]

        // create the ppu
        var ppu = PPU()
        memset(&ppu, 0, MemoryLayout<PPU>.size)
        ppu.pal = (shadowPalette as NSData).bytes
        
        ppu.layers.1.main = true
        ppu.layers.1.format = SNES_4BPP_TILE
        ppu.layers.1.gfx = (shadowGraphics as NSData).bytes
        ppu.layers.1.height = Int32(height / 8)
        ppu.layers.1.width = Int32(width / 8)
        ppu.layers.0.z.0 = Z_LEVEL_SPRITE_0
        ppu.layers.0.z.1 = Z_LEVEL_SPRITE_1
        ppu.layers.0.z.2 = Z_LEVEL_SPRITE_2
        ppu.layers.0.z.3 = Z_LEVEL_SPRITE_3
        ppu.layers.1.tiles = UnsafeRawPointer(shadowGraphicsTiles)

        // render the sprite
        let dataLength = height * width * MemoryLayout<UInt32>.size
        let data = UnsafeMutablePointer<UInt8>.allocate(capacity: dataLength)
        memset(data, 0, dataLength)
        gfx_render_ppu(data, ppu, Int32(dataLength), Int32(width))
        let imageData = Data(bytes: UnsafePointer<UInt8>(data), count: dataLength)
        data.deinitialize(); data.deallocate(capacity: dataLength)
        
        // create an image
        let provider = CGDataProvider(dataInfo: nil, data: (imageData as NSData).bytes, size: height * width * 4, releaseData: releaseImageData)!
        let bitsPerComponent = 8
        let bitsPerPixel = 32
        let bytesPerRow = 4 * width
        let colorSpaceRef = CGColorSpaceCreateDeviceRGB()
        let bitmapInfo = CGBitmapInfo(rawValue: CGBitmapInfo().rawValue | CGImageAlphaInfo.premultipliedLast.rawValue)
        let renderingIntent = CGColorRenderingIntent.defaultIntent
        let iref = CGImage(width: width, height: height, bitsPerComponent: bitsPerComponent, bitsPerPixel: bitsPerPixel, bytesPerRow: bytesPerRow, space: colorSpaceRef, bitmapInfo: bitmapInfo, provider: provider, decode: nil, shouldInterpolate: true, intent: renderingIntent)
        return iref
    }
}
