//
//  MonsterScript.swift
//  FF6Tools
//
//  Created by Evan Owen on 6/4/16.
//
//

import Foundation

class MonsterScriptEncoding: ROMScriptEncoding {
    
    override func willDisassembleScript(_ script: ROMScript) {
        
        for (index, object) in script.rom.mutableArrayValue(forKey: "monsterScriptPointers").enumerated() {
            guard let scriptPointerAssembly = object as? ROMAssembly,
                let scriptPointer = scriptPointerAssembly.value(forKeyPath: "scriptPointer") as? Int else { continue }
            
            // make a label for each monster
            let monsterName = script.rom.string(atIndex: index, inStringTable: "monsterNames") ?? ("Monster " + String(index))
            script.commandLabels[scriptPointer + script.pointerOffset] = monsterName
            
            // create a reference for each monster script
            script.addReference(forAssembly: scriptPointerAssembly, encodingName: "monsterScript", linkedKeyPath: "scriptPointer")
        }
        
        super.willDisassembleScript(script)
    }
    
    override func didDisassembleCommand(_ command: ROMCommand, fromData data: Data) {
        
        if command.identifier!.hasPrefix("monsterScript.if") {
            // conditional
            var currentOffset = command.range!.upperBound
            while currentOffset < command.script.range!.upperBound {
                guard let conditionalCommand = ROMCommand(command.rom, script: command.script, encodingName: command.childEncodingName, data: data, offset: currentOffset) else { break }
                conditionalCommand.parent = command
                command.childCommands.add(conditionalCommand)
                currentOffset += conditionalCommand.assembledLength
                if conditionalCommand.identifier!.hasPrefix("monsterScript.if") { break }
                if conditionalCommand.identifier == "monsterScript.endIf" { break }
                if conditionalCommand.identifier == "monsterScript.endOfScript" { break }
            }
            if command.childCommands.count == 0 {
                // if the child array is empty, add an end if command
                guard let endOfScriptCommand = ROMCommand(command.rom, script: command.script, identifier: "monsterScript.endIf") else { return }
                endOfScriptCommand.parent = command
                command.childCommands.add(endOfScriptCommand)
            }
        }
    }
    
    override func numLinesForCommandDescription(_ command: ROMCommand) -> Int {
        if command.identifier == "monsterScript.monsterDialog" {
            
            let dialogIndex = command.intValue(forKey: "dialog")
            guard let dialog = command.rom.value(at: dialogIndex, inPropertyWithKey: "monsterDialog") as? ROMText else { return 2 }
            return dialog.numLines + 1
        }
        return super.numLinesForCommandDescription(command)
    }
    
    override func valueForCommand(_ command: ROMCommand, forKey key: String) -> Any? {
        
        if key == "switch" {
            if let bit = super.valueForCommand(command, forKey: "bit") as? Int,
                let variable = super.valueForCommand(command, forKey: "variable") as? Int {
                return (variable * 8 + bit) as AnyObject?
            }
        } else if key == "allMonsters" {
            if let monsterSlots = super.valueForCommand(command, forKey: "monsterSlots") as? Int {
                return (monsterSlots == 0xFF) as AnyObject?
            }
            
        } else if key == "thisMonster" {
            if let monsterSlots = super.valueForCommand(command, forKey: "monsterSlots") as? Int {
                return (monsterSlots == 0xFF) as AnyObject?
            }
        }
        
        return super.valueForCommand(command, forKey: key)
    }
    
    override func setValue(_ value: Any?, forCommand command: ROMCommand, forKey key: String) {
        if key == "switch" {
            let intValue = value as? Int ?? 0
            super.setValue((intValue % 8) as AnyObject, forCommand: command, forKey: "bit")
            super.setValue((intValue / 8) as AnyObject, forCommand: command, forKey: "variable")
            return
            
        } else if key == "monsterSlots" {
            if let intValue = value as? Int {
                command.willChangeValue(forKey: "allMonsters")
                command.willChangeValue(forKey: "thisMonster")
                if intValue == 0xFF {
                    // leave the top two bits set only for "all monsters"
                    super.setValue(0xFF as AnyObject, forCommand: command, forKey: key)
                } else {
                    // otherwise, clear the top two bits
                    super.setValue((intValue & 0x3F) as AnyObject, forCommand: command, forKey: key)
                }
                command.didChangeValue(forKey: "allMonsters")
                command.didChangeValue(forKey: "thisMonster")
            }
            return

        } else if key == "allMonsters" {
            if let boolValue = value as? Bool , boolValue == true {
                command.willChangeValue(forKey: "allMonsters")
                command.willChangeValue(forKey: "thisMonster")
                super.setValue(0xFF as AnyObject, forCommand: command, forKey: "monsterSlots")
                command.didChangeValue(forKey: "allMonsters")
                command.didChangeValue(forKey: "thisMonster")
            }
            return
            
        } else if key == "thisMonster" {
            if let boolValue = value as? Bool , boolValue == true {
                command.willChangeValue(forKey: "allMonsters")
                command.willChangeValue(forKey: "thisMonster")
                super.setValue(0 as AnyObject, forCommand: command, forKey: "monsterSlots")
                command.didChangeValue(forKey: "allMonsters")
                command.didChangeValue(forKey: "thisMonster")
            }
            return
        }
        
        super.setValue(value, forCommand: command, forKey: key)
    }
    
    override func descriptionForCommand(_ command: ROMCommand) -> String {
        
        func attackName(_ key: String = "attack") -> String {
            let attack = command.intValue(forKey: key)
            return rom.string(atIndex: attack, inStringTable: "monsterAttacks") ?? "Battle Command " + String(attack)
        }
        
        func battleCommandName(_ key: String = "command") -> String {
            let battleCommand = command.intValue(forKey: key)
            return rom.string(atIndex: battleCommand, inStringTable: "battleCommandNames") ?? "Battle Command " + String(battleCommand)
        }
        
        func itemName(_ key: String = "item") -> String {
            let item = command.intValue(forKey: key)
            return rom.string(atIndex: item, inStringTable: "itemNames") ?? "Item " + String(item)
        }
        
        func targetName(_ key: String = "target") -> String {
            let target = command.intValue(forKey: key)
            return rom.string(atIndex: target, inStringTable: "targets") ?? "Target " + String(target)
        }
        
        func variableName(_ key: String = "variable") -> String {
            let variable = command.intValue(forKey: key)
            return rom.string(atIndex: variable, inStringTable: "battleVariables") ?? "Battle Variable " + String(variable)
        }
        
        func elementName(_ key: String = "element") -> String {
            let element = command.intValue(forKey: key)
            return rom.string(atIndex: element, inStringTable: "elements") ?? "Element " + String(element)
        }
        
        func elementNameFlags(_ key: String = "element") -> String {
            
            var value = command.intValue(forKey: key)
            if value == 0 { return "None" }
            
            // make an array of the bits that are set
            var elements = [String]()
            for i in 0..<8 {
                if (value & 1) == 1 {
                    elements.append(rom.string(atIndex: i, inStringTable: "elements") ?? "Element " + String(i))
                }
                value >>= 1
                if value == 0 { break }
            }
            
            // format a string of element names
            var elementString: String
            if elements.count == 1 {
                elementString = elements[0]
            } else if elements.count == 2 {
                elementString = elements[0] + " or " + elements[1]
            } else {
                elementString = ""
                while elements.count != 1 {
                    elementString += elements.remove(at: 0) + ", "
                }
                elementString += "or " + elements[0]
            }
            return elementString
        }
        
        func monsterSlotFlags(_ key: String = "monsterSlots") -> String {
            
            var value = command.intValue(forKey: key)
            if value == 0 { return "This Monster" }
            else if value == 0xFF { return "All Monsters" }
            
            // make an array of the bits that are set
            var monsters = [Int]()
            for i in 1...6 {
                if (value & 1) == 1 {
                    monsters.append(i)
                }
                value >>= 1
                if value == 0 { break }
            }
            
            // format a string of monster indexes
            var monsterString: String
            if monsters.count == 1 {
                monsterString = "Slot " + String(monsters[0])
            } else if monsters.count == 2 {
                monsterString = "Slot " + String(monsters[0]) + " or " + String(monsters[1])
            } else {
                monsterString = "Slot "
                while monsters.count != 1 {
                    monsterString += String(monsters.remove(at: 0)) + ", "
                }
                monsterString += "or " + String(monsters[0])
            }
            return monsterString
        }

        switch command.identifier! {

        case "monsterScript.attack":
            return "Use attack: " + attackName()

        case "monsterScript.randomAttack":
            let attack1 = command.intValue(forKey: "attack1")
            let attack2 = command.intValue(forKey: "attack2")
            let attack3 = command.intValue(forKey: "attack3")
            if attack1 == attack2 && attack2 == attack3 {
                return "Use attack: " + attackName("attack1")
            } else {
                return "Use attack: " + attackName("attack1") + ", " + attackName("attack2") + ", or " + attackName("attack3")
            }

        case "monsterScript.target":
            return "Change target: " + targetName()

        case "monsterScript.changeBattle":
            let battle = command.intValue(forKey: "battle")
            let battleName = rom.string(atIndex: battle, inStringTable: "battleProperties") ?? "Battle " + String(battle)
            return "Change battle: " + battleName
            
        case "monsterScript.monsterDialog":
            let dialogIndex = command.intValue(forKey: "dialog")
            guard let dialog = command.rom.value(at: dialogIndex, inPropertyWithKey: "monsterDialog") as? ROMText else {
                return "<Invalid dialog>"
            }
            return "Display monster dialog:\n" + dialog.formattedString()

        case "monsterScript.randomCommand":
            let command1 = command.intValue(forKey: "command1")
            let command2 = command.intValue(forKey: "command2")
            let command3 = command.intValue(forKey: "command3")
            if command1 == command2 && command2 == command3 {
                return "Use command: " + battleCommandName("command1")
            } else {
                return "Use command: " + battleCommandName("command1") + ", " + battleCommandName("command2") + ", or " + battleCommandName("command3")
            }

        case "monsterScript.entryExit":
            return "Monster entry/exit"

        case "monsterScript.item":
            let useThrow = command.boolValue(forKey: "throw")
            let item1 = command.intValue(forKey: "item1")
            let item2 = command.intValue(forKey: "item2")
            if item1 == item2 {
                return (useThrow ? "Throw" : "Use") + " item: " + itemName("item1")
            } else {
                return (useThrow ? "Throw" : "Use") + " item: " + itemName("item1") + " or " + itemName("item2")
            }

        case "monsterScript.battleScript":
            let event = command.intValue(forKey: "scriptIndex")
            let eventName = command.rom.string(atIndex: event, inStringTable: "battleScriptPointers") ?? "<Invalid event>"
            return "Battle Event: " + eventName

        case "monsterScript.variable":
            let value = command.intValue(forKey: "value")
            let operation = command.intValue(forKey: "operation")
            switch operation {
            case 0: return variableName() + " = " + String(value)
            case 1: return variableName() + " = " + String(value)
            case 2: return variableName() + " += " + String(value)
            case 3: return variableName() + " -= " + String(value)
            default: return "<Invalid battle variable operation>"
            }

        case "monsterScript.switch":
            let operation = command.intValue(forKey: "operation")
            let battleSwitch = command.intValue(forKey: "switch")
            let switchName = rom.string(atIndex: battleSwitch, inStringTable: "battleSwitches") ?? "Battle Switch " + String(battleSwitch)
            switch operation {
            case 0: return "Toggle " + switchName
            case 1: return switchName + " = On"
            case 2: return switchName + " = Off"
            default: return "<Invalid battle switch operation>"
            }
            
        case "monsterScript.monsterAnimation":
            let animation = command.intValue(forKey: "animation")
            let animationName = rom.string(atIndex: animation, inStringTable: "monsterAnimations") ?? "Animation " + String(animation)
            return "Monster Animation: " + animationName

        case "monsterScript.soundEffect":
            let soundEffect = command.intValue(forKey: "soundEffect")
            let soundEffectName = rom.string(atIndex: soundEffect, inStringTable: "gameSoundEffects") ?? "Sound Effect " + String(soundEffect)
            return "Play Sound Effect: " + soundEffectName

        case "monsterScript.resetMonsterTimer":
            return "Reset this Monster's Timer"

        case "monsterScript.resetBattleTimer":
            return "Reset the Battle Timer"

        case "monsterScript.terminateBattle":
            return "End Battle Immediately"

        case "monsterScript.addGau":
            let gauName = rom.string(atIndex: 11, inStringTable: "characterNames") ?? "GAU"
            return "Add " + gauName + " to the party"

        case "monsterScript.veldtGau":
            let gauName = rom.string(atIndex: 11, inStringTable: "characterNames") ?? "GAU"
            return gauName + " returns from the veldt"

        case "monsterScript.piranha":
            return "Hide this Monster (Piranha)"
            
        case "monsterScript.status":
            let status = command.intValue(forKey: "status")
            let statusName = rom.string(atIndex: status, inStringTable: "statusNamesReversed") ?? "Status " + String(status)
            switch command.intValue(forKey: "opcode") {
            case 0x0BFB: return "Set " + statusName + " Status (This Monster)"
            case 0x0CFB: return "Remove " + statusName + " Status (This Monster)"
            default: return "<Invalid Status Command>"
            }

        case "monsterScript.misc":
            switch command.intValue(forKey: "opcode") {
            case 0x01FB: return targetName() + " can't be damaged"
            case 0x05FB: return targetName() + " can be damaged"
            case 0x07FB: return targetName() + " can't be targeted"
            case 0x06FB: return targetName() + " can be targeted"
            case 0x08FB: return "If " + targetName() + " is a valid target, fill this monster's ATB gauge"
            default: return "<Invalid Misc. Command>"
            }

        case "monsterScript.ifNever":
            return "Never continue"

        case "monsterScript.ifCommand":
            if command.intValue(forKey: "command1") == command.intValue(forKey: "command2") {
                return "If " + battleCommandName("command1") + " was used against this monster"
            } else {
                return "If " + battleCommandName("command1") + " or " + battleCommandName("command2") + " was used against this monster"
            }
                
        case "monsterScript.ifAttack":
            if command.intValue(forKey: "attack1") == command.intValue(forKey: "attack2") {
                return "If " + attackName("attack1") + " was used against this monster"
            } else {
                return "If " + attackName("attack1") + " or " + attackName("attack2") + " was used against this monster"
            }

        case "monsterScript.ifItem":
            if command.intValue(forKey: "item1") == command.intValue(forKey: "item2") {
                return "If " + itemName("item1") + " was used against this monster"
            } else {
                return "If " + itemName("item1") + " or " + itemName("item2") + " was used against this monster"
            }

        case "monsterScript.ifElement":
            return "If " + elementNameFlags() + " was used against this monster"

        case "monsterScript.ifAnyAction":
            return "If any action was used against this monster"

        case "monsterScript.ifHP":
            let hp = command.intValue(forKey: "hp")
            return "If " + targetName() + "'s HP ≤ " + String(hp * 128)

        case "monsterScript.ifMP":
            let mp = command.intValue(forKey: "mp")
            return "If " + targetName() + "'s MP ≤ " + String(mp)

        case "monsterScript.ifStatus":
            let status = command.intValue(forKey: "status")
            let statusName = rom.string(atIndex: status, inStringTable: "statusNamesReversed") ?? "Status " + String(status)
            switch command.intValue(forKey: "opcode") {
            case 0x08FC: return "If " + targetName() + " has status: " + statusName
            case 0x09FC: return "If " + targetName() + " does not have status: " + statusName
            default: return "<Invalid Status Conditional>"
            }

        case "monsterScript.ifMonsterTimer":
            let value = command.intValue(forKey: "value")
            return "If this monster's timer has reached " + String(value * 2)

        case "monsterScript.ifVariable":
            let greater = command.boolValue(forKey: "greater")
            let value = command.intValue(forKey: "value")
            return "If " + variableName() + (greater ? " ≥ " : " < ") + String(value)

        case "monsterScript.ifLevel":
            let greater = command.boolValue(forKey: "greater")
            let level = command.intValue(forKey: "level")
            return "If " + targetName() + "'s level" + (greater ? " ≥ " : " < ") + String(level)

        case "monsterScript.ifAlone":
            return "If only one type of monster is remaining"

        case "monsterScript.ifNumTargets":
            let count = command.intValue(forKey: "count")
            if command.boolValue(forKey: "monsters") {
                return "If there are " + String(count) + " or fewer monsters remaining"
            } else {
                return "If there are " + String(count) + " or more characters remaining"
            }

        case "monsterScript.ifSwitch":
            let clear = command.boolValue(forKey: "clear")
            let battleSwitch = command.intValue(forKey: "switch")
            let switchName = rom.string(atIndex: battleSwitch, inStringTable: "battleSwitches") ?? "Battle Switch " + String(battleSwitch)
            return "If " + switchName + (clear ? " == Off" : " == On")

        case "monsterScript.ifBattleTimer":
            let value = command.intValue(forKey: "value")
            return "If the battle timer has reached " + String(value * 2)

        case "monsterScript.ifTargetValid":
            return "If " + targetName() + " is a valid target"

        case "monsterScript.ifGau":
            let gauName = (command.rom.value(at: 11, inPropertyWithKey: "characterNames") as AnyObject).description ?? "Gau"
            return "If " + gauName + " is in the party"

        case "monsterScript.ifMonsterSlot":
            switch command.intValue(forKey: "opcode") {
            case 0x11FC: return "If " + monsterSlotFlags() + " is alive"
            case 0x12FC: return "If " + monsterSlotFlags() + " is dead"
            case 0x19FC: return "If this monster is in " + monsterSlotFlags()
            default: return "<Invalid monster slot command>"
            }

        case "monsterScript.ifElementWeakness":
            return "If " + targetName() + " is weak against " + elementNameFlags()

        case "monsterScript.ifBattleIndex":
            let battleIndex = command.intValue(forKey: "battleIndex")
            return "If the current battle index is " + String(battleIndex)

        case "monsterScript.ifAlways":
            return "Always continue"
            
        case "monsterScript.wait":
            return "Wait until next turn"

        case "monsterScript.endIf":
            return "End if"

        case "monsterScript.endOfScript":
            return "End of Script"

        default:
            return super.descriptionForCommand(command)
        }
    }
}
