//
//  ROMGraphics.swift
//  FF6Tools
//
//  Created by Evan Owen on 5/22/17.
//  Copyright © 2017 Evan Owen. All rights reserved.
//

import Foundation

enum ROMGraphicsFormat: String {
    case Linear8bpp =       "Linear8bpp"
    case Linear2bpp =       "Linear2bpp"
    case Linear4bpp =       "Linear4bpp"
    case NES1bpp =          "NES1bpp"
    case NES2bpp =          "NES2bpp"
    case SNES2bpp =         "SNES2bpp"
    case SNES3bpp =         "SNES3bpp"
    case SNES4bpp =         "SNES4bpp"
    case Genesis4bpp =      "Genesis4bpp"
    case SNES8bpp =         "SNES8bpp"
    case SNESMode7 =        "SNESMode7"
    case SNESMode7Ex =      "SNESMode7Ex"
    case SNESMode7Zelda =   "SNESMode7Zelda"
    case Default =          "Default"
    
    var rawFormat: Int32 {
        switch self {
        case .Linear8bpp:       return LINEAR_8BPP
        case .Linear2bpp:       return LINEAR_2BPP
        case .Linear4bpp:       return LINEAR_4BPP
        case .NES1bpp:          return NES_1BPP
        case .NES2bpp:          return NES_2BPP
        case .SNES2bpp:         return SNES_2BPP
        case .SNES3bpp:         return SNES_3BPP
        case .SNES4bpp:         return SNES_4BPP
        case .Genesis4bpp:      return GENESIS_4BPP
        case .SNES8bpp:         return SNES_8BPP
        case .SNESMode7:        return SNES_MODE7
        case .SNESMode7Ex:      return SNES_MODE7_EX
        case .SNESMode7Zelda:   return SNES_MODE7_ZELDA
        case .Default:          return DEFAULT_GRAPHICS
        }
    }
    
    var bytesPerTile: Int {
        switch self {
        case .NES1bpp:
            return 8
        case .Linear2bpp, .NES2bpp, .SNES2bpp:
            return 16
        case .SNES3bpp:
            return 24
        case .Linear4bpp, .SNES4bpp, .Genesis4bpp:
            return 32
        case .Linear8bpp, .SNES8bpp, .SNESMode7, .SNESMode7Ex, .SNESMode7Zelda, .Default:
            return 64
        }
    }
}

class ROMGraphics: ROMData {
    
    var format = ROMGraphicsFormat.Default
    var palette: String?
    var wrapWidth: Int
        
    override init(_ rom: ROM, definition: NSDictionary?) {
        if let formatName = definition?["format"] as? String,
            let format = ROMGraphicsFormat(rawValue: formatName) {
            self.format = format
        }
        
        palette = definition?["palette"] as? String
        wrapWidth = Int(object: definition, keyPath: "wrapWidth") ?? 16

        super.init(rom, definition: definition)
    }
    
    override func createDefinition() -> NSMutableDictionary {
        let definition = super.createDefinition()
        
        definition["type"] = "ROMGraphics"
        
        if format != .Default { definition["format"] = format.rawValue }
        definition["palette"] = palette
        if wrapWidth != 16 {
            definition["wrapWidth"] = wrapWidth
        }
        
        return definition
    }
    
    var image: CGImage? {
        
        // image size (in pixels)
        let tileCount = (self.data.count + format.bytesPerTile - 1) / format.bytesPerTile
        let width = wrapWidth * 8
        let height = (tileCount + wrapWidth - 1) / wrapWidth * 8
        
        var palette = Data()
        if let paletteIdentifier = self.palette,
            let paletteData = rom.value(forKeyPath: paletteIdentifier.replacingOccurrences(of: "%n", with: String(index))) as? ROMData {
            let palLength = 0x100 * MemoryLayout<UInt32>.size
            let pal = UnsafeMutablePointer<UInt8>.allocate(capacity: palLength)
            memset(pal, 0, palLength)
            paletteData.data.withUnsafeBytes{(bytes: UnsafePointer<UInt8>)->Void in
                gfx_decode(pal, bytes, Int32(paletteData.data.count), BGR555)
            }
            palette = Data(bytes: UnsafePointer<UInt8>(pal), count: palLength)
            pal.deinitialize(); pal.deallocate(capacity: palLength)
        } else {
            // create a grayscale palette
            let colorCount = 1 << (format.bytesPerTile / 8)
            var palBuffer = [UInt32](repeating: 0, count: colorCount)
            for index in 0..<colorCount {
                let doubleValue = Double(colorCount - index - 1) / Double(colorCount - 1) * 255.0
                let intValue = Int(doubleValue) & 0xFF
                palBuffer[index] = UInt32(intValue | (intValue << 8) | (intValue << 16) | (0xFF << 24))
            }
            palette = Data(bytes: &palBuffer, count: colorCount * MemoryLayout<UInt32>.size)
        }
        
        // decode graphics
        let gfxBufferLength = tileCount * 64
        let gfxBuffer = UnsafeMutablePointer<UInt8>.allocate(capacity: gfxBufferLength)
        memset(gfxBuffer, 0, gfxBufferLength)
        self.data.withUnsafeBytes{ (bytes: UnsafePointer<UInt8>) in
            gfx_decode(gfxBuffer, bytes, Int32(self.data.count), format.rawFormat)
        }
        var graphics = Data(bytes: UnsafePointer<UInt8>(gfxBuffer), count: gfxBufferLength)
        graphics.count = width * height * 256
        gfxBuffer.deinitialize(); gfxBuffer.deallocate(capacity: gfxBufferLength)
        
        // render the graphics
        let dataLength = graphics.count * MemoryLayout<UInt32>.size
        let data = UnsafeMutablePointer<UInt8>.allocate(capacity: dataLength)
        memset(data, 0, dataLength)
        gfx_render(data, (graphics as NSData).bytes, (palette as NSData).bytes, Int32(graphics.count), Int32(width))
        var imageData = Data(bytes: UnsafePointer<UInt8>(data), count: dataLength)
        imageData.count = height * width * MemoryLayout<UInt32>.size
        data.deinitialize(); data.deallocate(capacity: dataLength)
        
        // create an image
        let provider = CGDataProvider(dataInfo: nil, data: (imageData as NSData).bytes, size: imageData.count, releaseData: releaseImageData)!
        let bitsPerComponent = 8
        let bitsPerPixel = 32
        let bytesPerRow = 4 * width
        let colorSpaceRef = CGColorSpaceCreateDeviceRGB()
        let bitmapInfo = CGBitmapInfo(rawValue: CGBitmapInfo().rawValue | CGImageAlphaInfo.premultipliedLast.rawValue)
        let renderingIntent = CGColorRenderingIntent.defaultIntent
        let iref = CGImage(width: width, height: height, bitsPerComponent: bitsPerComponent, bitsPerPixel: bitsPerPixel, bytesPerRow: bytesPerRow, space: colorSpaceRef, bitmapInfo: bitmapInfo, provider: provider, decode: nil, shouldInterpolate: true, intent: renderingIntent)
        return iref
    }
}
