//
//  ROMScriptController.swift
//  FF6Tools
//
//  Created by Evan Owen on 4/15/16.
//  Copyright © 2016 Evan Owen. All rights reserved.
//

// TODO: add a column to show the encoding of each command (icon?)
// TODO: observe all visible commands instead of just the selected command

import Cocoa

class ROMScriptController: NSResponder, NSOutlineViewDataSource, NSOutlineViewDelegate, NSMenuDelegate {

    @IBOutlet var scriptView: NSOutlineView!
    @IBOutlet var parentView: NSView!

    var draggedCommands = NSMutableArray()
    var selectedObject: AnyObject?
    @IBOutlet weak var windowController: ROMWindowController?
    static var context = 0

    var script: ROMScript? {
        didSet {
            if script == oldValue { return }
            scriptView.reloadData()
            scriptView.expandItem(nil, expandChildren: true)
            scriptView.scrollRowToVisible(0)
            scriptView.deselectAll(self)
        }
    }

    override func awakeFromNib() {
        
        // register for drag and drop
        scriptView.registerForDraggedTypes([NSPasteboard.PasteboardType(rawValue: "com.ekowen86.romtools.ROMCommand")])
        scriptView.setDraggingSourceOperationMask(NSDragOperation.every, forLocal: false)
        scriptView.doubleAction = #selector(ROMScriptController.doubleClickedCommand(_:))
        scriptView.target = self
        
        // insert the controller in the responder chain to intercept cut/copy/paste messages
        self.nextResponder = scriptView.nextResponder
        scriptView.nextResponder = self
    }
    
    @IBAction func cut(_ sender: AnyObject) {
        // copy the command
        copy(sender)
        
        let selectedIndexSet = scriptView.selectedRowIndexes
        if (scriptView.clickedRow == -1) || (selectedIndexSet.contains(scriptView.clickedRow)) {
            // if the clicked item is selected, delete all selected commands
            delete(sender)
        } else if let clickedCommand = scriptView.item(atRow: scriptView.clickedRow) as? ROMCommand {
            // otherwise, delete only the clicked command
            remove(commands: [clickedCommand], script: script)
        }
    }

    @IBAction func copy(_ sender: AnyObject) {
        
        // create an array of the selected commands
        let commandArray = NSMutableArray()
        let selectedIndexSet = scriptView.selectedRowIndexes
        if (scriptView.clickedRow == -1) || (selectedIndexSet.contains(scriptView.clickedRow)) {
            // if the clicked item is selected, copy all selected commands
            for i in 0..<scriptView.numberOfRows {
                if !selectedIndexSet.contains(i) { continue }
                guard let commandToCopy = scriptView.item(atRow: i) as? ROMCommand else { continue }
                commandArray.add(commandToCopy)
            }
        } else if let commandToCopy = scriptView.item(atRow: scriptView.clickedRow) as? ROMCommand {
            // otherwise, copy only the clicked command
            commandArray.add(commandToCopy)
        }
        guard let firstCommand = commandArray.firstObject as? ROMCommand else { return }
        
        // start with an empty NSMutableData
        var copiedData = Data()
        let copiedReferences = NSMutableDictionary()
        
        // assemble each child command
        var copyString = ""
        
        // recursive function to copy child commands
        func copyCommand(_ command: ROMCommand) {
            // copy the long description string
            copyString += command.longDescription
            
            // copy the command's data
            command.updateReferences()
            command.encoding.willAssembleCommand(command)
            copiedData.append(command.data)
            command.encoding.didAssembleCommand(command)
            
            // copy the command's references
            if let commandReferences = command.references {
                for (_, reference) in commandReferences {
                    guard let referencedCommand = reference.referencedAssembly as? ROMCommand,
                        let referenceOffset = referencedCommand.range?.lowerBound else { continue }
                    
                    copiedReferences[String(referenceOffset)] = referencedCommand.reference
                }
            }
            
            // copy child commands
            for object in command.childCommands {
                guard let childCommand = object as? ROMCommand else { continue }
                copyCommand(childCommand)
            }
        }

        // stop observing to avoid triggering undo when a command updates its references
        windowController?.stopObservingAllObjects()
        for object in commandArray {
            guard let command = object as? ROMCommand else { continue }
        
            // only assemble commands with the same parent as the first selected item
            if command.parent != firstCommand.parent { continue }
            
            // assemble the command
            copyCommand(command)
        }
        windowController?.startObservingAllObjects()
        
        // clear the pasteboard
        let pasteboard = NSPasteboard.general
        pasteboard.clearContents()
        
        // attach the assembled data to an NSPasteboard item and add it to the pasteboard
        // note: these commands will not retain any reference data since they are copies
        let pasteboardItem = NSPasteboardItem()
        pasteboardItem.setString(firstCommand.encodingName, forType: NSPasteboard.PasteboardType(rawValue: "com.ekowen86.romtools.ROMCommand.encoding"))
        pasteboardItem.setData(copiedData, forType: NSPasteboard.PasteboardType(rawValue: "com.ekowen86.romtools.ROMCommand"))
        pasteboardItem.setString(copyString, forType: NSPasteboard.PasteboardType.string)
        pasteboardItem.setPropertyList(copiedReferences, forType: NSPasteboard.PasteboardType(rawValue: "com.ekowen86.romtools.ROMCommand.references"))
        pasteboard.writeObjects([pasteboardItem])
    }

    @IBAction func paste(_ sender: AnyObject) {
        
        let previousCommand: ROMCommand
        if let clickedCommand = scriptView.item(atRow: scriptView.clickedRow) as? ROMCommand {
            // insert after the clicked command if it is valid
            previousCommand = clickedCommand
        } else if let selectedCommand = scriptView.item(atRow: scriptView.selectedRow) as? ROMCommand {
            // insert after the selected command if it is valid
            previousCommand = selectedCommand
        } else {
            // return if neither is valid
            NSSound.beep() // TODO: move this kind of thing to a log window controller
            return
        }
        
        // get the pasteboard
        guard let pastedItems = NSPasteboard.general.readObjects(forClasses: [NSPasteboardItem.self], options: nil) as? [NSPasteboardItem] else { return }
        if pastedItems.count == 0 { return }

        // make sure the encoding matches
        guard let encodingName = pastedItems[0].string(forType: NSPasteboard.PasteboardType(rawValue: "com.ekowen86.romtools.ROMCommand.encoding")) else { return }

        // check whether the previous command is an expanded parent - if it is, insert as a child command instead of a sibling
        let pasteAsChild = scriptView.isItemExpanded(previousCommand)
        
        // make sure the encoding matches
        if encodingName != (pasteAsChild ? previousCommand.childEncodingName : previousCommand.encodingNameForNextCommand) {
            script?.rom.log("Script encoding does not match")
            NSSound.beep()
            return
        }
        
        // create a script from the pasted bytes
        let dummyScript = ROMScript(previousCommand.rom, definition: ["encoding" : encodingName])
        dummyScript.pointerOffset = script?.pointerOffset
        
        // add pasted references to the dummy script
        if let pastedReferences = pastedItems[0].propertyList(forType: NSPasteboard.PasteboardType(rawValue: "com.ekowen86.romtools.ROMCommand.references")) as? [String : Int] {
            for (value, reference) in pastedReferences {
                guard let intValue = Int(object: value as AnyObject?) else { continue }
                dummyScript.commandReferences[intValue] = script?.commandReferences[reference]
            }
        }
        
        // disassemble the pasted data and add it to the script
        guard let pastedData = pastedItems.first?.data(forType: NSPasteboard.PasteboardType(rawValue: "com.ekowen86.romtools.ROMCommand")) else { return }
        dummyScript.disassembleFrom(pastedData)
        insert(commands: dummyScript.childCommands, after: previousCommand, child: pasteAsChild, script: script)
    }

    @IBAction func add(_ sender: AnyObject) {
        
        // return if no rows are selected
        if scriptView.selectedRow == -1 { return }
        
        // get the selected command and the menu item
        guard let selectedCommand = scriptView.item(atRow: scriptView.selectedRow) as? ROMCommand, let menuItem = sender as? NSMenuItem else { return }

        // get the appropriate encoding
        let pasteAsChild = scriptView.isItemExpanded(selectedCommand)
        let encodingName = (pasteAsChild ? selectedCommand.childEncodingName : selectedCommand.encodingNameForNextCommand)

        // insert the new command
        guard let newCommand = ROMCommand(selectedCommand.rom, script: selectedCommand.script, encodingName: encodingName, opcode: menuItem.tag) else { return }
        insert(commands: [newCommand], after: selectedCommand, child: pasteAsChild, script: script)
    }
    
    @IBAction func delete(_ sender: AnyObject) {
        
        // return if no rows are selected
        if scriptView.selectedRow == -1 { return }
        let selectedRows = scriptView.selectedRowIndexes
        if selectedRows.count == 0 { return }
        guard let firstCommand = scriptView.item(atRow: selectedRows.first!) as? ROMCommand else { return }
        
        // create an array of objects to delete
        let commandsToDelete = NSMutableArray()
        for row in selectedRows {
            if let command = scriptView.item(atRow: row) as? ROMCommand {

                // only delete commands with the same parent as the first selected item
                if command.parent != firstCommand.parent { continue }
                commandsToDelete.add(command)
            }
        }
        
        // remove all selected commands
        remove(commands: commandsToDelete, script: script)
    }
    
    @objc func doubleClickedCommand(_ sender: AnyObject) {
        
        guard let selectedCommand = selectedObject as? ROMCommand else { return }
        windowController?.deselectAllObjects()
        windowController?.selectObject(selectedCommand)
    }
    
    @objc func insert(commands: NSArray, after previousCommand: ROMCommand?, child: Bool, script: ROMScript?) {
        if commands.count == 0 { return }
        if script == nil { return }
        guard let firstCommand = commands.firstObject as? ROMCommand else { return }
        
        // insert the new commands
        script!.insertCommands(commands, after: previousCommand, child: child)
        if script === self.script {
            scriptView.beginUpdates()
            scriptView.insertItems(at: script!.indexesOfCommands(commands), inParent: firstCommand.parent, withAnimation: NSTableView.AnimationOptions())
            scriptView.endUpdates()
            select(commands: commands)
        }

        // save undo data
        if let undo = windowController?.undoManager {
            let target = undo.prepare(withInvocationTarget: self) as AnyObject
            target.remove(commands: commands, script: script)
            if (!undo.isUndoing) {
                if commands.count == 1 {
                    undo.setActionName("Insert Command")
                } else {
                    undo.setActionName("Insert Commands")
                }
            }
        }
    }
    
    @objc func remove(commands: NSArray, script: ROMScript?) {

        if commands.count == 0 { return }
        if script == nil { return }
        guard let firstCommand = commands.firstObject as? ROMCommand else { return }

        // never delete all of the siblings because that would leave an empty parent command
        let commandsToRemove = NSMutableArray(array: commands)
        if commands == firstCommand.siblings {
            commandsToRemove.removeLastObject()
            NSSound.beep() // TODO: move this kind of thing to a log window controller
            script!.rom.log("Unable to delete the last child command")
            if commandsToRemove.count == 0 { return }
        }
        
        // save undo data
        if let undo = windowController?.undoManager {
            let target = undo.prepare(withInvocationTarget: self) as AnyObject
            if let previousCommand = firstCommand.previousCommand {
                target.insert(commands: commandsToRemove, after: previousCommand, child: false, script: script)
            } else {
                target.insert(commands: commandsToRemove, after: firstCommand.parent, child: true, script: script)
            }
            if (!undo.isUndoing) {
                if commands.count == 1 {
                    undo.setActionName("Delete Command")
                } else {
                    undo.setActionName("Delete Commands")
                }
            }
        }

        if script === self.script {
            // remove the commands from the view
            scriptView.beginUpdates()
            scriptView.removeItems(at: script!.indexesOfCommands(commandsToRemove), inParent: firstCommand.parent, withAnimation: NSTableView.AnimationOptions())
            scriptView.endUpdates()
            
            // select the next command
            var lastSibling: ROMCommand? = commandsToRemove.lastObject as? ROMCommand
            while lastSibling?.parent !== firstCommand.parent {
                lastSibling = lastSibling?.parent
                if lastSibling == nil { break }
            }
            
            if let nextCommand = lastSibling?.nextCommand {
                select(command: nextCommand)
            } else if let previousCommand = firstCommand.previousCommand {
                select(command: previousCommand)
            } else if let parentCommand = firstCommand.parent {
                select(command: parentCommand)
            }
        }
        
        // remove the commands from the script
        script!.removeCommands(commandsToRemove)
    }
    
    @objc func move(commands: NSArray, after previousCommand: ROMCommand?, child: Bool, script: ROMScript?) {
        
        if commands.count == 0 { return }
        if script == nil { return }
        guard let firstCommand = commands.firstObject as? ROMCommand else { return }

        // never move all of the remaining siblings
        if Set(arrayLiteral: firstCommand.siblings).isSubset(of: Set(arrayLiteral: commands)) {
            NSSound.beep() // TODO: move this kind of thing to a log window controller
            script!.rom.log("Unable to move all child commands from a parent")
            return
        }

        // save undo data
        if let undo = windowController?.undoManager {
            let target = undo.prepare(withInvocationTarget: self) as AnyObject
            if let oldPreviousCommand = firstCommand.previousCommand {
                target.move(commands: commands, after: oldPreviousCommand, child: false, script: script)
            } else {
                target.move(commands: commands, after: firstCommand.parent, child: true, script: script)
            }
            if (!undo.isUndoing) {
                if commands.count == 1 {
                    undo.setActionName("Move Command")
                } else {
                    undo.setActionName("Move Commands")
                }
            }
        }

        // remove the commands from their old positions
        if script === self.script {
            scriptView.beginUpdates()
            scriptView.removeItems(at: script!.indexesOfCommands(commands), inParent: firstCommand.parent, withAnimation: NSTableView.AnimationOptions())
        }
        script!.removeCommands(commands)
        
        // insert the new commands in their new positions
        script!.insertCommands(commands, after: previousCommand, child: child)
        if firstCommand.script === self.script {
            scriptView.insertItems(at: script!.indexesOfCommands(commands), inParent: firstCommand.parent, withAnimation: NSTableView.AnimationOptions())
            scriptView.endUpdates()
            select(commands: commands)
        }
    }
    
//    func select(reference: Int) {
//        if let referencedCommand = script?.commandReferences[reference] as? ROMCommand {
//            select(command: referencedCommand)
//        }
//    }
    
    func select(command: ROMCommand?) {
        if command != nil {
            select(commands: [command!])
        } else {
            select(commands: [])
        }
    }
    
    func select(commands: NSArray) {
        
        let rows = NSMutableIndexSet()
        var selectionRect = NSZeroRect
        for object in commands {
            guard let command = object as? ROMCommand else { continue }
            
            // add the row
            let row = scriptView.row(forItem: command)
            if row == -1 { continue }
            rows.add(row)

            // add the rect
            let commandRect = scriptView.rect(ofRow: row)
            if commandRect.isEmpty { continue }
            if selectionRect.isEmpty {
                selectionRect = commandRect
            } else {
                selectionRect = selectionRect.union(commandRect)
            }
        }

        // clear the selection if no valid rows are selected
        if rows.count == 0 {
            scriptView.selectRowIndexes(IndexSet(), byExtendingSelection: false)
            return
        }
        
        // select the rows and scroll so they are visible
        scriptView.selectRowIndexes(rows as IndexSet, byExtendingSelection: false)
        
        guard let scrollView = scriptView.enclosingScrollView else { return }
        if !IndexSet(integersIn: Range.init(scriptView.rows(in: scrollView.documentVisibleRect)) ?? 0..<1).contains(integersIn: rows as IndexSet) {
            scriptView.scrollRowToVisible(scriptView.numberOfRows - 1)
            scriptView.scrollToVisible(selectionRect)
        }
    }
    
    @objc func setLabel(_ label: String, forCommand command: ROMCommand) {
        // save undo data
        if let undo = windowController?.undoManager {
            let target = undo.prepare(withInvocationTarget: self) as AnyObject
            let oldLabel = command.script?.commandLabels[command.reference] ?? ""
            target.setLabel(oldLabel, forCommand: command)
            if (!undo.isUndoing) {
                undo.setActionName("Change Command Label")
            }
        }
        
        // change the label and reload the command
        command.label = label
        if command.script !== self.script { return }
        scriptView.reloadItem(command)

        // reload descriptions for visible commands
        let visibleRows = scriptView.rows(in: scriptView.visibleRect)
        let commandColumn = scriptView.column(withIdentifier: NSUserInterfaceItemIdentifier(rawValue: "description"))
        scriptView.reloadData(forRowIndexes: IndexSet(integersIn: Range(visibleRows) ?? 0..<1), columnIndexes: IndexSet(integer: commandColumn))

    }
    
    // MARK: NSKeyValueObserving members
    
    override func observeValue(forKeyPath keyPath: String?, of object: Any?, change: [NSKeyValueChangeKey : Any]?, context: UnsafeMutableRawPointer?) {
        if context != &ROMScriptController.context {
            super.observeValue(forKeyPath: keyPath, of: object, change: change, context: context)
            return
        }
        
        // reload the selected command
        scriptView.reloadItem(object)
        scriptView.noteHeightOfRows(withIndexesChanged: IndexSet(integer: scriptView.row(forItem: object)))
    }
    
    // MARK: NSOutlineViewDataSource methods
    
    func outlineView(_ outlineView: NSOutlineView, child index: Int, ofItem item: Any?) -> Any {
        
        if let command = item as? ROMCommand {
            // script command, return the child command
            return command.childCommands[index]
        } else if let command = script?.childCommands[index] {
            // root command, return the script's child command
            return command
        }
        
        return "No Script Loaded"
    }
    
    func outlineView(_ outlineView: NSOutlineView, isItemExpandable item: Any) -> Bool {
        if let command = item as? ROMCommand {
            if command.childCommands.count != 0 {
                // return true only if this command has child commands
                return true
            }
        }
        return false
    }
    
    func outlineView(_ outlineView: NSOutlineView, numberOfChildrenOfItem item: Any?) -> Int {
        
        if script == nil {
            // no script loaded, return 1
            return 1
        } else if item == nil {
            // root command, return the number of commands in the event script
            return script?.childCommands.count ?? 1
        } else if let command = item as? ROMCommand {
            // script command, return the number of child commands
            return command.childCommands.count
        }
        
        return 1
    }
    
    func outlineView(_ outlineView: NSOutlineView, dataCellFor tableColumn: NSTableColumn?, item: Any) -> NSCell? {

        guard let command = item as? ROMCommand, let cell = tableColumn?.dataCell(forRow: 0) as? NSTextFieldCell else { return nil }

        if (tableColumn?.identifier)!.rawValue == "label" {
            if command.previousCommand?.encodingNameForNextCommand ?? command.parent?.childEncodingName ?? script?.encodingName != command.encodingName {
                cell.textColor = NSColor.red
            } else if (script?.commandLabels[command.reference] != nil) && (script?.commandReferences[command.reference] as? ROMCommand === command) {
                cell.textColor = NSColor.black
            } else {
                cell.textColor = NSColor.lightGray
            }
            
        }

        return cell
    }
    
    func outlineView(_ outlineView: NSOutlineView, objectValueFor tableColumn: NSTableColumn?, byItem item: Any?) -> Any? {
        guard let command = item as? ROMCommand else { return nil }
        if (tableColumn?.identifier)!.rawValue == "label" {
            // label column
            return command.label
            
        } else if (tableColumn?.identifier)!.rawValue == "description" {
            // command description
            return command.description
        }
        
        return "Invalid table column"
    }

    func outlineView(_ outlineView: NSOutlineView, setObjectValue object: Any?, for tableColumn: NSTableColumn?, byItem item: Any?) {
        if (tableColumn?.identifier)!.rawValue == "label" {
            guard let command = item as? ROMCommand, let label = object as? String else { return }
            
            // return if the label didn't change
            if command.label == label { return }
            
            // set the new label
            setLabel(label, forCommand: command)
        }
    }
   
    func outlineView(_ outlineView: NSOutlineView, draggingSession session: NSDraggingSession, willBeginAt screenPoint: NSPoint, forItems draggedItems: [Any]) {

        // add all of the dragged items to the draggedCommands array
        draggedCommands.removeAllObjects()
        for object in draggedItems {
            guard let command = object as? ROMCommand else { continue }
            draggedCommands.add(command)
        }
    }

    func outlineView(_ outlineView: NSOutlineView, validateDrop info: NSDraggingInfo, proposedItem item: Any?, proposedChildIndex index: Int) -> NSDragOperation {
        
        if index == NSOutlineViewDropOnItemIndex {
            // can never drop "on" another command
            return NSDragOperation()
        }
        
        // get the pasted items
        let pasteboard = info.draggingPasteboard()
        guard let pastedItems = pasteboard.readObjects(forClasses: [NSPasteboardItem.self], options: nil) as? [NSPasteboardItem] else { return NSDragOperation() }
        if pastedItems.count == 0 { return NSDragOperation() }
        
        // make sure the encoding matches
        guard let encodingName = pastedItems[0].string(forType: NSPasteboard.PasteboardType(rawValue: "com.ekowen86.romtools.ROMCommand.encoding")) else { return NSDragOperation() }
        if index == 0 {
            // dragged to index 0
            if let parent = item as? ROMCommand {
                if encodingName != parent.childEncodingName { return NSDragOperation() }
            } else {
                // dragged to index 0 of script root
                if encodingName != script?.encodingName { return NSDragOperation() }
            }
        } else if item == nil {
            // dragged to script root
            if let previousCommand = script?.childCommands[index - 1] as? ROMCommand,
                previousCommand.encodingNameForNextCommand != encodingName {
                return NSDragOperation()
            }
        } else {
            // dragged anywhere else
            guard let parent = item as? ROMCommand,
                let previousCommand = parent.childCommands[index - 1] as? ROMCommand else { return NSDragOperation() }
            if previousCommand.encodingNameForNextCommand != encodingName { return NSDragOperation() }
        }
        return NSDragOperation.generic
    }
    
    func outlineView(_ outlineView: NSOutlineView, acceptDrop info: NSDraggingInfo, item: Any?, childIndex index: Int) -> Bool {
        
        // TODO: move some of this validation stuff to validateDrop (above)
        
        // make sure at least one of the dragged commands is valid
        if draggedCommands.count == 0 { return false }
        guard let firstCommand = draggedCommands.firstObject as? ROMCommand else { return false }
        
        // copy the sibling commands to a new array
        let commandsToMove = NSMutableArray()
        for object in draggedCommands {
            guard let command = object as? ROMCommand else { continue }
            if command.parent != firstCommand.parent { continue }
            commandsToMove.add(command)
        }
        
        let childCommands: NSArray?
        if item == nil {
            // no parent, target is at root of script
            childCommands = script?.childCommands
        } else if let parent = item as? ROMCommand {
            // target is a valid parent command
            childCommands = parent.childCommands
            if draggedCommands.contains(parent) { return false }
        } else { return false } // invalid target
        
        if index == 0 {
            // dragged to root
            if firstCommand === childCommands?.firstObject as? ROMCommand { return false } // don't do anything if it's already there
            move(commands: commandsToMove, after: item as? ROMCommand, child: true, script: script)
            return true
            
        } else if childCommands != nil && index <= childCommands!.count {
            // dragged somewhere other than root
            guard let targetCommand = childCommands?[index - 1] as? ROMCommand else { return false }
            if draggedCommands.contains(targetCommand) { return false } // don't do anything if the target command is one of the dragged commands
            if targetCommand === firstCommand.previousCommand { return false } // don't do anything if it's already there
            move(commands: commandsToMove, after: targetCommand, child: false, script: script)
            return true
        }
        return false
    }
    
    func outlineView(_ outlineView: NSOutlineView, draggingSession session: NSDraggingSession, endedAt screenPoint: NSPoint, operation: NSDragOperation) {
        draggedCommands.removeAllObjects()
    }
    
    func outlineView(_ outlineView: NSOutlineView, pasteboardWriterForItem item: Any) -> NSPasteboardWriting? {
        
        guard let command = item as? ROMCommand else { return nil }
        
        // start with an empty Data
        var commandData = Data()
        command.assembleTo(&commandData)
        
        // attach the assembled data to an NSPasteboard item and add it to the pasteboard
        let pasteboardItem = NSPasteboardItem()
        pasteboardItem.setString(command.encodingName, forType: NSPasteboard.PasteboardType(rawValue: "com.ekowen86.romtools.ROMCommand.encoding"))
        pasteboardItem.setData(Data(buffer: UnsafeBufferPointer(start: &command.reference, count: 8)), forType: NSPasteboard.PasteboardType(rawValue: "com.ekowen86.romtools.ROMCommand.reference"))
        pasteboardItem.setData(commandData, forType: NSPasteboard.PasteboardType(rawValue: "com.ekowen86.romtools.ROMCommand"))

        // TODO: send the offset of the dragged command? or just make the script link field get the command directly from the NSDraggingInfo?
        
        return pasteboardItem
    }
    
    // MARK: NSOutlineViewDelegate methods
    
    func outlineView(_ outlineView: NSOutlineView, heightOfRowByItem item: Any) -> CGFloat {

        guard let command = item as? ROMCommand else { return outlineView.rowHeight }
        return CGFloat(command.numLinesForDescription) * outlineView.rowHeight
    }
    
    func outlineViewSelectionDidChange(_ notification: Notification) {
        
        // remove observers for each property
        if let properties = selectedObject?.value(forKey: "properties") as? [String : AnyObject] {
            for (propertyName, _) in properties {
                selectedObject!.removeObserver(self, forKeyPath: propertyName, context: &ROMScriptController.context)
            }
        }
        
        // get the object at the selected row
        selectedObject = scriptView.item(atRow: scriptView.selectedRow) as AnyObject?
        
        // TODO: add a delegate method to let the window know that the selection changed
//        if let identifier = selectedObject?.valueForKey("identifier") as? String {
////            let lastComponent = identifier.componentsSeparatedByString(".").last {
//            windowController?.selectObject(selectedObject, withIdentifier: identifier)
//        }

        // add observers for each property
        guard let properties = selectedObject?.value(forKey: "properties") as? NSDictionary else { return }
        for (key, _) in properties {
            guard let propertyName = key as? String else { continue }
            selectedObject?.addObserver(self, forKeyPath: propertyName, options: NSKeyValueObservingOptions.old, context: &ROMScriptController.context)
        }
    }
    
    func outlineView(_ outlineView: NSOutlineView, selectionIndexesForProposedSelection proposedSelectionIndexes: IndexSet) -> IndexSet {
        
        // only allow contiguous selection
        let selectedRanges = proposedSelectionIndexes.rangeView
        
        if selectedRanges.count == 0 {
            return proposedSelectionIndexes
        } else if selectedRanges.count > 1 {
            // return the longest contiguous range
            var longestRange = selectedRanges.first!
            for range in selectedRanges {
                if range.count > longestRange.count {
                    longestRange = range
                }
            }
            return IndexSet(integersIn: longestRange)
        }
        var selectedRange = selectedRanges[0]
        
        // don't allow a contiguous selection that extends from a child command past the end of its siblings
        guard let firstCommand = scriptView.item(atRow: selectedRange.lowerBound) as? ROMCommand,
            let parent = firstCommand.parent,
            let afterParent = parent.nextCommand else { return proposedSelectionIndexes }
        let maxIndex = scriptView.row(forItem: afterParent)
        if selectedRange.upperBound > maxIndex {
            // TODO: somehow limit the selected range based on the current range so that it doesn't clear everything when you select higher rows
            selectedRange = selectedRange.lowerBound..<maxIndex
            return IndexSet(integersIn: selectedRange)
        }
        
        return proposedSelectionIndexes
    }
    
    // MARK: NSMenuDelegate methods
    
    func menuNeedsUpdate(_ menu: NSMenu) {
        guard let addIcon = menu.item(at: 0) else { return }
        menu.removeAllItems()
        menu.addItem(addIcon)
        
        // get the categories for the encoding
        guard let selectedCommand = scriptView.item(atRow: scriptView.selectedRow) as? ROMCommand else { return }
        let encodingName = scriptView.isItemExpanded(selectedCommand) ? selectedCommand.childEncodingName : selectedCommand.encodingNameForNextCommand
        guard let encoding = selectedCommand.rom.scriptEncodings[encodingName] as? ROMScriptEncoding,
            let categories = encoding.commandTree.allKeys as? [String] else { return }
        
        for category in categories.sorted() {
            if let commandDictionary = encoding.commandTree[category] as? NSDictionary,
                let commandNames = commandDictionary.allKeys as? [String] {
                
                // add sub menus for each category
                let categoryMenu = NSMenu(title: category)
                let categoryMenuItem = NSMenuItem(title: category, action: nil, keyEquivalent: "")
                
                for name in commandNames.sorted() {
                    guard let tag = commandDictionary[name] as? Int else { continue }
                    let menuItem = NSMenuItem(title: name, action: #selector(ROMScriptController.add(_:)), keyEquivalent: "")
                    menuItem.target = self
                    menuItem.tag = tag
                    categoryMenu.addItem(menuItem)
                }
                menu.addItem(categoryMenuItem)
                menu.setSubmenu(categoryMenu, for: categoryMenuItem)
            } else if let tag = encoding.commandTree[category] as? Int {
                // add commands with no category to the root
                let menuItem = NSMenuItem(title: category, action: #selector(ROMScriptController.add(_:)), keyEquivalent: "")
                menuItem.target = self
                menuItem.tag = tag
                menu.addItem(menuItem)
            }
        }
    }
    
    // MARK: search methods
    
    @IBAction func search(_ sender: AnyObject) {
        
        guard let searchField = sender as? NSSearchField else { return }
        let searchString = searchField.stringValue
        if searchString.isEmpty { return }
        
        let descriptionColumnIndex = scriptView.column(withIdentifier: NSUserInterfaceItemIdentifier(rawValue: "description"))
        let descriptionColumn: NSTableColumn? = (descriptionColumnIndex != -1) ? scriptView.tableColumns[descriptionColumnIndex] : nil

        let labelColumnIndex = scriptView.column(withIdentifier: NSUserInterfaceItemIdentifier(rawValue: "label"))
        let labelColumn: NSTableColumn? = (labelColumnIndex != -1) ? scriptView.tableColumns[labelColumnIndex] : nil

        // start at the top if nothing is selected
        let startRow = scriptView.selectedRow == -1 ? 0 : scriptView.selectedRow
        guard let startCommand = scriptView.item(atRow: startRow) as? ROMCommand else { return }
        
        var searchCommand: ROMCommand? = startCommand
        repeat {
            
            // get the next command
            if let firstChild = searchCommand?.childCommands.firstObject as? ROMCommand {
                // if the current command has child commands
                searchCommand = firstChild
            } else if let nextCommand = searchCommand?.nextCommand {
                // if there is a command after the current command
                searchCommand = nextCommand
            } else {
                // try to find the next parent command
                var parent = searchCommand?.parent
                repeat {
                    if let nextParent = parent?.nextCommand {
                        searchCommand = nextParent
                        break
                    }
                    parent = parent?.parent
                } while parent != nil
                
                if parent == nil {
                    // made it back to the root, no next parent found
                    guard let firstCommand = searchCommand?.script.childCommands.firstObject as? ROMCommand else { break }
                    searchCommand = firstCommand
                }
            }
            
            guard let descriptionString = outlineView(scriptView, objectValueFor: descriptionColumn, byItem: searchCommand) as? String else { continue }
            guard let labelString = outlineView(scriptView, objectValueFor: labelColumn, byItem: searchCommand) as? String else { continue }
            if descriptionString.localizedCaseInsensitiveContains(searchString) || labelString.localizedCaseInsensitiveContains(searchString) {
                let selectRow = scriptView.row(forItem: searchCommand)
                if selectRow == -1 { break }
                scriptView.selectRowIndexes(IndexSet(integer: selectRow), byExtendingSelection: false)
                scriptView.scrollRowToVisible(selectRow)
                return
            }
        } while searchCommand !== startCommand
        
        // search string not found
        NSSound.beep()
    }

}
