//
//  ROMTileView.swift
//  FF6Tools
//
//  Created by Evan Owen on 6/4/17.
//  Copyright © 2017 Evan Owen. All rights reserved.
//

import Cocoa

class ROMTileView: ROMView {
    
    @IBOutlet weak var tilesetView: ROMTilesetView!

    var tileLayout: ROMTileLayout?
    var tiles = ROMTiles()
    var selectedTiles = ROMTiles() {
        didSet {
            needsDisplay = true
        }
    }
    
    var zoom: CGFloat = 2.0
    var tileImage: NSImage?
    var showCursor = false {
        didSet {
            needsDisplay = true
        }
    }
    
    override func draw(_ dirtyRect: NSRect) {
        super.draw(dirtyRect)

        // set the background color for the parent scroll view
        if let clipView = superview as? NSClipView, let scrollView = clipView.superview as? NSScrollView {
            scrollView.backgroundColor = NSColor.black
        }
        
        // fill the view with the background color
        backgroundColor.set()
        NSBezierPath.fill(dirtyRect)

        // return if there is no valid layout selected
        if tileLayout == nil { return }
        
        // update the tile image if needed
        updateTileImage()
        
        // return if the image could not be created
        if tileImage == nil { return }
        
        // draw the image
        NSGraphicsContext.current?.imageInterpolation = imageInterpolation
        tileImage?.draw(in: dirtyRect, from: dirtyRect, operation: NSCompositingOperation.sourceOver, fraction: 1.0)
        
        // draw the cursor
        if showCursor == true {
            let x = selectedTiles.x
            let y = selectedTiles.y
            let width = selectedTiles.width
            let height = selectedTiles.height
            drawCursor(x: x, y: y, width: width, height: height)
        }
    }
    
    func updateTileImage() {
        if tileImage != nil { return }
        
        // decode graphics
        let gfxLength = 0x10000 * MemoryLayout<UInt32>.size // 1024 8x8 tiles
        let gfx = UnsafeMutablePointer<UInt8>.allocate(capacity: gfxLength)
        memset(gfx, 0, gfxLength)
        for (key, object) in tileLayout!.graphics {

            guard let offset = Int(object: key as AnyObject) else { continue }

            if let identifier = object as? String,
                let graphicsData = rom.value(forKeyPath: identifier) as? ROMGraphics {

                // load graphics
                graphicsData.data.withUnsafeBytes{(bytes: UnsafePointer<UInt8>) -> Void in
                    gfx_decode(gfx.advanced(by: offset), bytes, Int32(graphicsData.data.count), graphicsData.format.rawFormat)
                }
            } else if let definition = object as? NSDictionary,
                let identifier = definition["identifier"] as? String,
                let graphicsData = rom.value(forKeyPath: identifier) as? ROMGraphics {
                
                // load a specific subrange of graphics
                var range = MakeRange(object: definition, keyPath: "range") ?? 0..<graphicsData.data.count
                if range.upperBound > graphicsData.count {
                    range = range.lowerBound..<graphicsData.count
                }
                
                graphicsData.data.subdata(in: range).withUnsafeBytes{(bytes: UnsafePointer<UInt8>) -> Void in
                    gfx_decode(gfx.advanced(by: offset), bytes, Int32(range.count), graphicsData.format.rawFormat)
                }
            }
        }
        let graphics = Data(bytes: UnsafePointer<UInt8>(gfx), count: gfxLength)
        gfx.deinitialize(); gfx.deallocate(capacity: gfxLength)
        
        // decode palette
        let palLength = 256 * MemoryLayout<UInt32>.size // 256 colors
        let pal = UnsafeMutablePointer<UInt8>.allocate(capacity: palLength)
        memset(pal, 0, palLength)
        for (key, object) in tileLayout!.palette {
            
            guard let offset = Int(object: key as AnyObject) else { continue }
            
            if let identifier = object as? String,
                let paletteData = rom.value(forKeyPath: identifier) as? ROMData {
                
                // load graphics
                paletteData.data.withUnsafeBytes{(bytes: UnsafePointer<UInt8>) -> Void in
                    gfx_decode(pal.advanced(by: offset), bytes, Int32(paletteData.data.count), BGR555)
                }
            } else if let definition = object as? NSDictionary,
                let identifier = definition["identifier"] as? String,
                let paletteData = rom.value(forKeyPath: identifier.replacingOccurrences(of: "%n", with: String(tileLayout!.index))) as? ROMData {
                
                // load a specific subrange of graphics
                var range = MakeRange(object: definition, keyPath: "range") ?? 0..<paletteData.data.count
                if range.upperBound > paletteData.count {
                    range = range.lowerBound..<paletteData.count
                }
                
                paletteData.data.subdata(in: range).withUnsafeBytes{(bytes: UnsafePointer<UInt8>) -> Void in
                    gfx_decode(pal.advanced(by: offset), bytes, Int32(range.count), BGR555)
                }
            }
        }

        // apply palette assignment to layout
        if let paletteAssignmentIdentifier = tileLayout!.paletteAssignment,
            let paletteAssignmentData = rom.value(forKeyPath: paletteAssignmentIdentifier) as? ROMData {
            
            // get palette assignment data
            var dummyData = Data(paletteAssignmentData.data)
            dummyData.count = 256
            var paletteAssignmentBytes: [UInt8] = dummyData.withUnsafeBytes{ [UInt8](UnsafeBufferPointer(start: $0, count: dummyData.count)) }
            
            // assign palette numbers to each tile
            for (index, tile) in tiles.tiles.enumerated() {
                let tileIndex = Int(tile & 0xFF)
                if tileLayout!.tileFormat == .SNES8bppMode7 {
                    let palette = Int(paletteAssignmentBytes[tileIndex] & 0x0F)
                    tiles.tiles[index] &= 0xFF00FFFF
                    tiles.tiles[index] |= UInt32(palette * tileLayout!.tileFormat.colorsPerPalette) << 16
                } else if tileLayout!.tileFormat == .SNES4bppMode7 {
                    var palette = Int(paletteAssignmentBytes[tileIndex >> 1])
                    if (tileIndex & 1) == 1 {
                        palette >>= 4
                    }
                    palette &= 0x0F
                    tiles.tiles[index] &= 0xFF00FFFF
                    tiles.tiles[index] |= UInt32(palette * tileLayout!.tileFormat.colorsPerPalette) << 16
                }
            }
        }
        
        let palette = Data(bytes: UnsafePointer<UInt8>(pal), count: palLength)
        pal.deinitialize(); pal.deallocate(capacity: palLength)
        
        // set up the PPU
        var ppu = PPU()
        memset(&ppu, 0, MemoryLayout<PPU>.size)
        
        ppu.pal = (palette as NSData).bytes
        ppu.layers.0.format = DEFAULT_TILE
        ppu.layers.0.height = Int32(tiles.height)
        ppu.layers.0.width = Int32(tiles.width)
        ppu.layers.0.v_scroll = 0
        ppu.layers.0.h_scroll = 0
        ppu.layers.0.z.0 = Z_LEVEL_BG1_LO
        ppu.layers.0.z.1 = Z_LEVEL_BG1_HI
        ppu.layers.0.gfx = (graphics as NSData).bytes
        ppu.layers.0.tiles = UnsafeRawPointer(tiles.tiles)
        ppu.layers.0.main = true
        
        // render the tiles
        let dataLength = tiles.height * tiles.width * 64 * MemoryLayout<UInt32>.size
        let data = UnsafeMutablePointer<UInt8>.allocate(capacity: dataLength)
        memset(data, 0, dataLength)
        gfx_render_ppu(data, ppu, Int32(dataLength), Int32(tiles.width * 8))
        let imageData = Data(bytes: UnsafePointer<UInt8>(data), count: dataLength)
        data.deinitialize(); data.deallocate(capacity: dataLength)
        
        // create an image
        let provider = CGDataProvider(dataInfo: nil, data: (imageData as NSData).bytes, size: imageData.count, releaseData: releaseImageData)!
        let bitsPerComponent = 8
        let bitsPerPixel = 32
        let bytesPerRow = tiles.width * 8 * MemoryLayout<UInt32>.size
        let colorSpaceRef = CGColorSpaceCreateDeviceRGB()
        let bitmapInfo = CGBitmapInfo(rawValue: CGBitmapInfo().rawValue | CGImageAlphaInfo.premultipliedLast.rawValue)
        let renderingIntent = CGColorRenderingIntent.defaultIntent
        let iref = CGImage(width: tiles.width * 8, height: tiles.height * 8, bitsPerComponent: bitsPerComponent, bitsPerPixel: bitsPerPixel, bytesPerRow: bytesPerRow, space: colorSpaceRef, bitmapInfo: bitmapInfo, provider: provider, decode: nil, shouldInterpolate: true, intent: renderingIntent)
        
        let imageSize = NSMakeSize(CGFloat(tiles.width * 8) * zoom, CGFloat(tiles.height * 8) * zoom)
        tileImage = NSImage(cgImage: iref!, size: imageSize)
    }
    
    func drawCursor(x: Int, y: Int, width: Int, height: Int) {
        
        // don't draw the cursor if the selection is empty
        if (width == 0) || (height == 0) { return }
        
        // draw the cursor
        let left = round(CGFloat(x) * 8 * zoom) + 0.5
        let right = round(CGFloat(x + width) * 8 * zoom) - 0.5
        let top = round(CGFloat(y) * 8 * zoom) + 0.5
        let bottom = round(CGFloat(y + height) * 8 * zoom) - 0.5
        var cursorRect = NSMakeRect(left, frame.height - bottom, right - left, bottom - top)
        
        // draw the outside border (black)
        NSColor.black.set()
        NSBezierPath.stroke(cursorRect)
        
        // draw the outer highlight
        cursorRect = cursorRect.insetBy(dx: 1.0, dy: 1.0)
        NSColor.green.set()
        NSBezierPath.stroke(cursorRect)
        
        // draw the inner highlight (white)
        cursorRect = cursorRect.insetBy(dx: 1.0, dy: 1.0)
        NSColor.white.set()
        NSBezierPath.stroke(cursorRect)
        
        // draw the inner border (black)
        cursorRect = cursorRect.insetBy(dx: 1.0, dy: 1.0)
        NSColor.black.set()
        NSBezierPath.stroke(cursorRect)
    }
    
    func select(tileLayout: ROMTileLayout?) {
        
        // set the tile layout
        self.tileLayout = tileLayout
        tiles = ROMTiles(tileLayout: tileLayout)
        
        // set the frame size and redraw the view
        setFrameSize(frame.size)
        scroll(NSMakePoint(0.0, frame.height))
        tileImage = nil
        needsDisplay = true

    }
    
    override func setFrameSize(_ newSize: NSSize) {
        
        // invalidate the tile image
        tileImage = nil
        
        // go to the super class if this is a tileset view
        if self is ROMTilesetView {
            super.setFrameSize(newSize)
            return
        }
        
        // calculate the frame size based on the tile layout
        let frameSize = NSMakeSize(CGFloat(tiles.width * 8) * zoom, CGFloat(tiles.height * 8) * zoom)
        super.setFrameSize(frameSize)
    }
    
    @objc func setData(data: Data, forTileLayout tileLayout: ROMTileLayout) {
        // return if the data hasn't changed
        if tileLayout.data == data { return }
        
        // save undo data
        if let undo = windowController.undoManager {
            let target = undo.prepare(withInvocationTarget: self) as AnyObject
            target.setData(data: tileLayout.data, forTileLayout: tileLayout)
            if (!undo.isUndoing) {
                undo.setActionName("Edit Tiles")
            }
        }

        if windowController.objectIsSelected(tileLayout) {
            windowController.stopObserving(object: tileLayout)
        }

        tileLayout.data = data
        
        if windowController.objectIsSelected(tileLayout) {
            windowController.startObserving(object: tileLayout)
        }

        // redraw if the changed layout is selected
        if self.tileLayout === tileLayout {
            tiles = ROMTiles(tileLayout: tileLayout)
            tileImage = nil
            needsDisplay = true
        }
    }
    
    func setTiles(tileSelection: ROMTiles) {

        // this line fixes a bug with Data.replaceSubrange
        tiles.tiles = [UInt32](tiles.tiles)

        // ignore tiles outside the layout
        let height = min(tileSelection.height, tiles.height - tileSelection.y)
        let width = min(tileSelection.width, tiles.width - tileSelection.x)
        for row in 0..<height {
            
            // get one row's worth of tiles
            let selectionStart = row * tileSelection.width
            let selectionEnd = selectionStart + width
            let rowData = selectedTiles.tiles[selectionStart..<selectionEnd]
            
            // copy the row to the layout
            let layoutStart = (tileSelection.y + row) * tiles.width + tileSelection.x
            let layoutEnd = layoutStart + width
            tiles.tiles.replaceSubrange(layoutStart..<layoutEnd, with: rowData)
        }
        tileImage = nil
        needsDisplay = true
    }

    // MARK: NSResponder methods
    
    override var acceptsFirstResponder: Bool { return true }
    
    override func mouseDown(with event: NSEvent) {
        
        window?.makeFirstResponder(self)
        
        var event: NSEvent! = event
        
        // get the point that was clicked
        let clickedPoint = convert(event.locationInWindow, from: nil)
        if !bounds.contains(clickedPoint) { return }
        
        showCursor = true
        
        while event != nil {
            
            switch event.type {
            case .leftMouseUp:
                // update the layout data in the rom when the mouse is released
                setData(data: tiles.encodedData(forTileLayout: tileLayout!), forTileLayout: tileLayout!)
                return
                
            case .leftMouseDragged, .leftMouseDown:
                // get the dragged/clicked point
                let draggedPoint = convert(event.locationInWindow, from: nil)
                if !bounds.contains(draggedPoint) { break }
                
                let draggedX = Int(draggedPoint.x / zoom / 8.0)
                let draggedY = Int((bounds.height - draggedPoint.y) / zoom / 8.0)
                
                // continue if the cursor hasn't moved
                let oldX = selectedTiles.x
                let oldY = selectedTiles.y
                if (oldX == draggedX) && (oldY == draggedY) && (event.type != .leftMouseDown) { break }
                
                // update the position of the selected tiles
                selectedTiles.x = draggedX
                selectedTiles.y = draggedY
                
                // pass a copy of the updated data to setTiles
                setTiles(tileSelection: selectedTiles)
                
            default:
                break
            }
            
            // get the next event
            event = window?.nextEvent(matching: [.leftMouseUp, .leftMouseDragged])
        }
    }
    
    override func rightMouseDown(with event: NSEvent) {
        window?.makeFirstResponder(self)
        
        var event: NSEvent! = event
        
        let clickedPoint = convert(event.locationInWindow, from: nil)
        if !bounds.contains(clickedPoint) { return }
        
        showCursor = true

        // get the clicked tile
        let clickedX = Int(clickedPoint.x / zoom / 8.0)
        let clickedY = Int((bounds.height - clickedPoint.y) / zoom / 8.0)
        let clickedTile = tiles.tileAt(x: clickedX, y: clickedY)

        // set the flags in the tileset view
        tilesetView.graphicsIndex = Int(clickedTile & 0x0000FF00) >> 8
        tilesetView.paletteIndex = Int(clickedTile & 0x00FF0000) >> 16
        tilesetView.priority = (clickedTile & 0x01000000) != 0
        tilesetView.horizontalFlip = (clickedTile & 0x40000000) != 0
        tilesetView.verticalFlip = (clickedTile & 0x80000000) != 0
        tilesetView.showCursor = true

        // update the selected tiles
        selectedTiles = ROMTiles(x: clickedX, y: clickedY, width: 1, height: 1, tiles: [clickedTile])
        
        // keep track of where the mouse gets dragged
        var draggedX = clickedX
        var draggedY = clickedY
        
        while event != nil {
            
            switch event.type {
            case .rightMouseUp:
                // return if the right mouse button is up
                return
                
            case .rightMouseDragged:
                let draggedPoint = convert(event.locationInWindow, from: nil)
                
                // don't update if the dragged coordinates didn't change
                let newX = Int(draggedPoint.x / zoom / 8.0)
                let newY = Int((bounds.height - draggedPoint.y) / zoom / 8.0)
                if (draggedX == newX) && (draggedY == newY) { break }
                draggedX = newX
                draggedY = newY
                
                // get the selected rectangle
                let minX = max(min(draggedX, clickedX), 0)
                let minY = max(min(draggedY, clickedY), 0)
                let maxX = min(max(draggedX, clickedX), tiles.width - 1)
                let maxY = min(max(draggedY, clickedY), tiles.height - 1)
                let width = maxX - minX + 1
                let height = maxY - minY + 1
                
                // create the new tile selection
                selectedTiles = tiles.subTilesAt(x: minX, y: minY, width: width, height: height)

                // show the cursor in the tileset view if only one tile is selected
                if selectedTiles.tiles.count == 1 {
                    tilesetView.showCursor = true
                } else {
                    tilesetView.showCursor = false
                }
                
                needsDisplay = true
                
            default: break
            }
            
            // get the next event
            event = window?.nextEvent(matching: [.rightMouseUp, .rightMouseDragged])
        }
    }
    
    override func mouseMoved(with event: NSEvent) {
        
        // get the cursor point
        let cursorPoint = convert(event.locationInWindow, from: nil)
        let x = Int(cursorPoint.x / 8 / zoom)
        let y = Int((bounds.height - cursorPoint.y) / 8 / zoom)

        showCursor = true

        if !bounds.contains(cursorPoint) {
            showCursor = false
        } else if selectedTiles.x != x || selectedTiles.y != y {
            selectedTiles.x = x
            selectedTiles.y = y
        }
        
        needsDisplay = true
    }
    
    override func mouseEntered(with theEvent: NSEvent) {
        window?.makeFirstResponder(self)
        showCursor = true
        needsDisplay = true
    }
    
    override func mouseExited(with theEvent: NSEvent) {
        showCursor = false
        needsDisplay = true
    }
    
    override func updateTrackingAreas() {
        for area in trackingAreas {
            removeTrackingArea(area)
        }
        window?.acceptsMouseMovedEvents = true
        
        let opts: NSTrackingArea.Options = [NSTrackingArea.Options.mouseEnteredAndExited, NSTrackingArea.Options.activeAlways]
        let trackingArea = NSTrackingArea(rect: self.bounds, options: opts, owner: self, userInfo: nil)
        addTrackingArea(trackingArea)
        super.updateTrackingAreas()
    }
}

struct ROMTiles {
    var x = 0
    var y = 0
    var width = 1
    var height = 1
    var tiles: [UInt32] = [0]
    
    init() { }
    
    init(x: Int, y: Int, width: Int, height: Int, tiles: [UInt32]) {
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.tiles = tiles
    }
    
    init(tileLayout: ROMTileLayout?) {
        
        if tileLayout == nil { return }
        
        // clear the tile data and set the size
        width = tileLayout!.width
        height = tileLayout!.height
        tiles = [UInt32](repeating: 0, count: width * height)
        let interlace = tileLayout?.interlace ?? "None"
        
        // de-interlace and decode the tile data
        var tileData = Data(tileLayout!.data)
        tileData.count = width * height * tileLayout!.tileFormat.bytesPerTile
        
        // map tileset interlacing
        if interlace == "FF6MapTileset" {
            let tileBytes: [UInt8] = tileData.withUnsafeBytes{ [UInt8](UnsafeBufferPointer(start: $0, count: tileData.count)) }
            var deinterlacedTileBytes = [UInt8](repeating: 0, count: tileBytes.count)
            for tileIndex in 0..<256 {
                let row = tileIndex / 16
                let column = tileIndex % 16
                deinterlacedTileBytes[row * 2 * 64 + column * 4] = tileBytes[tileIndex + 0]
                deinterlacedTileBytes[row * 2 * 64 + column * 4 + 2] = tileBytes[tileIndex + 256]
                deinterlacedTileBytes[row * 2 * 64 + column * 4 + 1] = tileBytes[tileIndex + 1024]
                deinterlacedTileBytes[row * 2 * 64 + column * 4 + 3] = tileBytes[tileIndex + 1280]
                deinterlacedTileBytes[(row * 2 + 1) * 64 + column * 4] = tileBytes[tileIndex + 512]
                deinterlacedTileBytes[(row * 2 + 1) * 64 + column * 4 + 2] = tileBytes[tileIndex + 768]
                deinterlacedTileBytes[(row * 2 + 1) * 64 + column * 4 + 1] = tileBytes[tileIndex + 1536]
                deinterlacedTileBytes[(row * 2 + 1) * 64 + column * 4 + 3] = tileBytes[tileIndex + 1792]
            }
            tileData = Data(bytes: deinterlacedTileBytes, count: deinterlacedTileBytes.count)
        }
        
        let tileBytes: [UInt8] = tileData.withUnsafeBytes{ [UInt8](UnsafeBufferPointer(start: $0, count: tileData.count)) }
        let length = min(tileBytes.count, width * height * MemoryLayout<UInt32>.size)
        gfx_decode(&tiles, tileBytes, Int32(length), tileLayout!.tileFormat.rawFormat)
        
        // deinterlace the tiles
        var deinterlacedTiles = [UInt32](tiles)
        
        switch interlace {
            
        case "HorizontalScreen":
            // horizontally interlaced screens
            deinterlacedTiles = [UInt32]()
            for screen in 0..<(height / 32) {
                for row in 0..<32 {
                    let left = (row + screen * 64) * 32
                    deinterlacedTiles.append(contentsOf: tiles[left..<(left + 32)])
                    let right = (row + 32 + screen * 64) * 32
                    deinterlacedTiles.append(contentsOf: tiles[right..<(right + 32)])
                }
            }
            
        case "VerticalScreen":
            // vertically interlaced screens
            // this line fixes a bug with Data.replaceSubrange
            let dummyTiles = [UInt32](tiles)
            deinterlacedTiles.replaceSubrange(1024..<2048, with: dummyTiles[2048..<3072])
            deinterlacedTiles.replaceSubrange(2048..<3072, with: dummyTiles[1024..<2048])
            
        case "FF6WorldTileset":
            // world map tileset interlacing
            for tileIndex in 0..<(tiles.count / 4) {
                let row = tileIndex / (width / 2)
                let column = tileIndex % (width / 2)
                deinterlacedTiles[row * 2 * width + column * 2] = tiles[tileIndex * 4]
                deinterlacedTiles[row * 2 * width + column * 2 + 1] = tiles[tileIndex * 4 + 1]
                deinterlacedTiles[(row * 2 + 1) * width + column * 2] = tiles[tileIndex * 4 + 2]
                deinterlacedTiles[(row * 2 + 1) * width + column * 2 + 1] = tiles[tileIndex * 4 + 3]
            }

        default: break
        }
        
        tiles = deinterlacedTiles
    }
    
    func encodedData(forTileLayout tileLayout: ROMTileLayout) -> Data {
        
        var interlacedTiles = [UInt32](tiles)
        let interlace = tileLayout.interlace ?? "None"
        
        switch interlace {
            
        case "HorizontalScreen":
            // horizontally interlaced screens
            interlacedTiles = [UInt32]()
            for screen in 0..<(tileLayout.height / 32) {
                for row in 0..<32 {
                    let offset = (row + screen * 32) * 64
                    interlacedTiles.append(contentsOf: tiles[offset..<(offset + 32)])
                }
                for row in 0..<32 {
                    let offset = (row + screen * 32) * 64 + 32
                    interlacedTiles.append(contentsOf: tiles[offset..<(offset + 32)])
                }
            }
            
        case "VerticalScreen":
            // vertically interlaced screens
            // this line fixes a bug with Data.replaceSubrange
            let dummyTiles = [UInt32](tiles)
            interlacedTiles.replaceSubrange(1024..<2048, with: dummyTiles[2048..<3072])
            interlacedTiles.replaceSubrange(2048..<3072, with: dummyTiles[1024..<2048])

        case "FF6WorldTileset":
            // world map tileset interlacing
            interlacedTiles = [UInt32](repeating: 0, count: tiles.count)
            for tile in 0..<(tiles.count / 4) {
                let row = tile / (tileLayout.width / 2)
                let column = tile % (tileLayout.width / 2)
                interlacedTiles[tile * 4] = tiles[row * 2 * tileLayout.width + column * 2]
                interlacedTiles[tile * 4 + 1] = tiles[row * 2 * tileLayout.width + column * 2 + 1]
                interlacedTiles[tile * 4 + 2] = tiles[(row * 2 + 1) * tileLayout.width + column * 2]
                interlacedTiles[tile * 4 + 3] = tiles[(row * 2 + 1) * tileLayout.width + column * 2 + 1]
            }

        default: break
        }
        
        var tileBytes = [UInt8](repeating: 0, count: interlacedTiles.count * tileLayout.tileFormat.bytesPerTile)
        gfx_encode(&tileBytes, interlacedTiles, Int32(interlacedTiles.count * MemoryLayout<UInt32>.size), tileLayout.tileFormat.rawFormat)
        
        // map tileset interlacing
        if interlace == "FF6MapTileset" {
            var interlacedTileBytes = [UInt8](repeating: 0, count: tileBytes.count)
            for tileIndex in 0..<256 {
                let row = tileIndex / 16
                let column = tileIndex % 16
                interlacedTileBytes[tileIndex + 0] = tileBytes[row * 2 * 64 + column * 4]
                interlacedTileBytes[tileIndex + 256] = tileBytes[row * 2 * 64 + column * 4 + 2]
                interlacedTileBytes[tileIndex + 1024] = tileBytes[row * 2 * 64 + column * 4 + 1]
                interlacedTileBytes[tileIndex + 1280] = tileBytes[row * 2 * 64 + column * 4 + 3]
                interlacedTileBytes[tileIndex + 512] = tileBytes[(row * 2 + 1) * 64 + column * 4]
                interlacedTileBytes[tileIndex + 768] = tileBytes[(row * 2 + 1) * 64 + column * 4 + 2]
                interlacedTileBytes[tileIndex + 1536] = tileBytes[(row * 2 + 1) * 64 + column * 4 + 1]
                interlacedTileBytes[tileIndex + 1792] = tileBytes[(row * 2 + 1) * 64 + column * 4 + 3]
            }
            tileBytes = interlacedTileBytes
        }
        
        return Data(bytes: tileBytes, count: tileBytes.count)
    }
    
    mutating func setGraphics(_ graphics: Int) {
        let shiftedGraphics = UInt32((graphics & 0xFF) << 8)
        for index in 0..<tiles.count {
            tiles[index] &= 0xFFFF00FF
            tiles[index] |= shiftedGraphics
        }
    }
    
    mutating func setPalette(_ palette: Int) {
        
        let shiftedPalette = UInt32((palette & 0xFF) << 16)
        for index in 0..<tiles.count {
            tiles[index] &= 0xFF00FFFF
            tiles[index] |= shiftedPalette
        }
    }
    
    mutating func setPriority(priority: Int) {
        let shiftedPriority = UInt32((priority & 0x0F) << 24)
        for index in 0..<tiles.count {
            tiles[index] &= 0xF0FFFFFF
            tiles[index] |= shiftedPriority
        }
    }
    
    mutating func togglePriority() {
        for index in 0..<tiles.count {
            tiles[index] ^= 0x01000000
        }
    }
    
    mutating func flipHorizontal() {
        var flippedLayout = [UInt32](repeating: 0, count: tiles.count)
        for index in 0..<tiles.count {
            // flip the tile and move it to the other side of the layout
            let x = index % width
            let y = index / width
            flippedLayout[y * width + (width - x - 1)] = tiles[index] ^ 0x40000000
        }
        tiles = flippedLayout
    }
    
    mutating func flipVertical() {
        var flippedLayout = [UInt32](repeating: 0, count: tiles.count)
        for index in 0..<tiles.count {
            // flip the tile and move it to the other side of the layout
            let x = index % width
            let y = index / width
            flippedLayout[(height - y - 1) * width + x] = tiles[index] ^ 0x80000000
        }
        tiles = flippedLayout
    }
    
    func tileAt(x: Int, y: Int) -> UInt32 {
        return tiles[y * width + x]
    }
    
    func subTilesAt(x: Int, y: Int, width: Int, height: Int) -> ROMTiles {
        
        var subTiles = [UInt32]()
        for row in 0..<height {
            for col in 0..<width {
                subTiles.append(tiles[x + col + (row + y) * self.width])
            }
        }
        return ROMTiles(x: x, y: y, width: width, height: height, tiles: subTiles)
    }
}
