//
//  ROMTilesetView.swift
//  FF6Tools
//
//  Created by Evan Owen on 6/10/17.
//  Copyright © 2017 Evan Owen. All rights reserved.
//

import Cocoa

class ROMTilesetView: ROMTileView {
    
    // force the zoom to match the frame size
    override var zoom: CGFloat { get { return frame.width / 128.0 } set { } }
    
    // link selectedTiles to the parent tileView
    override var selectedTiles: ROMTiles {
        get { return tileView.selectedTiles }
        set { tileView.selectedTiles = newValue }
    }
    
    @IBOutlet weak var tileView: ROMTileView!
    @IBOutlet var flipButton: NSSegmentedControl!
    @IBOutlet var graphicsSlider: NSSlider!
    @IBOutlet var paletteSlider: NSSlider!

    var graphicsIndex = 0 {
        didSet {
            graphicsSlider?.integerValue = graphicsIndex
            tiles.setGraphics(graphicsIndex)
            selectedTiles.setGraphics(graphicsIndex)
            tileImage = nil
        }
    }
    var paletteIndex = 0 {
        didSet {
            paletteSlider?.integerValue = paletteIndex / tileLayout!.tileFormat.colorsPerPalette
            if tileLayout!.tileFormat.rawFormat == SNES_MODE7_TILE {
                // TODO: edit palette assignment
                
            } else {
                tiles.setPalette(paletteIndex)
                selectedTiles.setPalette(paletteIndex)
                tileImage = nil
            }
        }
    }
    var horizontalFlip = false {
        didSet {
            if horizontalFlip != flipButton.isSelected(forSegment: 0) {
                // horizontal flip just changed
                flipButton?.setSelected(horizontalFlip, forSegment: 0)
                tiles.flipHorizontal()
                selectedTiles.flipHorizontal()
                tileImage = nil
            }
        }
    }
    var verticalFlip = false {
        didSet {
            if verticalFlip != flipButton.isSelected(forSegment: 1) {
                // vertical flip just changed
                flipButton?.setSelected(verticalFlip, forSegment: 1)
                tiles.flipVertical()
                selectedTiles.flipVertical()
                tileImage = nil
            }
        }
    }
    var priority = false {
        didSet {
            if priority != flipButton.isSelected(forSegment: 2) {
                // priority just changed
                flipButton?.setSelected(priority, forSegment: 2)
                tiles.togglePriority()
                selectedTiles.togglePriority()
            }
        }
    }
    
    @IBAction func changeFlags(_ sender: AnyObject) {
        if tileLayout?.tileFormat.rawFormat == SNES_MODE7_TILE { return }

        if horizontalFlip != flipButton.isSelected(forSegment: 0) {
            // horizontal flip just changed
            horizontalFlip = flipButton.isSelected(forSegment: 0)
            selectedTiles.flipHorizontal()
            tiles.flipHorizontal()
        }
        if verticalFlip != flipButton.isSelected(forSegment: 1) {
            // vertical flip just changed
            verticalFlip = flipButton.isSelected(forSegment: 1)
            selectedTiles.flipVertical()
            tiles.flipVertical()
        }
        if priority != flipButton.isSelected(forSegment: 2) {
            // priority just changed
            priority = flipButton.isSelected(forSegment: 2)
            selectedTiles.togglePriority()
            tiles.togglePriority()
        }
        if graphicsIndex != graphicsSlider.integerValue {
            // graphics just changed
            graphicsIndex = graphicsSlider.integerValue
            tiles.setGraphics(graphicsIndex)
        }
        if paletteIndex != paletteSlider.integerValue * tileLayout!.tileFormat.colorsPerPalette {
            // palette just changed
            paletteIndex = paletteSlider.integerValue * tileLayout!.tileFormat.colorsPerPalette
            tiles.setPalette(paletteIndex)
        }

        tileImage = nil
        needsDisplay = true
    }
    
    override func drawCursor(x: Int, y: Int, width: Int, height: Int) {
        if selectedTiles.tiles.count == 0 { return }
        let firstTile = selectedTiles.tiles.first!
        var x = Int(firstTile & 0x0F)
        var y = Int(firstTile & 0xF0) >> 4
        if firstTile & 0x40000000 != 0 { x = 15 - x }
        if firstTile & 0x80000000 != 0 { y = 15 - y }
        super.drawCursor(x: x, y: y, width: width, height: height)
    }
    
    override func select(tileLayout: ROMTileLayout?) {
        
        // select a copy of the tile layout (so we can safely fuck around with it)
        guard let tilesetLayout = tileLayout?.copy() as? ROMTileLayout else { return }
        tilesetLayout.width = 16
        tilesetLayout.height = 16
        tilesetLayout.data = Data(count: 256 * tilesetLayout.tileFormat.bytesPerTile)
        super.select(tileLayout: tileLayout)
        
        // reset all tile flags
        graphicsSlider?.integerValue = 0
        paletteSlider?.integerValue = 0
        graphicsIndex = 0
        paletteIndex = 0
        horizontalFlip = false
        verticalFlip = false
        priority = false
        
        // set up the tileset layout
        tiles.width = 16
        tiles.height = 16
        tiles.tiles = (0..<256).map{ UInt32($0) }
        
        // disable the sliders for mode 7 tiles
        if tilesetLayout.tileFormat.rawFormat == SNES_MODE7_TILE {
            graphicsSlider.isEnabled = false
            paletteSlider.isEnabled = false
            flipButton.isEnabled = false
        } else {
            graphicsSlider.isEnabled = true
            paletteSlider.isEnabled = true
            flipButton.isEnabled = true
        }
    }
        
    override func mouseDown(with theEvent: NSEvent) {
        
        tileView.showCursor = false
        
        let clickedPoint = convert(theEvent.locationInWindow, from: nil)
        if !bounds.contains(clickedPoint) { return }
        let clickedX = Int(clickedPoint.x / zoom / 8)
        let clickedY = Int((bounds.height - clickedPoint.y) / zoom / 8)
        let clickedTile = clickedX + clickedY * 8
        var draggedTile = clickedTile
        
        selectedTiles = tiles.subTilesAt(x: clickedX, y: clickedY, width: 1, height: 1)
        showCursor = true
        needsDisplay = true

        while true {
            let nextEvent = window!.nextEvent(matching: [.leftMouseUp, .rightMouseUp, .leftMouseDragged, .rightMouseDragged])!
            
            switch (nextEvent.type) {
            case .leftMouseUp, .rightMouseUp:
                return
                
            case .leftMouseDragged, .rightMouseDragged:
                let draggedPoint = convert(nextEvent.locationInWindow, from: nil)
                let draggedX = Int(min(max(draggedPoint.x / zoom / 8, 0), 15))
                let draggedY = Int(min(max((bounds.height - draggedPoint.y) / zoom / 8, 0), 15))
                if draggedTile == draggedX + draggedY * 16 { continue }
                draggedTile = draggedX + draggedY * 16
                
                // get the selected rectangle
                let minX = min(draggedX, clickedX)
                let minY = min(draggedY, clickedY)
                let maxX = max(draggedX, clickedX)
                let maxY = max(draggedY, clickedY)
                let width = maxX - minX + 1
                let height = maxY - minY + 1
                
                // create the new tile selection
                selectedTiles = tiles.subTilesAt(x: minX, y: minY, width: width, height: height)
                needsDisplay = true
                
            default:
                break
            }
        }
    }
    
    override func rightMouseDown(with theEvent: NSEvent) {
        // left and right click have the same effect
        mouseDown(with: theEvent)
    }
    
    // ignore mouse movement
    override var acceptsFirstResponder: Bool { return false }
    override func mouseMoved(with event: NSEvent) { return }
    override func mouseEntered(with theEvent: NSEvent) { return }
    override func mouseExited(with theEvent: NSEvent) { return }
    override func updateTrackingAreas() { return }
}
