//
//  ScriptLinkField.swift
//  CustomControls
//
//  Created by Evan Owen on 6/8/16.
//  Copyright © 2016 Evan Owen. All rights reserved.
//

import Cocoa

class ScriptLinkFormatter: Formatter {
    
    override func string(for obj: Any?) -> String? {
        
        guard let reference = obj as? ROMReference,
            let referencedCommand = reference.referencedAssembly as? ROMCommand else {
                return "<Script Not Loaded>"
        }
        
        return referencedCommand.label
    }

    override func getObjectValue(_ obj: AutoreleasingUnsafeMutablePointer<AnyObject?>?, for string: String, errorDescription error: AutoreleasingUnsafeMutablePointer<NSString?>?) -> Bool {
        return false
    }
}

class ScriptLinkField: NSTextField {

    @IBOutlet weak var windowController: ROMWindowController?
    var button: NSButton!
    @objc var scriptIdentifier: String?
    var script: ROMScript? {
        if scriptIdentifier == nil { return nil }
        return windowController?.rom.value(forKey: scriptIdentifier!) as? ROMScript
    }
    @objc var encodingName: String?

    var buttonFrame: NSRect {
        
        // TODO: right to left reading order
        if userInterfaceLayoutDirection == .leftToRight {
            
        } else {
            
        }
        
        switch controlSize {
        case .regular:
            return NSMakeRect(frame.width - 16, 3.0, 16.0, 16.0)
        case .small:
            return NSMakeRect(frame.width - 15, 2.0, 15.0, 15.0)
        case .mini:
            return NSMakeRect(frame.width - 12, 2.0, 12.0, 12.0)
        }
    }

    override init(frame frameRect: NSRect) {
        super.init(frame: frameRect)
        initButton()
    }
    
    required init?(coder: NSCoder) {
        super.init(coder: coder)
        initButton()
    }
    
    func initButton() {
        
        isSelectable = false
        let scriptLinkFormatter = ScriptLinkFormatter()
        formatter = scriptLinkFormatter
        
        button = NSButton(frame: buttonFrame)
        button.setButtonType(NSButton.ButtonType.momentaryChange)
        button.bezelStyle = NSButton.BezelStyle.inline
        button.isBordered = false
        button.title = ""
        button.target = self
        button.action = #selector(ScriptLinkField.buttonAction(_:))
        button.bind(NSBindingName(rawValue: "hidden"), to: self, withKeyPath: "hidden", options: nil)
        button.bind(NSBindingName(rawValue: "controlSize"), to: self, withKeyPath: "controlSize", options: nil)
        button.bind(NSBindingName(rawValue: "enabled"), to: self, withKeyPath: "enabled", options: nil)
        
        addSubview(button)
        
        registerForDraggedTypes([NSPasteboard.PasteboardType(rawValue: "com.ekowen86.romtools.ROMCommand")])
    }
    
    override func awakeFromNib() {
        let linkImage = NSImage(named: NSImage.Name(rawValue: "ScriptMenuTemplate"))!
        let buttonImage = NSImage(size: button.frame.size)
        buttonImage.lockFocus()
        linkImage.draw(in: NSMakeRect(0, 0, buttonImage.size.width, buttonImage.size.height), from: NSZeroRect, operation: NSCompositingOperation.sourceOver, fraction: 1.0)
        buttonImage.isTemplate = true
        buttonImage.unlockFocus()
        button.image = buttonImage
    }
    
    override func resizeSubviews(withOldSize oldSize: NSSize) {
        button.frame = buttonFrame
    }
    
    @IBAction func buttonAction(_ sender: AnyObject?) {
        
        if let reference = objectValue as? ROMReference,
            let referencedCommand = reference.referencedAssembly as? ROMCommand {
            windowController?.selectObject(referencedCommand.script)
            windowController?.scriptController?.script = referencedCommand.script
            windowController?.scriptController?.select(command: referencedCommand)
        } else if scriptIdentifier != nil {
            // update the property bound to this control's value
            if let script = windowController?.rom.value(forKey: scriptIdentifier!) as? ROMScript,
                let bindingInfo = infoForBinding(NSBindingName.value),
                let observedKeyPath = bindingInfo[NSBindingInfoKey.observedKeyPath] as? String,
                let observedObject = bindingInfo[NSBindingInfoKey.observedObject] as AnyObject?,
                let commandReference = observedObject.value(forKeyPath: observedKeyPath) as? ROMReference,
                let referencedCommand = commandReference.referencedAssembly as? ROMCommand {

                objectValue = commandReference
                
                windowController?.selectObject(script)
                windowController?.scriptController?.script = script
                windowController?.scriptController?.select(command: referencedCommand)
            }
        }
    }

    func commandForReference(_ reference: Int) -> ROMCommand? {
        return script?.commandReferences[reference] as? ROMCommand
    }

    override func draggingEntered(_ sender: NSDraggingInfo) -> NSDragOperation {
        return NSDragOperation.link
    }
    
    override func performDragOperation(_ sender: NSDraggingInfo) -> Bool {
        
        guard let pasteboardItem = sender.draggingPasteboard().pasteboardItems?.first,
            let encoding = pasteboardItem.string(forType: NSPasteboard.PasteboardType(rawValue: "com.ekowen86.romtools.ROMCommand.encoding")),
            let referenceData = pasteboardItem.data(forType: NSPasteboard.PasteboardType(rawValue: "com.ekowen86.romtools.ROMCommand.reference")) else { return false }
        
        // make sure the encoding matches (use the script identifier if no encoding is specified)
        if encoding != encodingName ?? scriptIdentifier { return false }
        
        // read the reference value (8 bytes for 64-bit machines, not sure if this will work for 32-bit)
        var reference = 0
        (referenceData as NSData).getBytes(&reference, length: MemoryLayout<Int>.size)
        
        // get the referenced command
        guard let referencedCommand = commandForReference(reference) else { return false }

        // update the property bound to this control's value
        if let bindingInfo = infoForBinding(NSBindingName.value),
            let observedKeyPath = bindingInfo[NSBindingInfoKey.observedKeyPath] as? String,
            let observedObject = bindingInfo[NSBindingInfoKey.observedObject] as AnyObject?,
            let observedReference = observedObject.value(forKeyPath: observedKeyPath) as? ROMReference,
            let reference = observedReference.copy() as? ROMReference {

            // linkedObject and referencedAssembly don't automatically get copied
            reference.linkedObject = observedReference.linkedObject
            reference.referencedAssembly = referencedCommand

            // set the new reference value (the old one gets lost in the void...)
            observedObject.setValue(reference, forKeyPath: observedKeyPath)
            objectValue = reference
        }

        return true
    }
}
