//
//  ROM.swift
//  ROMTools
//
//  Created by Evan Owen on 6/13/15.
//
//

import Foundation

class ROM: ROMObject {
    
    // MARK: instance variables
    
    @objc var data = Data()
    var subAssemblies = NSMutableDictionary()
    var system: System = .None
    var memoryMap: NSMutableDictionary!
    var crc32: UInt32!
    var mapMode: MapMode = .None
    var logURL: URL
    var stringTables = NSMutableDictionary()
    var textEncodings = NSMutableDictionary()
    var characterTables = NSMutableDictionary()
    var scriptEncodings = NSMutableDictionary()
    var hierarchy: NSArray?
    var unusedMemory = IndexSet()

    // MARK: instance methods
    
    required init?(contentsOfURL url: URL) {
        
        // create a log file
        if let libraryDirectory = NSSearchPathForDirectoriesInDomains(.libraryDirectory, .userDomainMask, true).first {
            let logFileName = url.deletingPathExtension().appendingPathExtension("log").lastPathComponent
            let logDirectory = libraryDirectory.appending("/Logs/FF6Tools/")
            let logPath = logDirectory.appending(logFileName)
            if FileManager.default.fileExists(atPath: logDirectory) == false {
                // create the log directory
                let _ = try? FileManager.default.createDirectory(atPath: logDirectory, withIntermediateDirectories: true, attributes: nil)
            }
            logURL = URL(fileURLWithPath: logPath)
        } else {
            logURL = url.deletingPathExtension().appendingPathExtension("log")
        }

        super.init()
        self.rom = self
        
        // load ROM data from the url
        guard let urlData = try? Data(contentsOf: url) else { return nil }
        if urlData.count == 0 {
            return nil
        } else if urlData.count % 0x0400 == 0x0200 {
            // remove copier header
            self.data = urlData.subdata(in: 0x0200..<urlData.count)
            log("Removed copier header")
        } else {
            self.data = urlData
        }

        // get the file length and calculate the crc32
        crc32 = ROM.calculateCRC32(data)

        log("Loaded ROM File: " + url.lastPathComponent)
        log(String(format: "CRC32: 0x%08X", crc32))
        log(String(format: "File Length: 0x%X", data.count))

        // load the ROM definition
        guard let definition = definition(forROMatURL: url) else {
            log("Unable to load a valid definition file")
            return nil
        }
        
        // load the memory map
        memoryMap = definition["memoryMap"] as? NSMutableDictionary ?? NSMutableDictionary()
        
        // set the system and map mode
        system = System(rawValue: definition["system"] as? String ?? System.None.rawValue) ?? .None
        mapMode = MapMode(rawValue: definition["mapMode"] as? String ?? MapMode.None.rawValue) ?? .None
        log("System: \(system)")
        log("Map Mode: \(mapMode.rawValue)")
        
        // print SNES checksum
        if system == .SFC {
            log(String(format: "SNES Checksum (calculated): 0x%04X", ROM.calculateChecksum(data)))
        }

        // load character tables
        characterTables = definition["characterTables"] as? NSMutableDictionary ?? NSMutableDictionary()
        
        // load text encodings
        if let textEncodingDefinitions = definition["textEncodings"] as? [String : NSMutableDictionary] {
            for (identifier, encodingDefinition) in textEncodingDefinitions {
                encodingDefinition["identifier"] = identifier
                log("Creating Text Encoding: \(encodingDefinition["name"] as? String ?? identifier)")
                textEncodings[identifier] = ROMTextEncoding(self, definition: encodingDefinition)
            }
        }

        // load script encodings
        scriptEncodings.removeAllObjects()
        if let scriptEncodingDefinitions = definition["scriptEncodings"] as? [String : NSMutableDictionary] {
            for (identifier, encodingDefinition) in scriptEncodingDefinitions {
                log("Creating Script Encoding: \(encodingDefinition["name"] as? String ?? identifier)")
                scriptEncodings[identifier] = ROMScriptEncoding(self, definition: encodingDefinition)
            }
        }
        
        // load string tables
        stringTables = definition["stringTables"] as? NSMutableDictionary ?? NSMutableDictionary()

        // load unused memory
        unusedMemory.removeAll()
        if let unusedMemoryArray = definition["unusedMemory"] as? [String] {
            for rangeString in unusedMemoryArray {
                if let range = MakeRange(object: rangeString as AnyObject?) {
                    unusedMemory.insert(integersIn: mapMode.mapRange(range))
                }
            }
        }
        
        // load hierarchy
        hierarchy = definition["hierarchy"] as? NSArray
    }
    
    func updateReferences() {
        for (_, value) in subAssemblies {
            guard let assembly = value as? ROMAssembly else { continue }
            assembly.updateReferences()
        }
    }
    
    func write(to url: URL) throws {
        try data.write(to: url)
    }
    
    func assemble() {
        
        // this line fixes a bug with Data.replaceSubrange
        data = Data(data)

        var modifiedAssemblies = [(assembly: ROMAssembly, length: Int)]()
        var relocatedAssemblies = [(assembly: ROMAssembly, length: Int)]()

        // get the assembled length of everything that has been accessed
        for (key, object) in subAssemblies {
            
            guard let assembly = object as? ROMAssembly else { continue }

            // can't assemble something without a range
            if assembly.range == nil {
                let assemblyName = assembly.name ?? (key as? String) ?? "Unnamed Assembly"
                log("Unable to assemble \(assemblyName), it has no range")
                continue
            }

            // add each assembly to the array
            modifiedAssemblies.append((assembly, assembly.assembledLength))
            
            // add the pointer table if there is one
            if let pointerTable = assembly.pointerTable,
                pointerTable.range != nil {
                modifiedAssemblies.append((pointerTable, pointerTable.assembledLength))
            }
        }
        
        // check for assemblies that changed size
        for (assembly, length) in modifiedAssemblies {
            
            let assemblyName = assembly.name ?? assembly.identifier ?? "Unnamed Assembly"
            let oldRange = CountableRange<Int>(assembly.range!)
            let start = oldRange.lowerBound
            let end = start + length
            let newRange = start..<end
            
            if newRange.count < oldRange.count {
                // shrink the range for assemblies that got smaller
                let freedRange = oldRange.suffix(from: newRange.upperBound)
                log("\(assemblyName) decreased in size by \(freedRange.count) bytes")
                unusedMemory.insert(integersIn: freedRange)
                assembly.range = start..<end
                
            } else if newRange.count > oldRange.count {
                // for assemblies that got bigger, either expand them in place or relocate
                let usedRange = newRange.suffix(from: oldRange.upperBound)
                log("\(assemblyName) increased in size by \(usedRange.count) bytes")
                
                // check if the adjacent memory is unused
                if unusedMemory.contains(integersIn: usedRange) {
                    // expand into unused memory
                    unusedMemory.remove(integersIn: usedRange)
                    assembly.range = start..<end
                } else {
                    // adjacent space is occupied, assembly needs to be relocated
                    relocatedAssemblies.append((assembly, length))
                }
            }
        }

        // relocate assemblies
        if autoRelocate(relocatedAssemblies) == false {
        
            if UserDefaults.standard.bool(forKey: autoExpandKey) == false {
                // automatic expansion is disabled
                log("Not enough room to relocate assemblies. Please expand the ROM.")
                
            } else if system == .SFC && data.count == 0x300000 {
                // expand the ROM and try again
                expand(toLength: 0x400000)
                if autoRelocate(relocatedAssemblies) == false {
                    // that didn't work either...
                    log("Unable to relocate assemblies")
                }
            } else {
                // ROM can't be expanded
                log("Unable to relocate assemblies")
            }
        }

        // update references (this is what allows the game's code to access relocated assemblies)
        updateReferences()

        // updating references may cause some the size of assemblies to change (if data is compressed or something)
        // repeat this subroutine and then return to update ranges
//        if relocatedAssemblies.count != 0 {
//            
//            assemble()
//            return
//        }
        
        // finally, assemble each sub-assembly and copy it into its validated range
        for (assembly, _) in modifiedAssemblies {
            
            let assemblyName = assembly.name ?? assembly.identifier ?? "Unnamed Assembly"
            log("Assembling: " + assemblyName)
            var assemblyData = Data()
            assembly.assembleTo(&assemblyData)
            // TODO: remove this once i figure out how to get fix the thing above
            // this should only affect the cinematic program if references are written to it during updateReferences()
            let start = assembly.range!.lowerBound
            let end = start + assemblyData.count
            assembly.range = start..<end
            unusedMemory.remove(integersIn: assembly.range!)
            // ***
            data.replaceSubrange(start..<end, with: assemblyData)
        }
        
        // pad unused space
        let unusedMemoryRanges = unusedMemory.rangeView
        for unusedRange in unusedMemoryRanges {
            let padBytes = [UInt8](repeating: 0xFF, count: unusedRange.count)
            data.replaceSubrange(unusedRange, with: padBytes)
        }
    }
    
//    func validateRanges() throws {
//        var modifiedAssemblies = [(assembly: ROMAssembly, length: Int)]()
//        var relocatedAssemblies = [(assembly: ROMAssembly, length: Int)]()
//        
//        // get the assembled length of everything that has been accessed
//        for (key, object) in subAssemblies {
//            
//            guard let assembly = object as? ROMAssembly else { continue }
//            
//            // can't assemble something without a range
//            if assembly.range == nil {
//                let assemblyName = assembly.name ?? (key as? String) ?? "Unnamed Assembly"
//                log("Unable to assemble \(assemblyName), it has no range")
//                continue
//            }
//            
//            // add each assembly to the array
//            modifiedAssemblies.append((assembly, assembly.assembledLength))
//            
//            // add the pointer table if there is one
//            if let pointerTable = assembly.pointerTable,
//                pointerTable.range != nil {
//                modifiedAssemblies.append((pointerTable, pointerTable.assembledLength))
//            }
//        }
//        
//        // check for assemblies that changed size
//        for (assembly, length) in modifiedAssemblies {
//            
//            let assemblyName = assembly.name ?? assembly.identifier ?? "Unnamed Assembly"
//            let oldRange = CountableRange<Int>(assembly.range!)
//            let start = oldRange.lowerBound
//            let end = start + length
//            let newRange = start..<end
//            
//            if newRange.count < oldRange.count {
//                // shrink the range for assemblies that got smaller
//                let freedRange = oldRange.suffix(from: newRange.upperBound)
//                log("\(assemblyName) decreased in size by \(freedRange.count) bytes")
//                unusedMemory.insert(integersIn: freedRange)
//                assembly.range = start..<end
//                
//            } else if newRange.count > oldRange.count {
//                // for assemblies that got bigger, either expand them in place or relocate
//                let usedRange = newRange.suffix(from: oldRange.upperBound)
//                log("\(assemblyName) increased in size by \(usedRange.count) bytes")
//                
//                // check if the adjacent memory is unused
//                if unusedMemory.contains(integersIn: usedRange) {
//                    // expand into unused memory
//                    unusedMemory.remove(integersIn: usedRange)
//                    assembly.range = start..<end
//                } else {
//                    // adjacent space is occupied, assembly needs to be relocated
//                    relocatedAssemblies.append((assembly, length))
//                }
//            }
//        }
//        
//        // relocate assemblies
//        if autoRelocate(relocatedAssemblies) == false {
//            
//            if UserDefaults.standard.bool(forKey: autoExpandKey) == false {
//                // automatic expansion is disabled
//                log("Not enough room to relocate assemblies. Please expand the ROM.")
//                throw NSError()
//                
//            } else if system == .SFC && data.count == 0x300000 {
//                // expand the ROM and try again
//                expand(toLength: 0x400000)
//                if autoRelocate(relocatedAssemblies) == false {
//                    // that didn't work either...
//                    log("Unable to relocate assemblies")
//                    throw NSError()
//                }
//            } else {
//                // ROM can't be expanded
//                log("Unable to relocate assemblies")
//                throw NSError()
//            }
//        }
//        
//        // update references (this is what allows the game's code to access relocated assemblies)
//        updateReferences()
//    }
    
    func autoRelocate(_ assemblies: [(assembly: ROMAssembly, length: Int)]) -> Bool {
        
        // return true if there's nothing to be relocated
        if assemblies.isEmpty { return true }

        // make a copy of the unused memory map
        var proposedUnusedMemory = unusedMemory

        // make a list of all assemblies that can be relocated
        var relocatedAssemblies = [(assembly: ROMAssembly, length: Int)]()
        for (assembly, _) in assemblies {

            // if this assembly has a previous assembly, find the first on in the chain and add that instead
            var firstAssembly: ROMAssembly! = assembly
            while firstAssembly.previousAssembly != nil {
                guard let previousAssembly = value(forKey: firstAssembly.previousAssembly!) as? ROMAssembly else { break }
                firstAssembly = previousAssembly
            }
            
            // assemblies can't be relocated if they don't have references
            if firstAssembly.references == nil {
                let assemblyName = firstAssembly.name ?? firstAssembly.identifier ?? "Unnamed Assembly"
                log("Unable to relocate assembly: " + assemblyName)
                continue
            }

            // calculate the total length of this assembly and all following it
            var totalLength = 0
            var nextAssembly: ROMAssembly! = firstAssembly
            while nextAssembly != nil {
                totalLength += nextAssembly!.assembledLength
                proposedUnusedMemory.insert(integersIn: nextAssembly.range!)
                guard let nextKey = nextAssembly.nextAssembly else { break }
                nextAssembly = value(forKey: nextKey) as? ROMAssembly
            }

            // only the first assembly is added to the array, but the length is the total length of all following assemblies
            relocatedAssemblies.append((firstAssembly, totalLength))
        }
        
        // dictionary to store the new ranges for each assembly while checking to see if everything fits
        var newRanges = [ROMAssembly : Range<Int>]()

        // sort the relocated assemblies in order from largest to smallest
        for (assembly, length) in relocatedAssemblies.sorted(by: { $0.length > $1.length } ) {

            // don't relocate an assembly more than once
            if newRanges[assembly] != nil { continue }
            
            // find the smallest empty block that the assembly will fit into
            var newRange: Range<Int> = 0..<data.count // start with the length of the entire ROM
            newRange = proposedUnusedMemory.rangeView.reduce(newRange, {
                var range = Range<Int>($1)
                
                // re-align the range if required by the assembly
                if assembly.align != 1 {
                    let end = range.upperBound
                    var start = range.lowerBound
                    if start % assembly.align != 0 {
                        // align range to the next valid block
                        start += assembly.align - (start % assembly.align)
                    }
                    if end <= start { return $0 }
                    range = start..<end
                }
                
                // if the data is smaller than one block make sure it fits all in one block
                let blockSize = 0x010000
                if length < blockSize {
                    let startBlock = range.lowerBound / blockSize
                    let endBlock = (range.lowerBound + length) / blockSize
                    let end = range.upperBound
                    if endBlock != startBlock {
                        if end > (endBlock * blockSize + length) {
                            range = (endBlock * blockSize)..<end
                        } else {
                            return $0
                        }
                    }
                }
                
                if (range.count >= length) && (range.count < $0.count) {
                    return range
                } else {
                    return $0
                }
            })
            if newRange.count == data.count {
                // fail if there are no blocks large enough to hold this assembly
                let assemblyName = assembly.name ?? assembly.identifier ?? "Unnamed Assembly"
                log("Not enough free space to relocate assembly: " + assemblyName)
                return false
            }

            // set the range length to match the assembly data
            newRange = newRange.lowerBound..<(newRange.lowerBound + length)
            
            // mark the new range as used
            newRanges[assembly] = newRange
            proposedUnusedMemory.remove(integersIn: newRange)
        }

        // update unused memory
        unusedMemory = proposedUnusedMemory
        
        for (assembly, range) in newRanges {
            unusedMemory.insert(integersIn: range)
            relocate(assembly, to: range.lowerBound)
        }
        
        return true
    }
    
    func relocate(_ assembly: ROMAssembly, to location: Int) {
        
        // note that this function does not remove the previous location from unused memory
        let newRange: Range<Int> = location..<(location + assembly.assembledLength)
        let assemblyName = assembly.name ?? assembly.identifier ?? "Unnamed Assembly"
        
        // make sure the proposed range is currently unused
        if !unusedMemory.contains(integersIn: newRange) {
            let rangeString = String(format: "0x%06X", mapMode.unmapAddress(newRange.lowerBound))
            log("Unable to Relocate " + assemblyName + ", Invalid Location: " + rangeString)
            return
        }
        
        log(String(format: "Relocating %@ from 0x%06X to 0x%06X", assemblyName, mapMode.unmapAddress(assembly.range!.lowerBound), mapMode.unmapAddress(location)))
        
        // update unused memory (only the new used memory, does not free old memory)
        unusedMemory.remove(integersIn: newRange)

        assembly.relocate(to: location)

        // if there is a next assembly, relocate it too
        // this will recursively relocate all attached assemblies
        if let nextAssemblyIdentifier = assembly.nextAssembly,
            let nextAssembly = value(forKey: nextAssemblyIdentifier) as? ROMAssembly {
            relocate(nextAssembly, to: assembly.range!.upperBound)
        }
    }
    
    func expand(toLength length: Int) {
        let lengthAdded = length - data.count
        if lengthAdded <= 0 {
            log("Expanded length is not greater than current length")
            return
        }
        unusedMemory.insert(integersIn: data.count..<length)
        willChangeValue(forKey: "data")
        data.count = length
        didChangeValue(forKey: "data")
        
        log(String(format: "ROM expanded to length 0x%06X", length))
    }
    
    func updateSNESChecksum() {
        
        // this line fixes a bug with Data.replaceSubrange
        data = Data(data)

        // get the SNES header assembly
        guard let snesHeader = value(forKey: "snesHeader") as? ROMData,
            let headerRange = snesHeader.range else {
            log("Unable to update SNES checksum: Header assembly not found")
            return
        }
        var headerData = snesHeader.data //self.data.subdata(in: headerRange)
        
        // clear the checksum in the snes header
        snesHeader.setValue(0x0000, forKeyPath: "checksum")
        snesHeader.setValue(0xFFFF, forKeyPath: "inverseChecksum")
        snesHeader.assembleTo(&headerData)
        data.replaceSubrange(headerRange, with: headerData)
        
        // calculate the actual checksum
        let calculatedChecksum = Int(ROM.calculateChecksum(self.data))
        
        // write the calculated checksum to the data
        snesHeader.setValue(calculatedChecksum, forKeyPath: "checksum")
        snesHeader.setValue(calculatedChecksum ^ 0xFFFF, forKeyPath: "inverseChecksum")
        snesHeader.assembleTo(&headerData)
        data.replaceSubrange(headerRange, with: headerData)
    }
    
    func definition(forROMatURL url: URL) -> NSDictionary? {
        
        guard let data = try? Data(contentsOf: url) else { return nil }
        let crc32 = ROM.calculateCRC32(data)
        
        let definitionURL = url.deletingPathExtension().appendingPathExtension("plist")
        if let userSpec = NSMutableDictionary(contentsOf: definitionURL) {
            // load a user definition from the same folder as the ROM file
            log("Loaded User Definition: " + definitionURL.lastPathComponent)
            return userSpec
        } else if let builtInSpec = builtInDefinition(withCRC32: crc32, fileLength: data.count) {
            return builtInSpec
        }
        
        return nil
    }
    
    func builtInDefinition(withCRC32 crc32: UInt32, fileLength: Int) -> NSDictionary? {
        
        // get an array of plist files from the main bundle
        guard let definitionURLs = Bundle.main.urls(forResourcesWithExtension: "plist", subdirectory: nil) else { return nil }
        
        // look for a definition with a matching CRC32 and file length
        for url in definitionURLs {
            
            guard let definition = NSDictionary(contentsOf: url),
                let definitionCRC32 = Int(object: definition, keyPath: "crc32"),
                let definitionLength = Int(object: definition, keyPath: "fileLength") else { continue }
            
            if (UInt32(definitionCRC32) == crc32) && (definitionLength == fileLength) {
                log("Loaded Built-In Definition: " + url.lastPathComponent)
                return definition
            }
            
        }
        return nil
    }

    override func createDefinition() -> NSMutableDictionary {
        let definition = super.createDefinition()
        
        // update header values
        crc32 = ROM.calculateCRC32(data)
        definition["crc32"] = String(format: "0x%08X", crc32)
        definition["fileLength"] = String(format: "0x%X", data.count)
        definition["system"] = system.rawValue
        definition["mapMode"] = mapMode.rawValue
        
        // update the memory map
        for (key, value) in subAssemblies {
            guard let name = key as? String, let assembly = value as? ROMAssembly else { continue }
            memoryMap[name] = assembly.createDefinition()
        }
        definition["memoryMap"] = memoryMap

        // update character tables
        definition["characterTables"] = characterTables
        
        // update text encodings
        let textEncodingsDefinitions = NSMutableDictionary()
        for (_, value) in textEncodings {
            guard let textEncoding = value as? ROMTextEncoding else { continue }
            textEncodingsDefinitions[textEncoding.identifier!] = textEncoding.createDefinition()
        }
        definition["textEncodings"] = textEncodingsDefinitions
        
        // update string tables
        for (key, value) in subAssemblies {
            guard let identifier = key as? String, let stringTable = value as? ROMStringTable else { continue }
            if stringTable.isLinked { continue }
            let identifierComponents = identifier.components(separatedBy: "-")
            if identifierComponents.count != 2 || identifierComponents[0] != "stringTable" { continue }
            let stringIdentifier = identifierComponents[1]
            
            stringTables[stringIdentifier] = stringTable.createDefinition()
        }
        definition["stringTables"] = stringTables
        
        // update unused memory
        let unusedMemoryRanges = unusedMemory.rangeView
        
        let unusedMemoryArray = NSMutableArray()
        for unusedRange in unusedMemoryRanges {
            let start = mapMode.unmapAddress(unusedRange.lowerBound)
            let end = mapMode.unmapAddress(unusedRange.upperBound)
            unusedMemoryArray.add(String(format: "0x%06X-0x%06X", start, end))
        }
        definition["unusedMemory"] = unusedMemoryArray
        
        // update script encodings
        let scriptEncodingArray = NSMutableDictionary()
        for (key, value) in scriptEncodings {
            guard let name = key as? String, let encoding = value as? ROMScriptEncoding else { continue }
            scriptEncodingArray[name] = encoding.createDefinition()
        }
        definition["scriptEncodings"] = scriptEncodingArray
        
        // update hierarchy (this shouldn't change)
        definition["hierarchy"] = hierarchy
        
        return definition
    }
    
    func log(_ message: String) {
        
        // make the timestamp
        let dateFormatter = DateFormatter()
        dateFormatter.dateFormat = "yyyy-MM-dd HH:mm:ss.SSS "
        let timestamp = dateFormatter.string(from: Date())

        // print the message to the debug console
        Swift.print(timestamp + message)

        // return if log is disabled
        if UserDefaults.standard.bool(forKey: createLogKey) == false { return }
        
        // write to the log file
        let logData = (timestamp + message + "\n").data(using: String.Encoding.unicode)!
        if let fileHandle = try? FileHandle(forWritingTo: logURL) {
            defer {
                fileHandle.closeFile()
            }
            fileHandle.seekToEndOfFile()
            fileHandle.write(logData)
        }
        else {
            try? logData.write(to: logURL, options: [.atomic])
        }
    }
    
    func createROMMap() {
        
        // create a list of memory map items with pointer tables and unused memory mixed in
        var romMap = [String : NSDictionary]()
        for (key, object) in memoryMap {
            guard let definition = object as? NSDictionary else { continue }
            guard let identifier = key as? String else { continue }
            romMap[identifier] = definition
            
            guard let pointerTable = definition["pointerTable"] as? NSDictionary else { continue }
            let name = definition["name"] as? String ?? identifier
            let pointerTableDefinition = NSMutableDictionary()
            pointerTableDefinition["name"] = "Pointers to " + name
            pointerTableDefinition["range"] = pointerTable["range"]
            romMap[identifier + ".pointerTable"] = pointerTableDefinition
        }
        
        // update unused memory
        let unusedMemoryRanges = unusedMemory.rangeView
        
        for unusedRange in unusedMemoryRanges {
            let start = mapMode.unmapAddress(unusedRange.lowerBound)
            let end = mapMode.unmapAddress(unusedRange.upperBound)
            let rangeString = String(format: "0x%06X-0x%06X", start, end)
            let unusedMemoryDefinition = NSMutableDictionary()
            unusedMemoryDefinition["name"] = "Unused Memory"
            unusedMemoryDefinition["range"] = rangeString
            romMap[rangeString] = unusedMemoryDefinition
        }
        
        // sort the memory map by assembly location
        let sortedMap = romMap.sorted{
            
            guard let firstRange = MakeRange(object: $0.value, keyPath: "range") else { return false }
            guard let secondRange = MakeRange(object: $1.value, keyPath: "range") else { return false }
            
            return firstRange.lowerBound < secondRange.lowerBound
        }
        
        // create a string with assembly locations and names
        var mapString = ""
        switch system {
        case .SFC, .None:
            mapString += "ROM Map:\n\nLocation      Description\n--------      -----------\n\n"
        case .GBA, .PSX:
            mapString += "ROM Map:\n\nLocation          Description\n--------          -----------\n\n"
        }
        
        for (identifier, definition) in sortedMap {
            let name = definition["name"] as? String ?? identifier
            guard let range = MakeRange(object: definition, keyPath: "range") else { continue }
            
            switch system {
            case .SFC, .None:
                mapString += String(format: "%06X-%06X %@\n", range.lowerBound, range.upperBound, name)
            case .GBA, .PSX:
                mapString += String(format: "%08X-%08X %@\n", range.lowerBound, range.upperBound, name)
            }
        }
        mapString += "\n"
        log(mapString)
    }
    
    func stringTable(forKey key: String) -> ROMStringTable? {
        
        var identifier = key
        if let prefixRange = key.range(of: "stringTable-") {
            identifier = String(key[prefixRange.upperBound...])
//            identifier = key.substring(from: prefixRange.upperBound)
        }
        
        if let stringTable = subAssemblies["stringTable-" + identifier] as? ROMStringTable {
            
            // string table is already loaded
            return stringTable
            
        } else if let definition = stringTables[identifier] as? NSMutableDictionary {
            
            // string table is in the ROM definition
            definition["identifier"] = "stringTable-" + identifier
            let newStringTable = ROMStringTable(self, definition: definition)
            subAssemblies["stringTable-" + identifier] = newStringTable
            
            log("Creating String Table: " + (newStringTable.name ?? identifier))
            
            return newStringTable
            
        } else if let assemblyDefinition = memoryMap[identifier] as? NSDictionary,
            let definition = assemblyDefinition["stringTable"] as? NSMutableDictionary {
            
            // string table is in an assembly
            definition["identifier"] = "stringTable-" + identifier
            let newStringTable = ROMStringTable(self, definition: definition)
            newStringTable.isLinked = true
            subAssemblies["stringTable-" + identifier] = newStringTable
            
            log("Creating String Table: " + (newStringTable.name ?? identifier))
            
            return newStringTable
        }
        return nil
    }
    
    func string(atIndex index: Int, inStringTable stringTableKey: String) -> String? {
        
        if let stringTable = stringTable(forKey: stringTableKey), let string = stringTable[index] as? ROMString {
            return string.formattedString
        }
        
        return nil
    }

    // MARK: NSObject methods
    
    override func value(forKey key: String) -> Any? {
        if let sub = subAssemblies.object(forKey: key) as AnyObject? {
            
            // sub-assembly is already loaded
            return sub
            
        } else if let definition = memoryMap[key] as? NSMutableDictionary {

            // key is in the memory map
            definition["identifier"] = key
            guard let newAssembly = ROMObject.createObject(self, definition: definition) as? ROMAssembly else { return nil }
            log("Disassembling: " + (newAssembly.name ?? key))
            newAssembly.disassembleFrom(data)
            subAssemblies[key] = newAssembly
            
            // disassemble next and previous assemblies
            if let nextAssembly = newAssembly.nextAssembly { let _ = value(forKey: nextAssembly) }
            if let previousAssembly = newAssembly.previousAssembly { let _ = value(forKey: previousAssembly) }
            
            return newAssembly
            
        } else if let stringTable = stringTable(forKey: key) {
            return stringTable
        }
        
        return super.value(forKey: key)
    }
    
    override func value(forUndefinedKey key: String) -> Any? {
//        log("Assembly \"\(key)\" not found")
        return nil
    }
    
    override func value(at index: Int, inPropertyWithKey key: String) -> Any? {
        if let array = value(forKey: key) as? ROMArray {
            if index < 0 { return nil }
            if index >= array.count { return nil }
            return array.objectAtIndex(index)
        } else if let stringTable = value(forKey: key) as? ROMStringTable {
            return stringTable[index]
        }
        
        return super.value(at: index, inPropertyWithKey: key)
    }
    
    override func mutableArrayValue(forKey key: String) -> NSMutableArray {
        if let array = value(forKey: key) as? ROMArray {
            return array.array
        } else if let pointerTable = value(forKey: key) as? ROMPointerTable {
            return NSMutableArray(array: pointerTable.pointers)
        }
        
        return super.mutableArrayValue(forKey: key)
    }
    
    override func setValue(_ value: Any?, forUndefinedKey key: String) {
        log("Assembly \"\(key)\" not found")
        return
    }

    // MARK: class methods
    
    class func calculateChecksum(_ data: Data) -> UInt16 {
        return calculate_checksum((data as NSData).bytes.bindMemory(to: UInt8.self, capacity: data.count), data.count)
    }
    
    class func calculateCRC32(_ data: Data) -> UInt32 {
        return CRC32((data as NSData).bytes.bindMemory(to: UInt8.self, capacity: data.count), data.count)
    }
}

extension Int {
    init?(object: AnyObject?, keyPath: String? = nil) {
        if object == nil {
            // nil value causes function to fail
            return nil
            
        } else if keyPath != nil, let value = object?.value(forKeyPath: keyPath!) as AnyObject? {
            // function is recursive if a keyPath is provided
            self.init(object: value)
            
        } else if let stringValue = object as? String {
            // string value
            self.init(expression: stringValue)
            
        } else if let intValue = object as? Int {
            // int value
            self.init(intValue)
            
        } else {
            return nil
        }
    }
    
    init?(expression: String) {
        
        if expression.isEmpty {
            self.init(0)
            return
        }
        
        // parse addition and subtraction
        if expression.characters.contains("+") {
            var intValue = 0
            for argumentString in expression.components(separatedBy: "+") {
                guard let argumentInt = Int(expression: argumentString) else { return nil }
                intValue += argumentInt
            }
            self.init(intValue)
            return
        } else if expression.characters.contains("-") {
            var intValue: Int?
            for argumentString in expression.components(separatedBy: "-") {
                guard let argumentInt = Int(expression: argumentString) else { return nil }
                if intValue == nil {
                    intValue = argumentInt
                } else {
                    intValue! -= argumentInt
                }
            }
            self.init(intValue!)
            return
        }
        
        let intString: String
        let radix: Int
        if expression.hasPrefix("0x") {
            intString = String(expression.characters.dropFirst(2))
            radix = 16
        } else if expression.hasPrefix("$") {
            intString = String(expression.characters.dropFirst())
            radix = 16
        } else if expression.hasSuffix("h") {
            intString = String(expression.characters.dropLast())
            radix = 16
        } else {
            intString = expression
            radix = 10
        }
        
        if let intValue = Int(intString, radix: radix) {
            self.init(intValue)
        } else {
            return nil
        }
    }
}

func MakeRange(object: AnyObject?, keyPath: String? = nil) -> Range<Int>? {
    
    if object == nil {
        // nil value causes function to fail
        return nil
        
    } else if keyPath != nil, let value = object?.value(forKeyPath: keyPath!) as AnyObject? {
        // function is recursive if a keyPath is provided
        return MakeRange(object: value)
        
    } else if let rangeString = object as? String {
        // string value
        guard let dashIndex = rangeString.index(of: "-"),
            let start = Int(object: String(rangeString[..<dashIndex]) as AnyObject?),
            let end = Int(object: String(rangeString[rangeString.index(after: dashIndex)...]) as AnyObject?) else {
                return nil
        }
        return start..<end
    }

    return nil
}

enum System: String {
    case SFC = "SFC"
    case GBA = "GBA"
    case PSX = "PSX"
    case None = "None"
}

enum MapMode: String {
    case LoROM = "LoROM"
    case HiROM = "HiROM"
    case GBA = "GBA"
//    case PSX = "PSX"
    case None = "None"
    
    func mapRange(_ range: Range<Int>) -> Range<Int> {
        let start = mapAddress(range.lowerBound)
        let end = start + range.count
        return start..<end
    }
    
    func mapAddress(_ address: Int) -> Int {
        
        switch self {
        case .LoROM:
            let bank = address & 0xFF0000;
            return (bank >> 1) | (address & 0x7FFF)
            
        case .HiROM:
            if address >= 0xC00000 {
                return address - 0xC00000
            }
            else if address >= 0x800000 {
                return address - 0x800000
            }
            
        case .GBA:
            if address >= 0x08000000 {
                return address - 0x08000000
//            } else if address < 0 {
//                return address - 0x08000000
            }
            
        case .None:
            break
        }
        return address
    }
    
    func unmapAddress(_ address: Int) -> Int {
        switch self {
        case .LoROM:
            let bank = address & 0x7F8000;
            return (bank << 1) | 0x8000 | (address & 0x7FFF)
            
        case .HiROM:
            return address + 0xC00000
            
        case .GBA:
            return address + 0x08000000
            
        case .None:
            break
        }
        
        return address
    }
}
