//
//  ROMData.swift
//  ROMTools
//
//  Created by Evan Owen on 2/23/15.
//
//

import Foundation

class ROMData: ROMAssembly {
    
    // MARK: instance properties

    @objc dynamic var data: Data {
        // invalidate lazy data any time this is accessed
        get {
            lazyData = nil
            return _data
        }
        set {
            lazyData = nil
            willChangeValue(forKey: "count")
            _data = newValue
            didChangeValue(forKey: "count")
        }
    }
    @objc var count: Int {
        get {
            return data.count
        }
        set {
            var newData = Data(data)
            newData.count = newValue
            data = newData
        }
    }
    fileprivate var _data = Data()
    fileprivate var lazyData: Data?

    var subAssemblies = NSMutableDictionary()
    var compression = ROMCompression.None
    override var description: String { return data.description }
    override var assembledLength: Int {
        
        // if the data has never been accessed, return the original length
        if let lazyLength = lazyData?.count {
            return lazyLength
        }
        
        // otherwise, assemble it to a dummy Data and return the length (this will take compression into account)
        var assembledData = Data()
        assembleTo(&assembledData)
        lazyData = assembledData
        return assembledData.count
    }
    
    // MARK: instance methods
    
    override init(_ rom: ROM, definition: NSDictionary? = nil) {
        
        if let subAssemblyDefinitions = definition?["subAssemblies"] as? NSDictionary {
            for (key, value) in subAssemblyDefinitions {
                guard let identifier = key as? String else { continue }
                
                // sub data definition
                if let subDefinition = value as? NSDictionary {
                    let subAssembly = ROMObject.createObject(rom, definition: subDefinition)
                    subAssembly.identifier = identifier
                    subAssemblies[identifier] = subAssembly
                }
            }
        }
        
        if let compressionName = definition?["compression"] as? String,
            let compression = ROMCompression(rawValue: compressionName) {
            self.compression = compression
        }
        
        super.init(rom, definition: definition)
        
        let properties = NSMutableDictionary(dictionary: self.properties ?? NSDictionary())
        properties["data"] = 0
        self.properties = properties
    }

    // MARK: ROMAssembly methods
    
    override func createDefinition() -> NSMutableDictionary {
        let definition = super.createDefinition()
        
        // type
        definition["type"] = "ROMData"
        
        // sub-assemblies
        if subAssemblies.count != 0 {
            let subDefinitions = NSMutableDictionary()
            for (name, sub) in subAssemblies {
                if let subAssembly = sub as? ROMAssembly {
                    subDefinitions[name as! String] = subAssembly.createDefinition()
                }
            }
            definition["subAssemblies"] = subDefinitions
        }
        if compression != .None {
            definition["compression"] = compression.rawValue
        }
        if let properties = definition["properties"] as? NSMutableDictionary {
            // copy the properties but remove "data"
            let definitionProperties = NSMutableDictionary(dictionary: properties)
            definitionProperties["data"] = nil
            definition["properties"] = definitionProperties
        }

        return definition
    }
    
    override func assembleTo(_ data: inout Data) {
        
        // this line fixes a bug with Data.replaceSubrange
        // i commented this one out because it slows things down too - hopefully it won't cause a problem
//        self.data = Data(self.data)

        // if the data was never accessed and there are no sub-assemblies, return lazy data
        if lazyData != nil && subAssemblies.count == 0 {
            data = lazyData!
            return
        }
        
        // assemble sub-assemblies
        for (_, sub) in subAssemblies {
            
            // assemble each sub-assembly
            guard let assembly = sub as? ROMAssembly else { continue }
            
            // assemble the data
            var assemblyData = self.data
            assembly.assembleTo(&assemblyData)
            if assembly.range != nil {
                if assembly.range!.count > assemblyData.count {
                    let start = assembly.range!.lowerBound
                    let end = start + assemblyData.count
                    assembly.range = start..<end
                }
                self.data.replaceSubrange(assembly.range!, with: assemblyData.subdata(in: 0..<assembly.range!.count))
            } else {
                self.data = assemblyData
            }
        }
        
        // this apparently creates a copy of self.data, because only the inout data gets compressed
        data = self.data
        
        // compress the data, if needed
        compression.compress(&data)
        
        // validate lazyData if there are no sub-assemblies
        if subAssemblies.count == 0 { lazyData = data }
    }
    
    override func disassembleFrom(_ data: Data) {
        
        // validate lazyData
        lazyData = data
        
        // if the range property is valid, copy only the specified range of data
        if range != nil {
            if range!.upperBound > lazyData!.count {
                // end of range is past the end of the data
                rom.log("Invalid Range (\(range!)) for data of length \(lazyData!.count)")
                range = range!.lowerBound..<lazyData!.count
                rom.log("Trimming range to (\(range!))")
                lazyData = lazyData!.subdata(in: range!)
            } else if range!.lowerBound > lazyData!.count {
                // beginning of range is past the end of the data
                rom.log("Invalid range (\(range!)) for data of length \(lazyData!.count)")
                range = nil
            } else {
                lazyData = lazyData!.subdata(in: range!)
            }
        }
        
        _data = lazyData!
        
        // decompress the data, if needed
        compression.decompress(&_data)
        
        // disassemble any sub-assemblies
        for (_, object) in subAssemblies {
            guard let assembly = object as? ROMAssembly else { continue }
            assembly.disassembleFrom(self.data) // this will invalidate lazyData
        }
    }
    
    override func updateReferences() {
        super.updateReferences()
        
        for (_, object) in subAssemblies {
            guard let subAssembly = object as? ROMAssembly else { continue }
            subAssembly.updateReferences()
        }
    }

    // MARK: NSKeyValueCoding methods
    
    func intValue(forKey key: String) -> Int {
        return value(forKey: key) as? Int ?? 0
    }

    func boolValue(forKey key: String) -> Bool {
        return intValue(forKey: key) != 0
    }

    override func value(forKey key: String) -> Any? {
        
        if let reference = references?[key] {
            return reference
            
        } else if let propertyDefinition = properties?[key] as? NSDictionary {
            
            let property = ROMProperty(rom, definition: propertyDefinition)
            property.disassembleFrom(data)
            return property.property
            
        } else if let sub = subAssemblies[key] as? ROMAssembly {
            return sub
            
        }
        return super.value(forKey: key)
    }

    override func value(forUndefinedKey key: String) -> Any? {
        return nil
    }
    
    override func setValue(_ value: Any?, forKey key: String) {
        
        if let referenceValue = value as? ROMReference {
            if references == nil { references = [String : ROMReference]() }
            willChangeValue(forKey: key)
            references![key] = referenceValue
            didChangeValue(forKey: key)

            return
            
        } else if let propertyDefinition = properties?[key] as? NSDictionary, let intValue = value as? Int {
            
            let property = ROMProperty(rom, definition: propertyDefinition)
            property.disassembleFrom(data)
            
            // return if the value didn't change
            if property.property == intValue { return }
            
            willChangeValue(forKey: key)
            property.property = intValue
            property.assembleTo(&_data)
            lazyData = nil
            didChangeValue(forKey: key)

            return

        } else if let propertyDefinition = properties?[key] as? NSDictionary, let boolValue = value as? Bool {
            
            let property = ROMProperty(rom, definition: propertyDefinition)
            property.disassembleFrom(data)
            
            // return if the value didn't change
            if property.property == (boolValue ? 1 : 0) { return }
            
            willChangeValue(forKey: key)
            property.property = boolValue ? 1 : 0
            property.assembleTo(&_data)
            lazyData = nil
            didChangeValue(forKey: key)
            
            return

        } else if let stringAssembly = subAssemblies[key] as? ROMText, let stringValue = value as? String {
            // return if the value didn't change
            if stringAssembly.string == stringValue { return }
            
            willChangeValue(forKey: key)
            stringAssembly.string = stringValue
            didChangeValue(forKey: key)
            
            return
        }
        super.setValue(value, forKey: key)
    }
    
    override func mutableArrayValue(forKey key: String) -> NSMutableArray {
        if let sub = value(forKey: key) as? ROMArray {
            return sub.array
        }
        
        return super.mutableArrayValue(forKey: key)
    }
    
    override func value(at index: Int, inPropertyWithKey key: String) -> Any? {
        if let array = value(forKey: key) as? ROMArray {
            return array[index]
        }
        
        return super.value(at: index, inPropertyWithKey: key)
    }

    subscript(index: String) -> AnyObject? {
        return value(forKey: index) as AnyObject?
    }
}

enum ROMCompression: String {
    case None = "None"
    case FF6 = "FF6"
    
    func compress(_ data: inout Data) {
        
        switch self {
            
        case .None:
            break

        case .FF6:
            let bufferSize = Int(MAX_COMPRESSION_BUFFER)
            let buffer = UnsafeMutablePointer<UInt8>.allocate(capacity: bufferSize)
            data.withUnsafeBytes{ (bytes: UnsafePointer<UInt8>) in
                let compressedSize = compress_ff3(buffer, bytes, data.count)
                data = Data(bytes: buffer, count: Int(compressedSize))
            }
            buffer.deinitialize()
            buffer.deallocate(capacity: bufferSize)
        }
    }
    
    func decompress(_ data: inout Data) {

        switch self {

        case .None:
            break
            
        case .FF6:
            if data.count < 2 {
                data = Data()
                return
            }
            
            let bufferSize = Int(MAX_COMPRESSION_BUFFER)
            let buffer = UnsafeMutablePointer<UInt8>.allocate(capacity: bufferSize)
            data.withUnsafeBytes{ (bytes: UnsafePointer<UInt8>) in
                let decompressedSize = decompress_ff3(buffer, bytes)
                data = Data(bytes: buffer, count: Int(decompressedSize))
            }
            buffer.deinitialize()
            buffer.deallocate(capacity: bufferSize)
        }
    }
}
