//
//  ROMString.swift
//  ROMTools
//
//  Created by Evan Owen on 9/8/16.
//  Copyright © 2016 Evan Owen. All rights reserved.
//

import Foundation

class ROMString : NSObject, ROMSelectable, NSCopying {
    
    let rom: ROM
    @objc let index: Int
    var identifier: String?
    var properties: NSDictionary? = ["rawString" : 0]
    var nib: String? { return nil }
    
    @objc dynamic var rawString: String {
        didSet {
            updateFormattedStrings()
        }
    }
    @objc dynamic var formattedString = ""
    @objc dynamic var numberedString = ""
    
    fileprivate var linkedStrings = NSMutableArray()
    
    override var description: String { return formattedString }
    
    init(_ rom: ROM, index: Int, rawString: String) {
        self.rom = rom
        self.index = index
        self.rawString = rawString
        super.init()
        self.updateFormattedStrings()
    }
    
    deinit {
        stopObservingAllStrings()
    }
    
    fileprivate func updateFormattedStrings() {
        stopObservingAllStrings()
        
        // replace "%n" with the string index
        formattedString = rawString.replacingOccurrences(of: "%n", with: String(index))
        
        // replace all links with their appropriate strings from the ROM
        while let startIndex = formattedString.range(of: "[")?.upperBound,
            let endIndex = formattedString.range(of: "]")?.lowerBound, startIndex < endIndex {
                
                
                let components = formattedString[startIndex..<endIndex].components(separatedBy: ".")
//                let components = formattedString.substring(with: startIndex..<endIndex).components(separatedBy: ".")
                if components[0] == "rom" && components.count > 2 {
                    // link to ROMText
                    if let index = Int(object: components[2] as AnyObject?),
                        let linkedObject = rom.value(at: index, inPropertyWithKey: components[1]) {
                        
                        if components.count > 3 {
                            
                            let keyPath = components[3..<components.count].joined(separator: ".")
                            if let linkedString = (linkedObject as AnyObject).value(forKeyPath: keyPath) as? ROMText {
                                let formattedLink = linkedString.formattedString(truncatedToLength: 40).replacingOccurrences(of: "\n", with: " ")
                                startObservingString(linkedString)
                                formattedString.replaceSubrange(formattedString.index(before: startIndex)...endIndex, with: formattedLink)
                                continue
                            }
                            
                        } else if let linkedString = linkedObject as? ROMText {
                            let formattedLink = linkedString.formattedString(truncatedToLength: 40).replacingOccurrences(of: "\n", with: " ")
                            startObservingString(linkedString)
                            formattedString.replaceSubrange(formattedString.index(before: startIndex)...endIndex, with: formattedLink)
                            continue
                        }
                    }
                    
                } else if let linkedStringTable = rom.value(forKey: "stringTable-" + components[0]) as? ROMStringTable {
                    // link to a different string table
                    if let index = Int(object: components[1] as AnyObject?),
                        let stringTableEntry = linkedStringTable[index] {
                        startObservingString(stringTableEntry)
                        formattedString.replaceSubrange(formattedString.index(before: startIndex)...endIndex, with: stringTableEntry.formattedString)
                        continue
                    }
                }
                
                // unable to parse link
                formattedString.replaceSubrange(formattedString.index(before: startIndex)...endIndex, with: "<Invalid Link>")
        }
        
        numberedString = String(index) + ": " + formattedString
    }
    
    fileprivate func startObservingString(_ string: AnyObject) {
        linkedStrings.add(string)
        if string is ROMText {
            string.addObserver(self, forKeyPath: "string", options: NSKeyValueObservingOptions.new, context: &ROMString.context)
        } else if string is ROMString {
            string.addObserver(self, forKeyPath: "formattedString", options: NSKeyValueObservingOptions.new, context: &ROMString.context)
        }
    }
    
    fileprivate func stopObservingAllStrings() {
        for object in linkedStrings {
            
            if object is ROMText {
                (object as AnyObject).removeObserver(self, forKeyPath: "string")
            } else if object is ROMString {
                (object as AnyObject).removeObserver(self, forKeyPath: "formattedString")
            }
            
        }
        linkedStrings.removeAllObjects()
    }
    
    override func observeValue(forKeyPath keyPath: String?, of object: Any?, change: [NSKeyValueChangeKey : Any]?, context: UnsafeMutableRawPointer?) {
        if context != &ROMString.context {
            super.observeValue(forKeyPath: keyPath, of: object, change: change, context: context)
            return
        }
        
        updateFormattedStrings()
    }
    
    static var context = 0
    
    func copy(with zone: NSZone?) -> Any {
        return ROMString(rom, index: index, rawString: rawString)
    }
}

class ROMStringTable : ROMObject {
    
    @objc var stringArray = NSMutableArray()
    let stringDictionary = NSMutableDictionary()
    let defaultString: String
    var isLinked: Bool = false
    
    override init(_ rom: ROM, definition: NSDictionary? = nil) {
        defaultString = definition?.object(forKey: "default") as? String ?? "String %n"
        let count = Int(object: definition, keyPath: "count") ?? 0
        
        super.init(rom, definition: definition)
        
        if definition == nil { return }
        
        // keep track of the highest defined string index
        var maxIndex = count
        
        // get defined strings
        let definedStrings = NSMutableDictionary()
        for (value, object) in definition! {
            guard let rawString = object as? String else { continue }
            
            if let range = MakeRange(object: value as AnyObject?) {
                // create a range of strings
                for index in range.lowerBound..<range.upperBound {
                    let newString = ROMString(rom, index: index, rawString: rawString)
                    newString.identifier = "string-" + (identifier ?? "unnamed")
                    definedStrings[index] = newString
                    maxIndex = max(index, maxIndex)
                }
            } else if let index = Int(object: value as AnyObject?) {
                // create a single string
                let newString = ROMString(rom, index: index, rawString: rawString)
                newString.identifier = "string-" + (identifier ?? "unnamed")
                definedStrings[index] = newString
                maxIndex = max(index, maxIndex)
            }
        }
        
        // populate the array and dictionary up to the maximum index
        for index in 0...maxIndex {
            if let definedString = definedStrings[index] as? ROMString {
                // add a custom string that we created above
                stringArray.add(definedString)
                stringDictionary[index] = definedString
            } else if index < count {
                // or create a default string
                let newString = ROMString(rom, index: index, rawString: defaultString)
                newString.identifier = "string-" + (identifier ?? "unnamed")
                stringArray.add(newString)
                stringDictionary[index] = newString
            }
        }
    }
    
    override func createDefinition() -> NSMutableDictionary {
        let definition = super.createDefinition()
        
        definition["default"] = defaultString
        definition["type"] = "ROMStringTable"
        
        var count = 0
        for object in stringArray {
            guard let string = object as? ROMString else { continue }
            
            if string.index == count {
                count = string.index + 1
            }
            
            if string.rawString == defaultString {
            } else {
                definition[String(string.index)] = string.rawString
            }
        }
        definition["count"] = count
        
        return definition
    }
    
    subscript(index: Int) -> AnyObject? {
        return stringAtIndex(index)
    }
    
    func stringAtIndex(_ index: Int) -> AnyObject? {
        return stringDictionary[index] as AnyObject?
    }
    
    func createNewStringAtIndex(_ index: Int) -> ROMString {
        // create a new default string
        let newString = ROMString(rom, index: index, rawString: defaultString)
        newString.identifier = "string-" + (identifier ?? "unnamed")
        return newString
    }
    
    func addString(_ string: ROMString) {

        // if there is already a string at that index, remove it
        if let oldString = stringDictionary[string.index] {
            stringArray.remove(oldString)
            stringDictionary[string.index] = nil
        }
        
        // insert the new string in the numerical order in the array
        var insertIndex = NSNotFound
        for object in stringArray {
            // find the first string with a larger index
            guard let existingString = object as? ROMString else { continue }
            if existingString.index < string.index { continue }
            insertIndex = stringArray.index(of: existingString)
            break
        }
        if insertIndex == NSNotFound {
            willChange(NSKeyValueChange.insertion, valuesAt: IndexSet(integer: stringArray.count), forKey: "stringArray")
            stringArray.add(string)
            didChange(NSKeyValueChange.insertion, valuesAt: IndexSet(integer: stringArray.count), forKey: "stringArray")
        } else {
            willChange(NSKeyValueChange.insertion, valuesAt: IndexSet(integer: insertIndex), forKey: "stringArray")
            stringArray.insert(string, at: insertIndex)
            didChange(NSKeyValueChange.insertion, valuesAt: IndexSet(integer: insertIndex), forKey: "stringArray")
        }
        
        // insert the new string in the string dictionary
        stringDictionary[string.index] = string
    }
    
    func removeString(_ string: ROMString) {
        guard let currentString = stringDictionary[string.index] as? ROMString else { return }
        if currentString === string  {
            stringDictionary[string.index] = nil
            let removeIndex = stringArray.index(of: string)
            if removeIndex != NSNotFound {
                willChange(NSKeyValueChange.removal, valuesAt: IndexSet(integer: removeIndex), forKey: "stringArray")
                stringArray.remove(string)
                didChange(NSKeyValueChange.removal, valuesAt: IndexSet(integer: removeIndex), forKey: "stringArray")
            }
        }
    }
}
