//
//  ROMText.swift
//  ROMTools
//
//  Created by Evan Owen on 3/4/15.
//
//

import Foundation

@objc class ROMText: ROMAssembly {
    
    // MARK: instance properties
    
    fileprivate var lazyString: String?
    var encodedData = NSMutableData()
    var byteIndex: Int
    var length: Int? // a nil length means the string is null-terminated
    var encoding: ROMTextEncoding?
    override var properties: NSDictionary? { get { return ["string" : 0] } set { } }
    override var assembledLength: Int {
        // fixed length strings are easy
        if length != nil { return length! }
        
        // for variable length strings, assemble to a dummy Data and return the length
        var assembledData = Data()
        assembleTo(&assembledData)
        return assembledData.count
    }

    @objc var string: String {
        get {
            if lazyString == nil {
                lazyString = encoding?.decode(data: encodedData as Data) ?? NSString(data: encodedData as Data, encoding: String.Encoding.utf8.rawValue) as String?
                if lazyString == nil {
                    rom.log("Unable to decode text")
                    lazyString = ""
                }
                
                // make sure null-terminated strings actually end with a null-terminator
                if length == nil {
                    if let nullTerminatorRange = lazyString!.range(of: "\\0") {
                        // delete everything after the first null-terminator
                        //println("String \"\(value)\" does not end with a null-terminator, removing trailing characters")
                        lazyString = lazyString!.substring(to: nullTerminatorRange.lowerBound)
                    }
                }
            }
            return lazyString!
        }
        set {
            // don't encode if the value is the same
            if lazyString == newValue { return }
            
            // encode the string
            if let data = encoding?.encode(string: newValue) {
                encodedData = data
            } else if let asciiData = newValue.data(using: String.Encoding.utf8) {
                // no encoding table, try to encode using ASCII character set
                encodedData = NSMutableData(data: asciiData)
            } else {
                rom.log("Unable to encode text: " + newValue)
                return
            }
            
            if (length == nil) {
                // null-terminated string
                
                // get the terminator value from the character set, use zero if not found (ascii null terminator)
                var terminatorValue = encoding?.encodingTable["\\0"] as? Int ?? 0
                
                // make sure the last character is a null-terminator
                if encodedData.length != 0 {
                    var lastValue: Int = 0
                    encodedData.getBytes(&lastValue, range: NSMakeRange(encodedData.length - 1, 1))
                    if lastValue != terminatorValue {
                        // string doesn't have a terminator, add one
                        encodedData.append(&terminatorValue, length: 1)
                    }
                } else {
                    // string is empty, add a terminator
                    encodedData.append(&terminatorValue, length: 1)
                }
                
            } else {
                // fixed-length string
                if encodedData.length > length! {
                    rom.log("Text \"\(newValue)\" too long, trimming to \(length!) characters")
                    encodedData.length = length!
                } else if encodedData.length < length! {
                    // get pad value if it can be found in the character set, otherwise use ascii space
                    let padValue = encoding?.encodingTable[""] as? Int ?? 0x20
                    var padBytes = [UInt8](repeating: UInt8(padValue), count: length! - encodedData.length)
                    encodedData.append(&padBytes, length: padBytes.count)
                }
            }
            
            // invalidate the value so it will have to be decoded again
            lazyString = nil
        }
    }
    
    override init(_ rom: ROM, definition: NSDictionary? = nil) {
        
        // get the byte index and bit mask
        byteIndex = Int(object: definition, keyPath: "byteIndex") ?? 0
        length = Int(object: definition, keyPath: "length")

        if let encoding = definition?["encoding"] as? String {
            self.encoding = rom.textEncodings[encoding] as? ROMTextEncoding
        }
        
        super.init(rom, definition: definition)
    }
    
    override func createDefinition() -> NSMutableDictionary {
        let definition = super.createDefinition()
        
        // type
        definition["type"] = "ROMText"

        // byte index
        if byteIndex != 0 {
            definition["byteIndex"] = byteIndex
        }
        
        // length
        if length != nil {
            definition["length"] = length!
        }
        
        // encoding
        definition["encoding"] = encoding?.identifier
        
        return definition
    }
    
    override func setValue(_ value: Any?, forKey key: String) {
        
        // make sure the string property never gets set to a nil value
        if (key == "string") && value == nil {
            string = ""
            return
        }
        super.setValue(value, forKey: key)
    }
    
    override func setNilValueForKey(_ key: String) {
        
        // make sure the string property never gets set to a nil value
        if key == "string" {
            string = ""
            return
        }
        super.setNilValueForKey(key)
    }
    
    // MARK: ROMAssembly methods
    
    override func assembleTo(_ data: inout Data) {
        
        if length == nil {
            
            // null-terminated string

            // get the terminator value from the character set, use zero if not found (ascii null terminator)
            var terminatorValue = encoding?.encodingTable["\\0"] as? Int ?? 0
            
            // make sure the last character is a null-terminator
            if encodedData.length != 0 {
                var lastValue: Int = 0
                encodedData.getBytes(&lastValue, range: NSMakeRange(encodedData.length - 1, 1))
                if lastValue != terminatorValue {
                    // string doesn't have a terminator, add one
                    encodedData.append(&terminatorValue, length: 1)
                }
            } else {
                // string is empty, add a terminator
                encodedData.append(&terminatorValue, length: 1)
            }

            data = data.subdata(in: 0..<byteIndex)
            data.append(encodedData as Data)
        } else {
            // fixed length string
            data.replaceSubrange(byteIndex..<(byteIndex + length!), with: encodedData as Data)
        }
        
    }
    
    override func disassembleFrom(_ data: Data) {
        lazyString = nil
        
        let start = byteIndex + (range?.lowerBound ?? 0)
        if start > data.count {
            rom.log("Invalid text range")
            return
        }
        
        if let fixedLength = length {
            // fixed-length string
            var end = fixedLength + start
            if end > data.count {
                rom.log("Text length exceeds data size")
                end = data.count
            }
            encodedData = NSMutableData(data: data.subdata(in: start..<end))
        } else {
            // null-terminated string
            let end = range?.upperBound ?? data.count
            encodedData = NSMutableData(data: data.subdata(in: start..<end))
        }
    }
    
    override var description: String {
        return string
    }
    
    func formattedString(truncatedToLength truncatedLength: Int = 0) -> String {
        
        if truncatedLength != 0 {
            return encoding?.format(string: self.string, truncatedLength: truncatedLength) ?? self.string
        } else {
            return encoding?.format(string: self.string) ?? self.string
        }
    }
    
    var numLines: Int {
        return encoding?.numLines(string: string) ?? 1
    }
}

class ROMTextEncoding: ROMObject {
    
    var encodingTable = NSDictionary()
    var decodingTable = NSDictionary()
    var characterTables: [String]?
    
    required override init(_ rom: ROM, definition: NSDictionary? = nil) {
        super.init(rom, definition: definition)
        
        // get the array of character tables included in this encoding
        characterTables = definition?["characterTables"] as? [String]
        if characterTables == nil { return }
        
        // create a dictionary to store the combined character table data
        var combinedTable = [String : String]()
        
        let nameString = name ?? identifier ?? "Unnamed Text Encoding"
        
        // add each included character table to the combined table
        for tableName in characterTables! {

            // get the character table from the ROM
            guard let characterTable = rom.characterTables[tableName] as? [String : String] else {
                // the character table was not found
                rom.log("Invalid character table \"\(tableName)\" in text encoding \(nameString)")
                continue
            }
            
            // add each entry from the character table to the combined table
            for (key, value) in characterTable {
                if combinedTable[key] != nil {
                    rom.log("Text Encoding \"\(nameString)\" already contains a character for value \(key)")
                }
                combinedTable[key] = value
            }
        }
        
        // add each character value to the decoding and encoding tables
        let mutableDecodingTable = NSMutableDictionary()
        let mutableEncodingTable = NSMutableDictionary()
        for (indexString, characterString) in combinedTable {
            
            guard let index = Int(object: indexString as AnyObject?) else {
                rom.log("Unable to parse character table entry \(indexString), \(characterString)")
                continue
            }
            
            if encodingTable[characterString] != nil {
                // encoding table contains more than one entry with the same string value
                rom.log("Text Encoding \(nameString) contains multiple entries with string value \"\(characterString)\"")
            } else {
                mutableDecodingTable[index] = characterString
                mutableEncodingTable[characterString] = index
            }
        }
        decodingTable = mutableDecodingTable
        encodingTable = mutableEncodingTable
    }
    
    override func createDefinition() -> NSMutableDictionary {
        let definition = super.createDefinition()
        definition["characterTables"] = characterTables
        definition["type"] = "ROMTextEncoding"
        return definition
    }
    
    func decode(data: Data) -> String {
        var decodedString = ""
        
        // copy the data into an array
        let rawString: [UInt8] = data.withUnsafeBytes{ [UInt8](UnsafeBufferPointer(start: $0, count: data.count)) }
        var rawStringPointer = 0
        
        // decode each byte
        while rawStringPointer < rawString.count {
            let characterValue = Int(rawString[rawStringPointer])
            rawStringPointer += 1
            var twoByteCharacter: String? = nil
            if (rawStringPointer < rawString.count) && (characterValue != 0) {
                let twoByteCharacterValue = (characterValue << 8) | Int(rawString[rawStringPointer])
                twoByteCharacter = decodingTable[twoByteCharacterValue] as? String
            }
            if twoByteCharacter != nil {
                // two-byte character
                decodedString += twoByteCharacter!
                rawStringPointer += 1

            } else if let character = decodingTable[characterValue] as? String {
                // character found in character set, append to decoded string
                decodedString += character
                if character == "\\0" {
                    // null-terminator
                    break
                } else if character.hasSuffix("[") && character.hasPrefix("\\") {
                    // escape sequence with a parameter
                    if (rawStringPointer >= rawString.count) {
                        rom.log("Escape sequence at end of string without expected parameter")
                        break
                    }
                    let parameter = rawString[rawStringPointer]
                    rawStringPointer += 1
                    decodedString += String(parameter) + "]"
                }
            } else {
                rom.log("Invalid character value \(characterValue)")
            }
        }
        return decodedString
    }
    
    func encode(string: String) -> NSMutableData {
        
        let encodedData = NSMutableData()
        var currentCharacter = string.startIndex
        let lastCharacter = string.endIndex
        let encodingKeys = encodingTable.allKeys as! [String]
        
        while currentCharacter != lastCharacter {
            var encodedValue: Int?
            
            // find matches in the encoding table
            var possibleMatches = encodingKeys
            let remainingString = String(string[currentCharacter...])
//            let remainingString = string.substring(from: currentCharacter)
            possibleMatches = possibleMatches.filter{ remainingString.hasPrefix($0) }
            
            // get the longest match
            let longestMatch = possibleMatches.reduce("", { $0.characters.count > $1.characters.count ? $0 : $1 })
            
            if longestMatch == "" {
                // no matches found, skip this character
                rom.log("Invalid character: " + String(string.characters[currentCharacter]))
                currentCharacter = string.index(after: currentCharacter)
                continue
                
            } else if longestMatch == "\\0" {
                // end of string
                break
            }
            encodedValue = encodingTable[longestMatch] as? Int
            currentCharacter = string.index(currentCharacter, offsetBy: longestMatch.characters.count)
            
            if longestMatch.hasSuffix("[") {
                // escape sequence with a parameter
                var parameterString = String(string[currentCharacter...])
//                var parameterString = string.substring(from: currentCharacter)
                
                // scan the parameter string for an integer value
                if let endOfParameter = parameterString.characters.index(of: "]"),
                    var parameterValue = Int(object: String(parameterString[..<endOfParameter]) as AnyObject?) {
//                    var parameterValue = Int(object: parameterString.substring(to: endOfParameter) as AnyObject?) {
                    parameterString = String(parameterString[..<endOfParameter])
//                    parameterString = parameterString.substring(to: endOfParameter)
                    if parameterValue > 0xFF {
                        // parameter value can't be more than one byte
                        rom.log("Parameter value too large: " + String(parameterValue))
                        parameterValue = 0
                    }
                    
                    // add the parameter value to the encoded value
                    encodedValue! <<= 8
                    encodedValue! |= parameterValue
                    
                    // advance the current character index
                    currentCharacter = string.index(currentCharacter, offsetBy: parameterString.characters.count + 1) // +1 accounts for "]"
                } else {
                    rom.log("Invalid parameter")
                }
            }
            
            if encodedValue == nil {
                continue
            }
            
            // write the encoded value
            if encodedValue! > 255 {
                // swap bytes before appending
                encodedValue = Int(UInt16(encodedValue!).byteSwapped)
                encodedData.append(&encodedValue, length: 2)
            } else {
                encodedData.append(&encodedValue, length: 1)
            }
        }
        return encodedData
    }
    
    func format(string: String) -> String {
        let formattedString = string.replacingOccurrences(of: "\\n", with: "\n").replacingOccurrences(of: "\\page", with: "\n")
        return formattedString
    }

    func format(string: String, truncatedLength: Int) -> String {
        let truncatedLength = min(truncatedLength, string.characters.count)
        let substring = String(string[..<string.index(string.startIndex, offsetBy: truncatedLength)])
//        let substring = string.substring(to: string.index(string.startIndex, offsetBy: truncatedLength))
        if string.characters.count > truncatedLength {
            let formattedString = format(string: substring)
            if let escapeSequenceRange = formattedString.range(of: "\\") {
                return String(formattedString[..<escapeSequenceRange.lowerBound]) + "…"
//                return formattedString.substring(to: escapeSequenceRange.lowerBound) + "…"
            } else {
                return formattedString + "…"
            }
        } else {
            return format(string: substring)
        }
    }

    func numLines(string: String) -> Int {
        return numLines(string: string, truncatedLength: string.characters.count)
    }
    
    func numLines(string: String, truncatedLength: Int) -> Int {
        let truncatedLength = min(truncatedLength, string.characters.count)
        var numLines = 1
        if let regex = try? NSRegularExpression(pattern: "\\\\n", options: []) {
            numLines += regex.numberOfMatches(in: string, options: [], range: NSMakeRange(0, truncatedLength))
        }
        if let regex = try? NSRegularExpression(pattern: "\\\\page", options: []) {
            numLines += regex.numberOfMatches(in: string, options: [], range: NSMakeRange(0, truncatedLength))
        }
        return numLines
    }

}
